/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"
#include "MGLTargaImage.h"
#include "fur_generator.h"

MGLApp::MGLApp()
{
	rotation = 0.0f;
	model = 0;
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{
	int i = 0;
	MGL::TargaImage img;

	// setup OpenGL
	glEnable(GL_CULL_FACE);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
	glShadeModel(GL_SMOOTH);
	glClearColor(0.5f,0.5f,0.5f,0.0f);
	glClearDepth(1.0f);
	glDepthFunc(GL_LESS);
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	if (!GLeeInit())
		return false;

	if (!GLEE_ARB_multitexture)
		return false;

	// load the model
	model = ASE_loadFilename("bear.ase");
	if (!model)
		return false;

	// load model textures
	for (i = 0; i < model->materialCount; i++)
	{
		GLuint tex;
		glGenTextures(1,&tex);
		glBindTexture(GL_TEXTURE_2D,tex);

		if (img.Load(model->materials[i].diffuseMaterialMap.image_path))
		{
			gluBuild2DMipmaps(GL_TEXTURE_2D,(img.GetFormat() == 0)? GL_RGB:GL_RGBA,
							  img.GetWidth(),img.GetHeight(),
							  (img.GetFormat() == 0)? GL_RGB:GL_RGBA,
							  GL_UNSIGNED_BYTE,
							  img.GetImage());
		

			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);

			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

			img.Release();
		}

		textureList.push_back(tex);
	}


	// generate model display list
	for (i = 0; i < model->objectCount; i++)
	{
		GLuint dl = glGenLists(1);
		glNewList(dl,GL_COMPILE);

			if (model->objs[i].material > -1)
			{
				glEnable(GL_TEXTURE_2D);
				glBindTexture(GL_TEXTURE_2D,textureList[model->objs[i].material]);
			} 
			else 
				glDisable(GL_TEXTURE_2D);

			glBegin(GL_TRIANGLES);
				for (int j = 0; j < model->objs[i].mesh.faceCount; j++)
				{
					for (int k = 0; k < 3; k++)
					{
						int vp = model->objs[i].mesh.faces[j].vertex[k];
						int tp = model->objs[i].mesh.faces[j].texture_coordinates[k];

						glNormal3f(model->objs[i].mesh.vertex_normals[vp].x,
								   model->objs[i].mesh.vertex_normals[vp].y,
								   model->objs[i].mesh.vertex_normals[vp].z);

						glTexCoord2f(model->objs[i].mesh.texture_coordinates[tp].x,
									 -model->objs[i].mesh.texture_coordinates[tp].y);

						glVertex3f(model->objs[i].mesh.vertices[vp].x,
								   model->objs[i].mesh.vertices[vp].y,
								   model->objs[i].mesh.vertices[vp].z);
					}
				}
			glEnd();
		glEndList();
		
		displayList.push_back(dl);
	}

	// load the fin texture
	glGenTextures(1,&finTexture);
	glBindTexture(GL_TEXTURE_2D,finTexture);
	if (img.Load("fin.tga"))
	{
		glTexImage2D(GL_TEXTURE_2D,0,(img.GetFormat() == 0)? GL_RGB:GL_RGBA,
					 img.GetWidth(),img.GetHeight(),0,
					 (img.GetFormat() == 0)? GL_RGB:GL_RGBA,
					 GL_UNSIGNED_BYTE,img.GetImage());
		
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

		img.Release();
	}

	// generate the shell textures
	generateShellTexture(shellTextures,14,128);


	// generate fins
	buildFins(model,textureList,displayList,0.75f,0.12f,finTexture);


	// generate shell
	buildShells(model,textureList,displayList,shellTextures,14,0.75f,0.12f);

	tim.Init();

	return true;
}

bool MGLApp::Shutdown()
{
	// destroy texture list
	glDeleteTextures(textureList.size(),&textureList[0]);
	textureList.clear();

	// destroy fin texture
	glDeleteTextures(1,&finTexture);

	// destroy shell texture
	glDeleteTextures(14,shellTextures);

	// destroy display lists
	for (int i = 0; i < displayList.size(); i++)
		glDeleteLists(1,displayList[i]);
	displayList.clear();

	if (model)
	{
		#ifdef ASE_unloadScene
			ASE_unloadScene(model);
		#endif
	}

	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 55.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
	glMatrixMode(GL_MODELVIEW);
}

void MGLApp::Update( double dt )
{
	rotation += tim.GetElapsedSeconds() * 10.0f;
	if (rotation > 360.0f) rotation -= 360.0f;
}

void MGLApp::Render()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// set up the camera
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glTranslatef(0.0f,0.0f,-55.0f);
	glRotatef(-90.0f,1,0,0);
	glRotatef(rotation,0,0,1);

	for (int i = 0; i < displayList.size(); i++)
		glCallList(displayList[i]);
}
