/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include <cmath>
#include <string>
#include "MGLApp.h"
#include "MGLVector.h"
#include "../_common/LoadProgram.h"
extern "C" void APIENTRY glutSolidTeapot(GLdouble scale);

GLuint fp;

//Camera and light positions (world space)
GLfloat cameraPos[4] = { 0.0f, 1.5f, 4.0f, 1.0f};
GLfloat lightPos[4] = { 2.0f, 1.5f, 5.0f, 1.0f};

double time = 0.0;

MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	if (!GLeeInit())
		return false;

	//Ensure that the necessary extensions are supported
	if(	!GLEE_ARB_fragment_program || !GLEE_ARB_texture_cube_map || !GLEE_ARB_multitexture)
		return false;

	//Load the fragment program
	fp = LoadProgram(GL_FRAGMENT_PROGRAM_ARB, "teapot.fp");

	if(fp == 0)
		return false;

	//Set OpenGL states
	glClearColor(0.0f, 0.0f, 1.0f, 1.0f);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);

	return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
}

void MGLApp::Update( double dt )
{
	//Update time
	time += dt;
}

void MGLApp::Render()
{
    //Clear buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	//Set up view transformation
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    gluLookAt(	cameraPos[0], cameraPos[1], cameraPos[2],
				0.0f, 0.0f, 0.0f,
				0.0f, 1.0f, 0.0f);

	//Calculate the eye-space light position
	GLfloat viewMatrix[16];
	GLfloat eyeLightPos[4];

	glGetFloatv(GL_MODELVIEW_MATRIX, viewMatrix);
	eyeLightPos[0] = viewMatrix[0]*lightPos[0]+viewMatrix[4]*lightPos[1]+viewMatrix[8]*lightPos[2]+viewMatrix[12]*lightPos[3];
	eyeLightPos[1] = viewMatrix[1]*lightPos[0]+viewMatrix[5]*lightPos[1]+viewMatrix[9]*lightPos[2]+viewMatrix[13]*lightPos[3];
	eyeLightPos[2] = viewMatrix[2]*lightPos[0]+viewMatrix[6]*lightPos[1]+viewMatrix[10]*lightPos[2]+viewMatrix[14]*lightPos[3];
	eyeLightPos[3] = viewMatrix[3]*lightPos[0]+viewMatrix[7]*lightPos[1]+viewMatrix[11]*lightPos[2]+viewMatrix[15]*lightPos[3];

	//Set up model transformation
	glRotated(20*time, 0.0f, 1.0f, 0.0f);

	//Use automatic texture coordinate generation
	//Texture coordinate set 0 holds object space position
	GLfloat xPlane[4] = { 1.0f, 0.0f, 0.0f, 0.0f};
	glTexGenfv(GL_S, GL_OBJECT_PLANE, xPlane);
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glEnable(GL_TEXTURE_GEN_S);

	GLfloat yPlane[4] = { 0.0f, 1.0f, 0.0f, 0.0f};
	glTexGenfv(GL_T, GL_OBJECT_PLANE, yPlane);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glEnable(GL_TEXTURE_GEN_T);

	GLfloat zPlane[4] = { 0.0f, 0.0f, 1.0f, 0.0f};
	glTexGenfv(GL_R, GL_OBJECT_PLANE, zPlane);
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glEnable(GL_TEXTURE_GEN_R);

	GLfloat wPlane[4] = { 0.0f, 0.0f, 0.0f, 1.0f};
	glTexGenfv(GL_Q, GL_OBJECT_PLANE, wPlane);
	glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
	glEnable(GL_TEXTURE_GEN_Q);

	//Texture coordinate set 1 holds object space normal
	glActiveTextureARB(GL_TEXTURE1_ARB);

	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
	glEnable(GL_TEXTURE_GEN_S);

	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
	glEnable(GL_TEXTURE_GEN_T);

	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB);
	glEnable(GL_TEXTURE_GEN_R);

	glActiveTextureARB(GL_TEXTURE0_ARB);

	//Bind and enable fragment program
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp);
	glEnable(GL_FRAGMENT_PROGRAM_ARB);

	//Send the planes which define the shape of the star as local parameters 0-4
	const GLfloat starCentreHeight = 1.5f;
	const GLfloat starCentreSize = 0.25f;
	GLfloat plane0[4] = { 0.0f, 0.0f, -1.0f, starCentreHeight+starCentreSize};
	GLfloat plane1[4] = {-sin(2*PI/5), 0.0f,-cos(2*PI/5), starCentreHeight*cos(2*PI/5)+starCentreSize};
	GLfloat plane2[4] = {-sin(4*PI/5), 0.0f,-cos(4*PI/5), starCentreHeight*cos(4*PI/5)+starCentreSize};
	GLfloat plane3[4] = {-sin(6*PI/5), 0.0f,-cos(6*PI/5), starCentreHeight*cos(6*PI/5)+starCentreSize};
	GLfloat plane4[4] = {-sin(8*PI/5), 0.0f,-cos(8*PI/5), starCentreHeight*cos(8*PI/5)+starCentreSize};
	glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 0, plane0);
	glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 1, plane1);
	glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 2, plane2);
	glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 3, plane3);
	glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 4, plane4);

	//Send the eye-space light position as local parameter 5
	glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, 5, eyeLightPos);

	//Draw the teapot
	glFrontFace(GL_CW);
	glutSolidTeapot(1.0f);
	glFrontFace(GL_CCW);

	//Restore states
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);
	glActiveTexture(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glActiveTexture(GL_TEXTURE0_ARB);
	glDisable(GL_FRAGMENT_PROGRAM_ARB);
}