/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include <cmath>
#include <string>
#include <vector>
#include "MGLApp.h"
#include "MGLTargaImage.h"
#include "MGLVector.h"
#include "../_common/Dolphin.h"
#include "../_common/LoadProgram.h"
#include "Torus.h"

#define	IMAGE_RGB       0
#define IMAGE_RGBA      1
#define IMAGE_LUMINANCE 2

GLuint cubeMapId;
GLuint normalMapId;
Torus torus(32);
GLuint fp;
double time = 0.0;

MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	//Ensure that the necessary extensions are supported
	if(	!GLEE_ARB_fragment_program || !GLEE_ARB_texture_cube_map || !GLEE_ARB_multitexture)
		return false;
	
	//Load the cube map images
	MGL::TargaImage cubeImages[6];
	if(	!cubeImages[0].Load("cube_face_negx.tga") ||
		!cubeImages[1].Load("cube_face_negy.tga") ||
		!cubeImages[2].Load("cube_face_negz.tga") ||
		!cubeImages[3].Load("cube_face_posx.tga") ||
		!cubeImages[4].Load("cube_face_posy.tga") ||
		!cubeImages[5].Load("cube_face_posz.tga"))
	{
		return false;
	}

	//Create the cube map texture
	glGenTextures(1, &cubeMapId);
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, cubeMapId);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	
	glTexImage2D(	GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB, 0, GL_RGBA8,
					cubeImages[0].GetWidth(), cubeImages[0].GetHeight(), 0,
					cubeImages[0].GetFormat(),
					GL_UNSIGNED_BYTE, cubeImages[0].GetImage());

	glTexImage2D(	GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB, 0, GL_RGBA8,
					cubeImages[1].GetWidth(), cubeImages[1].GetHeight(), 0,
					cubeImages[1].GetFormat(),
					GL_UNSIGNED_BYTE, cubeImages[1].GetImage());

	glTexImage2D(	GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB, 0, GL_RGBA8,
					cubeImages[2].GetWidth(), cubeImages[2].GetHeight(), 0,
					cubeImages[2].GetFormat(),
					GL_UNSIGNED_BYTE, cubeImages[2].GetImage());

	glTexImage2D(	GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB, 0, GL_RGBA8,
					cubeImages[3].GetWidth(), cubeImages[3].GetHeight(), 0,
					cubeImages[3].GetFormat(),
					GL_UNSIGNED_BYTE, cubeImages[3].GetImage());

	glTexImage2D(	GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB, 0, GL_RGBA8,
					cubeImages[4].GetWidth(), cubeImages[4].GetHeight(), 0,
					cubeImages[4].GetFormat(),
					GL_UNSIGNED_BYTE, cubeImages[4].GetImage());

	glTexImage2D(	GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB, 0, GL_RGBA8,
					cubeImages[5].GetWidth(), cubeImages[5].GetHeight(), 0,
					cubeImages[5].GetFormat(),
					GL_UNSIGNED_BYTE, cubeImages[5].GetImage());

	//Load the normal map image
	MGL::TargaImage normalImage;
	if(!normalImage.Load("normalmap.tga"))
		return false;
	
	//Create the normal map texture
	glGenTextures(1, &normalMapId);
	glBindTexture(GL_TEXTURE_2D, normalMapId);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	
	glTexImage2D(	GL_TEXTURE_2D, 0, GL_RGBA8,
					normalImage.GetWidth(), normalImage.GetHeight(), 0,
					normalImage.GetFormat(),
					GL_UNSIGNED_BYTE, normalImage.GetImage());

	//Load the program
	fp = LoadProgram(GL_FRAGMENT_PROGRAM_ARB, "torus.fp");
	if(fp == 0)
		return false;

    //Set OpenGL states
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glShadeModel(GL_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
}

void MGLApp::Update( double dt )
{
	//Update time
	time += dt;
}

void MGLApp::Render()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	//Draw the sky plane
	//Bind and enable cube map texture
	glBindTexture(GL_TEXTURE_CUBE_MAP_ARB, cubeMapId);
	glEnable(GL_TEXTURE_CUBE_MAP_ARB);

	glDisable(GL_DEPTH_TEST);

	glBegin(GL_TRIANGLE_STRIP);
	{
		glTexCoord3f(-10.0f,-10.0f,-1.0f);
		glVertex3f(-10.0f,-10.0f,-1.0f);
		
		glTexCoord3f( 10.0f,-10.0f,-1.0f);
		glVertex3f( 10.0f,-10.0f,-1.0f);
		
		glTexCoord3f(-10.0f, 10.0f,-1.0f);
		glVertex3f(-10.0f, 10.0f,-1.0f);
		
		glTexCoord3f( 10.0f, 10.0f,-1.0f);
		glVertex3f( 10.0f, 10.0f,-1.0f);
	}
	glEnd();

	//Restore states
	glEnable(GL_DEPTH_TEST);

	glDisable(GL_TEXTURE_CUBE_MAP_ARB);
	
	//Draw the torus
	//Set modelview transform
	glPushMatrix();
	glTranslatef(0.0f, 0.0f,-15.0f);
	glRotated(40*time, 1.0, 1.0, 0.0);
	
	//Bind textures
	glBindTexture(GL_TEXTURE_2D, normalMapId);
	glActiveTexture(GL_TEXTURE1);
	glBindTexture(GL_TEXTURE_CUBE_MAP, cubeMapId);
	glActiveTexture(GL_TEXTURE0);

	//Bind and enable fragment program
	glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, fp);
	glEnable(GL_FRAGMENT_PROGRAM_ARB);

	//Draw the torus
	glVertexPointer(3, GL_FLOAT, sizeof(TorusVertex), &torus.vertices[0].position);
	glEnableClientState(GL_VERTEX_ARRAY);
	
	glTexCoordPointer(3, GL_FLOAT, sizeof(TorusVertex), &torus.vertices[0].position);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE1);
	glTexCoordPointer(3, GL_FLOAT, sizeof(TorusVertex), &torus.vertices[0].tbn[0]);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE2);
	glTexCoordPointer(3, GL_FLOAT, sizeof(TorusVertex), &torus.vertices[0].tbn[1]);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE3);
	glTexCoordPointer(3, GL_FLOAT, sizeof(TorusVertex), &torus.vertices[0].tbn[2]);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE4);
	glTexCoordPointer(2, GL_FLOAT, sizeof(TorusVertex), &torus.vertices[0].s);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE0);
	
	glDrawElements(	GL_TRIANGLES, static_cast<GLsizei>(torus.indices.size()),
					GL_UNSIGNED_INT, &torus.indices[0]);
	
	//Restore states
	glDisable(GL_FRAGMENT_PROGRAM_ARB);

	glDisableClientState(GL_VERTEX_ARRAY);

  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE1);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE2);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE3);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE4);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glClientActiveTexture(GL_TEXTURE0);

	glPopMatrix();
}
