/*************************************************************************
MGLApp.cpp

Demo for "Projective Textures" section

Author   :   Andrei Gradinari
Based on Dave Astle's application template

Written for More OpenGL Game Programming
August, 2005
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include <stdio.h>

#include "MGLApp.h"
using namespace MGL;

MGLApp::MGLApp()
{
	m_dElapTime = 0;
	
	m_pMesh = m_pLampMesh[0] = m_pLampMesh[1] = NULL;

	lamp_pos_initial[0] = Vector3(0, 0, 3.5);
	lamp_pos_initial[1] = lamp_pos_initial[0] + Vector3(0,0,-2);
	lamp_pos_initial[2] = lamp_pos_initial[0] + Vector3(1,0,0);
	lamp_alpha = lamp_beta = 0;
}

MGLApp::~MGLApp()
{
	delete m_pMesh;
	for(int i=0; i<2; i++) delete m_pLampMesh[i];
}

bool MGLApp::Init()
{	
	if( g_bUseShaders ) 
		// CHECK IF HARDWARE SUPPORTS GLSL SHADERS 
		if (glUseProgramObjectARB == NULL)
		{
			MessageBox(NULL, "In order to run this demo in shaders mode\nyour hardware must support OpenGL shading language", "Error", MB_OK);
			return false;
		}

	// LOADING DATA
	if(! LoadData())
	{
		MessageBox(NULL, "Error loading data. Exiting", "Error", MB_OK);
		return false;
	}
		
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

    return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;

	m_nWindowWidth = width;
	m_nWindowHeight = height;

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective(60.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );

	glMatrixMode(GL_MODELVIEW);
}

void MGLApp::Update( double dt )
{
	int i;
	m_dElapTime += dt;
	float m_dElapTime1 = m_dElapTime;//28;

	float ampl_alpha = sin(m_dElapTime1*0.1);
	float ampl_beta = cos(m_dElapTime1*0.1);

	lamp_alpha = ampl_alpha * cos(m_dElapTime1*1.5);
	lamp_beta =  ampl_beta * sin(m_dElapTime1*1.5);// + PI*1.3;

	for(i=0; i<3; i++) lamp_pos[i] = lamp_pos_initial[i];
	
	Vector3 axis = Vector3(0,1,0);
	for(i=0; i<3; i++) rotate_point_around_axis(lamp_alpha, lamp_pos[i], lamp_pos[0], axis);
	
	axis = Vector3(lamp_pos[0], lamp_pos[2]);
	for(i=0; i<3; i++) rotate_point_around_axis(lamp_beta, lamp_pos[i], lamp_pos[0], axis);
}

void MGLApp :: SetupLighting()
{
	float pos[4] = {1, -0.5, 1, 0};
	float ambt[4] = {1, 1, 1, 0};
	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambt);
}

void MGLApp::UpdateShaderUniforms()
{
	//SETTING UNIFORMS
	glUseProgramObjectARB(m_iProjShader);
	GLint attr_loc = glGetUniformLocationARB(m_iProjShader, "mat4_view_inv");
	if(attr_loc!=-1) 
		glUniformMatrix4fv(attr_loc, 1, GL_FALSE, (GLfloat*)&mat4_view_inv);
		// attr loc
		// 1 - amount of matrices to modify
		// GL_FALSE - means that matrix is supplied in column-major order
}

void MGLApp::LookAt(float x0, float y0, float z0, 
						float x1, float y1, float z1, 
						float xn, float yn, float zn) 
{
	gluLookAt(x0, y0, z0, x1, y1, z1, xn, yn, zn);
	glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat*)&mat4_view);
	mat4_view_inv = mat4_view.inverse();

	if(g_bUseShaders) UpdateShaderUniforms();
}

void MGLApp :: RenderGeometry(bool b_first_pass)
{
	if(g_bUseShaders)
		if(b_first_pass) glUseProgramObjectARB(0);
			else glUseProgramObjectARB(m_iProjShader);
	
	if(b_first_pass) glBindTexture(GL_TEXTURE_2D, m_iFloorTexture);
	m_pMesh->Draw(); // FLOOR & WALLS
	
	if(b_first_pass) // render lamp mesh only in first pass
	{
		GLfloat emis[] = {1.0, 1.0, 1.0, 1.0}, zero[] = {0.0, 0.0, 0.0, 1.0};

		glPushMatrix();
			glTranslatef(lamp_pos[0].x, lamp_pos[0].y, lamp_pos[0].z);;
			glRotated(lamp_alpha * 180 / PI, 0, 1, 0);
			glRotated(lamp_beta * 180 / PI, 1, 0, 0);
			
			glBindTexture(GL_TEXTURE_2D, 0);
				m_pLampMesh[0]->Draw();
		
			glMaterialfv(GL_FRONT, GL_EMISSION, emis);
			glBindTexture(GL_TEXTURE_2D, m_iProjTexture);
				m_pLampMesh[1]->Draw();
			glMaterialfv(GL_FRONT, GL_EMISSION, zero);
			
		glPopMatrix();
	}

	Vector3 p(1,0,0);
	Vector3 o(0,0,0);
	Vector3 v(0,0,1);

	rotate_point_around_axis(90.0 * PI / 180, p, o, v);
}

void MGLApp::Render()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
		
	Vector3 pos = lamp_pos[0], dir = lamp_pos[1]-lamp_pos[0], up = lamp_pos[2] - lamp_pos[0];
	//LookAt(pos.x,pos.y,pos.z,	(pos+dir).x, (pos+dir).y, (pos+dir).z,	up.x, up.y, up.z);
	
	LookAt(5,-5,4, 0,0,1, 0,0,1);

// FIRST PASS: RENDER GEOMETRY
	SetupLighting();
	glEnable(GL_TEXTURE_2D);
	RenderGeometry();
	
// SECOND PASS: ADD PROJECTIVE TEXTURE	
	glMatrixMode(GL_TEXTURE); glPushMatrix(); glMatrixMode(GL_MODELVIEW);
	
	glPushAttrib(GL_TEXTURE_BIT | GL_LIGHTING_BIT | GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);	
	
	glDisable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE); // additive blending
	glDepthFunc(GL_EQUAL);
	
	m_Projector.LookAt(pos, pos+dir, up);
	m_Projector.SetProjection(1, 60);
	
	if(g_bUseShaders) m_Projector.SetupTexMatrix();
	else m_Projector.SetupTexGen(); 
	// set up texture generation after LookAt function and before any other modelview matrix operations
	
	glBindTexture(GL_TEXTURE_2D, m_iProjTexture);
	RenderGeometry(false);

	glDisable(GL_TEXTURE_GEN_S);
	glDisable(GL_TEXTURE_GEN_T);
	glDisable(GL_TEXTURE_GEN_R);
	glDisable(GL_TEXTURE_GEN_Q);

	glPopAttrib();
	glMatrixMode(GL_TEXTURE); glPopMatrix(); glMatrixMode(GL_MODELVIEW);
// **********************************

	glPopMatrix();
}


bool MGLApp::LoadData()
{
// LOADING TEXTURES	
	m_iFloorTexture = LoadTexture("data/floor.tga");
	m_iProjTexture = LoadTexture("data/spot.tga", GL_CLAMP, false);

	if(!m_iFloorTexture  || !m_iProjTexture) return false;
	
// LOADING MODELS	
	m_pMesh = CPrimitiveFactory::LoadPrimitive("Data/room.txt");
	m_pLampMesh[0] = CPrimitiveFactory::LoadPrimitive("Data/lamp0.txt");
	m_pLampMesh[1] = CPrimitiveFactory::LoadPrimitive("Data/lamp1.txt");

	if(m_pMesh == NULL) return false;
	for(int i=0; i<2; i++) if (m_pLampMesh[i] == NULL) return false;

// LOADING SHADERS
	if(g_bUseShaders)
	{
		m_iProjShader = LoadShader("data/shaders/proj.vert", "data/shaders/proj.frag");
		return m_iProjShader != 0;
	}

	return true;
}

GLuint MGLApp :: LoadTexture(char* fname, GLenum wrapst, bool mipmap)
{
	TargaImage img;
	if(!img.Load(fname)) return 0;
	img.FlipVertical();
	int width = img.GetWidth();
	int height = img.GetHeight();

	GLuint tex_id = 0;
	glGenTextures(1, &tex_id );
	glBindTexture(GL_TEXTURE_2D, tex_id);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mipmap ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, wrapst);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, wrapst);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, width, height, GL_RGB, GL_UNSIGNED_BYTE, img.GetImage() );
	
	return tex_id;
}

GLhandleARB MGLApp::LoadShader(const char* s_vert, const char* s_frag)
{
	GLhandleARB shader_object = 0;
	
	char *vs_src = textFileRead(s_vert);
	char *fs_src = textFileRead(s_frag);
	
	GLint vs_len, fs_len;

	if(vs_src==NULL || fs_src==NULL) goto label_end;
	
	vs_len = (GLint)strlen(vs_src);
	fs_len = (GLint)strlen(fs_src);

	GLhandleARB vertex_shader, fragment_shader;
	shader_object = glCreateProgramObjectARB();
	vertex_shader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	fragment_shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
	
	
	glShaderSourceARB(vertex_shader, 1, (const char**) &vs_src, NULL);
	glShaderSourceARB(fragment_shader, 1, (const char**) &fs_src, NULL);

	GLint t;
	glCompileShaderARB(vertex_shader);
	glGetObjectParameterivARB(vertex_shader, GL_OBJECT_COMPILE_STATUS_ARB, &t); 
	if(t==0)
	{
		glDeleteObjectARB(shader_object);
		shader_object = 0;
		goto label_end;
	}
	
	glCompileShaderARB(fragment_shader);
	glGetObjectParameterivARB(fragment_shader, GL_OBJECT_COMPILE_STATUS_ARB, &t); 
	if(t==0)
	{
		glDeleteObjectARB(shader_object);
		shader_object = 0;
		goto label_end;
	}
	
	glAttachObjectARB(shader_object, vertex_shader);
	glAttachObjectARB(shader_object, fragment_shader);

	 // LINKING
	glLinkProgramARB(shader_object);
	glGetObjectParameterivARB(shader_object, GL_OBJECT_LINK_STATUS_ARB, &t); 
	if(t==0)
	{
		glDeleteObjectARB(shader_object);
		shader_object = 0;
		goto label_end;
	}

	//SETTING UNIFORMS
	glUseProgramObjectARB(shader_object);
	GLint attr_loc;
	attr_loc = glGetUniformLocationARB(shader_object, "tex_unit_0");
	if(attr_loc!=-1) glUniform1iARB(attr_loc, 0);
	
label_end:
	if(vs_src!=NULL)delete vs_src;
	if(fs_src!=NULL)delete fs_src;
	return shader_object;
}