/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"
#include <gl/glut.h>


#define DYN_TEX_WIDTH		256
#define DYN_TEX_HEIGHT		256


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	// initialize extension manager
	if (!GLeeInit())
		return false;

	// initialize time and frame counters
	m_dElapTime = 0.0;
	m_dFPSTime = 0.0;
	m_nFrames = 0;
	m_dFPS = 0.0;

	// init model rotation
	m_fLeftX = 0.f;
	m_fLeftY = 0.f;
	m_fRightX = 0.f;
	m_fRightY = 0.f;

	return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid divide by zero
	if( height == 0 )
		height = 1;					

	m_nWindowWidth = width;
	m_nWindowHeight = height;
}

void MGLApp::Update( double dt )
{
	// accumulate our time counter
	m_dElapTime = m_dElapTime + dt;
}

void MGLApp::DrawDodecahedron()
{
	// this is the object to be rendered to a dynamic texture, so only 
	// render to a small portion of the screen
	glViewport( 0, 0, DYN_TEX_WIDTH, DYN_TEX_HEIGHT );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, 1.0f, 1.0f, 1000.0f );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0.0f, 0.0f, 5.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);

	float fHeight = (float)sin(5.0*m_dElapTime)*0.2f;
	glTranslatef( 0.f, fHeight, 0.f );

	glRotatef( m_fRightX/3.f, 0.f, 1.f, 0.f );
	glRotatef( m_fRightY/3.f, 1.f, 0.f, 0.f );

	glutSolidDodecahedron();
}

void MGLApp::DrawTeapot()
{
	// render to the whole screen
	glViewport( 0, 0, m_nWindowWidth, m_nWindowHeight );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)m_nWindowWidth/(GLfloat)m_nWindowHeight, 1.0f, 1000.0f );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	gluLookAt( 0.0f, 0.0f, 4.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
	glRotatef( m_fLeftX/3.f, 0.f, 1.f, 0.f );
	glRotatef( m_fLeftY/3.f, 1.f, 0.f, 0.f );

	glutSolidTeapot( 1.f );
}

void MGLApp::Render()
{
	static unsigned char pTempBuf[ 3 * DYN_TEX_WIDTH * DYN_TEX_HEIGHT ];
	float pfWhite[4] = { 1.f, 1.f, 1.f, 1.f };
	float pfGreen[4] = { 0.5f, 1.f, 0.5f, 1.f };

	// enable lighting to make things look a little less boring
	glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, pfWhite );
	glMaterialf( GL_FRONT_AND_BACK, GL_SHININESS, 50.f );
	glEnable( GL_LIGHT0 );
	glEnable( GL_LIGHTING );
	glEnable( GL_DEPTH_TEST );


	//
	// Render a mesh to a texture
	//

	// clear the screen
	glClearColor( 1.f, 0.5f, 0.5f, 1.f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// draw the first object to the window
	glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, pfGreen );
	glDisable( GL_TEXTURE_2D );
	DrawDodecahedron();

	// create a texture from the render
	glReadPixels( 0, 0, DYN_TEX_WIDTH, DYN_TEX_HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, pTempBuf );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, DYN_TEX_WIDTH, DYN_TEX_HEIGHT, 0, GL_RGB, GL_UNSIGNED_BYTE, pTempBuf );


	//
	// Apply the dynamic texture to another object
	//

	// clear the screen
	glClearColor( 0.7f, 0.3f, 0.3f, 1.f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// set up texture state for the second pass
	glEnable( GL_TEXTURE_2D );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY );

	// draw the second object with the first object textured onto it
	glMaterialfv( GL_FRONT_AND_BACK, GL_DIFFUSE, pfWhite );
	DrawTeapot();


	//
	// Draw the framerate
	//

	DrawFPS();
}

void MGLApp::DrawFPS()
{
	// update FPS value
	if( ++m_nFrames == 100 )
	{
		m_dFPS = m_nFrames / (m_dElapTime - m_dFPSTime);
		m_nFrames = 0;
		m_dFPSTime = m_dElapTime;
	}

	// draw the FPS to the screen
	glPushMatrix();
	{
		static char szText[32];
		float pfWhite[4] = { 1.f, 1.f, 1.f, 1.f };

		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glOrtho( 0.f, 100.f, 0.f, 100.f, -100.f, 100.f );

		glMatrixMode( GL_MODELVIEW );
		glLoadIdentity();

		glDisable( GL_LIGHTING );
		glDisable( GL_TEXTURE_2D );
		glColor3f( 1.0f, 1.0f, 1.0f );
		glRasterPos2f( 0, 0 );

		sprintf( szText, "FPS=%.2f", m_dFPS );
		for( char* p = szText ; *p ; p++ )
			glutBitmapCharacter( GLUT_BITMAP_HELVETICA_18, *p );
	}
	glPopMatrix();
}

void MGLApp::LeftMouseMove( int dx, int dy )
{
	m_fLeftX += (float)dx;
	m_fLeftY += (float)dy;
}

void MGLApp::RightMouseMove( int dx, int dy )
{
	m_fRightX += (float)dx;
	m_fRightY += (float)dy;
}

