#include "Cubemap.h"
#include <sstream>
#include <iostream>
#include <math.h>
#include "Endianness.h"

namespace ibl
{
	// some helper functions for loading vertical cross files easily
	void copyBlock(	char* src, unsigned int src_image_width, unsigned int src_row, unsigned int src_col,
					char* dst, unsigned int dst_image_width, unsigned int dst_row, unsigned int dst_col,
					unsigned int block_w, unsigned int block_h, unsigned int pixelsize	)
	{
		for( unsigned int row=0; row<block_h; ++row )
		{
			char* t_src = &( src[pixelsize*(src_image_width * (row+src_row) + src_col)] );
			char* t_dst = &( dst[pixelsize*(dst_image_width * (row+dst_row) + dst_col)] );
			memcpy( t_dst, t_src, pixelsize * block_w );
		}
	}
	
	void flipBlock( char* data, int width, int height, bool vertical, bool horizontal, int pixel_len )
	{
		int		row_len = width * pixel_len;
		
		if( vertical )
		{
			char*	t = new char[row_len];
			for( int i=0; i<height/2; ++i )
			{
				memcpy(t, data + i*row_len, row_len);
				memcpy(data + i*row_len, data + (height-1-i)*row_len, row_len);
				memcpy(data + (height-1-i)*row_len, t, row_len);
			}
			delete [] t;
		}
		
		if( horizontal )
		{
			for( int i=0; i<height; ++i )
			for( int j=0; j<width/2; ++j )
			for( int k=0; k<pixel_len; ++k )
			{
				char* a = &(data[i*row_len + j*pixel_len + k]);
				char* b = &(data[i*row_len + (width-1-j)*pixel_len + k]);
				*a ^= *b;
				*b ^= *a;
				*a ^= *b;
			}
		}
	}


	Cubemap::Cubemap() : mCubemapID( 0 ),
						 mTextureUnit( GL_TEXTURE0 )
	{
	}
	
	Cubemap::~Cubemap()
	{
	}

	void Cubemap::setTextureUnit( GLuint texUnit )
	{
		mTextureUnit = texUnit;
	}

	bool Cubemap::load( const std::string & file )
	{
		mCubemapID = 0;
		std::string ext = file.substr( file.size() - 3, 3 );
		if( ext == "tga" || ext == "TGA" ) return loadLDR( file );
		if( ext == "pfm" || ext == "PFM" ) return loadPFM( file );
		return false;
	}

	bool Cubemap::loadPFM( const std::string & filename )
	{	
		/* The PFM format. Source: http://netpbm.sourceforge.net/doc/pfm.html
		The PFM header is 3 consecutive "lines" of ASCII text. After each line is a white space 
		character. That character is typically a newline character, hence the term "line," but
		doesn't have to be.

		Identifier Line
		===============
		The identifier line contains the characters "PF" or "Pf". PF means it's a color PFM. Pf
		means it's a grayscale PFM.

		Dimensions Line
		===============
		The dimensions line contains two positive decimal integers, separated by a blank. The
		first is the width of the image; the second is the height. Both are in pixels.

		Scale Factor / Endianness
		=========================
		The Scale Factor / Endianness line is a queer line that jams endianness information into
		an otherwise sane description of a scale. The line consists of a nonzero decimal number,
		not necessarily an integer. If the number is negative, that means the PFM raster is
		little endian. Otherwise, it is big endian. The absolute value of the number is the
		scale factor for the image. The scale factor tells the units of the samples in the raster.
		You use it somehow along with some separately understood unit information to turn a
		sample value into something meaningful, such as watts per square meter. 
		*/
		FILE * file = fopen( filename.c_str(), "rb" );

		std::cout<<"loading texture file "<<filename<<" ... ";
		std::cout.flush();

		// this is all the information we need to figure out to create a texture
		unsigned int width;
		unsigned int height;
		unsigned int componentCount;
		float * data;
		GLenum format;
		GLint internalFormat;

		std::string line[4];
		char c;

		for( int i = 0; i<4; ++i )
		{
			//read up to whitespace
			do
			{
				fread( &c, sizeof(char), 1, file );
				line[i] += c;
			}while( c != '\n' && c!='\t' && c!=' ' && c!='\r' );

			//in between lines kill the whitespace
			if( i<3 ) while( c != '\n' && c!='\t' && c!=' ' && c!='\r' ) fread( &c, sizeof(char), 1, file );
		}
		
        if( line[0][0] != 'P' )
		{
			std::cerr<<"image not in PFM format"<<std::endl;
			return false;
		}

		// gotta love the PFM format, if the F is lower case it's 3 components, if it's lower case it's grayscale
		if( line[0][1] == 'F' )
		{
			componentCount = 3;
			format = GL_RGB;
			internalFormat = GL_RGB_FLOAT16_ATI;
		}
		else if( line[0][1] == 'f' )
		{
			componentCount = 1;
			format = GL_LUMINANCE;
			internalFormat = GL_LUMINANCE_FLOAT16_ATI;
		}
		else
		{
			std::cerr<<"image not in PFM format"<<std::endl;
			return false;
		}

        std::istringstream iss( line[1] );
		iss >> width;

		std::istringstream iss2( line[2] );
		iss2 >> height;	
	
		std::istringstream iss3( line[3] );
		float scale;
		iss3 >> scale;

		unsigned int dataSize = componentCount * width * height;
		data = new float[ dataSize ];
		fread( data, sizeof(float), dataSize, file );
		fclose( file );
		
		// sign on the scale determines endianness. if endianness is not what we need it to be, flip the byte order
#ifdef IBL_BIG_ENDIAN
		if( scale < 0 )
#else
		if( scale > 0 )
#endif
		{
			for( size_t i = 0; i<dataSize; ++i )
			{
				flipEndian( data[i] );
			}
		}

		//apply scale
		scale = fabs( scale );
		for( size_t n = 0; n< dataSize; ++n )
		{
			data[ n ] *= scale;
		}

		/* unshuffle the vertical cross image format
			 ___
			|+y |
		 ___|___|___
		|-x |-z |+x |
		|___|___|___|
		    |-y |
			|___|
			|+z |
			|___|
			
		becomes:
		 ___
		|+x | (rotated 180)
		|___|
		|-x | (rotated 180)
		|___|
		|+y |
		|___|
		|-y |
		|___|
		|+z |
		|___|
		|-z | (rotated 180)
		|___|
		
		*/
		
		int	pixelsize = componentCount * sizeof(float);
		char*	old = (char*)data;
		int		old_w = width;
		int		facesize = width/3;
		int		facelength = facesize*facesize*componentCount*sizeof(float);
		
		width = facesize;
		height += height/2;
		data = new float[ width * height * componentCount ];
		
		copyBlock( old, old_w, 2*facesize, 2*facesize,    (char *)data, width, 5*facesize,0*facesize,    facesize, facesize, pixelsize );	//pos x
		copyBlock( old, old_w, 2*facesize, 0*facesize,    (char *)data, width, 4*facesize,0*facesize,    facesize, facesize, pixelsize );	//neg x
		copyBlock( old, old_w, 3*facesize, 1*facesize,    (char *)data, width, 3*facesize,0*facesize,    facesize, facesize, pixelsize );	//pos y
		copyBlock( old, old_w, 1*facesize, 1*facesize,    (char *)data, width, 2*facesize,0*facesize,    facesize, facesize, pixelsize );	//neg y
		copyBlock( old, old_w, 0*facesize, 1*facesize,    (char *)data, width, 1*facesize,0*facesize,    facesize, facesize, pixelsize );	//pos z
		copyBlock( old, old_w, 2*facesize, 1*facesize,    (char *)data, width, 0*facesize,0*facesize,    facesize, facesize, pixelsize );	//neg z
		flipBlock( (char *)data, width, height/6, true, true, componentCount * sizeof(float) ); //rotate neg z by 180
		flipBlock( (char *)data + 5*facelength, width, height/6, true, true, componentCount * sizeof(float) ); //rotate pos x 180
		flipBlock( (char *)data + 4*facelength, width, height/6, true, true, componentCount * sizeof(float) ); //rotate neg x 180

		glPushAttrib( GL_ENABLE_BIT );
		glEnable( GL_TEXTURE_CUBE_MAP_ARB );
		glGenTextures( 1, &mCubemapID );
		glBindTexture( GL_TEXTURE_CUBE_MAP_ARB, mCubemapID );

		for( int k=0; k<6; ++k )
		{
			GLenum target = GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB + GLenum(k);
			
			// it is assumed data is stored in collumn format, top down: +x -x +y -y +z -z
			float * facedata = data + (5-k) * width * height * componentCount / 6;
			glTexImage2D( target, 0, internalFormat, width, height/6, 0, format, GL_FLOAT, facedata );			
		}

		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
		glTexParameteri(GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );

		glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glPopAttrib();
		
		delete [] old;
		delete [] data;
		std::cout<<"success"<<std::endl;
		return true;
	}

	bool Cubemap::loadLDR( const std::string & file )
	{
		//TODO: implement TGA loading and cross unshuffling same as above
		return false;
	}

	void Cubemap::bind()const
	{
		if( mCubemapID )
		{
			glActiveTexture( mTextureUnit );
			glEnable( GL_TEXTURE_CUBE_MAP_ARB );
			glBindTexture( GL_TEXTURE_CUBE_MAP_ARB, mCubemapID );
			glActiveTexture( GL_TEXTURE0 );
		}
	}

	void Cubemap::unbind()const
	{
		glActiveTexture( mTextureUnit );
		glDisable( GL_TEXTURE_CUBE_MAP_ARB );
		glActiveTexture( GL_TEXTURE0 );
	}
}
