/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	if (!GLeeInit())
	{
		std::cerr<<"ERROR: Failed to initialize GLee."<<std::endl;
		return false;
	}

	// check for ARB fragment program support
	if(!GLEE_ARB_fragment_program)
	{
		std::cerr<<"ERROR: This demo requires GL_ARB_fragment_program support."<<std::endl;
		return false;
	}

	mMenuAlpha = 0;
	mHelpAlpha = 0;
	mF1.load( "data/F1.tga" );
	mMenu.load( "data/menu.tga" );
	mTitle.load( "data/title.tga" );
	mMouseHelp.load( "data/mouse.tga" );
	mKeyHelp.load( "data/help.tga" );
	mFadeInTimer = 5.0f;


	// define which texture units are used for which components of our algorithm
	GLuint ambOccTexUnit = GL_TEXTURE0;
	GLuint diffIrrTexUnit = GL_TEXTURE1;
	GLuint reflectionTexUnit = GL_TEXTURE2;

	mCurrentEnv = 0;

	// load the environment textures as specified in environments.txt
	std::ifstream in;
	in.open( "data/environments.txt" );
	while( !in.eof() )
	{
		std::string dimPath="";
		std::string reflPath="";

		char c=in.get();
		if( in.eof() ) break;
		while( !in.eof() && c == '\n' || c == '\r' ) c = in.get();
		while( !in.eof() && c != '\n' && c!= '\r' )
		{
			dimPath += c;
			c = in.get();				
		}
		while( !in.eof() && c == '\n' || c == '\r' ) c = in.get();
		while( !in.eof() && c != '\n' && c!= '\r' )
		{
			reflPath += c;
			c = in.get();				
		}

		if( !dimPath.empty() && !reflPath.empty() && dimPath[0] != '#' && reflPath[0] != '#' )
		{
			Environment env;
			env.mDiffIrrMap.load( dimPath );
			env.mReflectionMap.load( reflPath );
			env.mDiffIrrMap.setTextureUnit( diffIrrTexUnit );
			env.mReflectionMap.setTextureUnit( reflectionTexUnit );
			mEnvironments.push_back( env );
		}
	}
	in.close();
	std::cout<<"done parsing environments.txt"<<std::endl;

	// load up the four pieces of the motorcycle model and their ambient occlusion textures
	mModels.resize( 4 );

	mModels[0].mMesh = new Mesh;
	mModels[0].mMesh->load( "data/cycle0.msh" );
	mModels[0].mTexture.load( "data/cycle0AO.tga" );
	mModels[0].mTexture.setTextureUnit( ambOccTexUnit );

	mModels[1].mMesh = new Mesh;
	mModels[1].mMesh->load( "data/cycle1.msh" );
	mModels[1].mTexture.load( "data/cycle1AO.tga" );
	mModels[1].mTexture.setTextureUnit( ambOccTexUnit );

	mModels[2].mMesh = new Mesh;
	mModels[2].mMesh->load( "data/cycle2.msh" );
	mModels[2].mTexture.load( "data/cycle2AO.tga" );
	mModels[2].mTexture.setTextureUnit( ambOccTexUnit );

	mModels[3].mMesh = new Mesh;
	mModels[3].mMesh->load( "data/cycle3.msh" );
	mModels[3].mTexture.load( "data/cycle3AO.tga" );
	mModels[3].mTexture.setTextureUnit( ambOccTexUnit );

	// load shaders for the models and the skybox
	mSkyboxShader.load( "data/skybox.fp" );

	mModelShaders.resize( 7 );
	mModelShaders[0].load( "data/ao.fp" );
	mModelShaders[1].load( "data/ldri_dim_ao.fp" );
	mModelShaders[2].load( "data/hdri_dim_ao.fp" );
	mModelShaders[3].load( "data/ldri_refl.fp" );
	mModelShaders[4].load( "data/hdri_refl.fp" );
	mModelShaders[5].load( "data/ldri_all.fp" );
	mModelShaders[6].load( "data/hdri_all.fp" );
	mCurrentShader = 6;

	//start out as pitch black
	mExposure[0] = mExposure[1] = mExposure[2] = mExposure[3] = 0;

	// miscellaneous GL stuff
	glDisable( GL_LIGHTING );
	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LEQUAL );
	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );		
	glColor4f(1,1,1,1);

	// generate texture coordinates for texture unit 1 based on normals 
	glActiveTexture( GL_TEXTURE1 );
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB );
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB );
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP_ARB );		
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glEnable(GL_TEXTURE_GEN_R);

	// generate texture coordinates for texture unit 2 as reflection vectors
	glActiveTexture( GL_TEXTURE2 );
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB );
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB );
	glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP_ARB );		
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glEnable(GL_TEXTURE_GEN_R);
	glActiveTexture( GL_TEXTURE0 );

	return true;
}

bool MGLApp::Shutdown()
{
	// free all memory
	for( size_t i = 0; i<mModels.size(); ++i )
	{
		mModels[i].destroy();
	}
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );

	mWindowSize[0] = width;
	mWindowSize[1] = height;
}

void MGLApp::ShowHelp()
{
	mShowHelp = true;
	mHelpAlpha = 1.0f;
}

void MGLApp::HideHelp()
{
	mShowHelp = false;
}

void MGLApp::RotateCamera( float dx, float dy )
{
	mCamera.rotate( dx, dy );
}

void MGLApp::ZoomCamera( float dz )
{
	mCamera.zoom( dz );
}

void MGLApp::ChangeExposure( float de )
{
	for( int i = 0; i<4; ++i ) mExposure[i] *= de;
}

void MGLApp::RotateEnvironment( float dy )
{
	mEnvironments[ mCurrentEnv ].mRotation += dy;
}

void MGLApp::NextEnv()
{
	mCurrentEnv++;
	if( mCurrentEnv >= mEnvironments.size() ) mCurrentEnv = 0;
}

void MGLApp::PrevEnv()
{	
	if( mCurrentEnv == 0 ) mCurrentEnv = mEnvironments.size()-1;
	else mCurrentEnv--;
}

void MGLApp::NextShader()
{
	mCurrentShader++;
	if( mCurrentShader >= mModelShaders.size() ) mCurrentShader = 0;
	mMenuAlpha = 2.0f;
}

void MGLApp::PrevShader()
{
	if( mCurrentShader == 0 ) mCurrentShader = mModelShaders.size()-1;
	else mCurrentShader--;
	mMenuAlpha = 2.0f;
}

void MGLApp::Update( double dt )
{	
	// fade out the menu
	if( mMenuAlpha > 0 ) mMenuAlpha -= 0.5f * (float)dt;

	// if help is no longer needed, fade it out
	if( !mShowHelp && mHelpAlpha > 0 ) mHelpAlpha -= 0.5f * (float)dt;

	// if still fading in ...
	if( mFadeInTimer > 0 )
	{
		mFadeInTimer -= (float)dt;
		mTitleAlpha = 1.0f;

		// leave the last three seconds for fading in from black
		if( mFadeInTimer < 3.f )
		{
			mTitleAlpha = mFadeInTimer/3;
			// after two seconds of nothing, exposure fades from 0 to 0.4 in 3 seconds
			mExposure[0]=
			mExposure[1]=
			mExposure[2]=
			mExposure[3]= (1 - mTitleAlpha) * 0.5f;
		}
	}
}

void MGLApp::Render()
{
	// no need to clear anything but the depth buffer, everything gets over-drawn
	// by the skybox anyway
	glClear( GL_DEPTH_BUFFER_BIT );
	glMatrixMode( GL_MODELVIEW );

	// pass the exposure parameter to the fragment shaders
	glProgramEnvParameter4fvARB( GL_FRAGMENT_PROGRAM_ARB, 0, mExposure );
	

	////////////
	// skybox //
	////////////

	// bind the environment cubemaps, the skybox works with the same textures
	mEnvironments[ mCurrentEnv ].bind();		
	mSkyboxShader.bind();		
	glColor4f(1,1,1,1);

	// put only the camera rotation into the modelview matrix
	// this way the skybox appears infinitely distant
	glLoadIdentity();
	mCamera.setupRotation();		

	glMatrixMode( GL_TEXTURE );	
	glRotatef( mEnvironments[ mCurrentEnv ].mRotation, 0,1,0 );
	glMatrixMode( GL_MODELVIEW );
	mSkybox.draw();
	glMatrixMode( GL_TEXTURE );
	glLoadIdentity();
	glMatrixMode( GL_MODELVIEW );

	///////////
	// model //
	///////////

	// now bind the model shader
	mModelShaders[ mCurrentShader ].bind();		

	// load the full camera view into the modelview matrix
	glLoadIdentity();
	mCamera.setup();

	// because the texture coordinates generated for sampling the environment maps
	// are in eye space, we rotate the texture 1 and texture 2 matrices by the 
	// inverse of the camera rotation, placing the texture coordinates in world space.
	// finally rotate the environment around the y axis according to user input
	glMatrixMode( GL_TEXTURE );
	glActiveTexture( GL_TEXTURE2 );
	glLoadIdentity();
	glRotatef( mEnvironments[ mCurrentEnv ].mRotation, 0,1,0 );
	mCamera.setupInverseRotation();

	glActiveTexture( GL_TEXTURE1 );
	glLoadIdentity();
	glRotatef( mEnvironments[ mCurrentEnv ].mRotation, 0,1,0 );
	mCamera.setupInverseRotation();

	glActiveTexture( GL_TEXTURE0 );
	glEnable( GL_TEXTURE_2D );
	glMatrixMode( GL_MODELVIEW );

	// the models are stored as z-up data but gluPerspective sets up a y-up system
	// this is a quick hack to fix it.
	glRotatef( 90,1,0,0 );

	// draw all models
	glColor4f(0.3f,0.3f,0.3f,0.3f);
	for( size_t i = 0; i<mModels.size(); ++i )
	{
		mModels[i].draw();
	}
	
	// clean up
	mModelShaders[ mCurrentShader ].unbind();
	mEnvironments[ mCurrentEnv ].unbind();

	DrawMenu();
}

//draw the shader menu
void MGLApp::DrawMenu()
{
	// make an orthographic 2D projection for nice screen aligned drawing
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	glViewport( 0, 0, mWindowSize[0], mWindowSize[1] );
	gluOrtho2D( 0, mWindowSize[0], mWindowSize[1], 0 );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA,GL_ONE );	
	glDisable( GL_DEPTH_TEST );

	unsigned size;

	if( mMenuAlpha > 0 && mFadeInTimer < 1.f )
	{
		// bind the menu texture and draw
		mMenu.bind();
		
		glColor4f(1,1,1,mMenuAlpha);
		
		size = 256;
		glBegin( GL_QUADS );
		glTexCoord2f(0,0);
		glVertex2i(0,20 + size);

		glTexCoord2f(1,0);
		glVertex2i(size,20 + size);	

		glTexCoord2f(1,1);
		glVertex2i(size,20);

		glTexCoord2f(0,1);
		glVertex2i(0,20);	
		glEnd();
		mMenu.unbind();

		// draw the selection box over the current shader
		int y = 20 + 40 * (int)mCurrentShader;
		glColor4f(1,1,1,mMenuAlpha*0.1f);
		glBegin( GL_QUADS );
		glVertex2i( 0, y );
		glVertex2i( 0, y + 20 );
		glVertex2i( size, y + 20 );
		glVertex2i( size, y );
		glEnd();

		// nice outline too
		glColor4f(1,1,1,mMenuAlpha*0.3f);
		glBegin( GL_LINE_LOOP );
		glVertex2i( 0, y );
		glVertex2i( 0, y + 20 );
		glVertex2i( size, y + 20 );
		glVertex2i( size, y );
		glEnd();
	}

	if( mHelpAlpha > 0 && mFadeInTimer < 1.f )
	{
		size = 512;
		mKeyHelp.bind();
		glColor4f(1,1,1,mHelpAlpha);

		glBegin( GL_QUADS );
		glTexCoord2f(0,0);
		glVertex2i(0,mWindowSize[1]);

		glTexCoord2f(1,0);
		glVertex2i(size,mWindowSize[1]);	

		glTexCoord2f(1,1);
		glVertex2i(size,mWindowSize[1]-size);

		glTexCoord2f(0,1);
		glVertex2i(0,mWindowSize[1]-size);	
		glEnd();
		mKeyHelp.unbind();

		size = 256;
		mMouseHelp.bind();
		glBegin( GL_QUADS );
		glTexCoord2f(0,0);
		glVertex2i(mWindowSize[0]-size,mWindowSize[1]);

		glTexCoord2f(1,0);
		glVertex2i(mWindowSize[0],mWindowSize[1]);	

		glTexCoord2f(1,1);
		glVertex2i(mWindowSize[0],mWindowSize[1]-size);

		glTexCoord2f(0,1);
		glVertex2i(mWindowSize[0]-size,mWindowSize[1]-size);	
		glEnd();
		mMouseHelp.unbind();
	}

	// draw the main title
	mTitle.bind();
	glColor4f(1,1,1,mTitleAlpha);

	glBegin( GL_QUADS );
	glTexCoord2f(0,0);
	glVertex2i(mWindowSize[0]/2-256,mWindowSize[1]/2+32);

	glTexCoord2f(1,0);
	glVertex2i(mWindowSize[0]/2+256,mWindowSize[1]/2+32);	

	glTexCoord2f(1,1);
	glVertex2i(mWindowSize[0]/2+256,mWindowSize[1]/2-32);

	glTexCoord2f(0,1);
	glVertex2i(mWindowSize[0]/2-256,mWindowSize[1]/2-32);	
	glEnd();
	mTitle.unbind();

	// draw "hold F1 for help" message
	glColor4f(1,1,1,0.7f * (1-mTitleAlpha) );

	mF1.bind();
	glBegin( GL_QUADS );
	glTexCoord2f(0,0);
	glVertex2i(mWindowSize[0]-256,32);

	glTexCoord2f(1,0);
	glVertex2i(mWindowSize[0],32);	

	glTexCoord2f(1,1);
	glVertex2i(mWindowSize[0],0);

	glTexCoord2f(0,1);
	glVertex2i(mWindowSize[0]-256,0);	
	glEnd();
	mF1.unbind();
	
    // clean up	
	glDisable( GL_BLEND );
	glEnable( GL_DEPTH_TEST );
	
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
	
}
