/*************************************************************************
MGLApp.h

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifndef MGLAPP_H_INCLUDED
#define MGLAPP_H_INCLUDED

#ifdef WIN32
  #define WIN32_LEAN_AND_MEAN
  #define WIN32_EXTRA_LEAN
  #include <windows.h>
#endif

#include "glee.h"
#include <gl/glu.h>
#include <vector>
#include <string>
#include <fstream>
#include <iostream>

#include "Model.h"
#include "Environment.h"
#include "Shader.h"
#include "Skybox.h"
#include "Camera.h"

using namespace ibl;

// values used to set up the window and rendering context
const char MGL_APP_NAME[]	= "MOGLGP - Chapter 6 - HDR Image Based Lighting";

const int MGL_WINDOW_WIDTH = 1000;
const int MGL_WINDOW_HEIGHT = 600;

const int MGL_COLOR_BITS = 32;
const int MGL_DEPTH_BITS = 24;
const int MGL_STENCIL_BITS = 0;

enum
{
  MGL_FULLSCREEN_FALSE = 0,
  MGL_FULLSCREEN_TRUE = 1,
  MGL_FULLSCREEN_CHOOSE = 2
};

const int MGL_FULLSCREEN_MODE = MGL_FULLSCREEN_CHOOSE;


class MGLApp
{
public:
	MGLApp();
	virtual ~MGLApp();

	bool Init();
	bool Shutdown();
	void SetupProjection( int width, int height );
	void Update( double dt );
	void Render();

	// these functions are used internally for input handling
	// adds dx and dy to mCamera's appropriate angles
	void RotateCamera( float dx, float dy );

	// adds dz to mCamera's zoom
	void ZoomCamera( float dz );

	// rotates the lighting environment around the y axis
	void RotateEnvironment( float dy );

	// cycle through the lighting environments
	void NextEnv();
	void PrevEnv();

	// multiplies the exposure multiplier by de the change the exposure exponentially
	void ChangeExposure( float de );

	// cycle through shaders
	void NextShader();
	void PrevShader();

	void ShowHelp();
	void HideHelp();
private:
	// a camera
	Camera mCamera;

	// four-component exposure color
	float mExposure[4];

	// background image
	Skybox mSkybox;

	// list of all the models drawn in the scene
	std::vector< Model > mModels;

	// list of environments to cycle through
	std::vector< Environment > mEnvironments;

	// index into mEnvironments defining the active environment
	size_t mCurrentEnv;

	// shader to sample the reflection map in the environment for the background
	Shader mSkyboxShader;

	// list of shaders for the model to cycle through
	std::vector< Shader > mModelShaders;

	// index to mModelShaders defining the active shader
	size_t mCurrentShader;

	// dimensions of the window in pixels
	unsigned mWindowSize[2];

	// transparency value for the shader menu
	float mMenuAlpha;
	// texture that has the shader menu on it
	Texture mMenu;

	// transparency value for the main title
	float mTitleAlpha;
	// title texture
	Texture mTitle;

	// texture for "press F1 for help" message
	Texture mF1;

	bool mShowHelp;
	// transparency value for the help menu
	float mHelpAlpha;
	// help textures
	Texture mKeyHelp;	
	Texture mMouseHelp;

	// timer for the fade in from black effect
	float mFadeInTimer;

	void DrawMenu();
};

#endif // MGLAPP_H_INCLUDED