#include "Shader.h"
#include <iostream>
#include <fstream>

namespace ibl
{
	Shader::Shader()
	{
		mType = GL_VERTEX_PROGRAM_ARB;
		mProgramID = 0;
		mProgramPath = "";
		mProgramString = "";
	}

	Shader::Shader( const Shader & src )
	{
		mType = src.mType;
		mProgramID = src.mProgramID;
		mProgramPath = src.mProgramPath;
		mProgramString = src.mProgramString;
	}

	Shader::~Shader()
	{
	}

	void Shader::load( const std::string & file )
	{
		std::cout<<"loading shader file "<<file<<" ... ";
		std::cout.flush();

		// temporary variables to store these until the load has been confirmed a success
		GLuint id;
		GLenum type;		
		std::string programString;

        // check for file extensions, if .vp it's a vertex program, .fp for fragment program, otherwise error out
		if( file.substr( file.size() - 3, 3 ) == ".vp" )		type = GL_VERTEX_PROGRAM_ARB;
		else if( file.substr( file.size() - 3, 3 ) == ".fp" )	type = GL_FRAGMENT_PROGRAM_ARB;
		else
		{
			std::cerr<<"shader error loading "<<file<<": file extension must be either .vp or .fp"<<std::endl;
			return;
		}

		if( type == GL_VERTEX_PROGRAM_ARB && !GLEE_ARB_vertex_program )
		{
			std::cerr<<"shader error loading "<<file<<": your hardware does not support ARB vertex programs (check for the extension GL_ARB_vertex_program)"<<std::endl;
		}

		if( type == GL_FRAGMENT_PROGRAM_ARB && !GLEE_ARB_fragment_program )
		{
			std::cerr<<"shader error loading "<<file<<": your hardware does not support ARB fragment programs (check for the extension GL_ARB_fragment_program)"<<std::endl;
		}

		// read the shader string from file
		std::ifstream in;
		in.open( file.c_str() );

		// something went wrong
		if( !in )
		{
			std::cerr<<"shader error loading "<<file<<": failed to open file"<<std::endl;
			return;
		}

		while( !in.eof() )
		{
			programString += in.get();
			in.peek();
		}
		in.close();

		// upload the program to the graphics context
		glPushAttrib( GL_ENABLE_BIT );
		glEnable( type );
		glGenProgramsARB( 1, &( id ) );
		glBindProgramARB( type, id );
		glProgramStringARB( type, GL_PROGRAM_FORMAT_ASCII_ARB, GLsizei( programString.size() )-1, programString.c_str() );
		glPopAttrib();

		// check to see if the programID generated is valid and everything is ok
		if( !glIsProgramARB( id ) )
		{
			// use openGL's error checking 
			GLenum error = glGetError();
			// if a shader compile error occurred, print it out
			if( error == GL_INVALID_OPERATION )
			{
				// get the compile error from the shader and print it out
				GLint errPos;
				glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &errPos );
				std::string errorString( (char *)glGetString( GL_PROGRAM_ERROR_STRING_ARB ) );
				std::cerr<<"shader error loading "<<file<<": "<<std::endl<<errorString<<std::endl;
			}

			// otherwise print a generic error
			else std::cerr<<"shader error loading "<<file<<": failed to load shader to video card"<<std::endl;
			return;
		}
		mProgramPath = file;
		mProgramString = programString;
		mProgramID = id;
		mType = type;
		
		std::cout<<"success"<<std::endl;
	}
	
    void Shader::bind()const
	{
		if( mProgramID )
		{
			glEnable( mType );
			glBindProgramARB( mType, mProgramID );
		}
	}

	void Shader::unbind()const
	{
		glDisable( mType );
	}
}
