#include "Texture.h"
#include "MGLTargaImage.h"
#include <sstream>
#include <iostream>
#include <math.h>
#include "Endianness.h"

namespace ibl
{
	Texture::Texture() : mTextureID( 0 ),
						 mTextureUnit( GL_TEXTURE0 )
	{
	}
	
	Texture::~Texture()
	{
	}
	
	void Texture::setTextureUnit( GLuint texUnit )
	{
		mTextureUnit = texUnit;
	}

	bool Texture::load( const std::string & file )
	{
		mTextureID = 0;
		std::string ext = file.substr( file.size() - 3, 3 );
		if( ext == "tga" || ext == "TGA" ) return loadLDR( file );
		if( ext == "pfm" || ext == "PFM" ) return loadPFM( file );
		return false;
	}

	bool Texture::loadPFM( const std::string & filename )
	{
		/* The PFM format. Source: http://netpbm.sourceforge.net/doc/pfm.html
		The PFM header is 3 consecutive "lines" of ASCII text. After each line is a white space 
		character. That character is typically a newline character, hence the term "line," but
		doesn't have to be.

		Identifier Line
		===============
		The identifier line contains the characters "PF" or "Pf". PF means it's a color PFM. Pf
		means it's a grayscale PFM.

		Dimensions Line
		===============
		The dimensions line contains two positive decimal integers, separated by a blank. The
		first is the width of the image; the second is the height. Both are in pixels.

		Scale Factor / Endianness
		=========================
		The Scale Factor / Endianness line is a queer line that jams endianness information into
		an otherwise sane description of a scale. The line consists of a nonzero decimal number,
		not necessarily an integer. If the number is negative, that means the PFM raster is
		little endian. Otherwise, it is big endian. The absolute value of the number is the
		scale factor for the image. The scale factor tells the units of the samples in the raster.
		You use it somehow along with some separately understood unit information to turn a
		sample value into something meaningful, such as watts per square meter. 
		*/

		FILE * file = fopen( filename.c_str(), "rb" );

		std::cout<<"loading texture file "<<filename<<" ... ";
		std::cout.flush();

		// this is all the information we need to figure out to create a texture
		unsigned int width;
		unsigned int height;
		unsigned int componentCount;
		float * data;
		GLenum format;
		GLint internalFormat;

		std::string line[4];
		char c;

		for( int i = 0; i<4; ++i )
		{
			//read up to whitespace
			do
			{
				fread( &c, sizeof(char), 1, file );
				line[i] += c;
			}while( c != '\n' && c!='\t' && c!=' ' && c!='\r' );

			//in between lines kill the whitespace
			if( i<3 ) while( c != '\n' && c!='\t' && c!=' ' && c!='\r' ) fread( &c, sizeof(char), 1, file );
		}

        if( line[0][0] != 'P' )
		{
			std::cerr<<"image not in PFM format"<<std::endl;
			return false;
		}

		// gotta love the PFM format, if the F is lower case it's 3 components, if it's lower case it's grayscale
		if( line[0][1] == 'F' )
		{
			componentCount = 3;
			format = GL_RGB;
			internalFormat = GL_RGB_FLOAT16_ATI;
		}
		else if( line[0][1] == 'f' )
		{
			componentCount = 1;
			format = GL_LUMINANCE;
			internalFormat = GL_LUMINANCE_FLOAT16_ATI;
		}
		else
		{
			std::cerr<<"image not in PFM format"<<std::endl;
			return false;
		}

        std::istringstream iss( line[1] );
		iss >> width;
		
		std::istringstream iss2( line[2] );
		iss2 >> height;
	
		std::istringstream iss3( line[3] );
		float scale;
		iss3 >> scale;

		unsigned int dataSize = componentCount * width * height;
		data = new float[ dataSize ];
		fread( data, sizeof(float), dataSize, file );
		fclose( file );
		
		// sign on the scale determines endianness. if endianness is not what we need it to be, flip the byte order
#ifdef IBL_BIG_ENDIAN
		if( scale < 0 )
#else
		if( scale > 0 )
#endif
		{
			for( size_t i = 0; i<dataSize; ++i )
			{
				flipEndian( data[i] );
			}
		}

		//apply scale
		scale = fabs( scale );
		for( size_t n = 0; n< dataSize; ++n )
		{
			data[ n ] *= scale;
		}

		// upload the texture to the GL context
		glPushAttrib( GL_ENABLE_BIT );
		glEnable( GL_TEXTURE_2D );
		glGenTextures( 1, &mTextureID );
		glBindTexture( GL_TEXTURE_2D, mTextureID );
	    glTexImage2D( GL_TEXTURE_2D, 0, internalFormat, width, height, 0, format, GL_FLOAT, data );			
				
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glPopAttrib();

		delete [] data;
		std::cout<<"success"<<std::endl;
		return true;
	}

	bool Texture::loadLDR( const std::string & file )
	{
		std::cout<<"loading texture file "<<file<<" ... ";
		std::cout.flush();

		MGL::TargaImage img;
		if( !img.Load( file.c_str() ) )
		{
			std::cerr<<"texture error loading "<<file<<std::endl;
			return false;
		}
		
		GLvoid * data = img.GetImage();
		GLuint width = img.GetWidth();
		GLuint height = img.GetHeight();
		GLenum datatype = img.GetType();		
		GLenum format;
		GLint internalFormat;

		format = img.GetFormat();
		internalFormat = format;
		
		glPushAttrib( GL_ENABLE_BIT );
		glEnable( GL_TEXTURE_2D );
		glGenTextures( 1, &mTextureID );
		glBindTexture( GL_TEXTURE_2D, mTextureID );
	    glTexImage2D( GL_TEXTURE_2D, 0, internalFormat, width, height, 0, format, datatype, data );			
				
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		
		glPopAttrib();

		std::cout<<"success"<<std::endl;
		return true;				
	}

	void Texture::bind()const
	{
		if( mTextureID )
		{
			glActiveTexture( mTextureUnit );
			glEnable( GL_TEXTURE_2D );
			glBindTexture( GL_TEXTURE_2D, mTextureID );		
			glActiveTexture( GL_TEXTURE0 );
		}
	}

	void Texture::unbind()const
	{
		glActiveTexture( mTextureUnit );
		glDisable( GL_TEXTURE_2D );
		glActiveTexture( GL_TEXTURE0 );
	}
}
