/*************************************************************************
MD2Model.cpp

Simple MD2 model loader.

Author   :   Josh Petrie

Written for More OpenGL Game Programming
*************************************************************************/

#include "Glee.h"

#include "MD2Model.h"

#include <cassert>
#include <cmath>


MGL::Vector3 InterpNormal(float t,const MGL::Vector3 &n0,const MGL::Vector3 &n1)
{
  // Compute the angle between n0 and n1 and the angle of interpolation.
float  angle        = n0.Angle(n1);
float  angle_interp = t * angle;

  // If they are parallel, return the initial vector.
  // Not that this is not handling the case where they are parallel
  // but opposite, in which case there are an infinite number of rotation
  // axes so one should just be chosen; we just assume that won't happen (not
  // suitable for production code, of course!)
  if(std::fabs(n0.x - n1.x) <= 0.01f &&
     std::fabs(n0.y - n1.y) <= 0.01f &&
     std::fabs(n0.z - n1.z) <= 0.01f)
    return (n0);

  // Compute the axis of rotation.
MGL::Vector3  axis = n0.CrossProduct(n1);

  axis.Normalize();
  
  // Rotate n0 about the axis vector.
MGL::Vector3  result = n0.Rotate(angle_interp,-axis);

  // Make sure the vector is normalized.
  result.Normalize();
  return (result);
}

MD2Model::MD2Model(const std::string &fn)
: mNumFrames(0),mNumTriangles(0),mNumTexcoords(0),mNumVertices(0),mFrameSize(0),
  mInterp(0),mThisFrame(0),mNextFrame(1),mFrameStart(0),mFrameEnd(39),
  mTriList(0),mAdjList(0),mTexList(0),mVtxList(0),mNrmList(0),mNrmFaceList(0),
  mInterpVtx(0),mInterpNrm(0),mInterpNrmFace(0)
{
  nLoadMD2Model(fn);
  
  // Compute attributes.
  nComputeAdjacency();
  
  mInterpVtx     = new MGL::Vector3[mNumVertices];
  mInterpNrm     = new MGL::Vector3[mNumVertices];
  mInterpNrmFace = new MGL::Vector3[mNumTriangles];
}

MD2Model::~MD2Model(void)
{
  // These were allocated by the model loader:
  delete[] mNrmFaceList;
  delete[] mAdjList;
  delete[] mTriList;
  delete[] mTexList;
  delete[] mNrmList;
  delete[] mVtxList; 
     
  // These were allocated in the constructor:
  delete[] mInterpNrmFace;
  delete[] mInterpNrm;
  delete[] mInterpVtx;
}


void MD2Model::Update(float percent)
{
  if(mInterp > 1.0f)
  {
    mInterp = 0.0f;
    mThisFrame++;

    // Wrap start frame.
    if(mThisFrame >= mFrameEnd)
      mThisFrame = mFrameStart;
    
    // Compute next frame and wrap if needed.
    mNextFrame = mThisFrame + 1;
    if(mNextFrame >= mFrameEnd)
      mNextFrame = mFrameStart;
  }
 
  // Update interpolation value.
  mInterp += percent;
  
MGL::Vector3 *vtxList         = &mVtxList[mNumVertices * mThisFrame];
MGL::Vector3 *vtxListNext     = &mVtxList[mNumVertices * mNextFrame];
MGL::Vector3 *nrmList         = &mNrmList[mNumVertices * mThisFrame];
MGL::Vector3 *nrmListNext     = &mNrmList[mNumVertices * mNextFrame];
MGL::Vector3 *nrmFaceList     = &mNrmFaceList[mNumTriangles * mThisFrame];
MGL::Vector3 *nrmFaceListNext = &mNrmFaceList[mNumTriangles * mNextFrame];

  // Compute interpolated vertices and vertex normals.
  for(int i = 0; i < mNumVertices; ++i)
  {
    mInterpVtx[i] = vtxList[i] + mInterp * (vtxListNext[i] - vtxList[i]);
    mInterpNrm[i] = InterpNormal(mInterp,nrmList[i],nrmListNext[i]);
  }
  
  // Compute interpolated face normals.
  for(int i = 0; i < mNumTriangles; ++i)
  { 
  MGL::Vector3  v0(mInterpVtx[mTriList[i].vtxIdx[1]] - mInterpVtx[mTriList[i].vtxIdx[0]]);
  MGL::Vector3  v1(mInterpVtx[mTriList[i].vtxIdx[2]] - mInterpVtx[mTriList[i].vtxIdx[1]]);
  MGL::Vector3  v2(mInterpVtx[mTriList[i].vtxIdx[0]] - mInterpVtx[mTriList[i].vtxIdx[2]]);
  MGL::Vector3  n = v1.CrossProduct(v0);

    // Normalize the normal and store it.
    n.Normalize();
    mInterpNrmFace[i] = n;
  }
}


void MD2Model::Render(void)
{
  glActiveTexture(GL_TEXTURE0);  
  glBegin(GL_TRIANGLES);
  for(int i = 0; i < mNumTriangles; i++)
  {
    glNormal3fv(mInterpNrm[mTriList[i].vtxIdx[0]].v);
    glMultiTexCoord2f(GL_TEXTURE0,mTexList[mTriList[i].texIdx[0]].s,mTexList[mTriList[i].texIdx[0]].t);
    glVertex3fv(mInterpVtx[mTriList[i].vtxIdx[0]].v);
    glNormal3fv(mInterpNrm[mTriList[i].vtxIdx[2]].v);
    glMultiTexCoord2f(GL_TEXTURE0,mTexList[mTriList[i].texIdx[2]].s,mTexList[mTriList[i].texIdx[2]].t);
    glVertex3fv(mInterpVtx[mTriList[i].vtxIdx[2]].v);
    glNormal3fv(mInterpNrm[mTriList[i].vtxIdx[1]].v);
    glMultiTexCoord2f(GL_TEXTURE0,mTexList[mTriList[i].texIdx[1]].s,mTexList[mTriList[i].texIdx[1]].t);
    glVertex3fv(mInterpVtx[mTriList[i].vtxIdx[1]].v);
  }
  glEnd();
}


void MD2Model::nLoadMD2Model(const std::string &fn)
{
FILE         *fp        = 0;
char         *buffer    = 0;
int           len       = 0;
MGL::Vector3 *vtxPtr    = 0;
Triangle     *triPtr    = 0;
TexCoordInt  *texPtr    = 0;
Header       *md2Header = 0;
Frame        *frame     = 0;

  // Open the file and get its length.
  fp = ::fopen(fn.c_str(),"rb");
  if(fp == 0)
    throw ("Failed to load model file.");
  ::fseek(fp,0,SEEK_END);
  len = ftell(fp);
  ::fseek(fp,0,SEEK_SET);

  // Read the entire file into a buffer.
  buffer = new char[len + 1];
  ::fread(buffer,sizeof(char),len,fp);
  ::fclose(fp);

  // Read the information we need from the model header.
  md2Header  = reinterpret_cast< Header* >(buffer);

  mNumVertices  = md2Header->numXYZ;
  mNumFrames    = md2Header->numFrames;
  mNumTexcoords = md2Header->numST;
  mNumTriangles = md2Header->numTris;
  mFrameSize    = md2Header->frameSize;

  // Allocate storage.
  mVtxList     = new MGL::Vector3[mNumVertices * mNumFrames];
  mNrmList     = new MGL::Vector3[mNumVertices * mNumFrames];
  mTexList     = new TexCoord[mNumTexcoords];
  mTriList     = new Triangle[mNumTriangles];
  mAdjList     = new Adjacency[mNumTriangles];
  mNrmFaceList = new MGL::Vector3[mNumTriangles * mNumFrames];
  
  // Build frames.
  for(int i = 0; i < mNumFrames; ++i)
  {
    frame = reinterpret_cast< Frame* >(&buffer[md2Header->offsetFrames + mFrameSize * i]);
    vtxPtr = reinterpret_cast< MGL::Vector3* >(&mVtxList[mNumVertices * i]);
    
    for(int j = 0; j < mNumVertices; ++j)
    {
      // When we apply the scale and translation to the coordinates, we also
      // bring them into the OpenGL coordinate frame by swapping Z and Y, and
      // negating the new Z (+Z is up in the MD2 coordinate frame).
      vtxPtr[j].x = frame->scale[1] * frame->fp[j].v[1] + frame->translate[1];
      vtxPtr[j].y = frame->scale[2] * frame->fp[j].v[2] + frame->translate[2];
      vtxPtr[j].z = frame->scale[0] * frame->fp[j].v[0] + frame->translate[0];
    }
  }

  // Build texture coordinates.
  texPtr = reinterpret_cast< TexCoordInt* >(&buffer[md2Header->offsetST]);
  for(int i = 0; i < mNumTexcoords; ++i)
  {
    mTexList[i].s = static_cast< float >(texPtr[i].s) / static_cast< float >(md2Header->skinWidth);
    mTexList[i].t = static_cast< float >(texPtr[i].t) / static_cast< float >(md2Header->skinHeight);
  }

  // Build triangle list.
  triPtr = reinterpret_cast< Triangle* >(&buffer[md2Header->offsetTris]);
  for(int i = 0; i < mNumFrames; ++i)         
  {
    for(int j = 0; j < mNumTriangles; ++j)
    {
      mTriList[j].vtxIdx[0] = triPtr[j].vtxIdx[0];
      mTriList[j].vtxIdx[1] = triPtr[j].vtxIdx[1];
      mTriList[j].vtxIdx[2] = triPtr[j].vtxIdx[2];
      mTriList[j].texIdx[0] = triPtr[j].texIdx[0];
      mTriList[j].texIdx[1] = triPtr[j].texIdx[1];
      mTriList[j].texIdx[2] = triPtr[j].texIdx[2];
    }
  }

  // Clean up.
  delete[] buffer;

  // Build face and vertex normals.
  nComputeNormals();
}


void MD2Model::nComputeNormals(void)
{
  // Zero the normals.
  std::fill_n(mNrmList,mNumVertices,MGL::Vector3());

  // For each frame...
  for(int frame = 0; frame < mNumFrames; ++frame)
  {
  MGL::Vector3 *vtxPtr     = &mVtxList[frame * mNumVertices];
  MGL::Vector3 *nrmPtr     = &mNrmList[frame * mNumVertices];
  MGL::Vector3 *nrmFacePtr = &mNrmFaceList[frame * mNumTriangles];
  
    // For each triangle...
    for(int i = 0; i < mNumTriangles; ++i)
    {
      // Find vectors between each vertex in a counter-clockwise fashion.
    MGL::Vector3  v0(vtxPtr[mTriList[i].vtxIdx[1]] - vtxPtr[mTriList[i].vtxIdx[0]]);
    MGL::Vector3  v1(vtxPtr[mTriList[i].vtxIdx[2]] - vtxPtr[mTriList[i].vtxIdx[1]]);
    MGL::Vector3  v2(vtxPtr[mTriList[i].vtxIdx[0]] - vtxPtr[mTriList[i].vtxIdx[2]]);
    
      // Crossing any two of those will yield the face normal.
    MGL::Vector3  n = v1.CrossProduct(v0);
      
      v0.Normalize();
      v1.Normalize();
      v2.Normalize();
      
      // Now add the face normal to the running average normal
      // at each vertex, weighted by the angle between them. 
    float  a0 = v0.Angle(-v2);
    float  a1 = v1.Angle(-v0);
    float  a2 = v2.Angle(-v1);
    
      // Normalize the normal and store it.
      n.Normalize();
      nrmFacePtr[i] = n;
    
      // Add the running summed normal.
      nrmPtr[mTriList[i].vtxIdx[0]] += a0 * n;
      nrmPtr[mTriList[i].vtxIdx[1]] += a1 * n;
      nrmPtr[mTriList[i].vtxIdx[2]] += a2 * n;
    }
  }
  
  // Then go through and make sure all the vectors are normalized.
  for(int i = 0; i < mNumVertices * mNumFrames; ++i)
    mNrmList[i].Normalize();
}


void MD2Model::nComputeAdjacency(void)
{  
  // For each triangle...
  for(int i = 0; i < mNumTriangles; ++i)
  {
    // -1 means this edge is not shared; this is the initial assumption about
    // all edges.
    mAdjList[i].neighbor[0] = mAdjList[i].neighbor[1] = mAdjList[i].neighbor[2] = -1;
    
    // For each edge...
    for(int j = 0; j < 3; ++j)
    {
      // The edge is from p0 to p1; on the neighboring triangle it will appear
      // as p1 to p0, so scan every other triangle for such an edge.
    unsigned short  p0 = mTriList[i].vtxIdx[j];
    unsigned short  p1 = mTriList[i].vtxIdx[(j + 1) % 3];
        
      for(int k = 0; k < mNumTriangles; ++k)
      {
        // For each edge...
        for(int l = 0; l < 3; ++l)
        {
          // If the edge is built from the same vertices but in opposite order,
          // we've found our neighbor triangle, so store it.
          if(p0 == mTriList[k].vtxIdx[(l + 1) % 3] && p1 == mTriList[k].vtxIdx[l])
            mAdjList[i].neighbor[j] = k;
        }
      }
    }
  }
}