/*************************************************************************
MD2Model.h

Simple MD2 model loader.

Author   :   Josh Petrie

Written for More OpenGL Game Programming
*************************************************************************/

#ifndef MD2MODEL_H
#define MD2MODEL_H


#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include <cstdio>
#include <string>

#include "MGLVector.h"


class MD2Model
{
  public:
    struct Triangle
    {
      unsigned short  vtxIdx[3];  //!< Vertex indices.
      unsigned short  texIdx[3];  //!< Texture coordinate indices.
    };
    
    struct Adjacency
    {
      int neighbor[3];  //!< Neighbor i is across edge i.
    };
    
    // ct and dt
    MD2Model(const std::string &fn);
    ~MD2Model(void);

    // accessors
    int                 GetTriangleCount(void) const        { return (mNumTriangles); }
    const Triangle&     GetTriangle(int i) const            { return (mTriList[i]); }
    const Adjacency&    GetTriangleAdj(int i) const         { return (mAdjList[i]); }
    int                 GetVertexCount(void) const          { return (mNumVertices); }
    const MGL::Vector3& GetVertex(int i) const              { return (mInterpVtx[i]); }
    const MGL::Vector3& GetNormal(int i) const              { return (mInterpNrm[i]); }
    const MGL::Vector3& GetFaceNormal(int i) const          { return (mInterpNrmFace[i]); }
    const MGL::Vector3& GetVertexOfFrame(int f,int i) const { return (mVtxList[f * mNumVertices + i]); }
    
    // update
    void Update(float percent);
    void Render(void);
        
  private:
    struct Header
    {
      int  ident;         //!< Should be 'IDP2' when viewed as a string.
      int  version;       //!< Should be 8.
      int  skinWidth;     //!< Texture width.
      int  skinHeight;    //!< Texture height.
      int  frameSize;     //!< Size of a single frame (in bytes).
      int  numSkins;      //!< Number of textures.
      int  numXYZ;        //!< Number of vertices (per frame).
      int  numST;         //!< Number of texture coordinates.
      int  numTris;       //!< Number of triangles.
      int  numGLCmds;     //!< Number of OpenGL commands.
      int  numFrames;     //!< Number of frames.
      int  offsetSkins;   //!< Offset to skin names (64 bytes each).
      int  offsetST;      //!< Offset of texture coordinate data.
      int  offsetTris;    //!< Offset of triangle data.
      int  offsetFrames;  //!< Offset of frame data.
      int  offsetGLCmds;  //!< Offset to OpenGL command data.
      int  offsetEnd;     //!< Offset to end of file.
    };
    
    struct TexCoord
    {
      float  s;  //!< S component.
      float  t;  //!< T component.
    };

    struct TexCoordInt
    {
      short  s;  //!< S component index.
      short  t;  //!< T component index.
    };

    struct FramePoint
    {
      unsigned char  v[3];  //!< Encoded vertex coordinates.
      unsigned char  n;     //!< Normal vector index (unused).
    };
    
    struct Frame
    {
      float       scale[3];      //!< This frame's scale.
      float       translate[3];  //!< This frame's translation.
      char        name[16];      //!< Frame name.
      FramePoint  fp[1];         //!< Frame point list base. 
    };

    // load
    void nLoadMD2Model(const std::string &fn);
    
    // attribute computation
    void nComputeNormals(void);
    void nComputeAdjacency(void);
    
    // data members
    int  mNumFrames;     // Total number of frames.
    int  mNumTriangles;  // Total number of triangles.
    int  mNumTexcoords;  // Total number of texture coordinates.
    int  mNumVertices;   // Total number of vertices and normals (per frame).
    int  mFrameSize;     // Size (in bytes) of a single frame.

    float  mInterp;      // Percentage between current and next frame.
    int    mThisFrame;   // Current key frame (interpolation source).
    int    mNextFrame;   // Next key frame (interpolation target).
    int    mFrameStart;  // Start frame ID for current animation.
    int    mFrameEnd;    // End frame ID for current animation.
    
    Triangle     *mTriList;      // Triangle list.
    Adjacency    *mAdjList;      // Adjacency list.
    TexCoord     *mTexList;      // Texture coordinate list.
    MGL::Vector3 *mVtxList;      // Vertex list (for all frames).
    MGL::Vector3 *mNrmList;      // Normal list (for all frames).
    MGL::Vector3 *mNrmFaceList;  // Face normal list (for all frames).
    
    MGL::Vector3 *mInterpVtx;      // List of interpolated positions for the current frame.
    MGL::Vector3 *mInterpNrm;      // List of interpolated normals for the current frame.
    MGL::Vector3 *mInterpNrmFace;  // List of interpolated face normals for the current frame.
};
 
#endif  /* MD2MODEL_H */