/*************************************************************************
Main application class.

Author   :   Lee Thomason

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include <stdlib.h>
#include <assert.h>
#include "MGLBillboards.h"
#include "SDL.h"

using namespace MGL;




MGLBillboards::MGLBillboards()
{
	mode = SHOW_BILLBOARD_TEXTURES;
	treeTexture = 0;
	symmetricBBTexture = 0;
	asymmetricBBTexture[0] = asymmetricBBTexture[1] = asymmetricBBTexture[2] = asymmetricBBTexture[3] = 0;
	transitionType = ALPHA1;
	billboardType = SYMMETRIC;
	animationType = RENDER_TRANSITION;
	view = LOW;
}

MGLBillboards::~MGLBillboards()
{
}

bool MGLBillboards::Init()
{	
	if (!GLeeInit())
		return false;

	if ( !(GLEE_SGIS_generate_mipmap) )
		return false;
	
	glEnable( GL_TEXTURE_2D );
    glEnable( GL_DEPTH_TEST );
    glDepthFunc( GL_LEQUAL );
	glEnable( GL_CULL_FACE );

	CreateTreeTexture();
	SetupLights();
	InitStipple();
	return true;
}

// Initializes the stipple pattern, and its pixel opposite "anti-stipple"
// pattern. This is a pretty crude stipple. It could be improved with more steps
// (this one uses 8 steps) and by using a diffusion patter rather than
// a regular pattern.
void MGLBillboards::InitStipple()
{
	const unsigned char stippleArray[16] = {
		0x01,	0x04,
		0x11,	0x44,
		0x15,	0x4a,
		0x55,	0xaa,
		0x57,	0xac,
		0x77,	0xcc,
		0x7f,	0xcf,
		0xff,	0xff
	};

	int i;
	for( i=0; i<8; ++i )
	{
		for( int j=0; j<32; ++j )
		{
			for ( int k=0; k<4; ++k )
			{
				stipplePattern[i][j*4+k] = stippleArray[i*2 + (j&0x01)];
				antiStipplePattern[i][j*4+k] = stipplePattern[i][j*4+k] ^ 0xff;
			}
		}
	}
}


/* Generates, programmatically, a random tree-ish texture. */
void MGLBillboards::CreateTreeTexture()
{
	glGenTextures( 1, &treeTexture );
	glBindTexture( GL_TEXTURE_2D, treeTexture );

	const int SIZE = 64;
	rgba_t textureMem[SIZE*SIZE] = { 0 };
	
	int i, j;

	for( j=0; j<SIZE; ++j ) {
		for( i=0; i<SIZE; ++i ) {
			textureMem[j*SIZE+i].r = rand()%20;
			textureMem[j*SIZE+i].g = 128 + rand()%127;
			textureMem[j*SIZE+i].b = rand()%128;
			textureMem[j*SIZE+i].a = 255;
		}
	}

	// For debugging, put in stripes.
	rgba_t red = { 255, 0, 0, 255 };
	rgba_t blue = { 0, 0, 255, 255 };
	rgba_t yellow = { 255, 255, 0, 255 };
	rgba_t pink = { 255, 0, 255, 255 };

//	for( j=0; j<SIZE; ++j )
//		textureMem[j*SIZE+j] = yellow;
//
	for( j=0; j<SIZE/2; ++j )		// South line red
		textureMem[j*SIZE + SIZE/2] = red;
	for( j=SIZE/2; j<SIZE; ++j )	// North line blue
		textureMem[j*SIZE + SIZE/2] = blue;
	for( i=0; i<SIZE/2; ++i )		// West line yellow
		textureMem[SIZE/2*SIZE + i] = yellow;
	for( i=SIZE/2; i<SIZE; ++i )	// East line pink
		textureMem[SIZE/2*SIZE + i] = pink;

	// Automatic MIP generation
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

	glTexImage2D(	GL_TEXTURE_2D,
					0,					// no mip mapping
					4,
					SIZE,
					SIZE,
					0,					// no border
					GL_RGBA,			// format
					GL_UNSIGNED_BYTE,
					textureMem );

	GLenum result = glGetError();
	assert ( result == GL_NO_ERROR );
};


void MGLBillboards::SetupLights()
{
	// Set up some lights. Very simple lighting.
	// That said, thanks to Steve Baker's great info pages for guidance on
	// even getting simple openGL lighting working:
	//		http://sjbaker.org/steve/omniv/opengl_lighting.html

	// Using the opengl lighting.
	glEnable( GL_LIGHTING );
	glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );

	float black[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
	float white[4] = { 1.0f, 1.0f, 1.0f, 1.0f };

	const float AMBIENT = 0.6f;
	const float DIFFUSE = 0.4f;

	float light0Color[4] = { DIFFUSE, DIFFUSE, DIFFUSE, 1.0f };
	float light0Vector[4] = { 0.0f, -0.50f, 0.87f, 0.0f };
	float ambient[4] = { AMBIENT, AMBIENT, AMBIENT, 1.0f };

	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, black );

	// Light 0.
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, light0Vector );
	glLightfv(GL_LIGHT0, GL_AMBIENT,  ambient );
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0Color );
	glLightfv(GL_LIGHT0, GL_SPECULAR, black );

	glEnable( GL_COLOR_MATERIAL );
	glColorMaterial( GL_FRONT, GL_DIFFUSE );

	glMaterialfv( GL_FRONT, GL_SPECULAR, black );
	glMaterialfv( GL_FRONT, GL_EMISSION, black );
	glMaterialfv( GL_FRONT, GL_AMBIENT,  white );
	glMaterialfv( GL_FRONT, GL_DIFFUSE,  white );
	
	glColor3f( 1.f, 1.f, 1.f );	
}


bool MGLBillboards::Shutdown()
{
	return true;
}


void MGLBillboards::SetMode( int _mode, int width, int height )
{
	mode = _mode;
	SetupProjection( width, height );
}

void MGLBillboards::SetupProjection(int width, int height)
{
	if ( mode == SHOW_BILLBOARD_TEXTURES )
		SetupProjectionBBTextures( width, height );
	else
		SetupProjectionAnimation( width, height );
}

void MGLBillboards::SetupProjectionBBTextures(int width, int height)
{
	// Set up the projection to be an orthographic unit cube. Since
	// there isn't any correct perspective projection for a billboard,
	// it is often easiest not to use it.
	// Use the default projection: the camera is looking down the
	// -z axis.
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho( 0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 1.0f );
}

void MGLBillboards::SetupProjectionAnimation(int width, int height)
{
	GLfloat ratio;
	ratio = ( float )width / ( float )height;
	glViewport( 0, 0, width, height );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 40.0f, ratio, 1.0f, MGL_BB_TREE_START + MGL_BB_TREE_RADIUS*2.0f );
}

void MGLBillboards::Update( double dt )
{
}


GLuint MGLBillboards::GetTextureFromColorBuffer( int x, int y )
{
	rgba_t colorBuffer[MGL_BB_TEXTURE_SIZE][MGL_BB_TEXTURE_SIZE];
	
	glReadPixels( x, y, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, GL_RGBA, GL_UNSIGNED_BYTE, colorBuffer );	

	GLuint id;	
	glGenTextures( 1, &id );
	glBindTexture( GL_TEXTURE_2D, id );

	if ( GLEE_SGIS_generate_mipmap)
	{
		// Automatic MIP generation
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
	}
	
	glTexImage2D(	GL_TEXTURE_2D,
					0,					// no mip mapping
					4,
					MGL_BB_TEXTURE_SIZE,
					MGL_BB_TEXTURE_SIZE,
					0,					// no border
					GL_RGBA,			// format
					GL_UNSIGNED_BYTE,
					colorBuffer );

	GLenum result = glGetError();
	assert ( result == GL_NO_ERROR );
	return id;
}


void MGLBillboards::Render()
{
	if ( mode == SHOW_BILLBOARD_TEXTURES ) {
		RenderBBTextures();

		if ( symmetricBBTexture == 0 ) {
			// If we don't have the billboard textures yet,
			// read the textures back from the framebuffer.

			symmetricBBTexture = GetTextureFromColorBuffer( 0, 0 );
			asymmetricBBTexture[0] = GetTextureFromColorBuffer( 0*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
			asymmetricBBTexture[1] = GetTextureFromColorBuffer( 1*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
			asymmetricBBTexture[2] = GetTextureFromColorBuffer( 2*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
			asymmetricBBTexture[3] = GetTextureFromColorBuffer( 3*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
		}
	}
	else
	{
		RenderAnimation();
	}

	GLenum result = glGetError();
	assert ( result == GL_NO_ERROR );
}

void MGLBillboards::RenderBBTextures()
{
	// Clear to a color that is approximately the average color for the tree.
	// Reduces color fringing when the texture is later rendered.
	glClearColor( 0.0f, 0.6f, 0.0f, 0.0f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	// Now set up the model matrix to transform the object to fill the unit cube.
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();

	glDisable( GL_LIGHTING );
	const float AMBIENT = 0.90f;	// Completely experimental.
	glColor3f( AMBIENT, AMBIENT, AMBIENT );
				
	/*	Draw 5 trees.
		One veiw of a symmetric tree, draw from the West, although being symmetric, this is arbitrary.
		4 veiws of an asymmetric tree.

			(North)

		(West) +---- X  (East)
			   |
			   |
			   Z
			(South)

		Note that the viewport is changed with every draw. This is to make sure we get
		perfect pixel snapping and that the textures don't effect one another.
	*/

	// The symmetric case.
	// Can look at the object from any axis (it is symmetric) but
	// this code chooses to look at an object positioned on the +X axis.
	glViewport( 0, 0, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (MGL_BB_TREE_RADIUS*2.0f), 1.0f / MGL_BB_TREE_HEIGHT, 1.0f / (MGL_BB_TREE_RADIUS*2.0f) );
	glTranslatef( MGL_BB_TREE_RADIUS, 0, -MGL_BB_TREE_RADIUS );
	glRotatef( 90, 0, 1, 0 );
	Vector3 origin( 0, 0, 0 );
	RenderTree( origin );	

	// The asymmetric case.
	// Object on +z axis.
	glViewport( 0*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (MGL_BB_TREE_RADIUS*2.0f), 1.0f / MGL_BB_TREE_HEIGHT, 1.0f / (MGL_BB_TREE_RADIUS*2.0f) );
	glTranslatef( MGL_BB_TREE_RADIUS, 0, -MGL_BB_TREE_RADIUS );
	glRotatef( 0, 0, 1, 0 );
	RenderTreeAsymmetric( origin );	

	// Object on +x axis.
	glViewport( 1*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (MGL_BB_TREE_RADIUS*2.0f), 1.0f / MGL_BB_TREE_HEIGHT, 1.0f / (MGL_BB_TREE_RADIUS*2.0f) );
	glTranslatef( MGL_BB_TREE_RADIUS, 0, -MGL_BB_TREE_RADIUS );
	glRotatef( 90, 0, 1, 0 );
	RenderTreeAsymmetric( origin );	

	// Object on -z axis.
	glViewport( 2*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (MGL_BB_TREE_RADIUS*2.0f), 1.0f / MGL_BB_TREE_HEIGHT, 1.0f / (MGL_BB_TREE_RADIUS*2.0f) );
	glTranslatef( MGL_BB_TREE_RADIUS, 0, -MGL_BB_TREE_RADIUS );
	glRotatef( 180, 0, 1, 0 );
	RenderTreeAsymmetric( origin );	

	// Object on -x axis.
	glViewport( 3*MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE, MGL_BB_TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (MGL_BB_TREE_RADIUS*2.0f), 1.0f / MGL_BB_TREE_HEIGHT, 1.0f / (MGL_BB_TREE_RADIUS*2.0f) );
	glTranslatef( MGL_BB_TREE_RADIUS, 0, -MGL_BB_TREE_RADIUS );
	glRotatef( 270, 0, 1, 0 );
	RenderTreeAsymmetric( origin );	
		
	// Unravel all the wackiness from above.
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	glMatrixMode( GL_MODELVIEW );
	glPopMatrix();
}


void MGLBillboards::RenderAnimation()
{
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	// Now set up the model matrix to transform the object to fill the unit cube.
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	Vector3 camera( 0, MGL_BB_EYE_HEIGHT, 0 );
	float range = 30.0f;
	
	switch ( view )
	{
		case HIGH:	camera.y = MGL_BB_EYE_HEIGHT + MGL_BB_TREE_HEIGHT*4.0f; range = 40.0f; break;
		case LEFT:	camera.z = -20.0f;	range = 40.0f;		break;
		case RIGHT: camera.z =  20.0f;	range = 40.0f;		break;
	}

	gluLookAt(	camera.x, camera.y, camera.z,		// eye
				range, MGL_BB_EYE_HEIGHT, 0,		// center
				0.0, 1.0, 0.0 );					// up vector
	
	Uint32 ticks = SDL_GetTicks();
	ticks = ticks % MGL_BB_TREE_PATH_TIME_MSEC;

	// 'distance' is from the camera to the tree
	float distance = MGL_BB_TREE_START - ( MGL_BB_TREE_START - MGL_BB_TREE_END ) * (float)ticks / MGL_BB_TREE_PATH_TIME_MSEC;
	assert( distance <= MGL_BB_TREE_START );
	assert( distance >= MGL_BB_TREE_END );

	// true if inside the model/billboard transition
	bool inTransition = ( distance > MGL_BB_NEAR_TRANSITION && distance < MGL_BB_FAR_TRANSITION );
	if ( animationType != RENDER_TRANSITION ) {
		inTransition = false;
	}

	float amountBillboard = ( distance - MGL_BB_NEAR_TRANSITION ) / ( MGL_BB_FAR_TRANSITION - MGL_BB_NEAR_TRANSITION );
	float amountModel = 1.0f - amountBillboard;

	// ALPHA1 is a standard crossover transition. ALHPA2 is a transition where either
	// the model or the billboard is opaque and the other is faded.
	if ( transitionType == ALPHA2 ) {
		amountBillboard *= 2.0f;
		amountModel     *= 2.0f;
	}

	if ( amountBillboard > 1.0f ) amountBillboard = 1.0f;
	if ( amountBillboard < 0.0f ) amountBillboard = 0.0f;
	if ( amountModel > 1.0f ) amountModel = 1.0f;
	if ( amountModel < 0.0f ) amountModel = 0.0f;

	glDisable( GL_ALPHA_TEST );

	if (    ( ( animationType == RENDER_TRANSITION ) && ( distance > MGL_BB_NEAR_TRANSITION ) )
	     || animationType == RENDER_BILLBOARD )
	{
		// The billboard needs to be drawn.
		glDisable( GL_LIGHTING );
		glDisable( GL_BLEND );
		glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );

		if ( inTransition ) {
			if ( transitionType == STIPPLE ) {
				StippleTransitionOn( amountBillboard, false );
			}
			else {
				// To get polygon sorting correct in the general case is
				// difficult. In this simple case, it works to check
				// the depth buffer but don't write it, since the billboard
				// is always rendered before the model.
				glDepthMask( GL_FALSE );
				AlphaTransitionOn( amountBillboard );
			}
		}
				
		Vector3 location( distance, 0, 0 );

		if ( billboardType == SYMMETRIC ) {
			RenderBillboard( location, camera );
		}
		else {
			RenderBillboard2( location, camera );
		}

		if ( inTransition ) {
			glDepthMask( GL_TRUE );
			TransitionOff();
		}
	}

	// The model needs to be drawn.
	if (    ( ( animationType == RENDER_TRANSITION ) && ( distance <= MGL_BB_FAR_TRANSITION ) )
	     || animationType == RENDER_MODEL )
	{
		glEnable( GL_LIGHTING );
		glDisable( GL_BLEND );
		glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );

		if ( inTransition ) {
			if ( transitionType == STIPPLE )
				StippleTransitionOn( amountBillboard, true );
			else
				AlphaTransitionOn( amountModel );
		}

		Vector3 location( distance, 0, 0 );
		if ( billboardType == SYMMETRIC ) {
			RenderTree( location );
		}
		else {
			RenderTreeAsymmetric( location );
		}

		if ( inTransition ) {
			TransitionOff();
		}
	}
}


void MGLBillboards::RenderCone( int sides, float yMin, float yMax, float radius, const MGL::Vector3& v )
{
	// Draw an approximate cone, but use quads - instead of the obvious tris -
	// to avoid lighting problems.

	glBegin( GL_QUADS );

	double delta = 2.0f * PI / (double)(sides);
	float INNER = 0.2f;

	// From the top:
	//
	// +---- x   2 1
	// |         3 0
	// |
	// z

	for( int i=0; i<sides; ++i ) {
		int next = (i+1)%sides;
		float sinA = (float)sin( delta*i );
		float cosA = (float)cos( delta*i );

		float sinB = (float)sin( delta*next );
		float cosB = (float)cos( delta*next );

		Vector3 x0( v.x + radius*cosA,	v.y + yMin,	v.z + -radius*sinA );
		Vector3 u0(	cosA*0.5f + 0.5f, -sinA*0.5f + 0.5f );
		Vector3 n0(	cosA,	-sinA );

		Vector3 x1(	v.x + radius*cosB,	v.y + yMin,	v.z + -radius*sinB );
		Vector3 u1( cosB*0.5f + 0.5f,		-sinB*0.5f + 0.5f );
		Vector3 n1(	cosB, -sinB );

		Vector3 x2(	v.x + INNER*radius*cosB,	v.y + yMax,	v.z + -INNER*radius*sinB );
		Vector3 u2(	INNER*cosB*0.5f + 0.5f,		-INNER*sinB*0.5f + 0.5f );
		Vector3 n2(	cosB, -sinB );

		Vector3 x3(	v.x + INNER*radius*cosA,	v.y + yMax,	v.z + -INNER*radius*sinA );
		Vector3 u3(	INNER*cosA*0.5f + 0.5f,		-INNER*sinA*0.5f + 0.5f );
		Vector3 n3(	cosA, -sinA );
		
		glNormal3fv( n0.v ); 
		glTexCoord2fv( u0.v );
		glVertex3fv( x0.v );

		glNormal3fv( n1.v ); 
		glTexCoord2fv( u1.v );
		glVertex3fv( x1.v );

		glNormal3fv( n2.v ); 
		glTexCoord2fv( u2.v );
		glVertex3fv( x2.v );

		glNormal3fv( n3.v ); 
		glTexCoord2fv( u3.v );
		glVertex3fv( x3.v );
	}
	glEnd();
}

void MGLBillboards::RenderTree( const MGL::Vector3& origin ) 
{
	glBindTexture( GL_TEXTURE_2D, treeTexture );

	RenderCone( 12, MGL_BB_TREE_HEIGHT*0.0f, MGL_BB_TREE_HEIGHT*0.6f, MGL_BB_TREE_RADIUS, origin );
	RenderCone( 12, MGL_BB_TREE_HEIGHT*0.4f, MGL_BB_TREE_HEIGHT*0.8f, MGL_BB_TREE_RADIUS*0.7f, origin );
	RenderCone( 12, MGL_BB_TREE_HEIGHT*0.7f, MGL_BB_TREE_HEIGHT*1.0f, MGL_BB_TREE_RADIUS*0.4f, origin );
}


void MGLBillboards::RenderTreeAsymmetric( const Vector3& _origin )
{
	Vector3 origin = _origin;

	glBindTexture( GL_TEXTURE_2D, treeTexture );

	RenderCone( 12, MGL_BB_TREE_HEIGHT*0.0f, MGL_BB_TREE_HEIGHT*0.4f, MGL_BB_TREE_RADIUS, origin );

	origin = _origin;
	origin.x += -0.1f * MGL_BB_TREE_RADIUS;
	origin.z +=  0.1f * MGL_BB_TREE_RADIUS;
	RenderCone( 12, MGL_BB_TREE_HEIGHT*0.2f, MGL_BB_TREE_HEIGHT*0.5f, MGL_BB_TREE_RADIUS*0.8f, origin );

	origin = _origin;
	origin.x += -0.08f * MGL_BB_TREE_RADIUS;
	origin.z += -0.12f * MGL_BB_TREE_RADIUS;
	RenderCone( 12, MGL_BB_TREE_HEIGHT*0.4f, MGL_BB_TREE_HEIGHT*0.8f, MGL_BB_TREE_RADIUS*0.6f, origin );

	origin = _origin;
	origin.x += -0.14f * MGL_BB_TREE_RADIUS;
	origin.z +=  0.07f * MGL_BB_TREE_RADIUS;
	RenderCone( 12, MGL_BB_TREE_HEIGHT*0.6f, MGL_BB_TREE_HEIGHT*1.0f, MGL_BB_TREE_RADIUS*0.5f, origin );
}

void MGLBillboards::StippleTransitionOn( float amount, bool useAntiStipple )
{
	int stipple = (int)(amount*8.0);

	if ( useAntiStipple )
		glPolygonStipple( &antiStipplePattern[stipple][0] );
	else
		glPolygonStipple( &stipplePattern[stipple][0] );
	glEnable( GL_POLYGON_STIPPLE );
}


void MGLBillboards::AlphaTransitionOn( float amount )
{
	glEnable( GL_BLEND );
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
	glColor4f( 1.f, 1.f, 1.f, amount );
}

void MGLBillboards::TransitionOff()
{
	glDisable( GL_BLEND );
	glColor4f( 1.f, 1.f, 1.f, 1.f );
	glDisable( GL_POLYGON_STIPPLE );
}


void MGLBillboards::RenderBillboard( const Vector3& location, const Vector3& camera )
{
	glBindTexture( GL_TEXTURE_2D, symmetricBBTexture );

	// The texture goes 'to the edge', so it is important to use clamp
	// mode to prevent the top from bleeding to the bottom and vice versa.
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

	// Remember to alpha test so you don't draw transparent pixels!
	// An alpha function of about 0.4 seems to give the smoothest animation
	// far away. Caveat however: a blend transition applies alpha, which
	// is then rejected by the alpha test. Therefore use a lower value to
	// compromise.
	glAlphaFunc ( GL_GREATER, 0.2f ) ;
	glEnable( GL_ALPHA_TEST );

	// Compute the 'right' vector. If the 'up' and 'n' are normal, and at right
	// angles, then 'right' will be a normal as well. This is easy to accomplish
	// by setting the y value of n to 0.
	Vector3 up( 0.0f, 1.0f, 0.0f );
	Vector3 n( camera.x - location.x, 0.0, camera.z - location.z );
	n.Normalize();

	Vector3 right = n.CrossProduct( up );

	glBegin( GL_QUADS );

	glTexCoord2f( 0.0f, 0.0f );
	glVertex3f( location.x + right.x*MGL_BB_TREE_RADIUS, 0.0f, location.z + right.z*MGL_BB_TREE_RADIUS );

	glTexCoord2f( 1.0f, 0.0f );
	glVertex3f( location.x - right.x*MGL_BB_TREE_RADIUS, 0.0f, location.z - right.z*MGL_BB_TREE_RADIUS );

	glTexCoord2f( 1.0f, 1.0f );
	glVertex3f( location.x - right.x*MGL_BB_TREE_RADIUS, MGL_BB_TREE_HEIGHT, location.z - right.z*MGL_BB_TREE_RADIUS );

	glTexCoord2f( 0.0f, 1.0f );
	glVertex3f( location.x + right.x*MGL_BB_TREE_RADIUS, MGL_BB_TREE_HEIGHT, location.z + right.z*MGL_BB_TREE_RADIUS );

	glEnd();
}


void MGLBillboards::RenderBillboard2( const Vector3& location, const Vector3& camera )
{
	// Remember to alpha test so you don't draw transparent pixels!
	// An alpha function of about 0.4 seems to give the smoothest animation
	// far away. Caveat however: a blend transition applies alpha, which
	// is then rejected by the alpha test. Therefore use a lower value to
	// compromise.
	//
	glAlphaFunc ( GL_GREATER, 0.2f ) ;
	glEnable( GL_ALPHA_TEST );
	
	Vector3 q(	location.x - camera.x,
				location.y - camera.y,
				location.z - camera.z );

	const Vector3 NORTH( 0, 0, -1 );
	const Vector3 EAST( 1, 0, 0 );

	if ( NORTH.DotProduct( q ) > 0.0f ) {
		// Use the South texture (2)
		glBindTexture( GL_TEXTURE_2D, asymmetricBBTexture[2] );

		// The texture goes 'to the edge', so it is important to use clamp
		// mode to prevent the top from bleeding to the bottom and vice versa.
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x - MGL_BB_TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x + MGL_BB_TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x + MGL_BB_TREE_RADIUS, MGL_BB_TREE_HEIGHT, location.z );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x - MGL_BB_TREE_RADIUS, MGL_BB_TREE_HEIGHT, location.z );

		glEnd();
	}
	else {
		// Use the North texture (0)
		glBindTexture( GL_TEXTURE_2D, asymmetricBBTexture[0] );

		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x + MGL_BB_TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x - MGL_BB_TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x - MGL_BB_TREE_RADIUS, MGL_BB_TREE_HEIGHT, location.z );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x + MGL_BB_TREE_RADIUS, MGL_BB_TREE_HEIGHT, location.z );

		glEnd();
	}

	if ( EAST.DotProduct( q ) > 0.0f ) {
		// Use the West texture (1)
		glBindTexture( GL_TEXTURE_2D, asymmetricBBTexture[1] );

		// The texture goes 'to the edge', so it is important to use clamp
		// mode to prevent the top from bleeding to the bottom and vice versa.
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z - MGL_BB_TREE_RADIUS );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z + MGL_BB_TREE_RADIUS );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x, MGL_BB_TREE_HEIGHT, location.z + MGL_BB_TREE_RADIUS );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x, MGL_BB_TREE_HEIGHT, location.z - MGL_BB_TREE_RADIUS );

		glEnd();
	}
	else {
		// Use the East texture (3)
		glBindTexture( GL_TEXTURE_2D, asymmetricBBTexture[3] );

		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z + MGL_BB_TREE_RADIUS );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z - MGL_BB_TREE_RADIUS );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x, MGL_BB_TREE_HEIGHT, location.z - MGL_BB_TREE_RADIUS );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x, MGL_BB_TREE_HEIGHT, location.z + MGL_BB_TREE_RADIUS );

		glEnd();
	}


}
