/*************************************************************************
Main application class.

Author   :   Lee Thomason

Written for More OpenGL Game Programming
*************************************************************************/

#ifndef MGLBillboards_H_INCLUDED
#define MGLBillboards_H_INCLUDED

#ifdef WIN32
  #define WIN32_LEAN_AND_MEAN
  #define WIN32_EXTRA_LEAN
  #include <windows.h>
#endif

#include "glee.h"
#include <gl/glu.h>

#include "MGLTypes.h"
#include "MGLVector.h"

// values used to set up the window and rendering context
const char MGL_APP_NAME[]	= "MOGLGP - Chapter 6 - Billboarding";

const int MGL_WINDOW_WIDTH = 800;
const int MGL_WINDOW_HEIGHT = 600;

const int MGL_COLOR_BITS = 32;
const int MGL_DEPTH_BITS = 32;
const int MGL_STENCIL_BITS = 0;

const bool MGL_FORCE_FULLSCREEN = false;


// ------------ control constants ---------- //
const unsigned	MGL_BB_TEXTURE_SIZE = 64;	// Size of the billboard textures (pixels)
const float		MGL_BB_EYE_HEIGHT = 1.8f;	// Height, off the ground, of the eye/camera (meters)

const float		MGL_BB_TREE_HEIGHT = 4.0f;						// The height of the tree (meters)
const float		MGL_BB_TREE_RADIUS = MGL_BB_TREE_HEIGHT * 0.4f;	// Tree radius (meters)

const float		MGL_BB_TREE_START = 200.0f;						// The beginning of tree motion, in meters from the camera
const float		MGL_BB_TREE_END = MGL_BB_TREE_START * 0.1f;		// The end of tree motion, in meters from the camera

const unsigned	MGL_BB_TREE_PATH_TIME_MSEC = 5000;

const float		MGL_BB_NEAR_TRANSITION = MGL_BB_TREE_START * 0.4f;	// When to start the billboard-model transition
const float		MGL_BB_FAR_TRANSITION  = MGL_BB_TREE_START * 0.8f;	// Where to end the transition


class MGLBillboards
{
public:
	MGLBillboards();
	virtual ~MGLBillboards();

	bool Init();
	bool Shutdown();
	void SetupProjection( int width, int height );
	void Update( double dt );
	void Render();

	/*	The point of this demo is billboards, not actually trees, so trees are
		approximated by one or more cones.
	*/
	void RenderCone( int sides, float yMin, float yMax, float radius, const MGL::Vector3& location );
	void RenderTree( const MGL::Vector3& location );
	void RenderTreeAsymmetric( const MGL::Vector3& location );

	/* Control functions. */
	// The demo starts in the SHOW_BILLBOARD_TEXTURES mode, and
	// then switches to SHOW_ANIMATION with any key press.
	enum { SHOW_BILLBOARD_TEXTURES, SHOW_ANIMATION };
	void SetMode( int _mode, int width, int height );
	int  GetMode()				{ return mode; }

	enum { ALPHA1, ALPHA2, STIPPLE };
	void SetTransition( int t )	{ transitionType = t; }

	enum { SYMMETRIC, ASYMMETRIC };
	void SetBBType( int t )		{ billboardType = t; }

	enum { RENDER_BILLBOARD, RENDER_MODEL, RENDER_TRANSITION };
	void SetAnimation( int t )	{ animationType = t; }

	enum { LEFT, RIGHT, LOW, HIGH };
	void SetView( int v )		{ view = v; }

private:

	// Programmatically create a tree texture.
	void CreateTreeTexture();
	// Configures the lighting
	void SetupLights();

	// Sets up a orthographic projection matrix used to create the billboards
	void SetupProjectionBBTextures( int width, int height );
	// Sets up a typical projection matrix for veiwing the model
	void SetupProjectionAnimation( int width, int height );

	void RenderBBTextures();	
	void RenderAnimation();

	// Renders a single billboard of a symmetric object
	void RenderBillboard( const MGL::Vector3& location, const MGL::Vector3& camera );
	// Renders 2 billboards, at 90 degrees, of an asymmetric object
	void RenderBillboard2( const MGL::Vector3& location, const MGL::Vector3& camera );

	void StippleTransitionOn( float amount, bool useAntiStipple );
	void AlphaTransitionOn( float amount );
	void TransitionOff();

	GLuint GetTextureFromColorBuffer( int x, int y );
	void InitStipple();


	int mode;
	int transitionType;		// alpha1,2 or stipple
	int billboardType;		// symmetric or asymmetric
	int animationType;		// normally transition
	int view;

	unsigned char stipplePattern[8][32*4];
	unsigned char antiStipplePattern[8][32*4];

	GLuint treeTexture;
	GLuint symmetricBBTexture;
	GLuint asymmetricBBTexture[4];
};

#endif // MGLBillboards_H_INCLUDED