#include "assert.h"

#include <math.h>

#include <windows.h>
#include "GLee.h"
#include "SDL.h"
#include "GL/GLU.h"

bool hackTest = false;

/* screen width, height, and bit depth */
const int SCREEN_WIDTH = 800;
const int SCREEN_HEIGHT = 600;
const int SCREEN_BPP = 32;

const float TREE_HEIGHT = 4.0f;	// meters
const float TREE_RADIUS = TREE_HEIGHT * 0.4f;

const float EYE_HEIGHT = 1.8f;		// meters
const float TREE_START = 200.0f;	// meters, max is the far plane: 400m
const float TREE_END = TREE_START * 0.1f;

const float TREE_METERS_PER_SECOND = 40.0f;
const unsigned TREE_PATH_TIME_MSEC = (unsigned)(1000 * (TREE_START-TREE_END) / TREE_METERS_PER_SECOND );

const float NEAR_TRANSITION = TREE_START * 0.4f;
const float FAR_TRANSITION  = TREE_START * 0.8f;

const int TEXTURE_SIZE = 64;



unsigned char stipplePattern[8][32*4];
unsigned char antiStipplePattern[8][32*4];
const double PI = 3.1415926535897932384626433832795;


struct Vector
{
	float x, y, z;
};

void CrossProduct( const Vector& u, const Vector& v, Vector* w ) 
{
	assert( &u != w && &v != w );

	w->x = u.y * v.z - u.z * v.y;
	w->y = u.z * v.x - u.x * v.z;
	w->z = u.x * v.y - u.y * v.x;
}

float DotProduct( const Vector& u, const Vector& v ) 
{
	return u.x*v.x + u.y*v.y + u.z*v.z;
}

void Normalize( Vector* v ) 
{
	float len = sqrtf( v->x*v->x + v->y*v->y + v->z*v->z );
	float lenInv = 1.0f / len;
	v->x *= lenInv;
	v->y *= lenInv;
	v->z *= lenInv;
}

struct Vector2
{
	float x, y;
};

struct RGBA
{
	unsigned char r, g, b, a;
};

struct RGB
{
	unsigned char r, g, b, a;
};

unsigned treeTextureId = 0;
unsigned billboardTextureId = 0;
unsigned asymmetricBillboardTextureId[4] = { 0, 0, 0, 0 };


void DrawCone( int sides, float yMin, float yMax, float radius, const Vector& v )
{
	// Draw an approximate cone, but use quads - instead of the obvious tris -
	// to avoid lighting problems.

	glBegin( GL_QUADS );

	double delta = 2.0f * PI / (double)(sides);
	float INNER = 0.2f;

	// From the top:
	//
	// +---- x   2 1
	// |         3 0
	// |
	// z

	for( int i=0; i<sides; ++i ) {
		int next = (i+1)%sides;
		float sinA = (float)sin( delta*i );
		float cosA = (float)cos( delta*i );

		float sinB = (float)sin( delta*next );
		float cosB = (float)cos( delta*next );

		Vector  x0 = {	v.x + radius*cosA,	v.y + yMin,	v.z + -radius*sinA };
		Vector2 u0 = {	cosA*0.5f + 0.5f,		-sinA*0.5f + 0.5f };
		Vector  n0 = {	cosA,	-sinA };

		Vector  x1 = {	v.x + radius*cosB,	v.y + yMin,	v.z + -radius*sinB };
		Vector2 u1 = {	cosB*0.5f + 0.5f,		-sinB*0.5f + 0.5f };
		Vector  n1 = {	cosB, -sinB };

		Vector  x2 = {	v.x + INNER*radius*cosB,	v.y + yMax,	v.z + -INNER*radius*sinB };
		Vector2 u2 = {	INNER*cosB*0.5f + 0.5f,		-INNER*sinB*0.5f + 0.5f };
		Vector  n2 = {	cosB, -sinB };

		Vector  x3 = {	v.x + INNER*radius*cosA,	v.y + yMax,	v.z + -INNER*radius*sinA };
		Vector2 u3 = {	INNER*cosA*0.5f + 0.5f,		-INNER*sinA*0.5f + 0.5f };
		Vector  n3 = {	cosA, -sinA };
		
		glNormal3fv( &n0.x ); 
		glTexCoord2fv( &u0.x );
		glVertex3fv( &x0.x );

		glNormal3fv( &n1.x ); 
		glTexCoord2fv( &u1.x );
		glVertex3fv( &x1.x );

		glNormal3fv( &n2.x ); 
		glTexCoord2fv( &u2.x );
		glVertex3fv( &x2.x );

		glNormal3fv( &n3.x ); 
		glTexCoord2fv( &u3.x );
		glVertex3fv( &x3.x );
	}
	glEnd();
}

void DrawTree( const Vector& origin ) 
{
	glBindTexture( GL_TEXTURE_2D, treeTextureId );

	DrawCone( 12, TREE_HEIGHT*0.0f, TREE_HEIGHT*0.6f, TREE_RADIUS, origin );
	DrawCone( 12, TREE_HEIGHT*0.4f, TREE_HEIGHT*0.8f, TREE_RADIUS*0.7f, origin );
	DrawCone( 12, TREE_HEIGHT*0.7f, TREE_HEIGHT*1.0f, TREE_RADIUS*0.4f, origin );
}

void DrawTreeAsymmetric( const Vector& _origin )
{
	Vector origin = _origin;

	glBindTexture( GL_TEXTURE_2D, treeTextureId );

	DrawCone( 12, TREE_HEIGHT*0.0f, TREE_HEIGHT*0.4f, TREE_RADIUS, origin );

	origin = _origin;
	origin.x += -0.1f * TREE_RADIUS;
	origin.z +=  0.1f * TREE_RADIUS;
	DrawCone( 12, TREE_HEIGHT*0.2f, TREE_HEIGHT*0.5f, TREE_RADIUS*0.8f, origin );

	origin = _origin;
	origin.x += -0.08f * TREE_RADIUS;
	origin.z += -0.12f * TREE_RADIUS;
	DrawCone( 12, TREE_HEIGHT*0.4f, TREE_HEIGHT*0.8f, TREE_RADIUS*0.6f, origin );

	origin = _origin;
	origin.x += -0.14f * TREE_RADIUS;
	origin.z +=  0.07f * TREE_RADIUS;
	DrawCone( 12, TREE_HEIGHT*0.6f, TREE_HEIGHT*1.0f, TREE_RADIUS*0.5f, origin );
}


void DrawBillboard( float height, float radius, const Vector& location, const Vector& camera )
{
	glBindTexture( GL_TEXTURE_2D, billboardTextureId );

	// The texture goes 'to the edge', so it is important to use clamp
	// mode to prevent the top from bleeding to the bottom and vice versa.
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

	// Remember to alpha test so you don't draw transparent pixels!
	// An alpha function of about 0.4 seems to give the smoothest animation
	// far away. Caveat however: a blend transition applies alpha, which
	// is then rejected by the alpha test. Therefore use a lower value to
	// compromise.
	glAlphaFunc ( GL_GREATER, 0.2f ) ;
	glEnable( GL_ALPHA_TEST );

	// Compute the 'right' vector. If the 'up' and 'n' are normal, and at right
	// angles, then 'right' will be a normal as well. This is easy to accomplish
	// by setting the y value of n to 0.
	Vector up = { 0.0f, 1.0f, 0.0f };
	Vector n = { camera.x - location.x, 0.0, camera.z - location.z };
	Normalize( &n );

	Vector right;
	CrossProduct( up, n, &right );

	glBegin( GL_QUADS );

	glTexCoord2f( 0.0f, 0.0f );
	glVertex3f( location.x - right.x*radius, 0.0f, location.z - right.z*radius );

	glTexCoord2f( 1.0f, 0.0f );
	glVertex3f( location.x + right.x*radius, 0.0f, location.z + right.z*radius );

	glTexCoord2f( 1.0f, 1.0f );
	glVertex3f( location.x + right.x*radius, height, location.z + right.z*radius );

	glTexCoord2f( 0.0f, 1.0f );
	glVertex3f( location.x - right.x*radius, height, location.z - right.z*radius );

	glEnd();
}


void DrawBillboard2( const Vector& location, const Vector& camera )
{
	// Remember to alpha test so you don't draw transparent pixels!
	// An alpha function of about 0.4 seems to give the smoothest animation
	// far away. Caveat however: a blend transition applies alpha, which
	// is then rejected by the alpha test. Therefore use a lower value to
	// compromise.
	glAlphaFunc ( GL_GREATER, 0.2f ) ;
	glEnable( GL_ALPHA_TEST );
	
	Vector q = {	location.x - camera.x,
					location.y - camera.y,
					location.z - camera.z };

	const Vector NORTH = { 0, 0, -1 };
	const Vector EAST  = { 1, 0, 0 };

	if ( DotProduct( NORTH, q ) > 0.0f ) {
		// Use the South texture (2)
		glBindTexture( GL_TEXTURE_2D, asymmetricBillboardTextureId[2] );

		// The texture goes 'to the edge', so it is important to use clamp
		// mode to prevent the top from bleeding to the bottom and vice versa.
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x - TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x + TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x + TREE_RADIUS, TREE_HEIGHT, location.z );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x - TREE_RADIUS, TREE_HEIGHT, location.z );

		glEnd();
	}
	else {
		// Use the North texture (0)
		glBindTexture( GL_TEXTURE_2D, asymmetricBillboardTextureId[0] );

		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x + TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x - TREE_RADIUS, 0.0f, location.z );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x - TREE_RADIUS, TREE_HEIGHT, location.z );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x + TREE_RADIUS, TREE_HEIGHT, location.z );

		glEnd();
	}

	if ( DotProduct( EAST, q ) > 0.0f ) {
		// Use the West texture (1)
		glBindTexture( GL_TEXTURE_2D, asymmetricBillboardTextureId[1] );

		// The texture goes 'to the edge', so it is important to use clamp
		// mode to prevent the top from bleeding to the bottom and vice versa.
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z - TREE_RADIUS );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z + TREE_RADIUS );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x, TREE_HEIGHT, location.z + TREE_RADIUS );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x, TREE_HEIGHT, location.z - TREE_RADIUS );

		glEnd();
	}
	else {
		// Use the East texture (3)
		glBindTexture( GL_TEXTURE_2D, asymmetricBillboardTextureId[3] );

		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

		glBegin( GL_QUADS );

		glTexCoord2f( 0.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z + TREE_RADIUS );

		glTexCoord2f( 1.0f, 0.0f );
		glVertex3f( location.x, 0.0f, location.z - TREE_RADIUS );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f( location.x, TREE_HEIGHT, location.z - TREE_RADIUS );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( location.x, TREE_HEIGHT, location.z + TREE_RADIUS );

		glEnd();
	}


}


unsigned CreateTreeTexture()
{
	unsigned id;	
	glGenTextures( 1, &id );
	glBindTexture( GL_TEXTURE_2D, id );

	const int SIZE = 64;
	RGBA textureMem[SIZE*SIZE] = { 0 };
	
	int i, j;

	for( j=0; j<SIZE; ++j ) {
		for( i=0; i<SIZE; ++i ) {
			textureMem[j*SIZE+i].r = rand()%20;
			textureMem[j*SIZE+i].g = 128 + rand()%127;
			textureMem[j*SIZE+i].b = rand()%128;
			textureMem[j*SIZE+i].a = 255;
		}
	}

	// For debugging, put in stripes.
	RGBA red = { 255, 0, 0, 255 };
	RGBA blue = { 0, 0, 255, 255 };
	RGBA yellow = { 255, 255, 0, 255 };
	RGBA pink = { 255, 0, 255, 255 };

//	for( j=0; j<SIZE; ++j )
//		textureMem[j*SIZE+j] = yellow;
//
	for( j=0; j<SIZE/2; ++j )		// South line red
		textureMem[j*SIZE + SIZE/2] = red;
	for( j=SIZE/2; j<SIZE; ++j )	// North line blue
		textureMem[j*SIZE + SIZE/2] = blue;
	for( i=0; i<SIZE/2; ++i )		// West line yellow
		textureMem[SIZE/2*SIZE + i] = yellow;
	for( i=SIZE/2; i<SIZE; ++i )	// East line pink
		textureMem[SIZE/2*SIZE + i] = pink;

	if (GLEE_SGIS_generate_mipmap)
	{
		// Automatic MIP generation
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
	}
	else
	{
		assert( 0 );	// It will look funny without mip maps
	}

	glTexImage2D(	GL_TEXTURE_2D,
					0,					// no mip mapping
					4,
					SIZE,
					SIZE,
					0,					// no border
					GL_RGBA,			// format
					GL_UNSIGNED_BYTE,
					textureMem );

	GLenum result = glGetError();
	assert ( result == GL_NO_ERROR );

	return id;
};


void DrawTreeBillboardToColorBuffer()
{
	glDisable( GL_LIGHTING );
	// Get new projection and modelview matrices to work with. Note
	// that the projection matrix stack may only be 2 deep, so this
	// is the matrix stack limit.
	int viewPort[4];
	glGetIntegerv(GL_VIEWPORT, viewPort);

	// Set up the projection to be an orthographic unit cube. Since
	// there isn't any correct perspective projection for a billboard,
	// it is often easiest not to use it.
	// Use the default projection: the camera is looking down the
	// -z axis.
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	glOrtho( 0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 1.0f );

	// Now set up the model matrix to transform the object to fill the unit cube.
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();

	// Clear to a color that is approximately the average color for the tree.
	// Reduces color fringing when the texture is later rendered.
    glClearColor( 0.0f, 0.6f, 0.0f, 0.0f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	const float AMBIENT = 0.90f;	// Completely experimental.

	glColor3f( AMBIENT, AMBIENT, AMBIENT );

	// The symmetric case.
	// Can look at the object from any axis (it is symmetric) but
	// this code chooses to look at an object positioned on the +X axis.
	glViewport( 0, 0, TEXTURE_SIZE, TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (TREE_RADIUS*2.0f), 1.0f / TREE_HEIGHT, 1.0f / (TREE_RADIUS*2.0f) );
	glTranslatef( TREE_RADIUS, 0, -TREE_RADIUS );
	glRotatef( 90, 0, 1, 0 );
	Vector origin = { 0, 0, 0 };
	DrawTree( origin );	

	// The asymmetric case.
	// Object on +z axis.
	glViewport( 0*TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (TREE_RADIUS*2.0f), 1.0f / TREE_HEIGHT, 1.0f / (TREE_RADIUS*2.0f) );
	glTranslatef( TREE_RADIUS, 0, -TREE_RADIUS );
	glRotatef( 0, 0, 1, 0 );
//	DrawTree( origin );	
	DrawTreeAsymmetric( origin );	

	// Object on +x axis.
	glViewport( 1*TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (TREE_RADIUS*2.0f), 1.0f / TREE_HEIGHT, 1.0f / (TREE_RADIUS*2.0f) );
	glTranslatef( TREE_RADIUS, 0, -TREE_RADIUS );
	glRotatef( 90, 0, 1, 0 );
//	DrawTree( origin );	
	DrawTreeAsymmetric( origin );	

	// Object on -z axis.
	glViewport( 2*TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (TREE_RADIUS*2.0f), 1.0f / TREE_HEIGHT, 1.0f / (TREE_RADIUS*2.0f) );
	glTranslatef( TREE_RADIUS, 0, -TREE_RADIUS );
	glRotatef( 180, 0, 1, 0 );
//	DrawTree( origin );	
	DrawTreeAsymmetric( origin );	

	// Object on -x axis.
	glViewport( 3*TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE );
	glLoadIdentity();
	glScalef( 1.0f / (TREE_RADIUS*2.0f), 1.0f / TREE_HEIGHT, 1.0f / (TREE_RADIUS*2.0f) );
	glTranslatef( TREE_RADIUS, 0, -TREE_RADIUS );
	glRotatef( 270, 0, 1, 0 );
//	DrawTree( origin );	
	DrawTreeAsymmetric( origin );	

	// Unravel all the wackiness from above.
	glViewport( viewPort[0], viewPort[1], viewPort[2], viewPort[3] );
    glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
	glPopMatrix();
//	glDisable( GL_NORMALIZE );
	glEnable( GL_LIGHTING );

	assert( glGetError() == GL_NO_ERROR );
}


unsigned GetTextureFromColorBuffer( int x, int y )
{
	RGBA colorBuffer[TEXTURE_SIZE][TEXTURE_SIZE];
	
	glReadPixels( x, y, TEXTURE_SIZE, TEXTURE_SIZE, GL_RGBA, GL_UNSIGNED_BYTE, colorBuffer );	

	unsigned id;	
	glGenTextures( 1, &id );
	glBindTexture( GL_TEXTURE_2D, id );

	if ( GLEE_SGIS_generate_mipmap)
	{
		// Automatic MIP generation
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
	}
	
	glTexImage2D(	GL_TEXTURE_2D,
					0,					// no mip mapping
					4,
					TEXTURE_SIZE,
					TEXTURE_SIZE,
					0,					// no border
					GL_RGBA,			// format
					GL_UNSIGNED_BYTE,
					colorBuffer );

	GLenum result = glGetError();
	assert ( result == GL_NO_ERROR );
	return id;
}


void InitializeStipple()
{
	const unsigned char stippleArray[16] = {
		0x01,	0x04,
		0x11,	0x44,
		0x15,	0x4a,
		0x55,	0xaa,
		0x57,	0xac,
		0x77,	0xcc,
		0x7f,	0xcf,
		0xff,	0xff
	};

	int i;
	for( i=0; i<8; ++i )
	{
		for( int j=0; j<32; ++j )
		{
			for ( int k=0; k<4; ++k )
			{
				stipplePattern[i][j*4+k] = stippleArray[i*2 + (j&0x01)];
				antiStipplePattern[i][j*4+k] = stipplePattern[i][j*4+k] ^ 0xff;
			}
		}
	}
}

void StippleTransitionOn( float amount, bool useAntiStipple )
{
	int stipple = (int)(amount*8.0);

	if ( useAntiStipple )
		glPolygonStipple( &antiStipplePattern[stipple][0] );
	else
		glPolygonStipple( &stipplePattern[stipple][0] );
	glEnable( GL_POLYGON_STIPPLE );
}


void AlphaTransitionOn( float amount )
{
	// Disable the depth test?? Cheap sleazy trick that won't 
	// generally work in a real situation. But this is a demo
	// on billboards, not the complexities of alpha sorting.
//	glDisable( GL_DEPTH_TEST );
	glEnable( GL_BLEND );
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//	glDepthMask( GL_FALSE );

	glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
	glColor4f( 1.f, 1.f, 1.f, amount );
}

void TransitionOff()
{
//	glDepthMask( GL_TRUE );
//    glEnable( GL_DEPTH_TEST );
	glDisable( GL_BLEND );
	glColor4f( 1.f, 1.f, 1.f, 1.f );
	glDisable( GL_POLYGON_STIPPLE );
}


int main( int argc, char **argv )
{
    bool done = false;
    SDL_Event event;
    const SDL_VideoInfo *videoInfo;
	SDL_Surface *surface = 0;

	InitializeStipple();

	// SDL initialization steps.
    if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE | SDL_INIT_TIMER ) < 0 )
	{
	    fprintf( stderr, "Video initialization failed: %s\n", SDL_GetError( ) );
		exit( 1 );
	}

    videoInfo = SDL_GetVideoInfo();

    if ( !videoInfo )
	{
	    fprintf( stderr, "Video query failed: %s\n",
		     SDL_GetError( ) );
	    exit( 1 );
	}

    int	videoFlags  = SDL_OPENGL;         
		videoFlags |= SDL_GL_DOUBLEBUFFER;

	SDL_EnableKeyRepeat( SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL );
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
    surface = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP, videoFlags );

    if ( !surface )
	{
	    fprintf( stderr,  "Video mode set failed: %s\n", SDL_GetError( ) );
	    exit( 1 );
	}

	glShadeModel( GL_SMOOTH );
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
    glClearDepth( 1.0f );
    glEnable( GL_DEPTH_TEST );
    glDepthFunc( GL_LEQUAL );
	glEnable( GL_CULL_FACE );
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	// ------- Set up the perspective -----------
	GLfloat ratio;
	ratio = ( float )SCREEN_WIDTH / ( float )SCREEN_HEIGHT;
	glViewport( 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT );

	glMatrixMode( GL_PROJECTION );
	glLoadIdentity( );
	gluPerspective( 40.0f, ratio, 1.0f, 400.0f );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity( );

	// Set up some lights. Very simple lighting.
	// That said, thanks to Steve Baker's great info pages:
	//		http://sjbaker.org/steve/omniv/opengl_lighting.html
	// Errors mine.

	// Using the opengl lighting.
	glEnable( GL_LIGHTING );
	glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );

	float black[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
	float white[4] = { 1.0f, 1.0f, 1.0f, 1.0f };

	const float AMBIENT = 0.6f;
	const float DIFFUSE = 0.4f;

	float light0Color[4] = { DIFFUSE, DIFFUSE, DIFFUSE, 1.0f };
	//float light0Vector[4] = { 0.0f, -.87f, .50f, 0.0f };
	float light0Vector[4] = { 0.0f, -0.50f, 0.87f, 0.0f };
	float ambient[4] = { AMBIENT, AMBIENT, AMBIENT, 1.0f };

	glLightModelfv( GL_LIGHT_MODEL_AMBIENT, black );

	// Light 0.
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, light0Vector );
	glLightfv(GL_LIGHT0, GL_AMBIENT,  ambient );
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0Color );
	glLightfv(GL_LIGHT0, GL_SPECULAR, black );

	glEnable( GL_COLOR_MATERIAL );
	glColorMaterial( GL_FRONT, GL_DIFFUSE );

	glMaterialfv( GL_FRONT, GL_SPECULAR, black );
	glMaterialfv( GL_FRONT, GL_EMISSION, black );
	glMaterialfv( GL_FRONT, GL_AMBIENT,  white );
	glMaterialfv( GL_FRONT, GL_DIFFUSE,  white );
	
	glColor3f( 1.f, 1.f, 1.f );

	gluLookAt(	0.0, EYE_HEIGHT, 0.0,	// eye
				20.0, EYE_HEIGHT, 0,	// center
				0.0, 1.0, 0.0 );		// up vector

	treeTextureId = CreateTreeTexture();
	glEnable( GL_TEXTURE_2D );

	DrawTreeBillboardToColorBuffer();
	billboardTextureId = GetTextureFromColorBuffer(0,0);

	asymmetricBillboardTextureId[0] = GetTextureFromColorBuffer(0*TEXTURE_SIZE, TEXTURE_SIZE );
	asymmetricBillboardTextureId[1] = GetTextureFromColorBuffer(1*TEXTURE_SIZE, TEXTURE_SIZE );
	asymmetricBillboardTextureId[2] = GetTextureFromColorBuffer(2*TEXTURE_SIZE, TEXTURE_SIZE );
	asymmetricBillboardTextureId[3] = GetTextureFromColorBuffer(3*TEXTURE_SIZE, TEXTURE_SIZE );

	bool drawBillboard = false;

	enum { ALL_MODEL, ALL_BILLBOARD, TRANSITION };
	int mode = TRANSITION;
	enum { ALPHA1, ALPHA2, STIPPLE };
	int transitionType = ALPHA2;
	bool simpleBillboard = true;

	Vector camera = { 0.0, EYE_HEIGHT, 0.0 };

	while ( !done )
	{
		while ( SDL_PollEvent( &event ) )
		{
			switch( event.type )
			{
				case SDL_QUIT:
					done = true;
					break;

				case SDL_KEYDOWN:
				{	
					switch ( event.key.keysym.sym )
					{
						// Symmetric tree:	
						// Irregular tree:
						//
						// draw billboard image:	i
						// animate:					n (normal)
						//
						// 1 Billboard:		r (radial symmetric)
						// 2 Billboards:	a (asymmetric)
						//
						// Model:			m
						// Billboard:		b
						// Transition:		t
						//
						// Alpha crossover  1
						// Alpha			2
						// Stipple			s

						case SDLK_r:	simpleBillboard = true;		break;
						case SDLK_a:	simpleBillboard = false;	break;

						case SDLK_1:	transitionType = ALPHA1;	break;
						case SDLK_2:	transitionType = ALPHA2;	break;
						case SDLK_s:	transitionType = STIPPLE;	break;
							
						case SDLK_m:	mode = ALL_MODEL;			break;
						case SDLK_b:	mode = ALL_BILLBOARD;		break;
						case SDLK_t:	mode = TRANSITION;			break;

						case SDLK_UP:
						case SDLK_DOWN:
						case SDLK_RIGHT:
						case SDLK_LEFT:
						{
							glMatrixMode( GL_MODELVIEW );
							glLoadIdentity();

							camera.x = 0.0f;
							camera.y = EYE_HEIGHT;
							camera.z = 0.0f;
							float range = 20.0f;

							if (  event.key.keysym.sym == SDLK_UP ) {
								camera.y = EYE_HEIGHT + TREE_HEIGHT*4.0f;
								range = 40.0f;
							}
							else if ( event.key.keysym.sym == SDLK_RIGHT ) {
								camera.z = 20.0f;
								range = 40.0f;
							}
							else if ( event.key.keysym.sym == SDLK_LEFT ) {
								camera.z = -20.0f;
								range = 40.0f;
							}

							gluLookAt(	camera.x, camera.y, camera.z,			// eye
										range, EYE_HEIGHT, 0,	// center
										0.0, 1.0, 0.0 );		// up vector
						}
						break;

						case SDLK_7:
							hackTest = !hackTest;
							break;

						case SDLK_i:
							drawBillboard = true;
							break;

						case SDLK_n:
							drawBillboard = false;
							break;

						case SDLK_ESCAPE:
							done = true;
							break;
					}
				}
				break;

				default:
					break;
			}
		}

		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		Uint32 ticks = SDL_GetTicks();

		if ( !drawBillboard )
		{
			ticks = ticks % TREE_PATH_TIME_MSEC;
			float distance = TREE_START - (float)ticks * 0.001f * TREE_METERS_PER_SECOND;
			bool inTransition = ( distance > NEAR_TRANSITION && distance < FAR_TRANSITION );

			float amountBillboard = ( distance - NEAR_TRANSITION ) / ( FAR_TRANSITION - NEAR_TRANSITION );
			float amountModel = 1.0f - amountBillboard;

			if ( transitionType == ALPHA2 ) {
				amountBillboard *= 2.0f;
				amountModel     *= 2.0f;
			}

			if ( amountBillboard > 1.0f ) amountBillboard = 1.0f;
			if ( amountBillboard < 0.0f ) amountBillboard = 0.0f;
			if ( amountModel > 1.0f ) amountModel = 1.0f;
			if ( amountModel < 0.0f ) amountModel = 0.0f;

			if (    mode == ALL_BILLBOARD 
			     || ( mode == TRANSITION && distance >= NEAR_TRANSITION ) )
			{
				glDisable( GL_LIGHTING );
				if ( inTransition ) {
					if ( transitionType == STIPPLE ) {
						StippleTransitionOn( amountBillboard, false );
					}
					else {
						glDepthMask( GL_FALSE );
						AlphaTransitionOn( amountBillboard );
					}
				}
				
				Vector location = { distance, 0, 0 };
				//DrawBillboard( TREE_HEIGHT, TREE_RADIUS, location, camera );
				DrawBillboard2( location, camera );

				if ( inTransition ) {
					glDepthMask( GL_TRUE );
					TransitionOff();
				}
				glEnable( GL_LIGHTING );
			}
			if (    mode == ALL_MODEL 
			          || ( mode == TRANSITION && distance <= FAR_TRANSITION ) )
			{
				if ( inTransition ) {
					if ( transitionType == STIPPLE )
						StippleTransitionOn( amountBillboard, true );
					else
						AlphaTransitionOn( amountModel );
				}

				Vector location = { distance, 0, 0 };
				//DrawTree( location );
				DrawTreeAsymmetric( location );

//				location.x += TREE_RADIUS*2;
//				location.z = TREE_RADIUS;
//				DrawTree( location );
//
//				location.x += TREE_RADIUS*0.5f;
//				location.z = -TREE_RADIUS*1.5f;
//				DrawTree( location );

				//DrawTreeAsymmetric();
//				glPopMatrix();

				if ( inTransition )
					TransitionOff();
			}
		}
		else 
		{
			DrawTreeBillboardToColorBuffer();
		}		
		assert( glGetError() == GL_NO_ERROR );
	    SDL_GL_SwapBuffers();
		assert( glGetError() == GL_NO_ERROR );

		// Cap frame rate to 33fps.
		while ( SDL_GetTicks() - ticks < 30 )	{};
	}

    SDL_Quit( );

    return( 0 );
}
