/*************************************************************************
MGLApp.cpp

Demo for "Planar Reflections" section

Author   :   Andrei Gradinari
Based on Dave Astle's application template

Written for More OpenGL Game Programming
August, 2005
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include <stdio.h>

#include "MGLApp.h"
using namespace MGL;

MGLApp::MGLApp()
{
	m_dElapTime = 0;
	for(int i=0; i<3; i++) m_pMesh[i] = NULL; 
}

MGLApp::~MGLApp()
{
	for(int i=0; i<3; i++) if(m_pMesh[i]) delete m_pMesh[i];
}

bool MGLApp::Init()
{	
	if(! LoadData()) // LOAD MODELS, SHADERS AND TEXTURES
	{
		MessageBox(NULL, "Error loading data. Exiting", "Error", MB_OK);
		return false;
	}

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glEnable(GL_LIGHTING);
    return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 0.5f, 100.0f );

	glMatrixMode(GL_MODELVIEW);
}

void MGLApp::Update( double dt )
{
	m_dElapTime += dt;
}

void MGLApp :: SetupLighting()
{
	float pos[4] = {-1, 1, 1, 0};
	float ambt[4] = {0.2f, 0.2f, 0.2f, 1.0f};
	float ambt1[4] = {0.5f, 0.5f, 0.5f, 1.0f};
	glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_POSITION, pos);
	glLightfv(GL_LIGHT0, GL_AMBIENT, ambt);
	glMaterialfv(GL_FRONT, GL_AMBIENT, ambt1);

	float diff[4] = {0.5f, 0, 0, 1};
	pos[0] = 1; pos[1] = -1;
	glEnable(GL_LIGHT1);
	glLightfv(GL_LIGHT1, GL_POSITION, pos);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diff);
}

void MGLApp :: RenderGeometry()
{
	float	diff1[] = {0.7f, 0.7f, 0.7f, 1.0f},
			diff2[] = {1.0f, 1.0f, 1.0f, 1.0f};
	for(int i=1; i<3; i++) 
	{
		if(i==1)glMaterialfv(GL_FRONT, GL_DIFFUSE, diff1);
		else	glMaterialfv(GL_FRONT, GL_DIFFUSE, diff2);
			
		m_pMesh[i]->Draw();
	}
}

void MGLApp::Render()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glPushMatrix();
//	float t = 7;//m_dElapTime;
//	float r = 17 + cos(t*0.1)*2;*/

	float t = m_dElapTime;
	float r = 14 + cos(t*0.1)*5;

	cam.LookAt(Vector3(r * -cos(t*0.2)+1, r * sin(t*0.2), 10), Vector3(3, 0, 0), Vector3(0,0,1));
	
	// CAM_REFL WILL REPRESENT OUR REFLECTED CAMERA
	CCamera cam_refl = cam;
		Vector3 a, b, c, n;
		GLdouble eq[4];
		// ANY MIRROR'S FACE LIES IN MIRROR PLANE,
		// THUS ANY FACE CAN DEFINE THIS PLANE
		if(m_pMesh[0]->nVertices > 2 && m_pMesh[0]->nTriangles > 1)
		{
			int ai = m_pMesh[0]->pIdx[0], 
				bi = m_pMesh[0]->pIdx[1], 
				ci = m_pMesh[0]->pIdx[2];
			
			a = Vector3(& m_pMesh[0]->pVert[ai*3]);
			b = Vector3(& m_pMesh[0]->pVert[bi*3]);
			c = Vector3(& m_pMesh[0]->pVert[ci*3]);

			get_plane_equation(c, b, a, eq);
			n = Vector3(eq[0], eq[1], eq[2]);
		}
		cam_refl.Refract(a, n, 2);
	
	// RENDERING SCENE
	cam.Apply(); // apply camera transformation
	
	// SETUP CLIP PLANE, BUT LEAVE IT TO BE DISABLED
	glClipPlane(GL_CLIP_PLANE0, eq);
	
	// RENDERING MIRROR MESH TO STENCIL BUFFER
	glDepthMask(GL_FALSE); // disable writuing to depth buffer
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE); // ... to color buffer
	
	glEnable(GL_STENCIL_TEST);
	glStencilMask(1); // enable writing to stencil buffer 1st bit plane
	glStencilFunc(GL_ALWAYS, 1, 1); // tell stencil test always to pass
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE); // tell opengl to replace stencil buffer value each time stencil test passes
		m_pMesh[0]->Draw(); // render mirror mesh
	// enable writing to color and depth buffers
	glDepthMask(GL_TRUE);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDisable(GL_STENCIL_TEST);
	
	// RENDERING REFLECTED GEOMETRY
	glPushMatrix();
	glPushAttrib(GL_POLYGON_BIT); // GL_FRONT_FACE
		glLoadIdentity();
		cam_refl.Apply();
		//glFrontFace(GL_CW); // in reflected world front face changes
			
		// CLIP PLANE SETUP
		glEnable(GL_CLIP_PLANE0);
		
		// STENCIL TEST SETUP
		glEnable(GL_STENCIL_TEST);
		glStencilFunc(GL_EQUAL, 1, ~0);
			// RENDERING GEOMETRY IN MIRROR
			SetupLighting(); // don't forget to call this each time before rendering
			RenderGeometry();	
		
		glDisable(GL_STENCIL_TEST);
		glDisable(GL_CLIP_PLANE0);
	
	glPopAttrib();
	glPopMatrix();
	
	// RENDER MIRROR MESH TO DEPTH BUFFER
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	m_pMesh[0]->Draw();
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	// RENDER GEOMETRY
	SetupLighting(); // don't forget to call this each time before rendering
	RenderGeometry();
	
	glPopMatrix();
}


bool MGLApp::LoadData()
{
// LOADING MODELS	
	m_pMesh[0] = CPrimitiveFactory::LoadPrimitive("Data/0.txt");
	m_pMesh[1] = CPrimitiveFactory::LoadPrimitive("Data/1.txt");
	m_pMesh[2] = CPrimitiveFactory::LoadPrimitive("Data/2.txt");

	for(int i=0; i<3; i++) if (m_pMesh[i] == NULL) return false;
	
	return true;
}
