/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#include "common.h"
#include "MGLTargaImage.h"

#include "Application.h"
#include "Mesh.h"

#ifdef WIN32
	#include <GL\glaux.h>
	#pragma comment(lib, "glaux")
#endif

cApplication::cApplication(int width, int height, int colorBits, int depthBits,
						   int stencilBits, bool fullscreen, const string& name) :
	Quit(false),
	Width(width),
	Height(height)
{
	int i;

	// Startup banner
	for (i = 2 * 4 + (int) name.length() - 1; i >= 0; i--)
		cout << '-'; cout << endl;
	cout << "--- " << name << " ---\n";
	for (i = 2 * 4 + (int) name.length() - 1; i >= 0; i--)
		cout << '-'; cout << endl;
	cout << "Initialization...\n";

	// Initialize SDL
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		string errorMessage = "Couldn't init SDL video, reason:";
		errorMessage += SDL_GetError();
		throw cException( errorMessage );
	}
	cout << "\tSDL video initialized..." << endl;

	// Grab and hide the mouse cursor so we always get relative motion
	// Note that this only works on Linux and Windows at the moment.
	SDL_WM_GrabInput(SDL_GRAB_ON);
	SDL_ShowCursor(0);

	// Create OpenGL window with 8 stencil bits
	unsigned int videoFlags = SDL_OPENGL;
	if (fullscreen)
		videoFlags |= SDL_FULLSCREEN;

	SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE, colorBits);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, depthBits);
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, stencilBits);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	if (SDL_SetVideoMode(width, height, 0, videoFlags) == 0)
	{
		string errorMessage = "Couldn't set requested video mode, reason:";
		errorMessage += SDL_GetError();
		throw cException( errorMessage );
	}

	cout << "\t\tVideo mode " << width << "x" << height << "x" << colorBits << " was succesfully set..." << endl;
	cout << "\t\tOpenGL version: " << glGetString(GL_VERSION) << endl;
	cout << "\t\tRenderer: " << glGetString(GL_RENDERER) << endl;
	cout << "\t\tVendor: " << glGetString(GL_VENDOR) << endl;

	SDL_WM_SetCaption(name.data(), name.data());

	// Set up timer
	Timer.Ticks = 0;
	Timer.FramesCounter = 0;
	Timer.FPS = 10;
	Timer.AveragedFrameTime = 0.1f;
	cout << "\tTimer initialized..." << endl;

	// Reset keys
	for (i = 0; i < NUM_KEYS; i++)
		LastPressedKeys[0] = 0;

	// Set up OpenGL viewport
	resizeViewport(width, height);

	// Build font
	buildFont();

	cout << "Application initialized succesfully..." << endl;
}

void cApplication::start()
{
	// Perform custom initialization (virtual call)
	this->init();

	try {

		// Enter main loop
		mainLoop();

	} catch (cException& e) {

		// Perform custom deinitilization (virtual call)
		this->shutdown();

		// Close SDL
		SDL_Quit();

		// Throw new exception up
		throw cException("Unexpected failure, reason: " + e.getMessage());
	}

	cout << "Closing application...\n";

	// Perform custom deinitilization (virtual call)
	this->shutdown();

	killFont();

	// Close SDL
	SDL_Quit();

	// Delete itself
	delete this;

	cout << "Application closed succesfully...\n";
}

void cApplication::resizeViewport(int width, int height)
{
	if (height == 0)
		height = 1;

	glViewport(0, 0, width, height);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	gluPerspective(45.0f, (GLfloat) width / (GLfloat) height, 0.1f, 1000.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();			
}

void cApplication::quit()
{
	Quit = true;
}

void cApplication::mainLoop()
{
	while (!Quit)
	{
		// Check incoming events
		SDL_Event event;
		while (SDL_PollEvent(&event))
		{
			switch (event.type) {
				case SDL_QUIT:
					Quit = true;
					break;
				default:
					break;
			}
		}

		// Grab keyboard (and mouse) buttons' state
		memcpy(PrevKeys, Keys, NUM_KEYS);
		memcpy(Keys, SDL_GetKeyState(NULL), NUM_KEYS);

		// Grab mouse movement vector
		int x, y;
		SDL_GetRelativeMouseState(&x, &y);
		MouseMovement = cVec2((float) x, (float) y);

		// Update timer
		const unsigned int time = SDL_GetTicks();
		const unsigned int deltaTime = time - Timer.Ticks;

		Timer.FrameTime = (time - Timer.PrevTicks) * 0.001f;
		Timer.PrevTicks = time;

		Timer.FramesCounter++;

		if (deltaTime > 100 || Timer.FramesCounter > 10)
		{
			Timer.AveragedFrameTime = (float) deltaTime * 0.001f / (float) Timer.FramesCounter;
			Timer.FPS = Timer.FramesCounter * 1000.0f / deltaTime;

			Timer.Ticks = time;
			Timer.FramesCounter = 0;
		}

		// Custom user update (rendering, scene update, anything...)
		this->update(deltaTime);

		// Custom user rendering function
		this->render();
	}
}

float cApplication::getAveragedFrameTime() const
{
	return Timer.AveragedFrameTime;
}

float cApplication::getFrameTime() const
{
	return Timer.FrameTime;
}

float cApplication::getFPS() const
{
	return Timer.FPS;
}

bool cApplication::isKeyPressed(unsigned char key, unsigned int minDelay)
{
	if ((Keys[key] != 0) && (SDL_GetTicks() - LastPressedKeys[key] >= minDelay))
	{
		LastPressedKeys[key] = SDL_GetTicks();
		return true;
	}
	return false;
}

unsigned int cApplication::createTexture(string fileName, bool mipmap)
{
	unsigned int textureId;

	MGL::TargaImage image;
	if (!image.Load((fileName.substr( 0, fileName.length() - 4 ) + ".tga").data()))
		throw cException("Failed to load texture " + fileName);

	glGenTextures(1, &textureId);
	glBindTexture(GL_TEXTURE_2D, textureId);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

	// Convert RGB -> BRG
	for (int i = 0; i < image.GetWidth(); i++)
		for (int j = 0; j < image.GetHeight(); j++)
		{
			unsigned char* r = image.GetImage() + 3 * (j * image.GetWidth() + i);
			unsigned char* g = r + 1;
			unsigned char* b = r + 2;

            unsigned char temp;
			
			temp = *g;
			*g = *r;
			*r = *b;
			*b = temp;
		}

	gluBuild2DMipmaps(
		GL_TEXTURE_2D, 
		3,
		image.GetWidth(),
		image.GetHeight(),
		GL_RGB,
		GL_UNSIGNED_BYTE,
		image.GetImage());

	return textureId;
}

void cApplication::buildFont()
{
	FontTexture = createTexture("../media/textures/font.bmp", true);

	FontBase = glGenLists(256);
	glBindTexture(GL_TEXTURE_2D, FontTexture);
	for (int loop1=0; loop1<256; loop1++)					// Loop Through All 256 Lists
	{
		float cx=float(loop1%16)/16.0f;						// X Position Of Current Character
		float cy=float(loop1/16)/16.0f;						// Y Position Of Current Character

		glNewList(FontBase+loop1,GL_COMPILE);					// Start Building A List
			glBegin(GL_QUADS);								// Use A Quad For Each Character
				glTexCoord2f(cx,1.0f-cy-0.0625f);			// Texture Coord (Bottom Left)
				glVertex2d(0,16);							// Vertex Coord (Bottom Left)
				glTexCoord2f(cx+0.0625f,1.0f-cy-0.0625f);	// Texture Coord (Bottom Right)
				glVertex2i(16,16);							// Vertex Coord (Bottom Right)
				glTexCoord2f(cx+0.0625f,1.0f-cy-0.001f);	// Texture Coord (Top Right)
				glVertex2i(16,0);							// Vertex Coord (Top Right)
				glTexCoord2f(cx,1.0f-cy-0.001f);			// Texture Coord (Top Left)
				glVertex2i(0,0);							// Vertex Coord (Top Left)
			glEnd();										// Done Building Our Quad (Character)
			glTranslated(14,0,0);							// Move To The Right Of The Character
		glEndList();										// Done Building The Display List
	}														// Loop Until All 256 Are Built
}

void cApplication::killFont()
{
	glDeleteLists(FontBase, 256);
}

void cApplication::printText(int x, int y, const char* format, ...)
{
	// Grab input params into single string
	static char text[1 << 12];
	va_list ap;

	if (format == NULL)
		return;

	va_start(ap, format);
	    vsprintf(text, format, ap);
	va_end(ap);

	// No italics (italics would be set = 1)
	const int set = 0;

	glDisable(GL_DEPTH_TEST);

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
	glEnable(GL_TEXTURE_2D);
	glLoadIdentity();
	glTranslated(x, y, 0);
	glListBase(FontBase - 32 + (128 * set));

	glScalef(0.8f, 1.0f, 0.5f);

	glBindTexture(GL_TEXTURE_2D, FontTexture);
	glCallLists((int) strlen(text), GL_UNSIGNED_BYTE, text);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	glEnable(GL_DEPTH_TEST);
}