/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _APPLICATION_H_
#define _APPLICATION_H_

/**
 *	Abstract OpenGL application being base for simple applications.
 *
 *	Application class handles ui events, keyboard and mouse input
 *	as well as simple timer.
 */
class cApplication
{
public:

	/**
	 *	Starts an application, runs it, and closes it.
	 */
	void start();

protected:

	/**
	 *	Creates application.
	 *	<p>
	 *	Throws exception on failure.
	 *	<p>
	 *	@param width width of the application window
	 *	@param height height of the application window
	 *	@param colorBits number of bits per pixel in color buffer (usually 16, 24 or 32)
	 *	@param depthBits depth buffer bits
	 *	@param stencilBits stencil buffer bits
	 *	@param fullscreen shall fullscreen mode be enabled?
	 *	@param name name of the application
	 */
	cApplication(int width, int height, int colorBits, int depthBits,
		int stencilBits, bool fullscreen, const string& name);

	// Forbidden destructor
	virtual ~cApplication() {}

	/*------------ Abstract operations -------------*/

	/**
	 *	Custom application initialization.
	 *	@return true on success, false otherwise.
	 */
	virtual bool init() = 0;

	/**
	 *	Custom application destroy function.
	 *	@return true on success, false otherwise.
	 */
	virtual bool shutdown() = 0;

	/**
	 *	Custom application update.
	 *	@param deltaTime delta time to update by
	 */
	virtual void update(double deltaTime) = 0;

	/**
	 *	Renders the scene and HUD.
	 */
	virtual void render() = 0;

	/*------------------- Public methods ------------------------*/

public:

	/**
	 *	Quits an application.
	 *	In fact, only sets 'flag' to leave main application loop
	 *	as soon as its end is reached.
	 */
	void quit();

	/**
	 *	Returns width of the window.
	 */
	inline int getWindowWidth() const
	{
		return Width;
	}

	/**
	 *	Returns height of the window.
	 */
	inline int getWindowHeight() const
	{
		return Height;
	}

	/**
	 *	Tells whether given key is pressed at the moment.
	 *	@param key key to be tested
	 *	@param minDelay minimal delay between key being pressed 2 times in a row
	 *	@return true in case key is pressed, false otherwise
	 */
	bool isKeyPressed(unsigned char key, unsigned int minDelay = 0);

	/**
	 *	Tells whether given key has just been pressed.
	 *	@param key key to be tested
	 *	@return true in case key has just been pressed, false otherwise
	 */
	inline bool justPressed(unsigned char key) const
	{
		return Keys[key] != 0 && PrevKeys[key] == 0;
	}

	/**
	 *	Returns mouse movement vector.
	 */
	inline const cVec2& getMouseMovement() const
	{
		return MouseMovement;
	}

	/**
	 *	Returns averaged frame time.
	 *	@return frame time as percentage of a second
	 */
	float getAveragedFrameTime() const;

	/**
	 *	Returns recent frame time.
	 *	@return frame time as percentage of a second
	 */
	float getFrameTime() const;

	/**
	 *	Returns frames-per-second count.
	 *	@return FPS
	 */
	float getFPS() const;

	/**
	 *	Prints given text to at given coordinates.
	 *	@param x x coordinate
	 *	@param y y coordinate
	 *	@param format format of the text (to parse following it arguments appropriately)
	 */
	void printText(int x, int y, const char* format, ...);

	/**
	 *	Creates texture from given file.
	 *	@param fileName file name of the bitmap to create texture from
	 *	@param mipmap tells whethre mipmapping should be enabled
	 */
	static unsigned int createTexture(string fileName, bool mipmap);

protected:

	/**
	 *	Main application loop with events handling, rendering etc.
	 */
	void mainLoop();

	/**
	 *	Resizes viewport after window's dimensions have been changed.
	 *	@param width width of the window
	 *	@param height of the window
	 */
	void resizeViewport(int width, int height);

	/**
	 *	Builds fonts into display lists.
	 */
	void buildFont();

	/**
	 *	Deletes font.
	 */
	void killFont();

	/*------------------ Protected attributes ----------------*/

	/// Number of keys we need to test (255 on keyboard + 3 mouse buttons)
	#define NUM_KEYS (256 + 3)

	/// Keys state
	unsigned char Keys[NUM_KEYS];

	/// Keys state from previous frame
	unsigned char PrevKeys[NUM_KEYS];

	/// Time at which each of the key was recently pressed
	unsigned int LastPressedKeys[NUM_KEYS];

	/// Mouse movement vector
	cVec2 MouseMovement;

	/// Tells whether to quit application next frame
	bool Quit;

	/// Width of the window
	int Width;

	/// Height of the window
	int Height;

	/**
	 *	Simple timer class.
	 */
	struct cTimer
	{
		/// Frames per second (updated once a while)
		float FPS;

		/// Time in miliseconds of the last frame
		float FrameTime;

		/// Averaged frame time among multiple frames
		float AveragedFrameTime;

		/// Ticks of last FPS (and AveragedFrameTime) calculation
		unsigned int Ticks;

		/// Previous frame ticks
		unsigned int PrevTicks;

		/// Counts frames for FPS/FrameTime calculation
		unsigned int FramesCounter;
	};

	/// Timer
	cTimer Timer;

	/// Font base for display list
	int FontBase;

	/// Font texture
	unsigned int FontTexture;

};

#endif //_APPLICATION_H_