/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#include "common.h"

#include "Application.h"
#include "Camera.h"
#include "Mirror.h"
#include "Renderer.h"

cFrustum::cFrustum(const cCamera* camera)
{
    unsigned int i;

	cMat4x4 mv;			// Model view matrix
	cMat4x4 proj;		// Projection matrix
	cMat4x4 clip;		// Clipping frustum matrix (equals: proj * mv)
	int viewport[4];	// Viewport (left, top, width, height)

	// Retrieve matrices and viewport from OpenGL
	glGetFloatv(GL_MODELVIEW_MATRIX, (float*) mv);
	glGetFloatv(GL_PROJECTION_MATRIX, (float*) proj);
	glGetIntegerv(GL_VIEWPORT, viewport);

	// Calculate multiplication of projection and model-view matrices
	clip = proj;
	clip *= mv;

	// Initialize space for 5 planes (do not use far plane)
	Planes.resize(5);

	// Calculate RIGHT plane
	Planes[0].Normal = cVec3(clip[3] - clip[0], clip[7] - clip[4], clip[11] - clip[8]);
	Planes[0].Offset = -clip[15] + clip[12];

	// Calculate BOTTOM plane
	Planes[1].Normal = cVec3(clip[3] + clip[1], clip[7] + clip[5], clip[11] + clip[9]);
	Planes[1].Offset = -clip[15] - clip[13];

	// Calculate LEFT plane
	Planes[2].Normal = cVec3(clip[3] + clip[0], clip[7] + clip[4], clip[11] + clip[8]);
	Planes[2].Offset = -clip[15] - clip[12];

	// Calculate TOP plane
	Planes[3].Normal = cVec3(clip[3] - clip[1], clip[7] - clip[5], clip[11] - clip[9]);
	Planes[3].Offset = -clip[15] + clip[13];

	// Calculate NEAR plane
	Planes[4].Normal = cVec3(clip[3] + clip[2], clip[7] + clip[6], clip[11] + clip[10]);
	Planes[4].Offset = -clip[15] - clip[14];

	/*
		Do not calculate FAR plane, which would normally be calculated as follows:

		Planes[5].Normal = cVec3(clip[3] - clip[2], clip[7] - clip[6], clip[11] - clip[10]);
		Planes[5].Offset = -clip[15] + clip[14];

		Far plane might easily be added though.
	*/

	// Normalize all planes
	for (i = 0; i < Planes.size(); i++)
		Planes[i].normalize();

	// Get camera position
	if (camera != NULL)
		Origin = camera->getCoords();

	// Assign scissor rectangle
	ScreenSpaceBoundingRectangle =
		cRectangle(
			(float) viewport[0],
			(float) viewport[1],
			(float) viewport[2] - viewport[0],
			(float) viewport[3] - viewport[1]);
}

cFrustum::cFrustum(const cFrustum& other)
{
	*this = other;
}

void cFrustum::transform(const cMat4x4& matrix)
{
	unsigned int i;

	// Transform clipping planes
	for (i = 0; i < Planes.size(); i++)
		matrix.transform(Planes[i]);

	// Transform frustum origin
	matrix.transform(Origin);
}

bool cFrustum::intersect(const cMirror* mirror)
{
	unsigned int i, j;

	// Get mirror bounding vertices (mirror portal)
	const vector<cVec3>& verts = mirror->getBoundingVerts();

	// Intersect bounding screen space rectangles for scissor test
	// Note that this implementation is highly unoptimized, but it doesn't really matter,
	// since the real rendering cost of the mirrors is not here - it's mainly fillrate.
	double x, y, z;
	double mv[16];
	double proj[16];
	int viewport[4];

	// It would be more efficient and elegant to move these 4 operations to the place,
	// where they'd be called as few times as actually needed, but even though they're
	// slow, they don't really cost much comparing to rendering cost. In more complex
	// environments (with thousands of portals) this might become a significant issue though.
	glGetDoublev(GL_MODELVIEW_MATRIX, mv);
	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetIntegerv(GL_VIEWPORT, viewport);
	cFrustum frustum(NULL);

	// Calculate bounding box of the mirror portal
	cRectangle mirrorRectangle;
	for (i = 0; i < verts.size(); i++)
	{
		// Project portal vertex
		if (gluProject(
			(double) verts[i][0], (double) verts[i][1], (double) verts[i][2],
			mv, proj, viewport,
			&x, &y, &z) == GL_TRUE)
		{
			// Clip vertex to screen dimensions
			if (x >= viewport[2] - viewport[0])
				x = viewport[2] - viewport[0] - 1;
			else if (x < viewport[0])
				x = viewport[0];

			if (y >= viewport[3] - viewport[1])
				y = viewport[3] - viewport[1] - 1;
			else if (y < viewport[1])
				y = viewport[1];

			// Extend rectangle by this projected vertex
			mirrorRectangle.extend(cVec2((float) x, (float) y));
		}

		// Check for special cases when vertex is not on the screen.
		// This is because gluProject() can return overflowed values in such cases.
		if (frustum.Planes[0].Normal.dot(verts[i]) - frustum.Planes[0].Offset < 0)
			mirrorRectangle.extendX( (float) viewport[2] - viewport[0] );
		if (frustum.Planes[2].Normal.dot(verts[i]) - frustum.Planes[2].Offset < 0)
			mirrorRectangle.extendX( (float) viewport[0] );

		if (frustum.Planes[1].Normal.dot(verts[i]) - frustum.Planes[1].Offset < 0)
			mirrorRectangle.extendY( (float) viewport[1] );
		if (frustum.Planes[3].Normal.dot(verts[i]) - frustum.Planes[3].Offset < 0)
			mirrorRectangle.extendY( (float) viewport[3] - viewport[1] );
	}

	// Intersect current bounding rectangle with bounding rectangle of the mirror portal
	ScreenSpaceBoundingRectangle.intersect(mirrorRectangle);

	// See if such intersection is not empty
	if (ScreenSpaceBoundingRectangle.isEmpty())
		return false;

	// Determine side of all 4 mirror verts with regard to all side frustum planes
	// (which there are 4 - except near plane)
	bool vertexInFront[4][4];

	for (i = 0; i < 4; i++)
	{
		bool allVertsOnPlanesBackside = true;

		for (j = 0; j < 4; j++)
		{
			vertexInFront[i][j] =
				(Planes[i].Normal.dot( verts[j] ) - Planes[i].Offset > 0);

			if (vertexInFront[i][j])
				allVertsOnPlanesBackside = false;
		}

		// All verts are on the backside of the plane means intersection is empty
		if (allVertsOnPlanesBackside)
			return false;
	}

	// Calculate 4 clipping planes for mirror portal
	cPlane newPlanes[4];
	for (j = 0; j < 4; j++)
		newPlanes[j].build(Origin, verts[(j + 1) % 4], verts[j % 4]);

	if (cRenderer::RenderMode.EnableFrustumIntersection)
		// For each of the side mirror planes: see if it can be replaced
		// by one of current planes and if so, replace it.
		// Note: This approximation sometimes is not as good as expected. It's definitely
		// possible to implement better, but surely more complex approximation to
		// frustum-portal intersection algorithm.
		for (j = 0; j < 4; j++)
			for (i = 0; i < 4; i++)
				if (!vertexInFront[i][j] && !vertexInFront[i][(j + 1) % 4])
				{
					newPlanes[j] = Planes[i];
					break;
				}

	// Set new planes as current planes
	for (i = 0; i < 4; i++)
		Planes[i] = newPlanes[i];

	// Set new "near" plane (don't mismatch with OpenGL near plane)
	Planes[4] = mirror->getPlane();

	// Flip near plane, since mirror is facing toward viewer and we are looking
	// through mirror
	Planes[4].flip();

	return true;
}

bool cFrustum::testVisibility(const cBoundingVolume& volume) const
{
	unsigned int i;

	// Perform appropriate test based on global (adjustable through GUI) setting
	if (cRenderer::RenderMode.TestBoundingBox)
	{
		// Perform more precise (and slightly more expensive) test with
		// volume's bounding box
		for (i = 0; i < Planes.size(); i++)
			/*
				The whole trick with bounding box test here requires some more explanation.

				Standard box-on-side (of the plane) visibility test involves 8 tests -
				one for each vertex of the bounding box.

                As you see, instead of testing 8 vertices of the box, we test only 1!
				Yes, it's possible.
				Moreover it's exactly same correct as method that tests 8 vertices.

				So, which vertex do we test?
				We find such corner(!) of the box that when we look from inside of the box
				in direction of the normal (plane's normal), this corner will be passed away
				as the last.

				So we take the most 'outstanding' box corner along plane's normal and we
				perform visibility test only for this single vertex. If this vertex is on the
				backside of the plane, it means the whole box is behind the plane - obviously
				because this vertex was the most outstanding along plane's normal.
			*/
			if (Planes[i].Normal.dot( volume.getMostOutstandingCorner(Planes[i].Normal) ) - Planes[i].Offset < 0)
				return false;
	}
	else
		// Test with bounding sphere (slightly faster, but a lot less accurate - especially in case
		// of flat objects liek walls)
		for (i = 0; i < Planes.size(); i++)
			if (Planes[i].Normal.dot( volume.getOrigin() ) - Planes[i].Offset + volume.getRadius() < 0)
				return false;

	return true;
}

bool cFrustum::testVisibilityPrecisely(const cBoundingVolume& volume, const vector<cVec3>& verts) const
{
	// Perform regular frustum culling first
	if (!testVisibility(volume))
		return false;

	unsigned int i;

	// Return success if at least one vertex is in front of the culling plane
	// Note: this test is performed only against near clipping plane (4th plane)
	for (i = 0; i < verts.size(); i++)
		if (Planes[4].Normal.dot( verts[i] ) - Planes[4].Offset > 0.01f)
			return true;

	return false;
}

bool cFrustum::testPlaneSide(const cPlane& plane) const
{
	return plane.Normal.dot(Origin) - plane.Offset > 0;
}

cFrustum& cFrustum::operator = (const cFrustum& other)
{
	Planes = other.Planes;
	Origin = other.Origin;
	ScreenSpaceBoundingRectangle = other.ScreenSpaceBoundingRectangle;
	return *this;
}