/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _FRUSTUM_H_
#define _FRUSTUM_H_

// Predefinitions
class cCamera;
class cMirror;

/**
 *	Frustum used for culling visible geometry.
 *	<p>
 *	Our custom frustum representation also includes viewer's origin,
 *	which will be necessary for appropriate frustum's intersection
 *	calculation.
 *	<p>
 *	It also contains scissor rectangle to be applied
 *	for visible scene part.
 *	<p>
 *	Tests against visibility are performed using one of 2 methods:
 *	- with bounding box (more expensive, more precise)
 *	- with bounding sphere (faster and less precise).
 *	Which kind of test is performed depends on global setting of renderer.
 */
class cFrustum
{
public:

	/**
	 *	Creates empty frustum.
	 */
	cFrustum();

	/**
	 *	Builds frustum based on current OpenGL model-view and projection matrices.
	 *	It also sets up scissor rectangle to the whole screen based on given camera.
	 *	@param camera camera object to create frustum from; camera set to NULL
	 *		indicates we only want to build clipping planes (no camera origin will
	 *		be used)
	 */
	cFrustum(const cCamera* camera);

	/**
	 *	Copy constructor.
	 *	@param other frustum to copy from
	 */
	cFrustum(const cFrustum& other);

	/*---------------- Operations -----------------*/

	/**
	 *	Transforms frustum by given matrix, e.g. reflection matrix.
	 *	@param matrix matrix to transform by
	 */
	void transform(const cMat4x4& matrix);

	/**
	 *	Performs a very approximate geometrical intersection between
	 *	this frustum and frustum built using this frustum's origin
	 *	and given mirror portal's verts. Note that such geometrical intersection
	 *	may result in an empty frustum - in such case false is returned
	 *	and, obviously, this means that all objects seen through given
	 *	portal are not visible.
	 *	<p>
	 *	The second important issue concerning frustum intersection is
	 *	intersection of bounding rectangles of current one and the one
	 *	built from mirror portal.
	 *	<p>
	 *	@param mirror mirror to calculate intersection with
	 *	@return true in case intersected frustum isn't
	 *		empty, false otherwise
	 */
	bool intersect(const cMirror* mirror);

	/**
	 *	Performs frustum test (visibility test) for a given bounding volume
	 *	used as a sphere.
	 *	@param sphere sphere to be tested against frustum
	 *	@return true if an object is, possibly partially, visible, false otherwise
	 */
	bool testVisibility(const cBoundingVolume& sphere) const;

	/**
	 *	Returns success if at least one vertex is in front of the near clipping plane
	 *	and volume's bounding box (or sphere - depending on global setting) was not
	 *	culled away (using regular testVisibility() method executed at the beginning).
	 *	<p>
	 *	@param volume bounding volume of the object to test
	 *	@param verts vertices of the object to test (used only in case bounding volume
	 *		test, performed before per-vertex test, fails)
	 *	@return true if an object is, possibly partially, visible, false otherwise
	 */
	bool testVisibilityPrecisely(const cBoundingVolume& volume, const vector<cVec3>& verts) const;

	/**
	 *	Performs test, which tells whether given plane is facing
	 *	frustum's (viewer's) origin or not.
	 *	<p>
	 *	When polygon's (or mirror's) plane is not facing frustum's
	 *	origin, it means that the polygon is facing back the user,
	 *	thus, in most common situations, it shouldn't be rendered.
	 *	<p>
	 *	@param plane plane to test
	 *	@return true if plane is facing frustum's (viewer's) origin,
	 *		false otherwise
	 */
	bool testPlaneSide(const cPlane& plane) const;

	/**
	 *	Assignment operator.
	 *	@param other other frustum to assign this frustum to
	 *	@return reference to this object
	 */
	cFrustum& operator = (const cFrustum& other);

	/**
	 *	Returns screen space bounding box of the frustum used for scissor test.
	 *	@return bounding screen space rectangle
	 */
	inline const cRectangle& getScreenSpaceBoundingRectangle() const
	{
		return ScreenSpaceBoundingRectangle;
	}

protected:

	/// Viewer's origin
	cVec3 Origin;

	/// Clipping planes of the frustum
	vector<cPlane> Planes;

	/// Rectangle used for scissor test
	cRectangle ScreenSpaceBoundingRectangle;

};

#endif //_FRUSTUM_H_