/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#include "common.h"

#include "Renderer.h"
#include "Scene.h"
#include "SimpleScene.h"
#include "Camera.h"
#include "Mesh.h"
#include "Material.h"
#include "Application.h"
#include "MGLApp.h"

void MGLApp::cSettings::setDefaults()
{
	ShowMirrorCube = true;
	RotateMirrorCube = true;
	ShowInfoScreen = false;
}

MGLApp::MGLApp(bool fullScreen) :
	cApplication(
		MGL_WINDOW_WIDTH,
		MGL_WINDOW_HEIGHT,
		MGL_COLOR_BITS,
		MGL_DEPTH_BITS,
		MGL_STENCIL_BITS,
		fullScreen,
		MGL_APP_NAME),
	Renderer(NULL),
	Scene(NULL),
	Camera(NULL)
{
	Settings.setDefaults();
}

bool MGLApp::init()
{
	// Create renderer
	Renderer = new cRenderer(this);

	// Create (empty) scene
	Scene = new cSimpleScene(this);

	// Build demo scene with rotating cube
	Scene->buildDemoScene();
	Scene->addMirrorCube(cVec3(5, 3, 5), 1.3f);

	// Uncomment to see 4 extra mirror cubes
	// (requires strong GPU)
	/*for (int i = 0; i < 2; i++)
		for (int j = 0; j < 2; j++)
			Scene->addMirrorCube(cVec3(2 + 6 * i, 3.5f, 2 + 6 * j), 1.0f);*/

	// Create camera (load its initial position from 'viewer.camera' file)
	Camera = new cCamera("viewer", this);

	return true;
}

void MGLApp::update(double deltaTime)
{
	// Handle quit
	if (justPressed(SDLK_ESCAPE))
		quit();

	// Show / hide help screen
	if (justPressed(SDLK_h))
		Settings.ShowInfoScreen = !Settings.ShowInfoScreen;

	// Handle render params changes
	if (justPressed(SDLK_v))
		cRenderer::RenderMode.IsWireMode = !cRenderer::RenderMode.IsWireMode;
	if (justPressed(SDLK_b))
		cRenderer::RenderMode.TestBoundingBox = !cRenderer::RenderMode.TestBoundingBox;
	if (justPressed(SDLK_g))
		cRenderer::RenderMode.EnableScissorTest = !cRenderer::RenderMode.EnableScissorTest;
	if (justPressed(SDLK_f))
		cRenderer::RenderMode.EnableFrustumIntersection = !cRenderer::RenderMode.EnableFrustumIntersection;
	if (justPressed(SDLK_j))
		cRenderer::RenderMode.EnableMirrorClipping = !cRenderer::RenderMode.EnableMirrorClipping;
	if (justPressed(SDLK_1))
		if (cRenderer::RenderMode.MaxReflectionLevel > 0)
			cRenderer::RenderMode.MaxReflectionLevel--;
	if (justPressed(SDLK_2))
		// Note: Don't set more than 30, because this limitation corresponds
		// with maximal OpenGL model-view matrix stack size.
		// If you want more, you have to implement your own model-view matrix stack.
		if (cRenderer::RenderMode.MaxReflectionLevel < 30)
			cRenderer::RenderMode.MaxReflectionLevel++;
	if (isKeyPressed(SDLK_3, 100))
		if (cRenderer::RenderMode.MinOpaqueness >= 0.01f)
			cRenderer::RenderMode.MinOpaqueness -= 0.01f;
	if (isKeyPressed(SDLK_4, 100))
		if (cRenderer::RenderMode.MinOpaqueness < 0.98f)
			cRenderer::RenderMode.MinOpaqueness += 0.01f;

	if (justPressed(SDLK_m))
		Settings.RotateMirrorCube = !Settings.RotateMirrorCube;
	if (justPressed(SDLK_n))
	{
		Settings.ShowMirrorCube = !Settings.ShowMirrorCube;

		// Sic! Instead of marking as invisible we delete/create mirror cube
		if (Settings.ShowMirrorCube)
			Scene->addMirrorCube(cVec3(5, 3, 5), 1.3f);
		else
			Scene->removeMirrorCubes();
	}

	if (isKeyPressed(SDLK_5, 100))
		if (Scene->getMaterial("cube mirror")->getTransparency() >= 0.01f)
			Scene->getMaterial("cube mirror")->setTransparency( Scene->getMaterial("cube mirror")->getTransparency() - 0.01f );
	if (isKeyPressed(SDLK_6, 100))
		if (Scene->getMaterial("cube mirror")->getTransparency() < 0.98f)
			Scene->getMaterial("cube mirror")->setTransparency( Scene->getMaterial("cube mirror")->getTransparency() + 0.01f );

	if (isKeyPressed(SDLK_7, 100))
		if (Scene->getMaterial("side mirror")->getTransparency() >= 0.01f)
			Scene->getMaterial("side mirror")->setTransparency( Scene->getMaterial("side mirror")->getTransparency() - 0.01f );
	if (isKeyPressed(SDLK_8, 100))
		if (Scene->getMaterial("side mirror")->getTransparency() < 0.98f)
			Scene->getMaterial("side mirror")->setTransparency( Scene->getMaterial("side mirror")->getTransparency() + 0.01f );

	if (isKeyPressed(SDLK_9, 100))
		if (Scene->getMaterial("floor mirror")->getTransparency() >= 0.01f)
			Scene->getMaterial("floor mirror")->setTransparency( Scene->getMaterial("floor mirror")->getTransparency() - 0.01f );
	if (isKeyPressed(SDLK_0, 100))
		if (Scene->getMaterial("floor mirror")->getTransparency() < 0.98f)
			Scene->getMaterial("floor mirror")->setTransparency( Scene->getMaterial("floor mirror")->getTransparency() + 0.01f );

	if (isKeyPressed(SDLK_i, 100))
		if (Scene->getMaterial("ceiling mirror")->getTransparency() >= 0.01f)
			Scene->getMaterial("ceiling mirror")->setTransparency( Scene->getMaterial("ceiling mirror")->getTransparency() - 0.01f );
	if (isKeyPressed(SDLK_o, 100))
		if (Scene->getMaterial("ceiling mirror")->getTransparency() < 0.98f)
			Scene->getMaterial("ceiling mirror")->setTransparency( Scene->getMaterial("ceiling mirror")->getTransparency() + 0.01f );

	// Update camera
	Camera->update();

	// Update mirror cube rotating
	if (Settings.RotateMirrorCube)
		Scene->updateMirrorCubes();
}

void MGLApp::render()
{
	// Render the scene seen from given camera
	Renderer->renderScene(Camera, Scene);

	// Renders demo information (options, keys, ...)
	renderInfo();

	// Swap color buffer
	SDL_GL_SwapBuffers();
}

bool MGLApp::shutdown()
{
	if (Renderer != NULL)
	{
		delete Renderer;
		Renderer = NULL;
	}

	if (Scene != NULL)
	{
		delete Scene;
		Scene = NULL;
	}

	if (Camera != NULL)
	{
		delete Camera;
		Camera = NULL;
	}

	return true;
}

void MGLApp::renderInfo()
{
	glDisable(GL_CULL_FACE);
	glDisable(GL_STENCIL_TEST);
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glDepthFunc(GL_ALWAYS);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDisable(GL_SCISSOR_TEST);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, 800, 600, 0, -1.0f, 1.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslated(0, 0, 0);

	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glColor4f(0.5f, 0.5f, 0.5f, 0.6f);
	glBegin(GL_QUADS);

		glVertex2f(0, 545);
		glVertex2f(0, 600);
		glVertex2f(420, 600);
		glVertex2f(420, 545);

		glVertex2f(0, 0);
		glVertex2f(0, 55);
		glVertex2f(150, 55);
		glVertex2f(150, 0);

		if (Settings.ShowInfoScreen)
		{
			glVertex2f(0, 100);
			glVertex2f(0, 500);
			glVertex2f(450, 500);
			glVertex2f(450, 100);
		}

	glEnd();
	glDisable(GL_BLEND);

	glEnable(GL_TEXTURE_2D);
	glColor3f(1, 1, 1);
	printText(10, 550, "Robust Multi-Level Reflections Demo");
	printText(10, 565, "Maciej Sawitus, 2005");
	printText(10, 580, "Press <H> to %s info screen", Settings.ShowInfoScreen ? "hide" : "show");

	printText(10, 5, "FPS %.1f", getFPS());
	printText(10, 20, "Meshes %d", cRenderer::RenderStats.NumRenderedMeshes);
	printText(10, 35, "Mirrors %d", cRenderer::RenderStats.NumRenderedMirrors);

	if (Settings.ShowInfoScreen)
	{
		const int rowHeight = 15;
		int height = 90;

		printText(10, height += rowHeight, "Movement controls:");
		printText(10, height += rowHeight, "------------------");
		printText(10, height += rowHeight, "S/X - front/back");
		printText(10, height += rowHeight, "Z/C - strafe left/right");
		printText(10, height += rowHeight, "Q/A - up/down");
		printText(10, height += rowHeight, "mouse - look around");
		printText(10, height += rowHeight, "");
		printText(10, height += rowHeight, "Options:");
		printText(10, height += rowHeight, "--------");
		printText(10, height += rowHeight, "Debug wire mode (V): %s", cRenderer::RenderMode.IsWireMode ? "YES" : "NO");
		printText(10, height += rowHeight, "Scissor test (G): %s", cRenderer::RenderMode.EnableScissorTest ? "YES" : "NO");
		printText(10, height += rowHeight, "Frustum intersect (F): %s", cRenderer::RenderMode.EnableFrustumIntersection ? "YES" : "NO");
		printText(10, height += rowHeight, "Frustum test against (B):");
		printText(10, height += rowHeight, cRenderer::RenderMode.TestBoundingBox ? " Bounding boxes" : " Bounding spheres");
		printText(10, height += rowHeight, "Mirror plane clipping (J): %s", cRenderer::RenderMode.EnableMirrorClipping ? "YES" : "NO");
		printText(10, height += rowHeight, "Max. reflection level (1-2): %d", cRenderer::RenderMode.MaxReflectionLevel);
		printText(10, height += rowHeight, "Min. reflection opaqueness (3-4): %.2f", cRenderer::RenderMode.MinOpaqueness);
		printText(10, height += rowHeight, "");
		printText(10, height += rowHeight, "Scene settings:");
		printText(10, height += rowHeight, "---------------");
		printText(10, height += rowHeight, "Show cube (N): %s", Settings.ShowMirrorCube ? "YES" : "NO");
		printText(10, height += rowHeight, "Rotate cube (M): %s", Settings.RotateMirrorCube ? "YES" : "NO");
		printText(10, height += rowHeight, "Cube specular (5-6): %.2f", Scene->getMaterial("cube mirror")->getTransparency());
		printText(10, height += rowHeight, "Side mirrors specular (7-8): %.2f", Scene->getMaterial("side mirror")->getTransparency());
		printText(10, height += rowHeight, "Floor specular (9-0): %.2f", Scene->getMaterial("floor mirror")->getTransparency());
		printText(10, height += rowHeight, "Ceiling specular (i-o): %.2f", Scene->getMaterial("ceiling mirror")->getTransparency());
	}

	glColor4f(1, 1, 1, 1);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
	glEnable(GL_CULL_FACE);
}