/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#include "common.h"

#include "Application.h"
#include "Material.h"

cMaterial::cMaterial(const string& name, const string& textureFileName,
					 float transparency, bool isMirrorSurface) :
	Name(name),
	Transparency(transparency),
	IsMirrorSurface(isMirrorSurface)
{
	TextureId = (unsigned int) cApplication::createTexture(textureFileName, true);

	cout << "\tMaterial \"" << textureFileName << "\" created...\n";
}

cMaterial::~cMaterial()
{
	glDeleteTextures(1, &TextureId);
}

void cMaterial::setRenderState(bool isMaxReflectionLevel) const
{
	// Texturing
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, TextureId);

	// Disable blending for opaque objects and mirror surfaces
	// if on max. reflection level
	if (Transparency == 0 || (IsMirrorSurface && isMaxReflectionLevel))
		glDisable(GL_BLEND);
	else
	{
		glEnable(GL_BLEND);
		glColor4f(1, 1, 1, Transparency);
		glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);
	}

	// Disable depth writes for mirror surfaces except for
	// lowest reflection level. Note that appropriate z values
	// should already be written during stencil decrementing pass
	if (IsMirrorSurface && !isMaxReflectionLevel)
	{
		glDepthMask(GL_FALSE);
		glDepthFunc(GL_EQUAL);
	}
	else
	{
		glDepthMask(GL_TRUE);
		glDepthFunc(GL_LESS);
	}
}

int cMaterial::compare(cMaterial* a, cMaterial* b)
{
	int cmp;

	// Mirror materials have to be first in rendering order
	// because they have to z-cover reflected scene, thus
	// not letting it to be overwritten by other meshes
	if (a->IsMirrorSurface && !b->IsMirrorSurface)
		return 1;
	else if (!a->IsMirrorSurface && b->IsMirrorSurface)
		return -1;

	// Sort by transparency
	cmp = (int) ((a->Transparency - b->Transparency) * 1000);
	if (cmp != 0)
		return cmp;

	return 0;
}