/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _MATERIAL_H_
#define _MATERIAL_H_

/**
 *	Very simple material representation.
 */
class cMaterial
{
public:

	/**
	 *	Creates material with single texture.
	 *
	 *	@param name name of the material (has to unique!)
	 *	@param textureFileName texture file name
	 *	@param transparency transparency factor within 0..1 (0 - opaque, 1 - invisible)
	 *	@param isMirrorSurface tells whether this material is used to render
	 *		mirror surface
	 */
	cMaterial(const string& name, const string& textureFileName,
		float transparency, bool isMirrorSurface = false);

	/**
	 *	Destroys material.
	 */
	~cMaterial();

	/**
	 *	Sets all render states necessary to render using this material.
	 *	@param isMaxReflectionLevel tells whether we're on maximal allowed
	 *		reflection level; this information is necessary for mirror surfaces
	 *		for two following reasons:
	 *		1. To know whether to disable / enable depth mask.
	 *			For all reflection levels except last one, we can disable depth
	 *			writes, because depth was already written during stencil
	 *			decrement pass. Note however that stencil decrement pass is
	 *			performed for all mirrors except for the case when we're on lowest
	 *			reflection level.
	 *		2. To disable blending for lowest reflection level.
	 *			This way we make lowest reflection level not catching the eyes
	 *			as it would when blended at some percentage with black background.
	 */
	void setRenderState(bool isMaxReflectionLevel) const;

	/**
	 *	Comparison used to sort meshes by material.
	 *	@param a first material to compare
	 *	@param b second material to compare
	 *	@return 1 if a is greater than b; -1 if b is greater than b;
	 *		o otherwise
	 */
	static int compare(cMaterial* a, cMaterial* b);

	/*---------------- Accessors -----------------*/

	/**
	 *	Returns name of the material.
	 *	@return name of the material
	 */
	inline string getName() const
	{
		return Name;
	}

	/**
	 *	Returns transparency of the material (value within 0..1).
	 *	0 - opaque, 1 - invisible
	 */
	inline float getTransparency() const
	{
		return Transparency;
	}

	/**
	 *	Sets new transparency of the material (value within 0..1).
	 *	@param transparency new transparency to be set
	 */
	inline void setTransparency(float transparency)
	{
		Transparency = transparency;
	}

	/**
	 *	Tells whether this material is used for rendering mirror.
	 */
	inline bool isMirrorSurface() const
	{
		return IsMirrorSurface;
	}

protected:

	/// Unique name of the material
	string Name;

	/// Transparency factor within 0..1
	/// (1 - fully transparent, 0.5 - semi transparent, 0 - opaque)
	float Transparency;

	/// Is this material used to render mirror surface?
	bool IsMirrorSurface;

	/// Texture id
	unsigned int TextureId;

};

#endif //_MATERIAL_H_