/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _MATH_H_
#define _MATH_H_

/*
 *	Bunch of linear algebra classes:
 *
 *	- cVec2 : vector of 2 floats
 *	- cVec3 : vector of 3 floats
 *	- cVec4 : vector of 4 floats
 *	- cMat4x4 : matrix 4x4
 *	- cPlane : plane consisting of normal and (non-abs'ed) distance from origin
 *	- cRectangle : rectangle in 2D space
 *	- cBoundingVolume : bounding box and bounding sphere at the same time
 */

const float PIOVER180 = 3.14f / 180.0f;

class cVec2
{
public:

	cVec2()
	{
	}
	
	cVec2(float v0, float v1)
	{
		V[0] = v0;
		V[1] = v1;
	}

	inline float& operator [] (int index) const
	{
		return (float&) V[index];
	}

	inline cVec2& operator *= (float v)
	{
		V[0] *= v;
		V[1] *= v;
		return *this;
	}

protected:

	float V[2];
};

class cVec3
{
public:

	cVec3()
	{
	}
	
	cVec3(float v0, float v1, float v2)
	{
		V[0] = v0;
		V[1] = v1;
		V[2] = v2;
	}

	inline float& operator [] (int index) const
	{
		return (float&) V[index];
	}

	float length() const
	{
		return sqrtf(V[0] * V[0] + V[1] * V[1] + V[2] * V[2]);
	}

	void normalize()
	{
		const float l = length();
		if (l == 0)
		{
			V[0] = V[1] = V[2] = 0;
			return;
		}
		V[0] /= l;
		V[1] /= l;
		V[2] /= l;
	}

	float dot(const cVec3& other) const
	{
		return
			V[0] * other.V[0] +
			V[1] * other.V[1] +
			V[2] * other.V[2];
	}

	cVec3& operator += (const cVec3& other)
	{
		V[0] += other.V[0];
		V[1] += other.V[1];
		V[2] += other.V[2];
		return *this;
	}

	cVec3& operator *= (const float value)
	{
		V[0] *= value;
		V[1] *= value;
		V[2] *= value;
		return *this;
	}

	cVec3& operator /= (const float value)
	{
		V[0] /= value;
		V[1] /= value;
		V[2] /= value;
		return *this;
	}

	cVec3& operator += (const float value)
	{
		V[0] += value;
		V[1] += value;
		V[2] += value;
		return *this;
	}

	cVec3 operator * (const cVec3& other) const
	{
		return cVec3(
			V[0] * other.V[0],
			V[1] * other.V[1],
			V[2] * other.V[2]);
	}

	cVec3 operator + (const cVec3& other) const
	{
		return cVec3(
			V[0] + other.V[0],
			V[1] + other.V[1],
			V[2] + other.V[2]);
	}

	cVec3 operator - (const cVec3& other) const
	{
		return cVec3(
			V[0] - other.V[0],
			V[1] - other.V[1],
			V[2] - other.V[2]);
	}

	cVec3 operator * (const float value) const
	{
		return cVec3(
			V[0] * value,
			V[1] * value,
			V[2] * value);
	}

	cVec3 operator - () const
	{
		return cVec3(-V[0], -V[1], -V[2]);
	}

	void cross(const cVec3& a, const cVec3& b)
	{
		V[0] = a.V[1] * b.V[2] - a.V[2] * b.V[1],
		V[1] = a.V[2] * b.V[0] - a.V[0] * b.V[2],
		V[2] = a.V[0] * b.V[1] - a.V[1] * b.V[0];
	}

	inline operator const float* () const
	{
		return V;
	}

protected:

	float V[3];
};

class cVec4
{
public:

	cVec4()
	{
	}
	
	cVec4(float v0, float v1, float v2, float v3)
	{
		V[0] = v0;
		V[1] = v1;
		V[2] = v2;
		V[3] = v3;
	}

	inline float& operator [] (int index) const
	{
		return (float&) V[index];
	}

	inline operator const float* () const
	{
		return V;
	}

protected:

	float V[4];
};

typedef vector<cVec2> cVec2Array;
typedef vector<cVec3> cVec3Array;
typedef vector<cVec4> cVec4Array;

class cPlane
{

public:

	void normalize()
	{
		const float length = Normal.length();
		if (length != 0)
		{
			Normal /= length;
			Offset /= length;
		}
		else
		{
			// Whatever, but valid
			Normal = cVec3(1, 0, 0);
			Offset = 0;
		}
	}

	void build(const cVec3& vertexA, const cVec3& vertexB, const cVec3& vertexC)
	{
		Normal.cross(vertexC - vertexB, vertexB - vertexA);
		Offset = vertexB.dot(Normal);
		
		normalize();
	}

	void flip()
	{
		Normal = -Normal;
		Offset = -Offset;
	}

	inline const cVec3& getNormal() const
	{
		return Normal;
	}

	inline float getOffset() const
	{
		return Offset;
	}

public:

	cVec3 Normal;
	float Offset;
};

/**
 *	Bounding volume represented in 2 ways: as bounding box and as bounding sphere.
 */
class cBoundingVolume
{
public:

	cBoundingVolume()
	{
	}

	void calculate(const cVec3Array& verts)
	{
		unsigned int i;

		if (verts.size() == 0)
		{
			Min = Max = cVec3(0, 0, 0);
			return;
		}
		
		// Calculate bounding box
		Min = verts[0];
		Max = verts[0];
		for (i = 1; i < verts.size(); i++)
		{
			// Extend in X
            if (verts[i][0] < Min[0])
				Min[0] = verts[i][0];
			else if (verts[i][0] > Max[0])
				Max[0] = verts[i][0];

			// Extend in Y
			if (verts[i][1] < Min[1])
				Min[1] = verts[i][1];
			else if (verts[i][1] > Max[1])
				Max[1] = verts[i][1];

			// Extend in Z
			if (verts[i][2] < Min[2])
				Min[2] = verts[i][2];
			else if (verts[i][2] > Max[2])
				Max[2] = verts[i][2];
		}

		// Calculate smallest sphere containing found box
		Origin = (Min + Max) * 0.5f;
		Radius = (Max - Min).length() * 0.5f;
	}

	/**
	 *	Calculates and returns the corner of the bounding box
	 *	which is most standing out when looking along given normal vector.
	 *	@param normal normal along which most outstanding corner should be
	 *		found
	 *	@return most outstanding corner of the box
	 */
	inline cVec3 getMostOutstandingCorner(const cVec3& normal) const
	{
		return cVec3(
			(normal[0] < 0) ? Min[0] : Max[0],
			(normal[1] < 0) ? Min[1] : Max[1],
			(normal[2] < 0) ? Min[2] : Max[2]);
	}

	/**
	 *	Calculates and returns the corner of the bounding box
	 *	which is least standing out when looking along given normal vector.
	 *	@param normal normal along which least outstanding corner should be
	 *		found
	 *	@return least outstanding corner of the box
	 */
	inline cVec3 getLeastOutstandingCorner(const cVec3& normal) const
	{
		return cVec3(
			(normal[0] > 0) ? Min[0] : Max[0],
			(normal[1] > 0) ? Min[1] : Max[1],
			(normal[2] > 0) ? Min[2] : Max[2]);
	}

	inline const cVec3& getMin() const
	{
		return Min;
	}

	inline const cVec3& getMax() const
	{
		return Max;
	}

	inline const cVec3& getOrigin() const
	{
		return Origin;
	}

	inline float getRadius() const
	{
		return Radius;
	}

protected:

	/// Minimal vertex of the box
	cVec3 Min;

	/// Maximal vertex of the box
	cVec3 Max;

	/// Origin of center of the bounding sphere
	cVec3 Origin;

	/// Radius of the center of the bounding sphere
	float Radius;

};

class cMat4x4
{
public:

	cMat4x4()
	{
		loadIdentity();
	}

	const float& operator[] (int i) const
	{
		return V[i];
	}

	float& operator[] (int i)
	{
		return V[i];
	}

	operator const float* (void)
	{
		return V;
	}

	operator float* (void)
	{
		return V;
	}

	void loadIdentity()
	{
		V[0] = V[5] = V[10] = V[15] = 1;
		V[1] = V[2] = V[3] = V[4] = V[6] = V[7] = V[8] = V[9] = V[11] = V[12] = V[13] = V[14] = 0;
	}

	/**
	 *	Builds reflection matrix for given plane.
	 *	@param plane reflection plane
	 */
	void buildReflectionTransformation(const cPlane& plane)
	{
		const cVec3 n = plane.getNormal();
		const float d = plane.getOffset();

		V[0] = 1 - 2 * n[0] * n[0];	V[1] = - 2 * n[0] * n[1];	V[2] = - 2 * n[0] * n[2];		V[3] = 0;
		V[4] = V[1];				V[5] = 1 - 2 * n[1] * n[1];	V[6] = - 2 * n[1] * n[2];		V[7] = 0;
		V[8] = V[2];				V[9] = V[6];				V[10] = 1 - 2 * n[2] * n[2];	V[11] = 0;
		V[12] = 2 * d * n[0];		V[13] = 2 * d * n[1];		V[14] = 2 * d * n[2];			V[15] = 1;
	}

	/**
	 *	Self-multiplication operator.
	 */
	cMat4x4& operator *= (const cMat4x4& b)
	{
		int i;
		cMat4x4 a;

		for (i = 0; i < 16; i++)
			a[i] = V[i];

		V[0] = b[0] * a[0] + b[1] * a[4] + b[2] * a[8] + b[3] * a[12];
		V[1] = b[0] * a[1] + b[1] * a[5] + b[2] * a[9] + b[3] * a[13];
		V[2] = b[0] * a[2] + b[1] * a[6] + b[2] * a[10] + b[3] * a[14];
		V[3] = b[0] * a[3] + b[1] * a[7] + b[2] * a[11] + b[3] * a[15];

		V[4] = b[4] * a[0] + b[5] * a[4] + b[6] * a[8] + b[7] * a[12];
		V[5] = b[4] * a[1] + b[5] * a[5] + b[6] * a[9] + b[7] * a[13];
		V[6] = b[4] * a[2] + b[5] * a[6] + b[6] * a[10] + b[7] * a[14];
		V[7] = b[4] * a[3] + b[5] * a[7] + b[6] * a[11] + b[7] * a[15];

		V[8] = b[8] * a[0] + b[9] * a[4] + b[10] * a[8] + b[11] * a[12];
		V[9] = b[8] * a[1] + b[9] * a[5] + b[10] * a[9] + b[11] * a[13];
		V[10] = b[8] * a[2] + b[9] * a[6] + b[10] * a[10] + b[11] * a[14];
		V[11] = b[8] * a[3] + b[9] * a[7] + b[10] * a[11] + b[11] * a[15];

		V[12] = b[12] * a[0] + b[13] * a[4] + b[14] * a[8] + b[15] * a[12];
		V[13] = b[12] * a[1] + b[13] * a[5] + b[14] * a[9] + b[15] * a[13];
		V[14] = b[12] * a[2] + b[13] * a[6] + b[14] * a[10] + b[15] * a[14];
		V[15] = b[12] * a[3] + b[13] * a[7] + b[14] * a[11] + b[15] * a[15];

		return *this;
	}

	void rotate(float angle, const cVec3& vertex)
	{
		float c = (float) cos(angle);
		float s = (float) sin(angle);
		float mc = 1 - c;
		float x = vertex[0];
		float y = vertex[1];
		float z = vertex[2];
		cMat4x4 tmpMatrix;
		
		tmpMatrix.V[0] = (x * x * mc) + c;
		tmpMatrix.V[1] = (y * x * mc) + (z * s);
		tmpMatrix.V[2] = (x * z * mc) - (y * s);
		tmpMatrix.V[4] = (x * y * mc) - (z * s);
		tmpMatrix.V[5] = (y * y * mc) + c;
		tmpMatrix.V[6] = (y * z * mc) + (x * s);
		tmpMatrix.V[8] = (x * z * mc) + (y * s);
		tmpMatrix.V[9] = (y * z * mc) - (x * s);
		tmpMatrix.V[10] = (z * z * mc) + c;

		(*this) *= tmpMatrix;
	}

	void translate(float x, float y, float z)
	{
		cMat4x4 tmpMatrix;

		tmpMatrix.V[12] = x;
		tmpMatrix.V[13] = y;
		tmpMatrix.V[14] = z;

		(*this) *= tmpMatrix;
	}

	// transforms vector by itself
	void transform(cVec3& vec) const
	{
		const cVec3 temp = vec;

		vec[0] = V[0] * temp[0] + V[4] * temp[1] + V[8] * temp[2] + V[12];
		vec[1] = V[1] * temp[0] + V[5] * temp[1] + V[9] * temp[2] + V[13];
		vec[2] = V[2] * temp[0] + V[6] * temp[1] + V[10] * temp[2] + V[14];
	}

	// transforms plane by matrix
	void transform(cPlane& p) const
	{
		const cPlane temp = p;

		// Figure out 2 points:
		// p1 - some point on a plane
		// p2 - point unit away from p1 in direction of plane normal
		cVec3 p1 = p.Normal * p.Offset;
		cVec3 p2 = p1 + p.Normal;

		// Transform both points by matrix
		transform(p1);
		transform(p2);

		// Construct new plane
		p.Normal = p2 - p1;
		p.Normal.normalize();
		p.Offset = p1.dot(p.Normal);
	}

protected:

	float V[16];
};

class cRectangle
{
public:

	cRectangle() :
		MinX(999999),
		MinY(999999),
		MaxX(-999999),
		MaxY(-999999)
	{
	}

	cRectangle(float minX, float minY, float maxX, float maxY) :
		MinX(minX),
		MinY(minY),
		MaxX(maxX),
		MaxY(maxY)
	{
	}

	cRectangle& operator = (const cRectangle& other)
	{
		MinX = other.MinX;
		MinY = other.MinY;
		MaxX = other.MaxX;
		MaxY = other.MaxY;
		return *this;
	}

	/**
	 *	Extends rectangle so that it contains given vertex.
	 *	@param vertex vertex to extend rectangle by
	 */
	void extend(const cVec2& vertex)
	{
		MinX = min(MinX, vertex[0]);
		MinY = min(MinY, vertex[1]);
		MaxX = max(MaxX, vertex[0]);
		MaxY = max(MaxY, vertex[1]);
	}

	/**
	 *	Extends rectangle so that it contains given x coordinate.
	 *	@param x x coordinate to extend rectangle by
	 */
	void extendX(const float x)
	{
		MinX = min(MinX, x);
		MaxX = max(MaxX, x);
	}

	/**
	 *	Extends rectangle so that it contains given y coordinate.
	 *	@param y y coordinate to extend rectangle by
	 */
	void extendY(const float y)
	{
		MinY = min(MinY, y);
		MaxY = max(MaxY, y);
	}

	/**
	 *	Intersects this rectangle with given one.
	 *	@param other rectangle to perform intersection with
	 */
	void intersect(const cRectangle& other)
	{
		MinX = max(MinX, other.MinX);
		MinY = max(MinY, other.MinY);
		MaxX = min(MaxX, other.MaxX);
		MaxY = min(MaxY, other.MaxY);
	}

	/**
	 *	Tells whether rectangle is empty or not.
	 *	Empty rectangle is such that MinX > MaxX or
	 *	MinY > MaxY.
	 */
	bool isEmpty() const
	{
		return MinX > MaxX || MinY > MaxY;
	}

	/*----------------- Accessors --------------------*/

	inline float getMinX() const { return MinX; }
	inline float getMinY() const { return MinY; }
	inline float getMaxX() const { return MaxX; }
	inline float getMaxY() const { return MaxY; }

	inline float getWidth() const { return MaxX - MinX + 1; }
	inline float getHeight() const { return MaxY - MinY + 1; }

protected:

	float MinX, MinY;
	float MaxX, MaxY;
};

#endif //_MATH_H_