/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _MESH_H_
#define _MESH_H_

// Predefinitions
class cMaterial;

/**
 *	Very simple mesh representation.
 *
 *	This mesh for educational purposes contains only what is really
 *	required for displaying simple mesh, that is:
 *	@li vertex coordinates
 *	@li texture coordinates
 *	@li texture id (OpenGL identifier)
 *	@li transparency (tells how transparent is mesh)
 */
class cMesh
{
	friend class cScene;
	friend class cSimpleScene;

public:

	/**
	 *	Default constructor.
	 */
	cMesh();

	/**
	 *	Renders mesh.
	 */
	void render() const;

	/**
	 *	Renders mesh using vertex coordinates only (no normals,
	 *	texture coords, etc.).
	 *	This rendering method is used to fill depth buffer
	 *	or stencil buffer with some values.
	 */
	void renderSimple() const;

	/**
	 *	Recalculates bounding volume.
	 */
	void recalculateBoundingVolume();

	/**
	 *	Flips triangles orientation.
	 *	Triangle (1,2,3) turns into (1,3,2). moreover all normal vectors get flipped.
	 */
	void flipTriangles();

	/**
	 *	Scales texture coords.
	 *	@param scale scale factor to multiply all texture coordinates by
	 */
	void scaleTextureCoords(float scale);

	/**
	 *	Comparison function used to sort meshes right before rendering.
	 *	@param a first mesh to compare
	 *	@param b second mesh to compare
	 *	@return true if mesh a should stand before mesh b in order,
	 *		false otherwise
	 */
	static bool greater(cMesh* a, cMesh* b);

	/*--------------- Accessors ----------------*/

	inline const cMaterial* getMaterial() const { return Material; }
	inline const vector<unsigned short>& getIndices() const { return Indices; }
	inline vector<cVec3>& getCoords() const { return (vector<cVec3>&) Coords; }
	inline const vector<cVec2>& getTexCoords() const { return TexCoords; }
	inline const cBoundingVolume& getBoundingVolume() const { return BoundingVolume; }

protected:

	/// Material used by this mesh
	cMaterial* Material;

	/// Mesh indices (lookup into Coords[] and TexCoords[])
	vector<unsigned short> Indices;

	/// Mesh vertex coordinates
	vector<cVec3> Coords;

	/// Texture coordinates
	vector<cVec2> TexCoords;

	/// Bounding volume of the mesh
	cBoundingVolume BoundingVolume;

};

/// Array of meshes
typedef vector<cMesh*> cMeshArray;

#endif //_MESH_H_