/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#include "common.h"

#include "Application.h"
#include "Frustum.h"
#include "Material.h"
#include "Mirror.h"
#include "Mesh.h"
#include "Camera.h"
#include "Scene.h"
#include "Renderer.h"

/*------------- Render stats --------------*/

cRenderer::cRenderStats cRenderer::RenderStats;

void cRenderer::cRenderStats::reset()
{
	NumRenderedMeshes = 0;
	NumRenderedMirrors = 0;
}

/*------------- Render mode --------------*/

cRenderer::cRenderMode cRenderer::RenderMode;

void cRenderer::cRenderMode::setDefaults()
{
	MaxReflectionLevel = 10;
	MinOpaqueness = 0.9f;
	IsWireMode = false;
	EnableScissorTest = true;
	EnableFrustumIntersection = true;
	EnableMirrorClipping = true;
	TestBoundingBox = true;
}

/*------------- Renderer --------------*/

cRenderer::cRenderer(cApplication* app) :
	App(app),
	Scene(NULL)
{
	// Set up basic OpenGL state
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glClearStencil(0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	// Set default render mode
	RenderMode.setDefaults();
}

cRenderer::~cRenderer()
{
}

void cRenderer::renderScene(const cCamera* camera, cScene* scene)
{
	Scene = scene;

	// Set viewer position and view direction
	camera->setView();

	// Clear color, depth and stencil buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	if (RenderMode.EnableScissorTest)
		glEnable(GL_SCISSOR_TEST);
	else
		glDisable(GL_SCISSOR_TEST);

	if (!RenderMode.IsWireMode)
		glEnable(GL_STENCIL_TEST);

	// Render grid (for testing purposes only)
//	this->renderGrid();

	// Reset renderer statistics
	RenderStats.reset();

	// Enable vertex arrays' modes
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	// Start rendering from level 0
	// (level 0 corresponds with geometry we can see directly)
	this->renderReflectionLevel(0, cFrustum(camera), NULL, 0);

	// Disable vertex arrays' modes
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

void cRenderer::renderReflectionLevel(int level, const cFrustum& frustum, const cMirror* mirror,
									  float currentOpaqueness)
{
	// Don't go into deeper reflection level than the maximal allowed
	if (level < RenderMode.MaxReflectionLevel && currentOpaqueness < RenderMode.MinOpaqueness)
	{
		unsigned int i;
		vector<cMirror*> visibleMirrors;

		// Figure out visible mirrors
		Scene->cullVisibleMirrors(frustum, visibleMirrors);

		// Process all mirrors (except the one we render reflection in)
		for (i = 0; i < visibleMirrors.size(); i++)
			if (visibleMirrors[i] != mirror)
				renderMirror(visibleMirrors[i], level, frustum, mirror, currentOpaqueness);
	}

	if (level > 0)
	{
		// For higher reflection levels than 0'th set stencil
		// test to pass only for values equal to 'level'
		glStencilFunc(GL_EQUAL, level, ~0);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

		// Enable mirror clipping (except for when there's no mirror)
		setMirrorClipping(mirror);
	}
	else
	{
		// Disable stencil test at first reflection level.
		// Since most of the geometry is usually visible on 0th
		// reflection level (not reflected in any mirror), disabling
		// stencil test is a slight optimization.
		glDisable(GL_STENCIL_TEST);

		// Disable near clipping plane
		glDisable(GL_CLIP_PLANE0);
	}

	// Set scissor test
	setScissorTest(frustum);

	// Enable depth and color writes (it might have been disabled by stencil passes)
	glDepthMask(GL_TRUE);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	// Set appropriate face culling
	glCullFace(level % 2 == 0 ? GL_FRONT : GL_BACK);

	// Render scene meshes except current mirror (if any) and without
	// reflections in remaining ones
	renderMeshes(level, frustum, mirror, currentOpaqueness);
}

void cRenderer::renderMirror(const cMirror* mirrorToRender, int level,
							 const cFrustum& frustum, const cMirror* mirror,
							 float currentOpaqueness)
{
	// Side test for mirror
	if (!frustum.testPlaneSide( mirrorToRender->getPlane() ))
		return;

	// Copy current frustum
	cFrustum newFrustum = frustum;

	// Intersect current frustum with frustum built
	// using mirror's portal (represented as quad)
	// ...and...
	// intersect current scissor rectangle with bounding
	// screen space rectangle of the mirror portal
	if (!newFrustum.intersect( mirrorToRender ))
		// Intersected frustum or scissor rectangle was empty,
		// so proceed to the next mirror
		return;

	// Update render statistics
	RenderStats.NumRenderedMirrors++;

	if (!RenderMode.IsWireMode)
	{
		// Set scissor test and mirror clipping
		setScissorTest(frustum);
		setMirrorClipping(mirror);

		// Increment stencil values under mirror polygons.
		// 1. Do it only inside previous reflection's region
		// 2. Disable depth writes
		// 3. Disable color writes
		// 4. Set approproate face culling
		// 5. Render mirror polygons (increment stencil values)
		glStencilFunc(GL_EQUAL, level, ~0);
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
		glDepthFunc(GL_LESS);
		glDepthMask(GL_FALSE);
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		glCullFace(level % 2 == 0 ? GL_FRONT : GL_BACK);

		mirrorToRender->getMesh()->renderSimple();

		// Set depth values under mirror polygons to maximum.
		// 1. Do it only for pixels tagged right before
		// 2. Don't change stencil buffer bits
		// 3. Enable depth writes
		// 4. Set depth range to always output 1 (max. allowed)
		// 5. Set depth test to always pass (only stencil test
		//	  can prevent us from writing to z-buffer)
		// 6. Render mirror polygons (set depth values to max)
		// 7. Turn back default depth range and depth comparison
		//    function
		glStencilFunc(GL_EQUAL, level + 1, ~0);
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
		glDepthMask(GL_TRUE);
		glDepthRange(1, 1);
		glDepthFunc(GL_ALWAYS);

		mirrorToRender->getMesh()->renderSimple();

		glDepthRange(0, 1);
		glDepthFunc(GL_LESS);
	}

	// Calculate reflection transformation from mirror's plane
	cMat4x4 matrix;
	matrix.buildReflectionTransformation( mirrorToRender->getPlane() );

	// Reflect frustum on mirror's plane
	newFrustum.transform(matrix);

	// Apply reflection transformation
	glPushMatrix();
	glMultMatrixf(matrix);

	// Calculate new opaqueness factor
	const float newOpaqueness =
		currentOpaqueness +
		(1.0f - currentOpaqueness) * 
		(1.0f - mirrorToRender->getMesh()->getMaterial()->getTransparency());

	// Render reflected scene (recursive call)
	// =======================================
	this->renderReflectionLevel(level + 1, newFrustum, mirrorToRender, newOpaqueness);
	// =======================================

	// Undo reflection transformation
	glPopMatrix();

	if (!RenderMode.IsWireMode)
	{
		// Set scissor test and mirror clipping
		setScissorTest(frustum);
		setMirrorClipping(mirror);

		// Decrement stencil values and fill z-buffer
		// under mirror polygons.
		// 1. Do it only for pixels under mirror polygons
		// 2. Set depth function to always pass
		// 3. Disable color writes
		// 4. Set appropriate face culling
		// 5. Render mirror polygons (decrement stencil values
		//    and write to depth buffer)
		// 6. Turn back default depth comparison function
		glStencilFunc(GL_EQUAL, level + 1, ~0);
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
		glDepthFunc(GL_ALWAYS);
		glDepthMask(GL_TRUE);
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		glCullFace(level % 2 == 0 ? GL_FRONT : GL_BACK);

		mirrorToRender->getMesh()->renderSimple();

		glDepthFunc(GL_LESS);
	}
}

void cRenderer::renderMeshes(int level, const cFrustum& frustum, const cMirror* mirrorToRender,
							 float currentOpaqueness)
{
	unsigned int i;
	const cMaterial* currentMaterial = NULL;

	// Figure out if this is the last reflection level
	const bool isLastLevel =
		level == RenderMode.MaxReflectionLevel ||
		currentOpaqueness >= RenderMode.MinOpaqueness;

	// Find visible meshes
	Scene->cullVisibleMeshes(frustum, MeshBuffer);

	// Do some state sorting optimization:
	// - first mirror meshes
	// - then opaque meshes
	// - then transparent meshes (back to front order)
	std::sort(MeshBuffer.begin(), MeshBuffer.end(), cMesh::greater);

	// Render meshes one by one
	for (i = 0; i < MeshBuffer.size(); i++)
	{
		const cMesh* currentMesh = MeshBuffer[i];

		// Don't render mesh belonging to mirror we're rendering to
		if (mirrorToRender != NULL && mirrorToRender->getMesh() == currentMesh)
			continue;

		// Switch current material if needed
		if (currentMesh->getMaterial() != currentMaterial)
		{
			currentMaterial = currentMesh->getMaterial();
			currentMaterial->setRenderState(isLastLevel);
		}

		// Render mesh
		currentMesh->render();
		RenderStats.NumRenderedMeshes++;
	}

	// Empty mesh buffer
	MeshBuffer.clear();
}

void cRenderer::setMirrorClipping(const cMirror* mirror)
{
	if (!RenderMode.EnableMirrorClipping)
		return;

	if (mirror == NULL)
	{
		glDisable(GL_CLIP_PLANE0);
		return;
	}
	else
		glEnable(GL_CLIP_PLANE0);

	const double plane[4] = {
		mirror->getPlane().getNormal()[0],
		mirror->getPlane().getNormal()[1],
		mirror->getPlane().getNormal()[2],
		-mirror->getPlane().getOffset()};
	glClipPlane(GL_CLIP_PLANE0, plane);
}

void cRenderer::setScissorTest(const cFrustum& frustum)
{
	if (!RenderMode.EnableScissorTest)
		return;

	const cRectangle& rect = frustum.getScreenSpaceBoundingRectangle();

	// Apply 1 pixel extended (in left, right, top & bottom) scissor test
	// because of raster precision rounding issues (when projecting
	// mirror portal's vertices)
	glScissor(
		(GLint) rect.getMinX() - 1,
		(GLint) rect.getMinY() - 1,
		(GLint) rect.getWidth() + 2,
		(GLint) rect.getHeight() + 2);
}

void cRenderer::renderGrid()
{
	const float gridSize = 1;
	const int numCells = 10;

	glDisable(GL_TEXTURE_2D);
	glColor3f(1, 1, 1);

	glBegin(GL_LINES);
	
	for (int i = 0; i <= numCells; i++)
	{
		// Line along Z axis
		glVertex3f(i * gridSize, 0, 0);
		glVertex3f(i * gridSize, 0, numCells * gridSize);
		
		// Line along X axis
		glVertex3f(0, 0, i * gridSize);
		glVertex3f(numCells * gridSize, 0, i * gridSize);
	}

	// Line along Y axis
	glVertex3f(0, 0, 0);
	glVertex3f(0, numCells * gridSize, 0);

	glEnd();
}