/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _RENDERER_H_
#define _RENDERER_H_

// Predefinitions
class cApplication;
class cCamera;
class cScene;
class cLight;
class cMesh;

/**
 *	Renderer is responsible for rendering scene and is capable of
 *	rendering multi-level mirror reflections.
 */
class cRenderer
{
public:

	/**
	 *	Constructor.
	 *	@param app application this renderer is to work for
	 */
	cRenderer(cApplication* app);

	/**
	 *	Destructs renderer.
	 */
	virtual ~cRenderer();

	/**
	 *	Renders given scene seen from given camera.
	 *	@param camera camera to render from
	 *	@param scene scene to be rendered
	 */
	void renderScene(const cCamera* camera, cScene* scene);

protected:

	/**
	 *	Renders requested reflection level.
	 *	This is probably the most important method in this whole demo program.
	 *
	 *	@param frustum frustum to render scene in
	 *	@param level level to be rendered
	 *	@param mirror this reflection is to be rendered in or NULL
	 *		in case this is initial 0-th reflection level
	 *	@param currentOpaqueness current opaqueness calculated based on
	 *		all previously visited mirrors along current recursion stack
	 */
	void renderReflectionLevel(int level, const cFrustum& frustum, const cMirror* mirror, float currentOpaqueness);

	/**
	 *	Renders appropriate values to stencil buffer and clears depth buffer
	 *	under mirror surface; then recursively renders scene reflected in this
	 *	mirror and finally writes appropriate depth values under mirror surface.
	 *
	 *	@param mirrorToRender currently processed mirror
	 *	@param frustum frustum to render scene in
	 *	@param level level to be rendered
	 *	@param mirror this reflection is to be rendered in or NULL
	 *		in case this is initial 0-th reflection level
	 *	@param currentOpaqueness current opaqueness calculated based on
	 *		all previously visited mirrors along current recursion stack
	 */
	void renderMirror(const cMirror* mirrorToRender, int level,
							 const cFrustum& frustum, const cMirror* mirror,
							 float currentOpaqueness);

	/**
	 *	Renders all meshes visible in frustum as a reflection in given mirror.
	 *
	 *	@param level current reflection level
	 *	@param frustum frustum to cull geometry within
	 *	@param mirror mirror this reflection is to be rendered in or NULL
	 *		in case this is initial 0-th reflection level; we pass it only to
	 *		preserve rendering of this mirror's mesh - we obviously don't want
	 *		this to happen (we want to render everything except for mirror mesh)
	 *	@param currentOpaqueness current opaqueness factor
	 */
	void renderMeshes(int level, const cFrustum& frustum, const cMirror* mirror, float currentOpaqueness);

	/**
	 *	Sets mirror clipping that is enables additional clipping plane equal
	 *	to mirror's plane.
	 *	@param mirror mirror whose clipping plane is to be set; NULL indicates
	 *		that mirror clipping shall be disabled
	 */
	void setMirrorClipping(const cMirror* mirror);

	/**
	 *	Sets scissor test for a given frustum.
	 *	@param frustum frustum for which scissor test should be enabled
	 */
	void setScissorTest(const cFrustum& frustum);

	/**
	 *	Renders squared grid - used for debug purposes.
	 */
	void renderGrid();

	/*------------- Attributes --------------*/

	/// Application this renderer works for
	cApplication* App;

	/// Currently rendered scene
	cScene* Scene;

	/// Buffer for meshes to be rendered
	vector<cMesh*> MeshBuffer;

public:

	/*------------- Adjustable rendering mode --------------*/

	/**
	 *	Render mode description.
	 *	All of the factors here are adjustable from user interface.
	 */
	struct cRenderMode
	{
		/**
		 *	Sets default render mode - best optimized.
		 */
		void setDefaults();

		/// Maximal allowed reflection level to be rendered (0 is the first)
		int MaxReflectionLevel;

		/// Maximal reflection opaqueness (value within 0..1)
		float MinOpaqueness;

		/// Wire mode (by default set to false)
		bool IsWireMode;

		/// Enable scissor test?
		bool EnableScissorTest;

		/// Enable frustum intersection?
		bool EnableFrustumIntersection;

		/// Enable OpenGL (per-pixel) mirror plane clipping?
		bool EnableMirrorClipping;

		/// Tells whether frustum tests with bounding volumes are
		/// performed using as with bounding box; otherwise they're performed
		/// as with bounding spheres
		bool TestBoundingBox;

	};

	/// Rendering mode
	static cRenderMode RenderMode;

	/*------------- Rendering statistics --------------*/

	/**
	 *	Per-frame rendering statistics.
	 */
	struct cRenderStats
	{
		/**
		 *	Resets statistics for next frame.
		 */
		void reset();

		/// Number of meshes rendered last frame
		int NumRenderedMeshes;

		/// Number of mirrors rendered (visible) last frame
		int NumRenderedMirrors;
	};

	/// Rendering statistics
	static cRenderStats RenderStats;

};

#endif //_RENDERER_H_