/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _SCENE_H_
#define _SCENE_H_

// Predefinitions
class cMesh;
class cMaterial;
class cMirror;
class cApplication;

typedef vector<cMesh*> cMeshArray;
typedef vector<cMaterial*> cMaterialArray;
typedef vector<cMirror*> cMirrorArray;

/**
 *	Abstract scene consisting of meshes and mirrors.
 */
class cScene
{
public:

	/**
	 *	Constructor.
	 *	@param app application
	 */
	cScene(cApplication* app);

	/**
	 *	Empty destructor.
	 */
	virtual ~cScene() {}

	/*------------- Abstract operations -------------*/

	/**
	 *	Loads scene from file of given name.
	 *	@param fileName scene file name
	 *	@return true on success, false otherwise
     */
	virtual bool load(const string& fileName) = 0;

	/**
	 *	Empties scene from any materials, lights, meshes and mirrors.
	 */
	virtual void empty() = 0;

	/**
	 *	Adds material to the scene.
	 *	@param material material to be added
	 */
	virtual void addMaterial(cMaterial* material) = 0;

	/**
	 *	Adds mesh to the scene.
	 *	@param mesh mesh to be addded.
	 */
	virtual void addMesh(cMesh* mesh) = 0;

	/**
	 *	Adds mirror to the scene.
	 *	@param mirror mirror to be added
	 */
	virtual void addMirror(cMirror* mirror) = 0;

	/**
	 *	Culls visible meshes.
	 *	@param frustum frustum to cull in
	 *	@param meshes culled meshes (result of culling)
	 */
	virtual void cullVisibleMeshes(const cFrustum& frustum,
		cMeshArray& meshes) = 0;

	/**
	 *	Culls visible mirrors.
	 *	@param frustum frustum to cull in
	 *	@param mirrors culled mirrors (result of culling)
	 */
	virtual void cullVisibleMirrors(const cFrustum& frustum,
		cMirrorArray& mirrors) = 0;

	/**
	 *	Returns material for name.
	 *	@param name name of the material
	 *	@return material or null if such doesn't exist
	 */
	virtual cMaterial* getMaterial(const string& name) const = 0;

protected:

	/// Application this scene was created for
	cApplication* App;

};

#endif //_SCENE_H_