/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#include "common.h"

#include "Application.h"
#include "Frustum.h"
#include "Mesh.h"
#include "Material.h"
#include "Mirror.h"
#include "Scene.h"
#include "SimpleScene.h"
#include "Renderer.h"

cSimpleScene::cSimpleScene(cApplication* app) :
	cScene(app)
{
}

cSimpleScene::~cSimpleScene()
{
	empty();
}

void cSimpleScene::addMirrorCube(const cVec3& center, float sideLength)
{
	unsigned int i;

	// Create material
	cMaterial* material = getMaterial("cube mirror");

	// Create cube
	cMirrorCube* cube = new cMirrorCube();
	cube->Center = center;
	cube->SideLength = sideLength;

	// Create 6 meshes for each side of the cube
	vector<cMesh*> meshes;
	for (i = 0; i < 6; i++)
		meshes.push_back( createQuad(cVec3(), cVec3(), cVec3(), cVec3(), material) );

	// Create mirror for each side of the cube
	for (i = 0; i < meshes.size(); i++)
	{
		this->addMesh(meshes[i]);
		
		cMirror* mirror = new cMirror(meshes[i]);
		cube->Mirrors.push_back( mirror );
		this->addMirror(mirror);
	}

	// Set up initial coordinates of the cube
	cube->update(10, 0, 0);

	// Add cube
	MirrorCubes.push_back(cube);
}

void cSimpleScene::removeMirrorCubes()
{
	unsigned int i, j, k;

	for (i = 0; i < MirrorCubes.size(); i++)
	{
		cMirrorCube* cube = MirrorCubes[i];

		for (j = 0; j < cube->Mirrors.size(); j++)
		{
			for (k = 0; k < Meshes.size(); k++)
				if (Meshes[k] == cube->Mirrors[j]->getMesh())
				{
					delete Meshes[k];
					Meshes[k] = Meshes[ Meshes.size() - 1 ];
					Meshes.resize( Meshes.size() - 1 );
					break;
				}

			for (k = 0; k < Mirrors.size(); k++)
				if (Mirrors[k] == cube->Mirrors[j])
				{
					delete Mirrors[k];
					Mirrors[k] = Mirrors[ Mirrors.size() - 1 ];
					Mirrors.resize( Mirrors.size() - 1 );
					break;
				}
		}

		delete MirrorCubes[i];
	}

	MirrorCubes.clear();
}

void cSimpleScene::cMirrorCube::update(float xRotation, float yRotation, float zRotation)
{
	unsigned int i;

	vector<cVec3> verts;
	cMat4x4 m;

	const float halfSide = SideLength * 0.5f;

	// Create non-transformed cube
	verts.resize(8);
	verts[0] = cVec3(-halfSide, -halfSide, -halfSide);
	verts[1] = cVec3(-halfSide, -halfSide, halfSide);
	verts[2] = cVec3(halfSide, -halfSide, halfSide);
	verts[3] = cVec3(halfSide, -halfSide, -halfSide);
	verts[4] = cVec3(-halfSide, halfSide, -halfSide);
	verts[5] = cVec3(-halfSide, halfSide, halfSide);
	verts[6] = cVec3(halfSide, halfSide, halfSide);
	verts[7] = cVec3(halfSide, halfSide, -halfSide);

	// Transform cube
	m.rotate(xRotation, cVec3(1, 0, 0));
	m.rotate(yRotation, cVec3(0, 1, 0));
	for (i = 0; i < verts.size(); i++)
	{
		m.transform(verts[i]);
		verts[i] += Center;
	}

	// Top cube side
	{
		vector<cVec3>& mirrorVerts = Mirrors[0]->getMesh()->getCoords();

		mirrorVerts[0] = verts[0];
		mirrorVerts[1] = verts[1];
		mirrorVerts[2] = verts[2];
		mirrorVerts[3] = verts[3];
	}

	// Bottom cube side
	{
		vector<cVec3>& mirrorVerts = Mirrors[1]->getMesh()->getCoords();

		mirrorVerts[0] = verts[7];
		mirrorVerts[1] = verts[6];
		mirrorVerts[2] = verts[5];
		mirrorVerts[3] = verts[4];
	}

	// Left cube side
	{
		vector<cVec3>& mirrorVerts = Mirrors[2]->getMesh()->getCoords();

		mirrorVerts[0] = verts[1];
		mirrorVerts[1] = verts[0];
		mirrorVerts[2] = verts[4];
		mirrorVerts[3] = verts[5];
	}

	// Right cube side
	{
		vector<cVec3>& mirrorVerts = Mirrors[3]->getMesh()->getCoords();

		mirrorVerts[0] = verts[2];
		mirrorVerts[1] = verts[1];
		mirrorVerts[2] = verts[5];
		mirrorVerts[3] = verts[6];
	}

	// Front cube side
	{
		vector<cVec3>& mirrorVerts = Mirrors[4]->getMesh()->getCoords();

		mirrorVerts[0] = verts[3];
		mirrorVerts[1] = verts[2];
		mirrorVerts[2] = verts[6];
		mirrorVerts[3] = verts[7];
	}

	// Back cube side
	{
		vector<cVec3>& mirrorVerts = Mirrors[5]->getMesh()->getCoords();

		mirrorVerts[0] = verts[0];
		mirrorVerts[1] = verts[3];
		mirrorVerts[2] = verts[7];
		mirrorVerts[3] = verts[4];
	}

	for (i = 0; i < Mirrors.size(); i++)
	{
		Mirrors[i]->getMesh()->recalculateBoundingVolume();
		Mirrors[i]->setBoundingVerts( Mirrors[i]->getMesh()->getCoords() );
	}
}

void cSimpleScene::updateMirrorCubes()
{
	unsigned int i;

	// Kind of hardcoded rotation scheme (only for educational purposes)
	static float xRotation = 0;
	static float yRotation = 0;
	
	const int time = SDL_GetTicks();

	xRotation = ((float) time) / 7000;
	yRotation = ((float) time) / 15000;

	for (i = 0; i < MirrorCubes.size(); i++)
		MirrorCubes[i]->update(xRotation, yRotation, 0);
}

void cSimpleScene::buildDemoScene()
{
	cout << "Building demo scene...\n";

	unsigned int i;
	vector<cMaterial*> materials;
	vector<cMesh*> meshes;
	vector<cMirror*> mirrors;

	// Create materials
	materials.push_back( new cMaterial("wooden wall", "../media/textures/wood.bmp", 0) );
	materials.push_back( new cMaterial("bricks wall", "../media/textures/bricks.bmp", 0) );
	materials.push_back( new cMaterial("floor mirror", "../media/textures/flower_motif_tiles.bmp", 0.4f, true) );
	materials.push_back( new cMaterial("side mirror", "../media/textures/clouds.bmp", 0.7f/*0.5f*/, true) );
	materials.push_back( new cMaterial("ceiling mirror", "../media/textures/art_noveau_glasswork.bmp", 0.5f, true) );
	materials.push_back( new cMaterial("cube mirror", "../media/textures/clouds.bmp", 0.3f/*0.5f*/, true) );

	// Create mirror meshes
	meshes.push_back( createQuad(cVec3(0, 0, 0), cVec3(10, 0, 0), cVec3(10, 0, 10), cVec3(0, 0, 10), materials[2]) );
	meshes.push_back( createQuad(cVec3(0, 10, 0), cVec3(0, 10, 10), cVec3(10, 10, 10), cVec3(10, 10, 0), materials[4]) );

	meshes.push_back( createQuad(cVec3(0.4f, 0.5f, 1), cVec3(0.4f, 0.5f, 7), cVec3(0.4f, 0.5f + 4, 7), cVec3(0.4f, 0.5f + 4, 1), materials[3]) );
	meshes.push_back( createQuad(cVec3(9.6f, 0.5f, 3), cVec3(9.6f, 0.5f + 4, 3), cVec3(9.6f, 0.5f + 4, 7), cVec3(9.6f, 0.5f, 7), materials[3]) );
	meshes.push_back( createQuad(cVec3(3, 0.5f, 0.4f), cVec3(3, 0.5f + 4, 0.4f), cVec3(7, 0.5f + 4, 0.4f), cVec3(7, 0.5f, 0.4f), materials[3]) );
	meshes.push_back( createQuad(cVec3(3, 0.5f, 9.6f), cVec3(7, 0.5f, 9.6f), cVec3(7, 0.5f + 4, 9.6f), cVec3(3, 0.5f + 4, 9.6f), materials[3]) );

	// Create mirrors in the scene (and associate geometry with them)
	for (i = 0; i < meshes.size(); i++)
		mirrors.push_back( new cMirror(meshes[i]->getCoords(), meshes[i]) );

	// Create regular meshes
	meshes.push_back( createQuad(cVec3(0, 0, 10), cVec3(0, 10, 10), cVec3(0, 10, 0), cVec3(0, 0, 0), materials[0]) );
	meshes.push_back( createQuad(cVec3(10, 0, 0), cVec3(10, 10, 0), cVec3(10, 10, 10), cVec3(10, 0, 10), materials[1]) );
	meshes.push_back( createQuad(cVec3(0, 0, 0), cVec3(0, 10, 0), cVec3(10, 10, 0), cVec3(10, 0, 0), materials[0]) );
	meshes.push_back( createQuad(cVec3(10, 0, 10), cVec3(10, 10, 10), cVec3(0, 10, 10), cVec3(0, 0, 10), materials[1]) );

	meshes.push_back( createCube(cVec3(3, 1.5f, 7.5f), cVec3(2, 1, 1), materials[0]) );
	meshes.push_back( createCube(cVec3(8, 0.5f, 2), cVec3(2, 1, 2), materials[0]) );
	meshes.push_back( createCube(cVec3(3, 0.5f, 7), cVec3(1, 1, 2), materials[0]) );

	// Add all objects to the scene
	for (i = 0; i < materials.size(); i++)
		this->addMaterial(materials[i]);
	for (i = 0; i < meshes.size(); i++)
		this->addMesh(meshes[i]);
	for (i = 0; i < mirrors.size(); i++)
		this->addMirror(mirrors[i]);

	cout << "Demo scene built succesfully...\n";
}

cMesh* cSimpleScene::createCube(const cVec3& center, const cVec3& size, cMaterial* material)
{
	const cVec3 offset = center - size * 0.5f;

	// 8 cube's vertex coordinates
	vector<cVec3> coords;
	coords.push_back(cVec3(0,		0,		 0)		  + offset);
	coords.push_back(cVec3(0,		0,		 size[2]) + offset);
	coords.push_back(cVec3(size[0], 0,		 size[2]) + offset);
	coords.push_back(cVec3(size[0], 0,		 0)		  + offset);
	coords.push_back(cVec3(0,		size[1], 0)		  + offset);
	coords.push_back(cVec3(0,		size[1], size[2]) + offset);
	coords.push_back(cVec3(size[0], size[1], size[2]) + offset);
	coords.push_back(cVec3(size[0], size[1], 0)		  + offset);

	cMesh* mesh = new cMesh();
	mesh->Material = material;
	
	addQuad(mesh, coords[0], coords[1], coords[2], coords[3]);
	addQuad(mesh, coords[7], coords[6], coords[5], coords[4]);
	addQuad(mesh, coords[0], coords[4], coords[5], coords[1]);
	addQuad(mesh, coords[1], coords[5], coords[6], coords[2]);
	addQuad(mesh, coords[2], coords[6], coords[7], coords[3]);
	addQuad(mesh, coords[3], coords[7], coords[4], coords[0]);
	
	return mesh;
}

cMesh* cSimpleScene::createQuad(const cVec3& coord0,
		const cVec3& coord1, const cVec3& coord2, const cVec3& coord3,
		cMaterial* material)
{
	cMesh* mesh = new cMesh();
	mesh->Material = material;
	addQuad(mesh, coord0, coord1, coord2, coord3);
	return mesh;
}

void cSimpleScene::addQuad(cMesh* mesh, const cVec3& coord0,
		const cVec3& coord1, const cVec3& coord2, const cVec3& coord3)
{
	const int numVertsBefore = (int) mesh->Coords.size();

	mesh->Indices.push_back(numVertsBefore + 0);
	mesh->Indices.push_back(numVertsBefore + 1);
	mesh->Indices.push_back(numVertsBefore + 2);
	mesh->Indices.push_back(numVertsBefore + 0);
	mesh->Indices.push_back(numVertsBefore + 2);
	mesh->Indices.push_back(numVertsBefore + 3);

	mesh->Coords.push_back(coord0);
	mesh->Coords.push_back(coord1);
	mesh->Coords.push_back(coord2);
	mesh->Coords.push_back(coord3);

	mesh->TexCoords.push_back(cVec2(0, 0));
	mesh->TexCoords.push_back(cVec2(0, 1));
	mesh->TexCoords.push_back(cVec2(1, 1));
	mesh->TexCoords.push_back(cVec2(1, 0));
}

bool cSimpleScene::load(const string& fileName)
{
	return false;
}

void cSimpleScene::empty()
{
	unsigned int i;

	for (i = 0; i < Materials.size(); i++)
		delete Materials[i];
	Materials.clear();

	for (i = 0; i < Meshes.size(); i++)
		delete Meshes[i];
	Meshes.clear();
	
	for (i = 0; i < Mirrors.size(); i++)
		delete Mirrors[i];
	Mirrors.clear();
}

void cSimpleScene::addMaterial(cMaterial* material)
{
	Materials.push_back(material);
}

void cSimpleScene::addMesh(cMesh* mesh)
{
	mesh->recalculateBoundingVolume();
	Meshes.push_back(mesh);
}

void cSimpleScene::addMirror(cMirror* mirror)
{
	Mirrors.push_back(mirror);
}

void cSimpleScene::cullVisibleMeshes(const cFrustum& frustum,
		cMeshArray& meshes)
{
	unsigned int i;

	// Perform per-vertex culling test for mirror surfaces
	// and regular bounding box/sphere culling for any other mesh
	for (i = 0; i < Meshes.size(); i++)
		if (Meshes[i]->getMaterial()->isMirrorSurface())
		{
			// It's important in our case to perform the same kind of frustum test 
			// as done in cullVisibleMirrors() method. Of course there is a little
			// redundancy in that we cull the same mirrors twice (once as a mirror from
			// cullVisibleMirrors() and once as a mesh from cullVisibleMeshes()), but
			// if we took care of all such aspects of this demo, we would have ended in
			// something close to a graphics engine, which would hide the main idea
			// behind this demo.
			if (frustum.testVisibilityPrecisely( Meshes[i]->getBoundingVolume(), Meshes[i]->getCoords() ))
				meshes.push_back(Meshes[i]);
		}
		else if (frustum.testVisibility( Meshes[i]->getBoundingVolume() ))
			meshes.push_back(Meshes[i]);
}

void cSimpleScene::cullVisibleMirrors(const cFrustum& frustum,
	cMirrorArray& mirrors)
{
	unsigned int i;
	for (i = 0; i < Mirrors.size(); i++)
		if (frustum.testVisibilityPrecisely( Mirrors[i]->getBoundingVolume(), Mirrors[i]->getMesh()->getCoords() ))
			mirrors.push_back(Mirrors[i]);
}

cMaterial* cSimpleScene::getMaterial(const string& name) const
{
	unsigned int i;
	for (i = 0; i < Materials.size(); i++)
		if (Materials[i]->getName() == name)
			return Materials[i];
	return NULL;
}