/*
 *	This file is part of demo program made to accompany
 *	"More OpenGL Game Programming" book's chapter
 *	titled "Robust Multi-Level Reflections".
 *
 *	Author: Maciej Sawitus, 2005
 *	E-mail: msawitus@gmail.com
 *
 */

#ifndef _SIMPLE_SCENE_H_
#define _SIMPLE_SCENE_H_

/**
 *	Simple scene implementation.
 *	<p>
 *	Simple scene is really simple. It stores all of its meshes on a list,
 *	which definitely is not good for spatial optimizations, but is much
 *	easier for educational purposes than implementing say octree based scene.
 *	<p>
 *	If you like an idea of abstract scene class and would like to implement
 *	your own optimized octree scene version, you're free to do so :-).
 */
class cSimpleScene : public cScene
{
public:

	cSimpleScene(cApplication* app);

	~cSimpleScene();

	/**
	 *	Builds demo sceen designed to especially to show reflections.
	 */
	void buildDemoScene();

	/**
	 *	Adds mirror cube to the scene.
	 *	@param center coordinates of the center of the cube
	 *	@param sideLength length of the cube side
	 */
	void addMirrorCube(const cVec3& center, float sideLength);

	/**
	 *	Removes all mirror cubes from the scene.
	 */
	void removeMirrorCubes();

	/**
	 *	Updates mirror cubes - rotations based on current miliseconds.
	 */
	void updateMirrorCubes();

	/*------------- Implementation of cScene -------------*/

	virtual bool load(const string& fileName);

	virtual void empty();

	virtual void addMaterial(cMaterial* material);

	virtual void addMesh(cMesh* mesh);

	virtual void addMirror(cMirror* mirror);

	virtual void cullVisibleMeshes(const cFrustum& frustum,
		cMeshArray& meshes);

	virtual void cullVisibleMirrors(const cFrustum& frustum,
		cMirrorArray& mirrors);

	virtual cMaterial* getMaterial(const string& name) const;

protected:

	/*------------- Utils -------------*/

	/**
	 *	Helper function for creating cube mesh.
	 */
	cMesh* createCube(const cVec3& center, const cVec3& size, cMaterial* material);

	/**
	 *	Helper function for creating mesh consisting of a single quad.
	 */
	cMesh* createQuad(const cVec3& coord0, const cVec3& coord1,
		const cVec3& coord2, const cVec3& coord3, cMaterial* material);

	/**
	 *	Helper function capable of adding single quad to given mesh.
	 */
	void addQuad(cMesh* mesh, const cVec3& coord0,
		const cVec3& coord1, const cVec3& coord2, const cVec3& coord3);

	/*------------- Attributes -------------*/

	/// Materials used in this scene
	vector<cMaterial*> Materials;

	/// Meshes in the scene
	vector<cMesh*> Meshes;

	/// Mirrors in the scene
	vector<cMirror*> Mirrors;

	/**
	 *	Mirror cube representation.
	 */
	class cMirrorCube
	{
	public:

		/**
		 *	Updates rotation of the mirror cube.
		 *	@param xRotation rotation around x axis
		 *	@param yRotation rotation around y axis
		 *	@param zRotation rotation around z axis
		 */
		void update(float xRotation, float yRotation, float zRotation);

		cVec3 Center;				///< Center of the cube
		float SideLength;			///< Length of the cube side
		vector<cMirror*> Mirrors;	///< 6 mirrors on cube's sides (each having its mesh)
	};

	/// Reflective cubes in the scene
	vector<cMirrorCube*> MirrorCubes;

};

#endif //_SIMPLE_SCENE_H_
