/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
-------------------------------------------------------------------------
| dd/mm/yyyy  |  Programmer        | Explanation of change              |
-------------------------------------------------------------------------
| 09/10/2005  |  J.Villagrana   @1 | -Added libraries links, texture    | 
|             |                    |  manager singleton def and a bunch |
|             |                    |  of changes(MGLApp,Render,Update..)| 
| 01/11/2005  |  J.Villagrana   @2 | -Changed hardcoded default psystems| 
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"

// ----------------------------@1A BEGINS ------------------------------
#include "./Texture/textureManager.h"

// SEARCH FOR LIBRARIES AT COMPILE TIME
// OPENGL LIBRARIES
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")
#pragma comment(lib, "glaux.lib")
// LIBRARY FOR timeGetTime
#pragma comment(lib, "winmm.lib")
// LIBRARY FOR checking OpenGL extensions
#pragma comment(lib, "glee.lib")

// EXPLICITLY DEFINE 
CTextureManager*	CTextureManager::m_pGTextureManager;
// ----------------------------@1A ENDS --------------------------------

MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	if (!GLeeInit())
		return false;

	// ****************** @1A BEGINS *************************
	// --------------------
	// Init texture manager	
	// --------------------
	CTextureManager::SetTextureManager(new CTextureManager);
	if(!_TextureManager())
	{
		MessageBox(NULL, "Cannot create texture manager class instance", "MGLApp::Init", MB_OK|MB_ICONSTOP);
		return false;
	}	

	// -----------
	// Init script
	// -----------
	m_pScript = new CParticleScript;
	assert(m_pScript!=NULL);
	if(m_pScript==NULL)
		return false;

	// ---------------------
	// Load and parse script
	// ---------------------
	m_currentScript = 0;
	char scriptFilename[30];
	memset(scriptFilename,0,sizeof(scriptFilename));
	memset(m_scriptTitle, 0, sizeof(m_scriptTitle)); //@2A
	// memcpy(scriptFilename, particle_scripts[m_currentScript], sizeof(scriptFilename)); @2D
	memcpy(scriptFilename, "data/test1.psys", 15); //@2A
	memcpy(m_scriptTitle, "test1.psys", 10); //@2A
	if(!(LoadScript(scriptFilename)))
	{
		m_pScript->DestroyAll();
		delete m_pScript;
		m_pScript = NULL;
		return false;
	}

	// --------------------
	// Init particle system
	// --------------------
	CParticleSys_Scriptable *newObjScr=NULL;	
	m_pParticleSystem  = new CParticleSys_Scriptable();
	if(m_pParticleSystem != NULL)
	{
  		newObjScr = (CParticleSys_Scriptable *)m_pParticleSystem;	
		newObjScr->AttachScript(m_pScript);
		BOOL scriptApplied = newObjScr->ApplyScript();
		if(!scriptApplied)
		{
			m_pParticleSystem->DestroyAll();
			delete m_pParticleSystem;
			m_pParticleSystem = NULL;
			return false;
		}
	}
	else
		return false;	

	// --------------------
	// Create font class
	// --------------------
	if(!(m_font.CreateGLFont("Data/Images/tx_font.tga")))
		return false;

	// --------------------
	// Init Timer
	// --------------------
	if(!(m_timer.Init()))
		return false;	
	// ****************** @1A ENDS *************************

	return true;
}

bool MGLApp::Shutdown()
{	
	// ****************** @1A BEGINS *************************
	// Cleanup texture manager
	if(_TextureManager())
	{
		_TextureManager()->DestroyAll();
		delete CTextureManager::GetTextureManager();
	}

	// Cleanup particle system
	// (particle system takes care of cleaning up script memory)
	if(m_pParticleSystem!=NULL)
	{
		m_pParticleSystem->DestroyAll();
		delete m_pParticleSystem;
	}

	// Cleanup font
	m_font.DestroyAll();
	// ****************** @1A ENDS *************************	

	return true;
}

// *********** @1A BEGINS *************
bool MGLApp::LoadScript(char *filename)
{
	FILE *pFile=NULL;
	int numRead=0, i=0, j=0, scriptLen=0, lineLength=0;
	char c[2], oneLine[100], scriptText[8000], *buffer=NULL;
	bool validScript = false, ScriptOK = false;	
	
	// -----------------------------
	// Read the script from the file
	// -----------------------------		
	if((pFile = fopen(filename, "r")) == NULL)
	{
		MessageBox(NULL,"Cannot open Script input file", "MGLApp::LoadScript", MB_OK);
		return false;
	}
	else
	{	
		memset(scriptText, 0, sizeof(scriptText));
		memset(c, 0, sizeof(c));
		numRead = fread(c, sizeof(char), 1, pFile);		

		while(numRead>0)
		{
			memset(oneLine,0,sizeof(oneLine));
			i=0;

			// GET ALL LINES
			while((numRead>0) && (c[0]!='\n') && (c[0]!='\r') && (i<98))
			{
				oneLine[i]=c[0];					
				i++;
				memset(c, 0, sizeof(c));
				numRead = fread(c, sizeof(char), 1, pFile);
			}			
			oneLine[i++]=13;
			oneLine[i++]=10;
			oneLine[i]='\0';
			
			scriptLen = strlen(scriptText);			
			if(scriptLen < 7900)
				memcpy(&scriptText[scriptLen], oneLine, strlen(oneLine));			

			if(numRead>0)
			{
				memset(c, 0, sizeof(c));
				numRead = fread(c, sizeof(char), 1, pFile);
			}
		}

		fclose(pFile);
	} // end file open ok	

	// -----------------------------
	// Add all lines to the script
	// -----------------------------
	m_pScript->DestroyAll();
	lineLength = m_pScript->GetScriptStringLength();
	assert(lineLength>0);
	buffer = new char[lineLength];
	assert(buffer != NULL && "LoadScript()");
	if(buffer!=NULL)
	{			
		scriptLen = strlen(scriptText);	
		memset(buffer,0,sizeof(char)*lineLength);

		// COPY ALL LINES			
		for(i=0; i<scriptLen; i++)
		{
			c[0] = scriptText[i];
			if((c[0]!=10) && (c[0]!='\0'))
			{
				if((c[0]!=13) && (j < (lineLength-2)))
					buffer[j++]=c[0];
				else
				{
					buffer[j++]='\n';
					buffer[j]='\0';

					m_pScript->AddScriptLine(buffer);
				
					memset(buffer,0,sizeof(char)*lineLength);
					j=0;
				}
			}
		}

		delete [] buffer;
		buffer = NULL;
	}
	else
		return false;

	// -----------------------------
	// Validate the script
	// This is done in two steps
	// 1- Tokenize the script
	// 2- Verify it's valid
	// -----------------------------		

	// Tokenize Lines
	ScriptOK = m_pScript->ProcessScriptLines();		
	if(ScriptOK)
	{
		// Validate Script
		ScriptOK = m_pScript->ParseScript(); 
		
		if(ScriptOK)
			validScript = true;
		else // Show errors
		{
			int errorsLen = 0, available=0, tmpLen=0;
			char errorsText[3000], *errorLine=NULL;

			memset(errorsText,0,sizeof(errorsText));
			memcpy(errorsText,"<--- Errors in script --->",26);
			errorsText[26]=13;
			errorsText[27]=10;
			errorsLen = strlen(errorsText);			
			errorLine = new char [lineLength+3];

			while((m_pScript->HasMoreErrors(errorLine)) && (errorsLen < sizeof(errorsText)))
			{
				tmpLen = strlen(errorLine);
				errorLine[tmpLen]   = 13;
				errorLine[tmpLen+1] = 10;
				errorLine[tmpLen+2] = '\0';				

				available = sizeof(errorsText) - errorsLen;
				if(available >= strlen(errorLine))
					memcpy(&errorsText[errorsLen], errorLine, strlen(errorLine));
				else
					memcpy(&errorsText[errorsLen], errorLine, available);

				errorsLen += strlen(errorLine);
			}

			delete [] errorLine;
			MessageBox(NULL,errorsText,"Invalid Script",MB_OK|MB_ICONSTOP);
		}
	}

	return validScript;
}
// ************ @1A ENDS* *************

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
}

void MGLApp::Update( double dt )
{
	// ************* @1A BEGINS *****************
	float currentTime = timeGetTime() / 1000.0f;
	
	if(m_pParticleSystem!=NULL)
	{
		if(m_pParticleSystem->IsSysActive())		
			m_pParticleSystem->Update(currentTime);		
	}
	// ************** @1A ENDS ******************
}

void MGLApp::Render()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// set up the camera
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	// ************* @1A BEGINS *****************
	// @1A - Render Particle System		
	glPushMatrix();
		glTranslatef(0.0f, 0.0f, -10.0f);
		if(m_pParticleSystem!=NULL)
		{
			if(m_pParticleSystem->IsSysActive())
				m_pParticleSystem->Render();
		}
	glPopMatrix();	

	// Display info
	RenderInfo();
	// ************** @1A ENDS ******************
}

// ************************** @1A BEGINS ***************************
void MGLApp::Draw2DBegin()
{
	glPushAttrib(GL_CURRENT_BIT);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, 800, 0, 600, -1, 1);
	glMatrixMode(GL_MODELVIEW);
}

void MGLApp::Draw2DEnd()
{	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopAttrib();
}

VOID MGLApp::RenderInfo()
{	
	char	str[80]	= "";
	char    script[40] = "";
	GLfloat fps=0;
	GLfloat minX	= 0.0f;
	GLfloat	minY	= 0.0f;
	GLfloat	maxX	= 800.0f;
	GLfloat	maxY	= 600.0f;	

	memset(script,0,sizeof(script));

	if(m_pParticleSystem==NULL)
		sprintf(script,"none");
	else
		// memcpy(script, particle_scripts[m_currentScript], 30); @2D
		memcpy(script, m_scriptTitle, 15); //@2A		

	fps = m_timer.GetFPS();
	sprintf(str,"fps: %4.2f",fps);
	glDisable(GL_BLEND);

	Draw2DBegin();
		
	glPushMatrix();
		glColor4f(1.0f,1.0f,1.0f,1.0f);
		glBegin(GL_LINE_LOOP);			
			glVertex3f(1, minY+55, 0.0f);
			glVertex3f(minX + 370, minY+55, 0.0f);
			glVertex3f(minX + 370, 1, 0.0f);
			glVertex3f(1, 1, 0.0f);
		glEnd();		
		
		m_font.PrintScreen(0.13f, 0.43f, 0.82f, 1.0f, 5, 40, "particle system loaded:");
		m_font.PrintScreen(0.78f, 0.78f, 0.78f, 1.0f, 5, 22, "F2 - restart particle system");
		m_font.PrintScreen(0.78f, 0.78f, 0.78f, 1.0f, 5, 7, "F3 - load particle system..."); //@2C
		m_font.PrintScreen(0.37f, 0.63f, 0.98f, 1.0f, 220, 40, script);
		m_font.PrintScreen(0.13f, 0.53f, 0.87f, 1.0f, maxX-100, maxY-25, str);
	glPopMatrix();

	Draw2DEnd();
}

void MGLApp::ReinitParticleSystem()
{
	if(m_pParticleSystem!=NULL)
	{
		CParticleSys_Scriptable *newObjScr=NULL;	
		newObjScr = (CParticleSys_Scriptable *)m_pParticleSystem;
		newObjScr->ReinitSystem();
	}
}

void MGLApp::NextScript(char *pstrFileName, char *pstrTitleName) //@2C
{
	// Next script
	
	assert(pstrFileName[0]!='\0'); //@2A

	/* -----------------------------------------
	m_currentScript++;	
	if(m_currentScript>MAX_INDEX_DEMO_SCRIPTS)
		m_currentScript = 0; 
	--------------------------------------- @2D*/

	if(m_pParticleSystem!=NULL)
	{
		m_pParticleSystem->DestroyAll();
		delete m_pParticleSystem;
		m_pParticleSystem = NULL;
	}
	
	// ---------------------
	// Load and parse script
	// ---------------------
	m_pScript = new CParticleScript;
	assert(m_pScript!=NULL);
	if(m_pScript!=NULL)
	{
		// -------------------------------------
		// Cleanup textures used for last system
		// -------------------------------------
		_TextureManager()->DestroyAll();

		// --------------------
		// Reload font 
		// --------------------
		if(!(m_font.CreateGLFont("Images/tx_font.tga")))
			return;

		// ---------------------
		// Load and parse script
		// ---------------------
		char scriptFilename[200];
		memset(scriptFilename,0,sizeof(scriptFilename));
		memset(m_scriptTitle,0,sizeof(m_scriptTitle)); //@2A
		// memcpy(scriptFilename, particle_scripts[m_currentScript], sizeof(scriptFilename)); @2D
		memcpy(scriptFilename, pstrFileName, sizeof(scriptFilename)); //@2A
		memcpy(m_scriptTitle, pstrTitleName, sizeof(m_scriptTitle)); //@2A
		if(!(LoadScript(scriptFilename)))
		{
			m_pScript->DestroyAll();
			delete m_pScript;
			m_pScript = NULL;	
			return;
		}

		// --------------------
		// Init particle system
		// --------------------
		CParticleSys_Scriptable *newObjScr=NULL;	
		m_pParticleSystem  = new CParticleSys_Scriptable();
		if(m_pParticleSystem != NULL)
		{
  			newObjScr = (CParticleSys_Scriptable *)m_pParticleSystem;	
			newObjScr->AttachScript(m_pScript);
			BOOL scriptApplied = newObjScr->ApplyScript();
			if(!scriptApplied)
			{
				m_pParticleSystem->DestroyAll();
				delete m_pParticleSystem;
				m_pParticleSystem = NULL;			
			}
		}
	}
}

// ************************** @1A ENDS ***************************







