/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: CParticleSys_Scriptable.cpp                               |*/
/*| Author: Joel Villagrana    (joel.villagrana@gmail.com)              |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yyyy | Programmer        | Modifications                      |*/
/*|=====================================================================|*/
/*| 12/07/2005 | JVillagrana       | Initial Version                    |*/
/*|---------------------------------------------------------------------|*/
/*************************************************************************/

#include "CParticleSys_Scriptable.h"

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ========================================================================
// IMPLEMENTATION OF CPARTICLESYS_SCRIPTABLE CLASS 
// ========================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

// 
// Constructor
// 
CParticleSys_Scriptable::CParticleSys_Scriptable()
{
	// CALL DEFAULT CONSTRUCTOR
	InitAll(PSYS_SCRIPTABLE,// SYSTEM TYPE
			1000);			// DEFAULT MAX PARTICLES
	m_scriptApplied = FALSE;
	m_pScript = NULL;
	m_events.reserve(20);
}

// 
// Destructor
// 
CParticleSys_Scriptable::~CParticleSys_Scriptable()
{
}

// 
// Cleanup
// 
VOID CParticleSys_Scriptable::DestroyAll()
{	
	// FREE MEMORY USED FOR SCRIPT
	if(m_pScript != NULL)
	{
		m_pScript->DestroyAll(); 
		delete m_pScript;
		m_pScript = NULL;
	}

	// FREE MEMORY USED FOR EVENTS
	for(m_events_it = m_events.begin(); m_events_it != m_events.end();)
	{
		(*m_events_it)->DestroyAll();
		delete (*m_events_it);
		m_events_it = m_events.erase(m_events_it);
	}

	// FREE MEMORY USED FOR EMITTERS
	DestroySystem();
}

// 
// Reinit
// 
VOID CParticleSys_Scriptable::ReinitSystem()
{
	// FREE MEMORY USED FOR EVENTS
	for(m_events_it = m_events.begin(); m_events_it != m_events.end();)
	{
		(*m_events_it)->DestroyAll();
		delete (*m_events_it);
		m_events_it = m_events.erase(m_events_it);
	}

	// FREE MEMORY USED FOR EMITTERS
	DestroySystem();

	// APPLY SCRIPT
	ApplyScript();
}

// 
// Update a particle
// 
VOID CParticleSys_Scriptable::Update(GLfloat currentTime)
{		
	int i=0;

	if(m_scriptApplied && IsSysActive())
	{
		// UPDATE EMITTERS
		UpdateSystem((currentTime - m_startTime));

		// PROCESS EVENTS IF NEEDED
		for(m_events_it = m_events.begin(); 
			m_events_it != m_events.end();
			++m_events_it)
		{
			// IF EVENT NOT PROCESSED
			if( !((*m_events_it)->AlreadyProcessed()) )
			{
				// PROCESS
				(*m_events_it)->ProcessIfNeeded(m_startTime, currentTime, (VOID *)(this));
			}
			i++;
		}
	}
}

// 
// Render
// 
VOID CParticleSys_Scriptable::Render()
{ 			
	if(m_scriptApplied && IsSysActive())
	{
		// RENDER
		RenderParticleSystem();
	}
}

// 
// Attach a script
// 
VOID CParticleSys_Scriptable::AttachScript(CParticleScript *script)
{
	if(m_pScript!=NULL)
		delete m_pScript;
	m_pScript = script;
}

// 
// Apply the script
// 
BOOL CParticleSys_Scriptable::ApplyScript()
{	
	m_scriptApplied = AddSystemProperties();
	if(!m_scriptApplied)
		MessageBox(NULL, "Could not add the system properties", "Error", MB_OK);
	else	
	{
		m_scriptApplied = AddEmitters();
		if(!m_scriptApplied)
			MessageBox(NULL, "Could not add the emitters", "Error", MB_OK);
		else
		{
			m_scriptApplied = AddEvents();	
			if(!m_scriptApplied)
				MessageBox(NULL, "Could not add the events", "Error", MB_OK);
		}
	}

	if(m_scriptApplied)
	{
		SetStartTime();		
		SetActive(TRUE);
	}

	return m_scriptApplied;
}

// 
// Add System Properties
// 
BOOL CParticleSys_Scriptable::AddSystemProperties()
{
	BOOL added = FALSE, gotValue = FALSE;
	SCR_TOKEN_TYPE propertyType = NONE;

	assert(m_pScript!=NULL && "CParticleSys_Scriptable::AddSystemProperties");

	m_pScript->SetIterator_SystemProperties();
	while(m_pScript->MoreSystemProperties())
	{
		propertyType = m_pScript->Get_PropertyType();

		if(propertyType == KEYWORD_SYSP_MAXPARTICLES)
		{
			int maxP=0;
			gotValue = m_pScript->Get_IntValues(&maxP, 1);
			if(gotValue)
			{
				Set_MaxParticles(maxP);
				added = TRUE;
			}
			else
				added = FALSE;

		}
		else if(propertyType == KEYWORD_SYSP_FORCE)
		{
			int index=0;
			float forceVec[3];

			gotValue = m_pScript->Get_IntValues(&index, 1);
			if(gotValue)
				gotValue = m_pScript->Get_FloatValues(forceVec, 3);
			if(gotValue)
			{
				SetSysForce(index, forceVec);
				added = TRUE;
			}
			else
				added = FALSE;
		}	
	}

	return added;
}

// 
// Add Emitters
// 
BOOL CParticleSys_Scriptable::AddEmitters()
{
	CParticle_Emitter *newEmitter=NULL;
	int intValues[5], tokenLength=0;
	float floatValues[10];	
	char literal[SCRIPT_TOKEN_LENGTH];
	BOOL added = FALSE, gotValue = FALSE;
	SCR_TOKEN_TYPE propertyType = NONE;	
	PARTICLE_REP particleRep;
	PSYS_EMITTERTYPE emitterType;
	GLenum blends[2];
	MGL::Vector3 vector3, forcePos;

	assert(m_pScript!=NULL && "CParticleSys_Scriptable::AddEmitters");	
	memset(literal,0,sizeof(literal));

	m_pScript->SetIterator_Emitters();
	while(m_pScript->MoreEmitters())
	{
		// CREATE NEW EMITTER
		newEmitter = new CParticle_Emitter();		
		assert(newEmitter != NULL && "CParticleSys_Scriptable::AddEmitters");

		// GET EMITTER ID
		gotValue = m_pScript->Get_Identifier(literal, 1);
		if(gotValue)
			newEmitter->Set_ID(literal);

		while(m_pScript->MoreEmitterProperties())
		{
			memset(intValues,0,sizeof(intValues));
			memset(floatValues,0,sizeof(floatValues));
			memset(literal,0,sizeof(literal));

			propertyType = m_pScript->Get_PropertyType();

			switch(propertyType)
			{
				case KEYWORD_EM_POSITION:
											gotValue = m_pScript->Get_FloatValues(floatValues, 3);
											if(gotValue)
											{
												vector3.Set(floatValues[0], floatValues[1], floatValues[2]);
												newEmitter->Set_Position(vector3);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_TYPE:
											gotValue = m_pScript->Get_TypeValues(&emitterType, 1);
											if(gotValue)
											{
												newEmitter->Set_Type(emitterType);
												added = TRUE;
											}										
											break;
				case KEYWORD_EM_PARTICLEREP:
											gotValue = m_pScript->Get_RepValues(&particleRep, 1);
											if(gotValue)
											{
												newEmitter->Set_ParticleRep(particleRep);
												added = TRUE;
											}
											break;
				case KEYWORD_SYSP_MAXPARTICLES:
											gotValue = m_pScript->Get_IntValues(intValues, 1);
											if(gotValue)
											{
												intValues[0] = abs(intValues[0]);
												newEmitter->Set_MaxParticles(intValues[0]);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_TEXTURE:
											gotValue = m_pScript->Get_StringValues(literal, 1);
											if(gotValue)
											{
												newEmitter->Set_Texture(literal);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_EMISSIONRATE:
											gotValue = m_pScript->Get_FloatValues(floatValues, 1);
											if(gotValue)
											{
												floatValues[0] = fabs(floatValues[0]);
												newEmitter->Set_EmissionRate(floatValues[0]);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_BATCHSIZERANGE:
											gotValue = m_pScript->Get_IntValues(intValues, 1);
											if(gotValue)
											{
												intValues[0] = abs(intValues[0]);
												newEmitter->Set_BatchSize(intValues[0]);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_BLENDINGMODES:
											gotValue = m_pScript->Get_BlendValues(blends, 2);
											if(gotValue)
											{
												newEmitter->Set_Blends(blends[0], blends[1]);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_VERTEXSHADER:
											gotValue = m_pScript->Get_StringValues(literal, 1);
											if(gotValue)
											{
												newEmitter->Set_VxShader(literal);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_PIXELSHADER:
											gotValue = m_pScript->Get_StringValues(literal, 1);
											if(gotValue)
											{
												newEmitter->Set_PxShader(literal);
												added = TRUE;
											}
											break;

				case KEYWORD_SYSP_FORCE:
											gotValue = m_pScript->Get_IntValues(intValues, 1);
											gotValue = m_pScript->Get_FloatValues(floatValues, 7);
											if(gotValue)
											{
												intValues[0] = abs(intValues[0]);
												vector3.Set(floatValues[0], floatValues[1], floatValues[2]);
												forcePos.Set(floatValues[3], floatValues[4], floatValues[5]);
												newEmitter->Set_Force(intValues[0], vector3, forcePos, floatValues[6]);
												added = TRUE;
											}
											break;

				case KEYWORD_EM_LIFERANGE:
											gotValue = m_pScript->Get_FloatValues(floatValues, 2);
											if(gotValue)
											{
												newEmitter->Set_Life(floatValues[0], floatValues[1]);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_SIZERANGE:
											gotValue = m_pScript->Get_FloatValues(floatValues, 2);
											if(gotValue)
											{
												newEmitter->Set_Size(floatValues[0], floatValues[1]);
												added = TRUE;
											}
											break;
				case KEYWORD_EM_STATE:
											gotValue = m_pScript->Get_StateValues(intValues, 1);
											if(gotValue)
											{
												intValues[0] = abs(intValues[0]);
												bool val = (intValues[0]==1 ? true : false);
												newEmitter->Set_EmitterState(val);
												added = TRUE;
											}
											break;

				case KEYWORD_PARTICLE_COLOR:
											// GET RGBA
											gotValue = m_pScript->Get_IntValues(intValues, 4);
											if(gotValue)
											{
												intValues[0] = abs(intValues[0]);
												intValues[1] = abs(intValues[1]);
												intValues[2] = abs(intValues[2]);
												intValues[3] = abs(intValues[3]);
												// ADD CHANGE
												CColour newColor;
												newColor.Set_ColourRGBA(intValues[0], intValues[1],
																		intValues[2], intValues[3]);
												newEmitter->Set_ParticleColor(newColor);
												added = TRUE;
											}
											break;

			} // SWITCH

		} // MORE EMITTER PROPERTIES

		// ADD EMITTER TO THE VECTOR
		AddEmitter(newEmitter);		
		newEmitter = NULL;

	} // MORE EMITTERS

	return added;
}

// 
// Add Events
// 
BOOL CParticleSys_Scriptable::AddEvents()
{
	CParticle_Event *newEvent=NULL;
	BOOL added = TRUE, gotValue = FALSE, moreEvents = FALSE;
	SCR_TOKEN_TYPE propertyType = NONE;
	float eventTime=0.0f, floatValues[10], fadeTime=0.0f;
	int intValues[5], tokenL = 0, emitterIndex=0, fadeProperty=0, i=-1;	
	char *emitterID = NULL, literal[SCRIPT_TOKEN_LENGTH];
	GLenum blends[2];	

	memset(floatValues,0,sizeof(floatValues));
	memset(intValues,0,sizeof(intValues));	

	assert(m_pScript!=NULL && "CParticleSys_Scriptable::AddEvents");

	tokenL = m_pScript->GetTokenLength();
	emitterID = new char[tokenL];
	memset(emitterID,0,tokenL);

	moreEvents = m_pScript->SetIterator_Events();
	while(moreEvents)
	{		
		added = FALSE;

		// CREATE NEW EVENT
		// ================
		newEvent = new CParticle_Event();		
		assert(newEvent != NULL && "CParticleSys_Scriptable::AddEvents");
		
		// GET THE EVENT TIME
		// ==================
		gotValue = m_pScript->Get_FloatValues(floatValues, 1);
		if(gotValue)
			eventTime = floatValues[0];

		// SET EVENT TIME
		newEvent->Set_EventTime(eventTime);

		// ADD THE CHANGES TO THE EVENT CHANGE...		
		// ADD SYSTEM PROPERTIES CHANGES FIRST
		// ===================================
		while(m_pScript->MoreSystemProperties())
		{
			propertyType = m_pScript->Get_PropertyType();

			// ADD SYS FORCE CHANGE
			// 
			if(propertyType == KEYWORD_SYSP_FORCE)
			{
				int index=0;
				gotValue = m_pScript->Get_IntValues(&index, 1);
				if(gotValue)gotValue = m_pScript->Get_FloatValues(floatValues, 3);
				if(gotValue)
				{
					// Add index as another float value
					floatValues[3] = (float)index;
					// ADD CHANGE
					newEvent->AddPropChange_Float(0, SYS_FORCE, eventTime, eventTime, floatValues, 4);
				}
			}	
		}

		// ADD EMITTER CHANGES
		// ===================
		while(m_pScript->MoreEmitterChanges(emitterID))
		{
			i++;
			emitterIndex = -1;

			// FIGURE OUT EMITTERID INDEX IN EMITTERS VECTOR (IN CPARTICLE_SYS)
			// ================================================================
			emitterIndex = GetEmitterIndex(emitterID);		

			if(emitterIndex >= 0)
			{
				fadeProperty = 0;
				//while(m_pScript->MoreEmitterProperties())
				while(m_pScript->More_EmitterORParticle_Changes())
				{
					propertyType = m_pScript->Get_PropertyType();
			
					switch(propertyType)
					{
						case KEYWORD_EM_POSITION:
													gotValue = m_pScript->Get_FloatValues(floatValues, 3);
													if(gotValue)
													{			
														// ADD CHANGE
														newEvent->AddPropChange_Float(emitterIndex, EM_POSITION, 
																		eventTime, eventTime, floatValues, 3);	
													}
													break;
						case KEYWORD_EM_EMISSIONRATE:
													gotValue = m_pScript->Get_FloatValues(floatValues, 1);
													if(gotValue)
													{
														floatValues[0] = fabs(floatValues[0]);
														// ADD CHANGE
														newEvent->AddPropChange_Float(emitterIndex, EM_EMISSIONRATE,
																		eventTime, eventTime, floatValues, 1);
													}
													break;
						case KEYWORD_EM_BATCHSIZERANGE:
													gotValue = m_pScript->Get_IntValues(intValues, 1);
													if(gotValue)
													{
														intValues[0] = abs(intValues[0]);
														// ADD CHANGE
														newEvent->AddPropChange_Int(emitterIndex, EM_BATCHSIZE,
																		eventTime, eventTime, intValues, 1);
													}
													break;
						case KEYWORD_EM_TEXTURE:
													gotValue = m_pScript->Get_StringValues(literal, 1);
													if(gotValue)
													{				
														// ADD CHANGE
														newEvent->AddPropChange_Str(emitterIndex, EM_TEXTURE,
																		eventTime, eventTime, literal, 1);
													}
													break;
						case KEYWORD_EM_BLENDINGMODES:
													gotValue = m_pScript->Get_BlendValues(blends, 2);
													if(gotValue)
													{				
														// ADD CHANGE
														newEvent->AddPropChange_Enum(emitterIndex, EM_BLENDING,
																		eventTime, eventTime, blends, 2);
													}
													break;
						case KEYWORD_EM_STATE:
													gotValue = m_pScript->Get_StateValues(intValues, 1);
													if(gotValue)
													{
														intValues[0] = abs(intValues[0]);
														// ADD CHANGE														
														newEvent->AddPropChange_Int(emitterIndex, EM_STATE,
																		eventTime, eventTime, intValues, 1);
													}
													break;
						case KEYWORD_EM_VERTEXSHADER:
													gotValue = m_pScript->Get_StringValues(literal, 1);
													if(gotValue)
													{
														// ADD CHANGE
														newEvent->AddPropChange_Str(emitterIndex, EM_VERTEXSH,
																		eventTime, eventTime, literal, 1);
													}
													break;
						case KEYWORD_EM_PIXELSHADER:
													gotValue = m_pScript->Get_StringValues(literal, 1);
													if(gotValue)
													{
														// ADD CHANGE
														newEvent->AddPropChange_Str(emitterIndex, EM_PIXELSH,
																		eventTime, eventTime, literal, 1);
													}
													break;

						case KEYWORD_SYSP_FORCE:
													gotValue = m_pScript->Get_IntValues(intValues, 1);
													gotValue = m_pScript->Get_FloatValues(floatValues, 7);
													if(gotValue)
													{
														intValues[0] = abs(intValues[0]);
														// Add index as another float value
														floatValues[7] = (float)intValues[0];														
														// Dist shouldnt be negative...
														floatValues[6] = fabs(floatValues[6]);
														// ADD CHANGE
														newEvent->AddPropChange_Float(emitterIndex, EM_FORCE, 
																		eventTime, eventTime, floatValues, 8);
													}
													break;


						case KEYWORD_PARTICLE_FADE: // GET FADE TIME
													gotValue = m_pScript->Get_FloatValues(&fadeTime, 1);
													fadeTime += eventTime;
													if(gotValue)
														fadeProperty = true;
													break;

						case KEYWORD_PARTICLE_COLOR:
													// IF NOT FADE, SET SAME TIME
													if(!fadeProperty)
														fadeTime = eventTime;
													else
														fadeProperty = false; // IF FADE, RESET VARIABLE

													// GET RGBA
													gotValue = m_pScript->Get_IntValues(intValues, 4);
													if(gotValue)
													{
														intValues[0] = abs(intValues[0]);
														intValues[1] = abs(intValues[1]);
														intValues[2] = abs(intValues[2]);
														intValues[3] = abs(intValues[3]);
														// ADD CHANGE
														newEvent->AddPropChange_Int(emitterIndex, PA_COLOR, 
																		eventTime, fadeTime, intValues, 4);
													}
													break;

						case KEYWORD_PARTICLE_SIZE:
													// IF NOT FADE, SET SAME TIME
													if(!fadeProperty)
														fadeTime = eventTime;
													else
														fadeProperty = false; // IF FADE, RESET VARIABLE

													// GET SIZE
													gotValue = m_pScript->Get_FloatValues(floatValues, 1);
													if(gotValue)
													{
														floatValues[0] = fabs(floatValues[0]);
														// ADD CHANGE
														newEvent->AddPropChange_Float(emitterIndex, PA_SIZE, 
																		eventTime, fadeTime, floatValues, 1);
													}

													break;

						case KEYWORD_PARTICLE_WEIGHT:
													// IF NOT FADE, SET SAME TIME
													if(!fadeProperty)
														fadeTime = eventTime;
													else
														fadeProperty = false; // IF FADE, RESET VARIABLE

													// GET WEIGHT
													gotValue = m_pScript->Get_FloatValues(floatValues, 1);
													if(gotValue)
													{
														floatValues[0] = fabs(floatValues[0]);
														// ADD CHANGE
														newEvent->AddPropChange_Float(emitterIndex, PA_WEIGHT, 
																		eventTime, fadeTime, floatValues, 1);
													}

													break;

					} // SWITCH

				} // while(m_pScript->More_EmitterORParticle_Changes())

			} // emitterIndex >= 0

			//m_pScript->GetNextToken(); ??

		} // while(m_pScript->MoreEmitterChanges(emitterID))

		// ADD EVENT TO THE VECTOR
		m_events.push_back(newEvent);
		added = TRUE;

		// MORE EVENTS TO PROCESS?
		moreEvents = m_pScript->MoreEvents();
	}
	
	// FREE MEMORY
	delete [] emitterID;

	return added;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Fade_ParticleSize(int index, float newSize, 
										float elapsedTime, float timeRange)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Fade_ParticleSize");
	//if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Fade_Particles_Size(newSize, elapsedTime, timeRange);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Fade_ParticleColor(int index, CColour newColor, 
										float elapsedTime, float timeRange)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Fade_ParticleColor");
	//if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Fade_Particles_Color(newColor, elapsedTime, timeRange);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Fade_ParticleWeight(int index, float newWeight, 
										float elapsedTime, float timeRange)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Fade_ParticleWeight");
	//if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Fade_Particles_Weight(newWeight, elapsedTime, timeRange);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Set_EmitterPos(int index, MGL::Vector3 newPos)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Set_EmitterPos");
	if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Set_Position(newPos);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Set_EmitterEmissionRate(int index, float newRate)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Set_EmitterEmissionRate");
	if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Set_EmissionRate(newRate);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Set_EmitterBatchSize(int index, int newBatchSize)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Set_EmitterBatchSize");
	if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Set_BatchSize(newBatchSize);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Set_EmitterBlends(int index, GLenum src, GLenum dst)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Set_EmitterBlends");
	if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Set_Blends(src, dst);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Set_EmitterState(int index, bool newState)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Set_EmitterState");	
	m_emitters[index]->Set_EmitterState(newState);
	return TRUE;
}

// 
// 
// 
BOOL CParticleSys_Scriptable::Set_EmitterForce(int index, int forceIndex, 
					MGL::Vector3 newForce, MGL::Vector3 forcePos, float dst)
{
	assert(index < m_emitters.size() && "CParticleSys_Scriptable::Set_EmitterForce");		
	if(m_emitters[index]->IsEmitterActive())
		m_emitters[index]->Set_Force(forceIndex, newForce, forcePos, dst);
	return TRUE;
}

// ------------------------- END OF FILE -----------------------------------
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


