/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: CParticle_Emitter.cpp                                     |*/
/*| Author: Joel Villagrana    (joel.villagrana@gmail.com)              |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yyyy | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 17/05/2005 | JVillagrana       | Initial Version                    |*/
/*|---------------------------------------------------------------------|*/
/*************************************************************************/

#include "CParticle_Emitter.h"
#include <mmsystem.h>

extern float RangedRandom(float minV, float maxV);
extern int RangedRandomInt(int v1, int v2);
extern GLfloat distance2Points(GLVector3 v1, GLVector3 v2);

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ========================================================================
// IMPLEMENTATION OF CPARTICLE_EMITTER CLASS 
// ========================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

// 
// Constructor
// 
CParticle_Emitter::CParticle_Emitter()
{
	m_particles.reserve(1000);
	InitAll();
}

// 
// Destructor
// 
CParticle_Emitter::~CParticle_Emitter()
{
}

// 
// Init
// 
HRESULT CParticle_Emitter::InitAll()
{
	memset(m_ID,0,sizeof(m_ID));

	m_position.Set(0.0f, 1.5f, 0.0f);
	m_lineEndPoint.Set(0.0f, 0.0f, 0.0f);
	m_totalLocalForces.Set(0.0f, 0.0f, 0.0f);
	m_totalGlobalForces.Set(0.0f, 0.0f, 0.0f);
	m_particleInitialColor.Set_ColourRGBA(255,255,255,255);
	
	memset(m_texture,0,sizeof(m_texture));
	memset(m_vxShader,0,sizeof(m_vxShader));
	memset(m_pxShader,0,sizeof(m_pxShader));
	memset(m_forcesDst,0,sizeof(m_forcesDst));

	m_type				= E_POINT;
	m_representation	= R_QUADS;
	m_maxParticles		= 2500;
	m_numParticles		= 0;
	m_emissionRate		= 0.25f;
	m_lastEmission		= 0.0f;
	m_batchSize			= 20;
	m_srcBlend			= GL_ONE;
	m_dstBlend			= GL_ONE_MINUS_SRC_ALPHA;
	m_numForces			= 0;
	m_minLife			= 1.0f;
	m_maxLife			= 10.0f;
	m_minSize			= 1.0f;
	m_maxSize			= 2.5f;
	m_emitterON			= true;	
	m_textureID			= -1;
	m_textureLoaded		= false;
	m_useTexture		= false;	
	m_numDead			= 0;
	m_lastEmission		= -10000.0f;

	for(int i=0; i<MAX_EMITTER_FORCES; i++)
	{
		m_forces[i].Set(0.0f, 0.0f, 0.0f);
		m_forcesPos[i].Set(0.0f, 0.0f, 0.0f);
		m_forcesAct[i] = false;
	}
	
	// Use GLee to check if point sprites are supported	 
	if(GLEE_VERSION_2_0 || (GLEE_ARB_point_sprite && GLEE_ARB_point_parameters))
		m_pointSpritesSupported = true;
	else
		m_pointSpritesSupported = false;

	float x = 100*((float)rand())/((float)RAND_MAX);
	float y = 100*((float)rand())/((float)RAND_MAX);
	float z = 100*((float)rand())/((float)RAND_MAX);
	MGL::Vector3 dir;
	dir.Set(x, y, z);
	dir.Normalize();

	// If emitter is line, calculate endpoint
	m_lineEndPoint = m_position + ((dir) * 10.0f);

	return S_OK;
}

// 
// Cleanup
// 
VOID CParticle_Emitter::DestroyAll()
{
	// Delete all particles	
	for(m_particles_it = m_particles.begin(); 
		m_particles_it != m_particles.end();)
	{
		delete(*m_particles_it);
		m_particles_it = m_particles.erase(m_particles_it);
	}
}

// 
// Set Particle Representation
// 
void CParticle_Emitter::Set_ParticleRep(PARTICLE_REP rep)
{
	if(rep==R_POINT_SPRITES)
	{
		if(m_pointSpritesSupported)
			m_representation = rep;
		else
			m_representation = R_POINTS;
	}
	else
		m_representation = rep;
}

// 
// Render
// 
void CParticle_Emitter::Render()
{
	MGL::Vector3 billbX, billbY;
	GLfloat vm[16];
	GLMatrix44 viewMatrix, viewTranspose;

	// ENABLE BLENDING 
	glEnable(GL_BLEND);
	// SET BLENDING MODE
	glBlendFunc(m_srcBlend, m_dstBlend);

	if(m_representation==R_POINTS || m_representation==R_LINES)
		m_useTexture = false;

	// ENABLE TEXTURING IF NEEDED
	if((m_useTexture)&&(m_texture[0]!='\0'))
	{
		if(!m_textureLoaded)
			LoadTexture();

		if(m_textureID>=0)
		{
			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, _TextureManager()->Get_texture(m_textureID));				
		}
	}
	
	// SET PARAMETERS FOR POINT SPRITES
	if(m_representation==R_POINT_SPRITES)
	{
        float quadratic[] =  { 1.0f, 0.0f, 0.01f };
        glPointParameterfvARB( GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic );
        glPointParameterfARB( GL_POINT_FADE_THRESHOLD_SIZE_ARB, 1.0f );
	    glPointParameterfARB( GL_POINT_SIZE_MIN_ARB, 1.0f );
	    glPointParameterfARB( GL_POINT_SIZE_MAX_ARB, 5.0f ); 	
        glTexEnvf( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );
        glEnable( GL_POINT_SPRITE_ARB ); 
	}

	// SETUP FOR MANUAL BILLBOARDS
	glGetFloatv(GL_MODELVIEW_MATRIX, vm);
	viewMatrix.Set(vm);
	viewMatrix.Transpose(&viewTranspose);
	billbX.Set(viewTranspose.elem[0], viewTranspose.elem[1], viewTranspose.elem[2]);
	billbY.Set(viewTranspose.elem[4], viewTranspose.elem[5], viewTranspose.elem[6]);

	// RENDER
	for(m_particles_it = m_particles.begin(); 
		m_particles_it != m_particles.end();
		++m_particles_it)	
	{
		if((*m_particles_it)->IsAlive())
		{
			(*m_particles_it)->Render(m_representation, billbX, billbY, m_useTexture);	
		}
	}

	// DISABLE POINT SPRITES IF USED
	if(m_representation==R_POINT_SPRITES)
		glDisable( GL_POINT_SPRITE_ARB );

	// DISABLE TEXTURING
	if(m_useTexture)
		glDisable(GL_TEXTURE_2D);
	// DISABLE BLENDING
	glDisable(GL_BLEND);
}

// 
// Load Texture
// 
void CParticle_Emitter::LoadTexture()
{	
	m_textureID = _TextureManager()->LoadTexture(m_texture, GL_LINEAR);
	m_textureLoaded = true;
}

// 
// Update
// 
void CParticle_Emitter::Update(float elapsedTime, float startTime)
{	
	int deadParticle=0, i=0;
	MGL::Vector3 externalForce, particlePos;
	float distance=0.0f;

	// UPDATE PARTICLES	
	for(m_particles_it = m_particles.begin(); 
		m_particles_it != m_particles.end();
		++m_particles_it)	
	{
		// Only update particle if alive
		if((*m_particles_it)->IsAlive())
		{
			externalForce.Set(0.0f, 0.0f, 0.0f);
			deadParticle = 0;

			// Check if this particle is affected by local forces
			(*m_particles_it)->Get_Position(&particlePos);
			for(i=0; i<MAX_EMITTER_FORCES; i++)
			{
				// If force active...
				if(m_forcesAct[i])
				{
					// Get distance between particle position and force position
					distance = distance2Points(GLVector3(particlePos.x, particlePos.y, particlePos.z), 
											   GLVector3(m_forcesPos[i].x, m_forcesPos[i].y, m_forcesPos[i].z));
					// If distance is equal or less than max dist...
					if(distance <= m_forcesDst[i])
						externalForce += m_forces[i];
				}
			}

			// UPDATE POSITION, DIRECTION, ALPHA, LIFE...
			(*m_particles_it)->Update(	elapsedTime,	
										&deadParticle, 
										(externalForce + m_totalGlobalForces),
										m_maxLife );
			if(deadParticle)
			{
				++m_numDead;
			}
		}
	}

	// EMIT NEW PARTICLES??
	if(m_emitterON)	
	{
		if((elapsedTime - m_lastEmission) >= m_emissionRate)
		{
			EmitParticles(startTime);
			m_lastEmission = elapsedTime;
		}
	}
}

// 
// Initialize
// 
void CParticle_Emitter::InitializeParticle(CParticle *pParticle, float startTime)
{
	MGL::Vector3 position, direction;
	float life=0.0f, size=0.0f;
	float now = timeGetTime() / 1000.0f;

	// Position and direction depend on emitter type
	CalculateInitialPosition(&position, &direction);

	// Generate random for life and size
	life = RangedRandom(m_minLife, m_maxLife);
	size = RangedRandom(m_minSize, m_maxSize);

	// Init particle
	pParticle->InitParticle(m_particleInitialColor, position, direction, 
												life, size, (now - startTime));
}

// 
// Calculate Initial Position According to Emitter Type
// (Emitter type parameters should be scriptable in next version...)
// 
void CParticle_Emitter::CalculateInitialPosition(MGL::Vector3 *pos, MGL::Vector3 *dir)
{
	float x=0.0f, y=0.0f, z=0.0f;

	// Random direction	
	x = (100*((float)rand())/((float)RAND_MAX))-50.0f;
	y = (100*((float)rand())/((float)RAND_MAX))-50.0f;
	z = (100*((float)rand())/((float)RAND_MAX))-50.0f;
	dir->Set(x, y, z);
	dir->Normalize();

	if(m_type==E_POINT)
	{
		// Same position as emitter position
		*pos = m_position;
	}
	else if(m_type==E_BEAM)
	{
		// Vector from position to endpoint
		*dir = m_lineEndPoint - m_position;
		dir->Normalize();		
		float amount = (10.0f)*((float)rand())/((float)RAND_MAX);
		*pos = m_position + ((*dir) * amount);
	}
	else if(m_type==E_LINE)
	{
		float amount = (100.0f)*((float)rand())/((float)RAND_MAX);
		*pos = (m_lineEndPoint - m_position) * amount;
	}
	else if(m_type==E_SPHERE)
	{		
		// Sum normalized direction to emitter position
		*pos = m_position + (*dir);
	}
	else if(m_type==E_RING)
	{
		int maxSlices = 100;
		int radPos = rand() % maxSlices;
		float d = PI_2 * radPos;
		float ring_width = 1.5f; 
		float x_2d = (cos(d/maxSlices) * ring_width);
		float y_2d = (sin(d/maxSlices) * ring_width);

		pos->Set(x_2d, y_2d, 0.0f);
		*pos = *pos + m_position;
		*dir = *pos - m_position;
		dir->Normalize();
	}
}

// 
// Emit
// 
int CParticle_Emitter::EmitParticles(float startTime)
{
	int emit=0, i=0, deadIndex=0, j=0, replaced=0;	
	bool deadIndexFound = false;
	CParticle *newParticle = NULL;

	if((m_numParticles - m_numDead) < m_maxParticles)
	{	
		if((m_numParticles + m_batchSize) <= m_maxParticles)		
			emit = m_batchSize;
		else			
			emit = abs((m_numParticles + m_batchSize) - m_maxParticles);

		// If able to emit more particles
		if(emit > 0)
		{
			for(i=0; i<emit; i++)
			{
				deadIndex = 0;
				deadIndexFound = false;

				// Allocate memory
				newParticle = new CParticle();
				assert(newParticle != NULL && "CParticle_Emitter::EmitParticles");

				// Initialize particle properties
				InitializeParticle(newParticle, startTime);
				
				// Replace dead particles if we can
				if(m_numDead > 0)
				{					
					// Find one free index
					j = 0;
					STLparticleVec_it tempIter = m_particles.begin();
					while((tempIter != m_particles.end()) && !deadIndexFound)
					{
						if( (*tempIter)->IsDead() )
						{
							deadIndex = j;
							deadIndexFound = true;
						}
						++j;
						++tempIter;
					}
		
					// Replace particle
					if(deadIndexFound)
					{
						delete (m_particles[deadIndex]);
						m_particles[deadIndex] = newParticle;
						--m_numDead;						
						++replaced;
					}
				}
				
				// If no index found to replace, push new particle
				if(!deadIndexFound)
					m_particles.push_back(newParticle);	
		
				newParticle = NULL;
			}
			m_numParticles += (emit - replaced);		
		}
	}

	return emit;
}

// 
// Set Force
// 
void CParticle_Emitter::Set_Force(int index, MGL::Vector3 force, 
											MGL::Vector3 forcePos, float dst)
{
	if(index < MAX_EMITTER_FORCES)
	{
		m_totalLocalForces -= m_forces[index];

		m_forces[index] = force;
		m_forcesPos[index] = forcePos;
		m_forcesDst[index] = dst;

		if(force.x==0.0f && force.y==0.0f && force.z==0.0f)
			m_forcesAct[index] = false;
		else
			m_forcesAct[index] = true;

		// Calculate total local forces at this point
		// so in Update we don't need to calculate every time...
		m_totalLocalForces += force;
	}
}

// 
// Get Force
// 
void CParticle_Emitter::Get_Force(int index, MGL::Vector3 *force, 
											MGL::Vector3 *forcePos, float *dst)
{
	if((index < m_numForces) && (index < MAX_EMITTER_FORCES))
	{
		*force = m_forces[index];
		*forcePos = m_forcesPos[index];
		*dst = m_forcesDst[index];
	}
}

// 
// Fade Size
// 
void CParticle_Emitter::Fade_Particles_Size(float finalSize,
										float elapsedEventTime, float timeRange)
{
	for(m_particles_it = m_particles.begin(); 
		m_particles_it != m_particles.end();
		++m_particles_it)
	{
		(*m_particles_it)->Fade_Size(finalSize, elapsedEventTime, timeRange);
	}
}

// 
// Fade Color
// 
void CParticle_Emitter::Fade_Particles_Color(CColour finalColor, 
										float elapsedEventTime, float timeRange)
{
	m_particleInitialColor = finalColor;

	for(m_particles_it = m_particles.begin(); 
		m_particles_it != m_particles.end();
		++m_particles_it)
	{
		(*m_particles_it)->Fade_Color(finalColor, elapsedEventTime, timeRange);
	}
}

// 
// Fade Weight
// 
void CParticle_Emitter::Fade_Particles_Weight(float finalWeight, 
										float elapsedEventTime, float timeRange)
{
	for(m_particles_it = m_particles.begin(); 
		m_particles_it != m_particles.end();
		++m_particles_it)
	{
		(*m_particles_it)->Fade_Weight(finalWeight, elapsedEventTime, timeRange);
	}
}

// ------------------------- END OF FILE -----------------------------------
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


