/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: CParticle_Sys.cpp                                         |*/
/*| Author: Joel Villagrana    (joel.villagrana@gmail.com)              |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yyyy | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 17/05/2005 | JVillagrana       | Initial Version                    |*/
/*|---------------------------------------------------------------------|*/
/*************************************************************************/

#include "CParticle_Sys.h"
#include <mmsystem.h>

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ========================================================================
// IMPLEMENTATION OF CPARTICLE_SYS CLASS 
// ========================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

// 
// CONSTRUCTOR
// 
CParticle_Sys::CParticle_Sys()
{
	m_sysType		= PSYS_INVALIDSYS;
	m_active		= FALSE;
	m_startTime		= 0.0f;	
	m_maxParticles	= MAX_PARTICLES;
	m_totalForces	= 0;
	m_emitters.reserve(10);
	for(int i=0; i<MAX_FORCES; i++)
		m_forces[i].Set(0.0f, 0.0f, 0.0f);
}

// 
// INITIALIZATION
// 
VOID CParticle_Sys::InitAll(PARTICLESYS_TYPE sType, int maxParticles)
{
	m_sysType		= sType;
	m_active		= FALSE;	
	m_startTime		= timeGetTime() / 1000.0f;
	m_maxParticles	= maxParticles;
}

// 
// UPDATE SYSTEM
// 
VOID CParticle_Sys::UpdateSystem(GLfloat elapsedTime)
{
	// UPDATE EMITTERS
	for(m_emitters_it = m_emitters.begin(); 
		m_emitters_it != m_emitters.end();
		++m_emitters_it)
	{
		(*m_emitters_it)->Update(elapsedTime, m_startTime);
	}
}

// 
// FREE MEMORY
// 
VOID CParticle_Sys::DestroySystem()
{
	m_active = FALSE;

	// FREE MEMORY USED FOR EMITTERS
	for(m_emitters_it = m_emitters.begin(); 
		m_emitters_it != m_emitters.end();)
	{
		(*m_emitters_it)->DestroyAll(); // FREE PARTICLES MEMORY
		delete (*m_emitters_it);
		m_emitters_it = m_emitters.erase(m_emitters_it);
	}
}

// 
// ADD AN EMITTER TO THE VECTOR
// 
VOID CParticle_Sys::AddEmitter(CParticle_Emitter *newEmitter)
{
	assert(newEmitter!=NULL && "CParticle_Sys::AddEmitter");
	
	// ADD GLOBAL FORCES...
	for(int i=0; i<m_totalForces; i++)
		newEmitter->AddGlobalForce(m_forces[i]);

	// SET MAX PARTICLES
	//newEmitter->Set_MaxParticles(m_maxParticles);

	// ADD EMITTER
	m_emitters.push_back(newEmitter);
}

// 
// RENDER
// 
VOID CParticle_Sys::RenderParticleSystem()
{
	glDepthMask(GL_FALSE);

	// RENDER EMITTERS
	for(m_emitters_it = m_emitters.begin(); 
		m_emitters_it != m_emitters.end();
		++m_emitters_it)
	{
		(*m_emitters_it)->Render();
	}

	glDepthMask(GL_TRUE);
}

// 
// SET SYS FORCE
// 
VOID CParticle_Sys::SetSysForce(int forceIndex, float force[3])
{
	if((forceIndex>=0) && (forceIndex < MAX_FORCES))
	{
		MGL::Vector3 reverseForce = (m_forces[forceIndex] * (-1.0f));
		m_forces[forceIndex].Set(force[0], force[1], force[2]);
		m_totalForces = (forceIndex + 1);

		// UPDATE EMITTERS WITH FORCE	
		for(m_emitters_it = m_emitters.begin(); 
			m_emitters_it != m_emitters.end();
			++m_emitters_it)
		{
			// Clear the global force that was added before...
			(*m_emitters_it)->AddGlobalForce(reverseForce);
			// Add new force
			(*m_emitters_it)->AddGlobalForce(m_forces[forceIndex]);
		}
	}
}

// 
// GET THE EMITTER INDEX FOR A SPECIFIED EMITTER ID
// 
int CParticle_Sys::GetEmitterIndex(char *emitterID)
{
	int index=-1, i=0;
	char thisID[40];

	m_emitters_it = m_emitters.begin();
	while( (m_emitters_it != m_emitters.end()) && (index==-1))
	{
		(*m_emitters_it)->Get_ID(thisID);
		if(0==memcmp(thisID, emitterID, sizeof(thisID)))
			index = i;
		
		++i;
		++m_emitters_it;
	}

	return index;
}

// 
// SET START TIME
// 
VOID CParticle_Sys::SetStartTime()
{
	m_startTime = timeGetTime() / 1000.0f;
}

// ------------------------- END OF FILE -----------------------------------
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




