/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: CParticleScript.h                                         |*/
/*| Author: Joel Villagrana    (joel.villagrana@gmail.com)              |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yyyy | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 11/07/2005 | JVillagrana       | Initial Version                    |*/
/*|---------------------------------------------------------------------|*/
/*| 04/08/2005 | JVillagrana    @1 | Added some optimization methods    |*/
/*|---------------------------------------------------------------------|*/
/*************************************************************************/

#ifndef _PSCRIPTING_H
#define _PSCRIPTING_H		

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include <vector>
#include <string>
using namespace std;

#include "../Util/helper_defs.h"

#define SCRIPT_LINE_LENGTH	100
#define SCRIPT_TOKEN_LENGTH	40

// OUR ALLOWED STATES
enum FSM_STATE
{
	STATE_NONE,
	STATE_IDENTIFIER,
	STATE_NUMBER,
	STATE_NUMBERSIGN,
	STATE_NUMBERDOT,
	STATE_DIGITONLY,
	STATE_LITERAL,
	STATE_COMMENT
};

#define NUM_KEYWORDS 30

// OUR ALLOWED TOKEN TYPES
enum SCR_TOKEN_TYPE
{
	// ==============
	// GENERAL TOKENS
	// ==============
	KEYWORD_PARTICLESYSTEM = 0,
	KEYWORD_SYSTEMPROPERTIES,
	KEYWORD_EMITTER,
	KEYWORD_EVENTAT,
	KEYWORD_SYSPROPCHG,
	KEYWORD_EMITTERPROPCHG,
	KEYWORD_RANDOM,
	KEYWORD_VECTOR,
	KEYWORD_COLOR,
	KEYWORD_ATPOS,
	KEYWORD_DIST,
	CURLYBRACE_OPEN,
	CURLYBRACE_CLOSE,
	BRACKET_OPEN,
	BRACKET_CLOSE,
	PARENTHESIS_OPEN,
	PARENTHESIS_CLOSE,
	BLENDINGMODE_ID,
	EMITTERTYPE_ID,
	PARTICLEREP_ID,
	COMMA,
	IDENTIFIER,
	LITERAL,
	NUMBER,
	SEMI_COLON,
	ASSIGNMENT_OPERATOR,
	NONE,
	// ======================
	// PARTICLE SYSTEM TOKENS
	// ======================
	KEYWORD_SYSP_MAXPARTICLES,
	KEYWORD_SYSP_FORCE,
	// ==============
	// EMITTER TOKENS
	// ==============
	KEYWORD_EM_POSITION,
	KEYWORD_EM_TYPE,
	KEYWORD_EM_PARTICLEREP,
	KEYWORD_EM_TEXTURE,
	KEYWORD_EM_EMISSIONRATE,
	KEYWORD_EM_BATCHSIZERANGE,
	KEYWORD_EM_BLENDINGMODES,
	KEYWORD_EM_VERTEXSHADER,
	KEYWORD_EM_PIXELSHADER,
	KEYWORD_EM_FORCE,
	KEYWORD_EM_LIFERANGE,
	KEYWORD_EM_SIZERANGE,
	KEYWORD_EM_STATE,
	// ===============
	// PARTICLE TOKENS
	// ===============
	KEYWORD_PARTICLE_COLOR,
	KEYWORD_PARTICLE_SIZE,
	KEYWORD_PARTICLE_WEIGHT,
	KEYWORD_PARTICLE_FADE
};

/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
=====================================================================================
					DEFINITION OF THE TokenNode CLASS
=====================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
class CTokenNode
{
	private:
		char			m_token[SCRIPT_TOKEN_LENGTH];
		SCR_TOKEN_TYPE	m_type;
		int				m_scriptLine;

	public:
		CTokenNode(char *tk, SCR_TOKEN_TYPE tt, int line);
		~CTokenNode();
		
		int		GetScrLine()		{	return m_scriptLine;						}
		void	SetScriptLine(int s){	m_scriptLine = s;							}
		void	GetToken(char *tk)	{	memcpy(tk, m_token, SCRIPT_TOKEN_LENGTH);	}
		void	SetToken(char *tk)	{	memcpy(m_token, tk, SCRIPT_TOKEN_LENGTH);	}
		SCR_TOKEN_TYPE	GetTokenType()					{	return m_type;	}
		void			GetTokenTypeStr(char *str);
		void			SetTokenType(SCR_TOKEN_TYPE t)	{	m_type = t;		}
};

// ========================================================================
// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^|
// STL DEFINITIONS                                                        |
// ------------------------------------------------------------------------
// VECTOR OF TOKENS
typedef vector<CTokenNode *> STL_tokenList;
typedef STL_tokenList::iterator STL_tokenIter;
// VECTOR OF STRINGS
typedef vector<char *> STL_stringList;
typedef STL_stringList::iterator STL_strIter;

/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
=====================================================================================
					DEFINITION OF THE CParticleScript CLASS
=====================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
class CParticleScript
{
	private:
		FSM_STATE		m_state;
		int				m_lineNumber, 
						m_numErrors, 
						m_currError;

		char			m_filename[SCRIPT_LINE_LENGTH],
						m_errorString[SCRIPT_LINE_LENGTH], 
						m_lex[1000], 
						m_token[SCRIPT_TOKEN_LENGTH],
						m_line[SCRIPT_LINE_LENGTH],
						m_keywords[NUM_KEYWORDS][SCRIPT_TOKEN_LENGTH];
		
		STL_tokenList	m_tokenList;
		STL_tokenIter	m_tokenIterator;

		STL_stringList	m_scriptLines,
						m_errorsList,
						m_emitterIDs;

		STL_strIter		m_scriptIterator;

	public:

		CParticleScript();
		~CParticleScript();

		// INIT AND CLEANUP
		void	InitAll();
		void	DestroyAll();

		// ========================================================================================

		// THESE METHODS GET THE SCRIPT LINES STL LIST POPULATED
		bool	GetScriptFromFile(char *filename);
		bool	AddScriptLine(char *line);
		
		// THESE TWO MAIN METHODS PROCESS THE SCRIPT AND POPULATE THE TOKENS LIST
		bool	ProcessScriptLines();
		bool	TokenizeScriptLine();
		
		bool	AddToken(char *token, SCR_TOKEN_TYPE type);	// ADD TOKEN TO LIST
		void	LogExpectedToken(char *eToken, bool goBack);// LOG AN SPECIFIC TOKEN WAS EXPECTED
		void	AddError(char *error);						// ADD ERROR TO LIST		
		bool	HasMoreErrors(char *error);					// GET NEXT ERROR
		void	SaveIdentifier();							// SAVE IDENTIFIER
		bool	IdentifierDefined();						// CHECK IF IDENTIFIER IS DEFINED
		bool	IsAlpha(char c);							// CHECK IF CHAR IS A..Z
		bool	IsValidChar(char c);						// CHECKS IF VALID CHAR
		bool	IsKeyword(char *word);						// CHECKS IF IT'S A KEYWORD
		bool	IsBlendingMode(char *word, GLenum *b);		// CHECKS IF IT'S A BLENDING MODE
		bool	IsEmitterType(char *word);					// CHECKS IF IT'S AN EMITTER TYPE
		bool	IsParticleRep(char *word);					// CHECKS IF IT'S A PARTICLE REP
		bool	IsSpace(char c);							// CHECKS IF SPACE (\n, \t, \r, ' ')
		bool	CheckToken(SCR_TOKEN_TYPE t,				// CHECKS CURRENT TOKEN
						   char *expected);
		void	OutputTokensToFile(FILE *pFile);			// CREATE TOKENS FILE (TEMP)
		SCR_TOKEN_TYPE	KeywordType(char *word);			// GET KEYWORD TYPE
				
		// ========================================================================================

		// ONCE WE HAVE ALL OUR TOKENS THIS METHOD CHECKS IF THE SCRIPT IS VALID
		bool	ParseScript();
		
		// METHODS TO GET THE POINTER TO OUT NEXT/PREVIOUS TOKEN
		bool	GetNextToken();
		bool	GetPreviousToken();

		bool	ParseNumber();								// CHECK IF NUMBER IS VALID
		bool	ParseVector();								// CHECK IF VECTOR IS VALID
		bool	ParseColor();								// CHECK IF COLOR IS VALID
		
		// PARSING METHODS 
		// THESE ARE BASED ON HOW WE DEFINED OUR GRAMMAR
		bool	ParseSystemProperties();
		bool	ParseEmitters();
		bool	ParseEvents();
		// for EVENTS
		bool	ParseSystemPropChg();
		bool	ParseEmitterPropChg();
		// for PARTICLE SYSTEMS
		bool	Parse_SYSP_MaxParticles();
		bool	Parse_SYSP_Force(); 
		// for EMITTERS
		bool	Parse_EM_Position();
		bool	Parse_EM_Type();
		bool	Parse_EM_ParticleRep();	
		bool	Parse_EM_Texture();
		bool	Parse_EM_EmissionRate();
		bool	Parse_EM_BatchSize();
		bool	Parse_EM_BlendingModes();
		bool	Parse_EM_VertexShader();
		bool	Parse_EM_PixelShader();
		bool	Parse_EM_Force();
		bool	Parse_EM_LifeRange();
		bool	Parse_EM_SizeRange();
		bool	Parse_EM_EmitterState();		
		// for PARTICLES
		bool	Parse_Particle_Color();
		bool	Parse_Particle_Size();
		bool	Parse_Particle_Weight();
		bool	Parse_Particle_Fade();		

		// THERE ARE QUITE A FEW PROPERTIES THAT ARE ONLY ASSIGNED ONLY A NUM, A LITERAL
		// OR A VECTOR, SO FOR THOSE METHODS, CALL THESE ONES INSTEAD TO AVOID DUPLICATING
		// CODE ---- @1
		bool	ParseAssignmentNum();						
		bool	ParseAssignmentLit();
		bool	ParseAssignmentVec();
		// ------------

		// ========================================================================================

		SCR_TOKEN_TYPE Get_PropertyType();
		bool	SetIterator_SystemProperties();
		bool	SetIterator_Emitters();
		bool	SetIterator_Events();
		bool	MoreSystemProperties();
		bool	MoreEmitters();
		bool	MoreEmitterProperties();
		bool	MoreEvents();
		bool	MoreEmitterChanges(char *emitterID);
		bool	More_EmitterORParticle_Changes();
		bool	Get_IntValues(int *intVar, int numIntsToGet);
		bool	Get_FloatValues(float *floatVar, int numFloatsToGet);
		bool	Get_StringValues(char *literal, int numStringsToGet);
		bool	Get_Identifier(char *identifier, int numIDsToGet);
		bool	Get_TypeValues(PSYS_EMITTERTYPE *emitterType, int numTypesToGet);
		bool	Get_RepValues(PARTICLE_REP *particleRep, int numRepToGet);
		bool	Get_BlendValues(GLenum *blends, int numBlendsToGet);
		bool	Get_StateValues(int *intVar, int numIntsToGet);

		// ========================================================================================

		int		GetScriptStringLength()
				{	return SCRIPT_LINE_LENGTH;								}
		int		GetTokenLength()
				{	return SCRIPT_TOKEN_LENGTH;								}
		void	GetLastError(char *error)
				{	memcpy(error, m_errorString, sizeof(m_errorString));	}
		void	ClearLastError()
				{	memset(m_errorString, 0, sizeof(m_errorString));		}
};

#endif _PSCRIPTING_H

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% END OF FILE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



