/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"
#include "../../_common/MGLTargaImage.h"
#include <cmath>

GLfloat LIGHT_POS[4] = { -55.0, 25.0, 55.0, 1.0 };
GLfloat CAMERA_POS[4] = { 0.0, 100.0, 150.0, 1.0 };

GLfloat WHITE[] = { 1.0, 1.0, 1.0, 1.0 };
GLfloat BLACK[] = { 0.0, 0.0, 0.0, 0.0 };
GLfloat GREY[] = { 0.3f, 0.3f, 0.3f, 0.3f };
GLfloat DARK_GREY[] = { 0.1f, 0.1f, 0.1f, 0.1f };

MGLApp::MGLApp()
{
  m_bounce = 0.0;
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	if (!GLeeInit())
		return false;

  m_quadric = gluNewQuadric();
  gluQuadricNormals(m_quadric, GLU_SMOOTH);

  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);

  // set up the shadow texture in texture unit 1
//  glActiveTexture(GL_TEXTURE1);
  glGenTextures(1, &m_shadowMap);
  glBindTexture(GL_TEXTURE_2D, m_shadowMap);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_INTENSITY);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);

  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
  glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
  glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
  glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);


  CPrimitiveFactory fact;
  m_ground = fact.MakePatch(100, 100, 20, 20);

  glEnable(GL_LIGHT0);
  glEnable(GL_COLOR_MATERIAL);
  glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);

  glPolygonOffset(4.0f, 0.0f);

  return true;
}

bool MGLApp::Shutdown()
{
  CPrimitiveFactory fact;
  fact.DeletePrimitive(&m_ground);
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;

  m_windowHeight = height;
  m_windowWidth = width;

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
}

void MGLApp::Update( float dt )
{
  m_bounce += 10 * dt;
}

bool showShadowMap = false;
bool noShadows = false;
bool ambientShadowAvailable = false;

void MGLApp::Render()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  GenerateShadowMap();

  // Track camera angle
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0f, 1.0f, 1.0f, 1000.0f);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(CAMERA_POS[0], CAMERA_POS[1], CAMERA_POS[2], 
    0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
  glViewport(0, 0, m_windowWidth, m_windowHeight);

  // Track light position
  glLightfv(GL_LIGHT0, GL_POSITION, LIGHT_POS);

  // Clear the window with current clearing color
  glClear(GL_DEPTH_BUFFER_BIT);

  if (showShadowMap)
  {
    // Display shadow map for educational purposes
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_TEXTURE);
    glPushMatrix();
    glLoadIdentity();
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_NONE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    // Show the shadowMap at its actual size relative to window
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f);
    glVertex2f(-1.0f, -1.0f);
    glTexCoord2f(1.0f, 0.0f);
    glVertex2f(((GLfloat)512/(GLfloat)m_windowWidth)*2.0-1.0f, 
      -1.0f);
    glTexCoord2f(1.0f, 1.0f);
    glVertex2f(((GLfloat)512/(GLfloat)m_windowWidth)*2.0-1.0f, 
      ((GLfloat)512/(GLfloat)m_windowHeight)*2.0-1.0f);
    glTexCoord2f(0.0f, 1.0f);
    glVertex2f(-1.0f, 
      ((GLfloat)512/(GLfloat)m_windowHeight)*2.0-1.0f);
    glEnd();
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    gluPerspective(45.0f, 1.0f, 1.0f, 1000.0f);
    glMatrixMode(GL_MODELVIEW);
  }
  else if (noShadows)
  {
    // Set up some simple lighting
    glLightfv(GL_LIGHT0, GL_AMBIENT, GREY);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, WHITE);

    // Draw objects in the scene
    RenderObjects(false);
  }
  else
  {
    GLfloat sPlane[4] = {1.0f, 0.0f, 0.0f, 0.0f};
    GLfloat tPlane[4] = {0.0f, 1.0f, 0.0f, 0.0f};
    GLfloat rPlane[4] = {0.0f, 0.0f, 1.0f, 0.0f};
    GLfloat qPlane[4] = {0.0f, 0.0f, 0.0f, 1.0f};

    if (!ambientShadowAvailable)
    {
      GLfloat lowAmbient[4] = {0.1f, 0.1f, 0.1f, 1.0f};
      GLfloat lowDiffuse[4] = {0.35f, 0.35f, 0.35f, 1.0f};

      // Because there is no support for an "ambient"
      // shadow compare fail value, we'll have to
      // draw an ambient pass first...
      glLightfv(GL_LIGHT0, GL_AMBIENT, GREY);
      glLightfv(GL_LIGHT0, GL_DIFFUSE, DARK_GREY);

      // Draw objects in the scene
      RenderObjects(false);

      // Enable alpha test so that shadowed fragments are discarded
      glAlphaFunc(GL_GREATER, 0.9f);
      glEnable(GL_ALPHA_TEST);
    }

    glLightfv(GL_LIGHT0, GL_AMBIENT, GREY);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, WHITE);

    // Set up shadow comparison
    glEnable(GL_TEXTURE_2D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, 
      GL_COMPARE_R_TO_TEXTURE);

    // Set up the eye plane for projecting the shadow map on the scene
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
    glEnable(GL_TEXTURE_GEN_R);
    glEnable(GL_TEXTURE_GEN_Q);
    glTexGenfv(GL_S, GL_EYE_PLANE, sPlane);
    glTexGenfv(GL_T, GL_EYE_PLANE, tPlane);
    glTexGenfv(GL_R, GL_EYE_PLANE, rPlane);
    glTexGenfv(GL_Q, GL_EYE_PLANE, qPlane);

    // Draw objects in the scene
    RenderObjects(false);

    glDisable(GL_ALPHA_TEST);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);
    glDisable(GL_TEXTURE_GEN_Q);
  }

/*
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  // render the shadow map
  GenerateShadowMap();

	// set up the camera
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

  gluLookAt(CAMERA_POS[0], CAMERA_POS[1], CAMERA_POS[2], 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);


  // render the scene with only ambient and minor diffuse lighting
  glEnable(GL_TEXTURE_2D);
  glEnable(GL_LIGHTING);
  glLightfv(GL_LIGHT0, GL_POSITION, LIGHT_POS);
  glLightfv(GL_LIGHT0, GL_AMBIENT, GREY);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, DARK_GREY);
  glLightfv(GL_LIGHT0, GL_SPECULAR, BLACK);
  RenderObjects(false);


  // now render the scene with diffuse and specular lighting
  glLightfv(GL_LIGHT0, GL_POSITION, LIGHT_POS);
  glLightfv(GL_LIGHT0, GL_AMBIENT, BLACK);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, WHITE);
  glLightfv(GL_LIGHT0, GL_SPECULAR, WHITE);

  glActiveTexture(GL_TEXTURE1);
  glEnable(GL_TEXTURE_2D);
  GLfloat sPlane[4] = {1.0f, 0.0f, 0.0f, 0.0f};
  GLfloat tPlane[4] = {0.0f, 1.0f, 0.0f, 0.0f};
  GLfloat rPlane[4] = {0.0f, 0.0f, 1.0f, 0.0f};
  GLfloat qPlane[4] = {0.0f, 0.0f, 0.0f, 1.0f};
  glEnable(GL_TEXTURE_GEN_S);
  glEnable(GL_TEXTURE_GEN_T);
  glEnable(GL_TEXTURE_GEN_R);
  glEnable(GL_TEXTURE_GEN_Q);
  glTexGenfv(GL_S, GL_EYE_PLANE, sPlane);
  glTexGenfv(GL_T, GL_EYE_PLANE, tPlane);
  glTexGenfv(GL_R, GL_EYE_PLANE, rPlane);
  glTexGenfv(GL_Q, GL_EYE_PLANE, qPlane);
  glActiveTexture(GL_TEXTURE0);

  RenderObjects(false);

  glActiveTexture(GL_TEXTURE1);
  glDisable(GL_TEXTURE_2D);
  glActiveTexture(GL_TEXTURE0);
*/
}

void MGLApp::RenderObjects(bool isShadowPass)
{
  glPushMatrix();
  glColor3f(0.2f, 0.9f, 0.3f);
  glTranslatef(0.0, 5 + 5 * sinf(m_bounce), 0.0);
  glRotatef(-90.0, 1.0, 0.0, 0.0);
  gluCylinder(m_quadric, 10, 0, 20, 20, 20);
  glPopMatrix();

  glPushMatrix();
  glColor3f(0.75f, 0.2f, 0.1f);
  glTranslatef(15.0, 5.0, 12 * cosf(m_bounce/3.0f));
  gluSphere(m_quadric, 5, 20, 20);
  glPopMatrix();

  glPushMatrix();
  glColor3f(0.5f, 0.1f, 0.9f);
  glTranslatef(28.0, 6.0, -24.0);
  glRotatef(30.0, 0.0, 1.0, 0.0);
  gluCylinder(m_quadric, 6, 6, 20, 20, 20);
  glPopMatrix();

  glColor3f(0.36f, 0.34, 0.29);
  m_ground->Draw(false, true, false, false);
}

void MGLApp::GenerateShadowMap()
{
  GLfloat lightToSceneDistance, nearPlane, fieldOfView;
  GLfloat lightModelview[16], lightProjection[16];

  // Save the depth precision for where it's useful
  lightToSceneDistance = sqrt(LIGHT_POS[0] * LIGHT_POS[0] + 
    LIGHT_POS[1] * LIGHT_POS[1] + 
    LIGHT_POS[2] * LIGHT_POS[2]);
  nearPlane = lightToSceneDistance - 150.0f;
  if (nearPlane < 50.0f)
    nearPlane = 50.0f;
  // Keep the scene filling the depth texture
  fieldOfView = 17000.0f / lightToSceneDistance;

  glViewport(0, 0, 512, 512);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-12.0, 12.0, -20.0, 5.0, 13.0, 150.0);
  glGetFloatv(GL_PROJECTION_MATRIX, lightProjection);
  // Switch to light's point of view
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(LIGHT_POS[0], LIGHT_POS[1], LIGHT_POS[2], 
    0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
  glGetFloatv(GL_MODELVIEW_MATRIX, lightModelview);

  // Clear the window with current clearing color
  glClear(GL_DEPTH_BUFFER_BIT);

  // All we care about here is resulting depth values
  glShadeModel(GL_FLAT);
  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_NORMALIZE);
  glColorMask(0, 0, 0, 0);

  // Overcome imprecision
  glEnable(GL_POLYGON_OFFSET_FILL);

  // Draw objects in the scene
  RenderObjects(false);

  // Copy depth values into depth texture
  glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, 
    0, 0, 512, 512, 0);

  // Restore normal drawing state
  glShadeModel(GL_SMOOTH);
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_NORMALIZE);
  glColorMask(1, 1, 1, 1);
  glDisable(GL_POLYGON_OFFSET_FILL);

  // Set up texture matrix for shadow map projection
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  glTranslatef(0.5f, 0.5f, 0.5f);
  glScalef(0.5f, 0.5f, 0.5f);
  glMultMatrixf(lightProjection);
  glMultMatrixf(lightModelview);
  /*
  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
  glViewport( 0, 0, 512, 512 );
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glFrustum(-0.25, 0.25, -0.5, 0.1, 0.3, 150.0);

  // set up the camera to be at the light's position
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  gluLookAt(LIGHT_POS[0], LIGHT_POS[1], LIGHT_POS[2], 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

  // set up the texture matrix
  GLfloat modelview[16], projection[16];
  glGetFloatv(GL_PROJECTION_MATRIX, projection);
  glGetFloatv(GL_MODELVIEW_MATRIX, modelview);
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  glTranslatef(0.5f, 0.5f, 0.5f);
  glScalef(0.5f, 0.5f, 0.5f);
  glMultMatrixf(projection);
  glMultMatrixf(modelview);
  glMatrixMode(GL_MODELVIEW);

  glDisable(GL_LIGHTING);
  glEnable(GL_POLYGON_OFFSET_FILL);
  RenderObjects(true);

  glActiveTexture(GL_TEXTURE1);
  glBindTexture(GL_TEXTURE_2D, m_shadowMap);
  glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT24, 0, 0, 512, 512, 0);
  glActiveTexture(GL_TEXTURE0);

  SetupProjection(m_windowWidth, m_windowHeight);
  glClear( GL_DEPTH_BUFFER_BIT );
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
  glDisable(GL_POLYGON_OFFSET_FILL);
  */
}