#include "prim.h"


//////////////////////////////////////////////////////////////////////////
// Primitive struct implementation
//////////////////////////////////////////////////////////////////////////
Primitive::Primitive( GLenum nType, int nVerts, int nTris )
	: nElemType( nType )
	, nVertices( nVerts )
	, nTriangles( nTris )
	, pVert( new GLfloat[3 * nVerts] )
	, pCol( new GLubyte[4 * nVerts] )
	, pNormal( new GLfloat[3 * nVerts] )
	, pTexCoord( new GLfloat[2 * nVerts] )
	, pIdx( new GLushort[3 * nTris] )
{}

Primitive::~Primitive()
{
	if( pVert )		delete[] pVert;
	if( pCol )		delete[] pCol;
	if( pNormal )	delete[] pNormal;
	if( pTexCoord )	delete[] pTexCoord;
	if( pIdx )		delete[] pIdx;
}

void Primitive::Draw( bool bWire, bool bNormalArray, bool bColorArray, bool bTexCoordArray )
{
	glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_COLOR_ARRAY );
	glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

	if( pVert )
	{
		glEnableClientState( GL_VERTEX_ARRAY );
		glVertexPointer( 3, GL_FLOAT, 0, pVert );
	}

	if( pCol && bColorArray )
	{
		glEnableClientState( GL_COLOR_ARRAY );
		glColorPointer( 4, GL_UNSIGNED_BYTE, 0, pCol );
	}

	if( pNormal && bNormalArray )
	{
		glEnableClientState( GL_NORMAL_ARRAY );
		glNormalPointer( GL_FLOAT, 0, pNormal );
	}

	if( pTexCoord && bTexCoordArray )
	{
    glClientActiveTexture(GL_TEXTURE1);
		glEnableClientState( GL_TEXTURE_COORD_ARRAY );
		glTexCoordPointer( 2, GL_FLOAT, 0, pTexCoord );
    glClientActiveTexture(GL_TEXTURE0);
	}

	glDrawElements( bWire ? GL_LINE_STRIP : nElemType, 3 * nTriangles, GL_UNSIGNED_SHORT, pIdx );
}


//////////////////////////////////////////////////////////////////////////
// CPrimitiveFactory class implementation
//////////////////////////////////////////////////////////////////////////
#define SETVERT(p,x,y,z)	{ (p)[0]=(x); (p)[1]=(y); (p)[2]=(z); (p)+=3;				}
#define SETCOL(p,r,g,b,a)	{ (p)[0]=(r); (p)[1]=(g); (p)[2]=(b); (p)[3]=(a); (p)+=4;	}
#define SETNORMAL(p,x,y,z)	{ (p)[0]=(x); (p)[1]=(y); (p)[2]=(z); (p)+=3;				}
#define SETTEXCOORD(p,u,v)	{ (p)[0]=(u); (p)[1]=(v); (p)+=2;							}
#define SETTRI(p,i1,i2,i3)	{ (p)[0]=(i1); (p)[1]=(i2); (p)[2]=(i3); (p) += 3;			}

void CPrimitiveFactory::DeletePrimitive( Primitive **ppPrim )
{
	if( ppPrim && *ppPrim )
	{
		delete (*ppPrim);
		(*ppPrim) = NULL;
	}
}

Primitive* CPrimitiveFactory::MakePatch( float fWidth, float fHeight, int nCols, int nRows )
{
	if( fWidth < 0.f || fHeight < 0.f || nCols < 1 || nRows < 1 )
		return NULL;

	Primitive *pPrim = new Primitive( GL_TRIANGLES, (nCols+1)*(nRows+1), nCols*nRows*2 );
	Primitive temp = *pPrim; // copy and use as an iterator
	int i = 0;

	// loop over all patch vertices
	float u = 0.f;
	for( int r=0 ; r<=nRows ; r++, u += 1.f/nRows )
	{
		float v = 0.f;
		for( int c=0 ; c<=nCols ; c++, v += 1.f/nCols )
		{
			// set up this vertex
			SETVERT( temp.pVert, (u-0.5f) * fWidth, 0.f, (v-0.5f) * fHeight );
			SETCOL( temp.pCol, 0xFF, 0xFF, 0xFF, 0xFF );
			SETNORMAL( temp.pNormal, 0.f, 1.f, 0.f );
			SETTEXCOORD( temp.pTexCoord, u, v );

			// form the quad for this row and column
			if( c<nCols && r<nRows )
			{
				SETTRI( temp.pIdx, i, i+nCols+1, i+1 );
				SETTRI( temp.pIdx, i+1, i+nCols+1, i+nCols+2 );
			}

			// increment the index counter
			i++;
		}
	}
	
	// prevent a memory error in temp's destructor
	memset( &temp, 0, sizeof(Primitive) );

	return pPrim;
}
