/*************************************************************************
MGLApp.cpp

Main application class.

Authors   :  Dave Astle (framework) & Angus Dorbie (water render etc.)

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#define  _USE_MATH_DEFINES
#include <math.h>
#include "MGLApp.h"
#include "gui.h"
#include <stdio.h>
#include <MGLVector.h>
#include <MGLTargaImage.h>


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	int i;
	char fname[100];
	if (!GLeeInit())
		return false;

	mode_polarize = false;
	mode_hedgehog = false;
	mode_points = false;
	mode_use_vertex_fog = false;

	turbidity = 0.2f;
	wave_amplitude = 0.4f;
	//sky_color.Set(1.0f, 1.0f, 1.0f);
	sky_color.Set(0.85f, 0.9f, 1.0f);
	clear_color.Set(0.2f, 0.22f, 0.02f);
	//clear_color.Set(0.25f, 0.17f, 0.02f);
	//clear_color.Set(0.02f, 0.05f, 0.15f);
	sprintf( message_str, "Test");

	glFogf(GL_FOG_MODE, GL_LINEAR);
	if(mode_use_vertex_fog && GLEE_EXT_fog_coord)
	{
		glFogf(GL_FOG_COORD_SRC, GL_FOG_COORD);
	}
	glFogf(GL_FOG_START, 10000.0f);
	glFogf(GL_FOG_END, 1000000.0f);
	glFogfv(GL_FOG_COLOR, &(clear_color[0]));

	glGenTextures(1, &uw_tex);
	glBindTexture(GL_TEXTURE_2D, uw_tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	if(GLEE_EXT_texture_filter_anisotropic)
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 4);


	for(i = 0; i<10; i++) // load the image and all MIP levels (they should be on disc
	{
		MGL::TargaImage groundImage;
		sprintf(fname, "..\\Textures\\pebbles%d.tga", i);
		if(groundImage.Load(fname))
		{
			glTexImage2D(GL_TEXTURE_2D, i, GL_RGB, groundImage.GetWidth(), groundImage.GetHeight(), 0, GL_RGB, GL_UNSIGNED_BYTE, groundImage.GetImage());
		}
		groundImage.Release();
	}

	GenerateMeshes();

	text_display = new fontprint();

	GUI_Init(text_display);
	GUI_SetPolarize(&mode_polarize);
	GUI_SetAmplitude(&wave_amplitude);
	GUI_SetSkyColor(&sky_color[0]);
	GUI_SetWaterColor(&clear_color[0]);
	GUI_SetFog(&turbidity);
	return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );
	app_width = width;
	app_height = height;

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
	glPointSize(5.0f);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
}

void MGLApp::Update( double dt )
{
	float phase;
	static double dphase=195.0;
	static double angle = 0.0f;

	MGL::Vector3 upvec;

	// use elapsed time to update some animation parameters
	dphase += dt*1.5;
	if(dphase > M_PI *200.0)
		dphase -= M_PI *200.0;
	phase = (float)dphase;
	angle += dt*0.1f;
	if(angle > M_PI *2.0)
		angle -= M_PI *2.0;
	phase = (float)dphase;

	// update viewpoint animation
	upvec.Set(0.0f, 1.0f, 0.0f);
	eye_pos.Set((float)water->w_xsize*.5f, (float)water->w_xsize*(.15f +.1f * sinf(phase*.5f)), (float)water->w_xsize*1.05f);
	eye_heading = (float)angle;
	eye_pos.x -= water->w_xsize*0.5f;
	eye_pos.z -= water->w_ysize*0.5f;
	eye_pos = eye_pos.Rotate(-eye_heading, upvec);
	eye_pos.x += water->w_xsize*0.5f;
	eye_pos.z += water->w_ysize*0.5f;
	eye_pitch = -.35f-1.1f*atanf(eye_pos.y/ (water->w_xsize*0.65f));

	// update water mesh animation
	water->CalculateWave(eye_pos, 0.0f, wave_amplitude*0.25f, phase, mode_polarize, &sky_color[0]);
	if(mode_use_vertex_fog) // no need to recompute this if we have the alpha multipass depth + standard range fog hack
		ground->CalculateFog(eye_pos, true, turbidity, &clear_color[0]);
	else
		ground->CalculateFog(eye_pos, false, turbidity, &clear_color[0]);
}

void MGLApp::GenerateMeshes()
{
	water = new gmesh(80, 80, 0.0f, 0.0f, 1.0f, sky_color, 0);
	ground = new gmesh(40, 40, -60.0f, -60.0f, 4.0f, clear_color, uw_tex);
	ground->CalculateWave(eye_pos, -10.0f, 1.5f, 0.0f, false, &sky_color[0]);
	ground->CalculateFog(eye_pos, false, turbidity, &clear_color[0]);
}


void MGLApp::Clear(bool polarized_region)
{
	glClearColor(clear_color[0], clear_color[1], clear_color[2], 0.0f);
	if(polarized_region)
	{
		if(!mode_polarize)
			return;
		//glScissor(app_width/6, app_height/4, app_width-2*app_width/6, app_height-2*app_height/4);
		glScissor(0, 0, app_width/2, app_height);
		glEnable(GL_SCISSOR_TEST);
	}
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	if(!polarized_region)
		glFlush();
	glDisable(GL_SCISSOR_TEST);
}

void MGLApp::Render_sky(void)
{
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);

	glBegin(GL_TRIANGLE_FAN);
	glVertex3f(0.0f, 750.0f, 0.0f);
	glVertex3f(750.0f, 0.0f, 0.0f);
	glVertex3f(500.0f, 0.0f, 500.0f);
	glVertex3f(0.0f, 0.0f, 750.0f);
	glVertex3f(-500.0f, 0.0f, 500.0f);
	glVertex3f(-750.0f, 0.0f, 0.0f);
	glVertex3f(-500.0f, 0.0f, -500.0f);
	glVertex3f(0.0f, 0.0f, -750.0f);
	glVertex3f(500.0f, 0.0f, -500.0f);
	glVertex3f(750.0f, 0.0f, 0.0f);
	glEnd();

	glBegin(GL_TRIANGLE_STRIP);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(750.0f, 0.0f, 0.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(750.0f, -20.0f, 0.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(500.0f, 0.0f, 500.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(500.0f, -20.0f, 500.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(0.0f, 0.0f, 750.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(0.0f, -20.0f, 750.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(-500.0f, 0.0f, 500.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(-500.0f, -20.0f, 500.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(-750.0f, 0.0f, 0.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(-750.0f, -20.0f, 0.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(-500.0f, 0.0f, -500.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(-500.0f, -20.0f, -500.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(0.0f, 0.0f, -750.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(0.0f, -20.0f, -750.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(500.0f, 0.0f, -500.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(500.0f, -20.0f, -500.0f);
	glColor3f(sky_color[0], sky_color[1], sky_color[2]);
	glVertex3f(750.0f, 0.0f, 0.0f);
	glColor3f(clear_color[0], clear_color[1], clear_color[2]);
	glVertex3f(750.0f, -20.0f, 0.0f);
	glEnd();
}

void MGLApp::Render(bool polarized_region)
{
	static float heading=0.0f;

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);

	if(polarized_region)
	{
		if(!mode_polarize)
			return;

		//glScissor(app_width/6, app_height/4, app_width-2*app_width/6, app_height-2*app_height/4);
		glScissor(0, 0, app_width/2, app_height);
		glEnable(GL_SCISSOR_TEST);
	}

	// set up the camera
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glPushMatrix();
	glRotatef(-eye_pitch*180.0f/(float)M_PI, 1.0f, 0.0f, 0.0f);
	glRotatef(-eye_heading*180.0f/(float)M_PI, 0.0f, 1.0f, 0.0f);
	glTranslatef(-eye_pos.x, -eye_pos.y, -eye_pos.z);

	Render_sky();

	ground->Render(polarized_region, mode_polarize, false, false, false, false, true, true, true, mode_use_vertex_fog);
	ground->Render(polarized_region, mode_polarize, false, false, true, false, false, false, false, mode_use_vertex_fog);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	water->Render(polarized_region, mode_polarize, mode_hedgehog, mode_points, true, true, false, false, false, mode_use_vertex_fog);

	// boost sky brightness, note that this doesn't darken UW contribution any more
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	//water->Render(polarized_region, mode_polarize, mode_hedgehog, mode_points, true, true, false, false, false, mode_use_vertex_fog);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glDisable(GL_SCISSOR_TEST);
	glPopMatrix();
	glDisable(GL_FOG);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glTranslatef(-15.0f, 10.0f, -30.0f);
	glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	/*
	text_display->displayStr("Water Demo");
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	glTranslatef(0.0f, -1.5f, 0.0f);
	text_display->displayStr(message_str);
	*/

	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
    glOrtho(gui_left, gui_right, gui_bottom, gui_top, -100.0, 100.0);
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	GUI_Draw();

	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );

	glDisable(GL_TEXTURE_2D);
}

gmesh::gmesh(int x, int y, float xpos, float ypos, float scale, MGL::Vector3 color, GLuint tex)
{
	w_xsize = x;
	w_ysize = y;

	w_verts = new GLfloat[w_xsize*w_ysize*3];
	w_normals = new GLfloat[w_xsize*w_ysize*3];
	w_color = new GLfloat[w_xsize*w_ysize*4];
	w_color_p = new GLfloat[w_xsize*w_ysize*4];
	w_tex = new GLfloat[w_xsize*w_ysize*2];
	w_fog = new GLfloat[w_xsize*w_ysize];
	w_indices = new GLshort[(w_xsize-1)*w_ysize*2];
	tex_handle = tex;


	for(x=0; x<w_xsize; x++)
	{
		for(y=0; y<w_ysize; y++)
		{
			w_verts[(x*w_ysize+y)*3  ] = xpos + scale * (float)x;
			w_verts[(x*w_ysize+y)*3+1] = 0.0f;
			w_verts[(x*w_ysize+y)*3+2] = ypos + scale * (float)y;

			w_normals[(x*w_ysize+y)*3] = 0.0f;
			w_normals[(x*w_ysize+y)*3+1] = 1.0f;
			w_normals[(x*w_ysize+y)*3+2] = 0.0f;

			w_color[(x*w_ysize+y)*4] = color[0];
			w_color[(x*w_ysize+y)*4+1] = color[1];
			w_color[(x*w_ysize+y)*4+2] = color[2];
			w_color[(x*w_ysize+y)*4+3] = 1.0f;

			w_color_p[(x*w_ysize+y)*4] = color[0];
			w_color_p[(x*w_ysize+y)*4+1] = color[1];
			w_color_p[(x*w_ysize+y)*4+2] = color[2];
			w_color_p[(x*w_ysize+y)*4+3] = 1.0f;

			w_tex[(x*w_ysize+y)*2] = 0.5f*(float)x;
			w_tex[(x*w_ysize+y)*2+1] = 0.5f*(float)y;

			if(x < w_xsize-1)
			{
				w_indices[(x*w_ysize+y)*2  ] = x*w_ysize+y;
				w_indices[(x*w_ysize+y)*2+1 ] = (x+1)*w_ysize+y;
			}
		}
	}
}

gmesh::~gmesh()
{
	delete[] w_verts;
	delete[] w_normals;
	delete[] w_color;
	delete[] w_color_p;
	delete[] w_tex;
	delete[] w_fog;
	delete[] w_indices;
}

void gmesh::Render(bool polarized_region, bool mode_polarize, bool mode_hedgehog, bool mode_points, bool transparent, bool zfillfirst, bool lit, bool tex, bool fog, bool vfog)
{
	int i;

	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);

	if(tex)
		glBindTexture(GL_TEXTURE_2D, tex_handle);

	if(fog)
	{
		if(vfog && GLEE_EXT_fog_coord)
		{
			glEnableClientState(GL_FOG_COORD);
			glFogCoordPointer(GL_FLOAT, 0, w_fog);
		}
		//glEnable(GL_FOG);
	}

	if(lit)
	{
		GLfloat worldpos[] = {0.0f, 1.0f, 1.0f, 0.0f};
		GLfloat ucol[] = {0.5f, 0.5f, 0.5f, 1.0f};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, ucol);
		glLightfv(GL_LIGHT0, GL_POSITION, worldpos);
		glEnable(GL_LIGHT0);
		glEnable(GL_LIGHTING);

		glDisableClientState(GL_COLOR_ARRAY);
	}
	else
	{
		glEnableClientState(GL_COLOR_ARRAY);
	}

	if(tex)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glEnable(GL_TEXTURE_2D);
	}
	else
	{
		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	}

	glVertexPointer(3, GL_FLOAT, 0, w_verts);
	glNormalPointer(GL_FLOAT, 0, w_normals);
	if(polarized_region)
		glColorPointer(4, GL_FLOAT, 0, w_color_p);
	else
		glColorPointer(4, GL_FLOAT, 0, w_color);
	glTexCoordPointer(2, GL_FLOAT, 0, w_tex);

	if(zfillfirst)
	{
		// write to depth buffer only for order invariant transparent self occlusion
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		for(i=0;i<w_xsize-1;i++)
		{
			glDrawElements(GL_TRIANGLE_STRIP, w_ysize*2, GL_UNSIGNED_SHORT, w_indices+i*w_ysize*2);
		}
		// pass 2 assumes state invariant z fragment values (would use glPolygonOffset to fix on older cards)
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	}
	if(transparent)
	{
		glEnable(GL_BLEND);
	}
	for(i=0;i<w_xsize-1;i++)
	{
		glDrawElements(GL_TRIANGLE_STRIP, w_ysize*2, GL_UNSIGNED_SHORT, w_indices+i*w_ysize*2);
	}
	if(transparent)
	{
		glDisable(GL_BLEND);
	}
	if(lit)
	{
		glDisable(GL_LIGHTING);
	}

	if(mode_points)
	{
		// visualize vertices as points
		glDisable(GL_COLOR_ARRAY);
		glColor4f(1.0f, 0.25f, 0.0f, 1.0f);
		glDrawArrays(GL_POINTS, 0, w_xsize*w_ysize);
	}

	if(mode_hedgehog)
	{
		glDisable(GL_COLOR_ARRAY);
		glColor4f(0.0f, 0.5f, 1.0f, 1.0f);

		// visualize normals hedgehog style
		int x, y;
		for(x=0;x<w_xsize;x++)
		{
			glBegin(GL_LINES);
			for(y=0; y<w_ysize; y++)
			{
				glVertex3f(w_verts[(x*w_ysize+y)*3],w_verts[(x*w_ysize+y)*3+1],w_verts[(x*w_ysize+y)*3+2]);
				glVertex3f(w_verts[(x*w_ysize+y)*3]+w_normals[(x*w_ysize+y)*3],w_verts[(x*w_ysize+y)*3+1]+w_normals[(x*w_ysize+y)*3+1],w_verts[(x*w_ysize+y)*3+2]+w_normals[(x*w_ysize+y)*3+2]);
			}
			glEnd();
		}
	}
	if(tex)
	{
		glDisable(GL_TEXTURE_2D);
	}
	if(fog)
	{
		glDisable(GL_FOG);
		if(vfog && GLEE_EXT_fog_coord)
		{
			glDisableClientState(GL_FOG_COORD);
		}
	}
}

void gmesh::CalculateFog(MGL::Vector3 eye_pos, bool use_vertex_fog, float turbidity, float *color)
{
	MGL::Vector3 view_vec, upvec;
	int vert_offset3, vert_offset4;

	turbidity *= .1f;

	upvec.Set(0.0f, 1.0f, 0.0f);
	int x, y;
	for(x=0; x<w_xsize; x++)
	{
		for(y=0; y<w_ysize; y++)
		{
			if(use_vertex_fog)
			{
				// abandoned this, may revisit this if vertex fog is found to be functional
				// simple depth basis rather than range
				//w_fog[x*w_ysize+y] = 2.0f*-w_verts[(x*w_ysize+y)*3+1];
			}
			else
			{
				vert_offset3 = x*w_ysize+y;
				vert_offset4 = vert_offset3 *4;
				vert_offset3 *= 3;
				// calculate the view vector
				view_vec.Set(	eye_pos.x-w_verts[vert_offset3],
								eye_pos.y-w_verts[vert_offset3+1],
								eye_pos.z-w_verts[vert_offset3+2]	);
				view_vec.Normalize();


				w_color_p[vert_offset4] =
				w_color[vert_offset4] = *color;
				w_color_p[vert_offset4+1] =
				w_color[vert_offset4+1] = *(color+1);
				w_color_p[vert_offset4+2] =
				w_color[vert_offset4+2] = *(color+2);

				w_color_p[vert_offset4+3] =
				w_color[vert_offset4+3] = turbidity*-w_verts[vert_offset3+1]/view_vec.DotProduct(upvec);


				//0.065f * -w_verts[vert_offset3+1];
			}
		}
	}
}

void gmesh::CalculateWave(MGL::Vector3 eye_pos, float level, float amplitude, float phase, bool mode_polarize, float *color)
{
	int x, y;
	float *v1, *v2, *v3, *v4;
	int vert_offset3, vert_offset4;
	float fx, fy;
	float incident_angle, refract_angle, coef_Rs, coef_Rp;
	MGL::Vector3 vec1, vec2, normal_vec, view_vec;

	// calculate displacement based on superimposed animated sine waves
	for(x=0, fx = 0.0; x<w_xsize; x++, fx+=0.5f)
	{
		for(y=0, fy=0; y<w_ysize; y++, fy+=0.5f)
		{
			// way more expensive than it needs to be but this is just for interesting waves
			w_verts[(x*w_ysize+y)*3+1] = level + amplitude *(
				0.75f * powf(sin(1.3f*fx+phase), 4.0f)+ 0.5f * powf(sin(1.2f*fy+fx*.3f+phase), 4.0f)+
				1.5f * sin(fx*.4f+fy*.8f+phase) + 1.5f * sin(fx*.7f+fy*.7f+phase)+
				0.35f * (sin(2.0f*(fx-phase))+sin(2.0f*(fy-phase)))+
				0.5f * powf(0.5f * (sin(1.6f*(fx+fy+phase)) + sin(1.6f*(fx-fy+phase)) ), 4.0f)+
				0.5f * (sin(1.6f*(fx+fy-phase)) + sin(1.6f*(fx-fy-phase)) )+
				0.75f * (sin(0.75f*(fx-phase))+sin(0.75f*(fy-phase)))+
				0.45f * (sin(1.15f*(fx+fy+phase)) + sin(1.15f*(fx-fy+phase)) )
				);
		}
	}

	// calculate normals using cross produce of adjacent vertices
	for(x=1; x<w_xsize-1; x++)
	{
		for(y=1; y<w_ysize-1; y++)
		{
			vert_offset3 = x*w_ysize+y;
			vert_offset4 = vert_offset3 *4;
			vert_offset3 *= 3;

			v1 = w_verts+(vert_offset3-3);
			v2 = w_verts+(vert_offset3+3);
			v3 = w_verts+(vert_offset3-w_xsize*3);
			v4 = w_verts+(vert_offset3+w_xsize*3);
			vec1.Set((*v2) - (*v1), (*(v2+1)) - (*(v1+1)), (*(v2+2)) - (*(v1+2)));
			vec2.Set((*v4) - (*v3), (*(v4+1)) - (*(v3+1)), (*(v4+2)) - (*(v3+2)));
			normal_vec = vec1.CrossProduct(vec2);
			normal_vec.Normalize();
			w_normals[vert_offset3] = normal_vec.x;
			w_normals[vert_offset3+1] = normal_vec.y;
			w_normals[vert_offset3+2] = normal_vec.z;

			if(0)
			{
				// visualize normals as colors
				w_color[vert_offset4] = normal_vec.x*0.5f+0.5f;
				w_color[vert_offset4+1] = normal_vec.y*0.5f+0.5f;
				w_color[vert_offset4+2] = normal_vec.z*0.5f+0.5f;
			}
			else
			{
				// calculate the view vector
				view_vec.Set(	eye_pos.x-w_verts[vert_offset3],
								eye_pos.y-w_verts[vert_offset3+1],
								eye_pos.z-w_verts[vert_offset3+2]	);

				view_vec.Normalize();

				// calculate the incidence and refraction angles
				incident_angle = acosf(view_vec.DotProduct(normal_vec));
				refract_angle = asinf(sinf(incident_angle)/1.333f);

				// calculate the Fresnel coefficients for both polarization orientations 
				coef_Rs = powf( (sinf(incident_angle-refract_angle) / sinf(incident_angle+refract_angle)), 2.0f);
				coef_Rp = powf( (1.333f*cosf(incident_angle) - cosf(refract_angle) ) / (1.33f*cosf(incident_angle) + cosf(refract_angle) ), 2.0f);

				// sky color
				w_color[vert_offset4] = *color;
				w_color[vert_offset4+1] = *(color+1);
				w_color[vert_offset4+2] = *(color+2);
				// apply the average of the polarization coefficients to the alpha channel
				w_color[vert_offset4+3] = (coef_Rs + coef_Rp) *0.5f;

				if(mode_polarize)
				{
					// sky color
					w_color_p[vert_offset4] = *color;
					w_color_p[vert_offset4+1] = *(color+1);
					w_color_p[vert_offset4+2] = *(color+2);
					// coefficient
					w_color_p[vert_offset4+3] = coef_Rp;
				}
			}

			// handle edge with simple padding
			if(y==1)
			{
				memcpy(w_normals+(vert_offset3-3), w_normals+vert_offset3, sizeof(GLfloat)*3);
				memcpy(w_color+(vert_offset4-4), w_color+vert_offset4, sizeof(GLfloat)*4);
				memcpy(w_color_p+(vert_offset4-4), w_color_p+vert_offset4, sizeof(GLfloat)*4);

				// corners
				if(x==1)
				{
					memcpy(w_normals+(vert_offset3-w_xsize*3-3), w_normals+vert_offset3, sizeof(GLfloat)*3);
					memcpy(w_color+(vert_offset4-w_xsize*4-4), w_color+vert_offset4, sizeof(GLfloat)*4);
					memcpy(w_color_p+(vert_offset4-w_xsize*4-4), w_color_p+vert_offset4, sizeof(GLfloat)*4);
				}
				else
					if(x==w_xsize-2)
					{
						memcpy(w_normals+(vert_offset3+w_xsize*3-3), w_normals+vert_offset3, sizeof(GLfloat)*3);
						memcpy(w_color+(vert_offset4+w_xsize*4-4), w_color+vert_offset4, sizeof(GLfloat)*4);
						memcpy(w_color_p+(vert_offset4+w_xsize*4-4), w_color_p+vert_offset4, sizeof(GLfloat)*4);
					}

			}
			if(y==w_ysize-2)
			{
				memcpy(w_normals+(vert_offset3+3), w_normals+vert_offset3, sizeof(GLfloat)*3);
				memcpy(w_color+(vert_offset4+4), w_color+vert_offset4, sizeof(GLfloat)*4);
				memcpy(w_color_p+(vert_offset4+4), w_color_p+vert_offset4, sizeof(GLfloat)*4);

				// corners
				if(x==1)
				{
					memcpy(w_normals+(vert_offset3-w_xsize*3+3), w_normals+vert_offset3, sizeof(GLfloat)*3);
					memcpy(w_color+(vert_offset4-w_xsize*4+4), w_color+vert_offset4, sizeof(GLfloat)*4);
					memcpy(w_color_p+(vert_offset4-w_xsize*4+4), w_color_p+vert_offset4, sizeof(GLfloat)*4);
				}
				else
					if(x==w_xsize-2)
					{
						memcpy(w_normals+(vert_offset3+w_xsize*3+3), w_normals+vert_offset3, sizeof(GLfloat)*3);
						memcpy(w_color+(vert_offset4+w_xsize*4+4), w_color+vert_offset4, sizeof(GLfloat)*4);
						memcpy(w_color_p+(vert_offset4+w_xsize*4+4), w_color_p+vert_offset4, sizeof(GLfloat)*4);
					}

			}

			if(x==1)
			{
				memcpy(w_normals+(vert_offset3-w_xsize*3), w_normals+vert_offset3, sizeof(GLfloat)*3);
				memcpy(w_color+(vert_offset4-w_xsize*4), w_color+vert_offset4, sizeof(GLfloat)*4);
				memcpy(w_color_p+(vert_offset4-w_xsize*4), w_color_p+vert_offset4, sizeof(GLfloat)*4);
			}
			else
				if(x==w_xsize-2)
				{
					memcpy(w_normals+(vert_offset3+w_xsize*3), w_normals+vert_offset3, sizeof(GLfloat)*3);
					memcpy(w_color+(vert_offset4+w_xsize*4), w_color+vert_offset4, sizeof(GLfloat)*4);
					memcpy(w_color_p+(vert_offset4+w_xsize*4), w_color_p+vert_offset4, sizeof(GLfloat)*4);
				}
		}
	}
}