/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"

MGL::Matrix4 MGLTree::modelView;
MGLTree Trees[TREE_GRID_SIZE*TREE_GRID_SIZE];

MGLApp::MGLApp()
{
    pos.Set(0.0f, 2.0f, 0.0f);
    keyarraysize = 512;
    lastx = lasty = 0;
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
    SetupProjection(512, 512);

    glMatrixMode(GL_MODELVIEW);
    
	if (!GLeeInit())
		return false;

    int Maxlevel = 5;
    float Length = 1.0f;
    float LengthMod = 0.9f;
    float AngleMod = 30.0f;
    float LeafSize = 1.0f;
    float AngleRand = 10.0f;
    float Twist = 90.0f;
    float branchwidth = 0.125f;
    int NumSprites = 8;

    MGL::TargaImage image;

    if(!image.Load("grass.tga"))
    {
	fprintf(stderr, "Error loading grass.tga");
	return false;
    }

    glGenTextures(1, &grass);
    glBindTexture(GL_TEXTURE_2D, grass);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA,
		 image.GetWidth(), image.GetHeight(),
		 GL_RGBA, GL_UNSIGNED_BYTE,
		 image.GetImage());

    freopen("log.txt", "w", stdout);

    noise.Generate();

    srand(time(NULL));
    
    glCullFace(GL_BACK);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    TreeGeometry *realtrees = new TreeGeometry[10];

    for(int i=0; i < 10; i++)
    	realtrees[i].SetupTex("bark.tga", "leaf.tga", NumSprites);

    for(int i=0; i < 10; i++)
    	realtrees[i].Generate(Maxlevel, AngleMod, Twist, Length, LengthMod, AngleRand, branchwidth, LeafSize);

    for(int i=0; i < TREE_GRID_SIZE*TREE_GRID_SIZE; i++)
    {
	Trees[i].tree = &realtrees[rand()%10];
	Trees[i].pos = MGL::Vector3(6.0f*(float(i % TREE_GRID_SIZE)-16.0f), 0.0f, 6.0f*(float(i / TREE_GRID_SIZE)-16.0f)) +
	    MGL::Vector3(RAND(-2.5f, 2.5f), 0.0f, RAND(-2.5f, 2.5f));
	Trees[i].enabled = true;
#define MAP(x, y) int(((x) * float(256/TREE_GRID_SIZE)) + 256*((y) * float(256/TREE_GRID_SIZE)))

	if(noise[MAP(i%TREE_GRID_SIZE, i/TREE_GRID_SIZE)] == 0.0f)
	    Trees[i].enabled = false;
    }

    SetupProjection(MGL_WINDOW_WIDTH, MGL_WINDOW_HEIGHT);
	return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
}

void MGLApp::Update( double dt )
{
    const float speed = 50.0f;

    // MOUSE INPUT
    if(lastx == 0 || lasty == 0)
    {
 	lastx = mousex, lasty = mousey;
    }

    // KEYBOARD INPUT
    if(GetKey(KEY_ARROW_UP) || GetKey('w'))
    {
 	pos.x += speed * dt * (float)sinf(DEG2RAD(rot.y));
 	pos.y -= speed * dt * (float)sinf(DEG2RAD(rot.x));
 	pos.z -= speed * dt * (float)cosf(DEG2RAD(rot.y));
    }
    if(GetKey(KEY_ARROW_LEFT) || GetKey('a'))
    {
 	pos.x += speed * dt * (float)sinf(DEG2RAD(rot.y-90.0f));
 	pos.z -= speed * dt * (float)cosf(DEG2RAD(rot.y-90.0f));
    }
    if(GetKey(KEY_ARROW_RIGHT) || GetKey('d'))
    {
 	pos.x += speed * dt * (float)sinf(DEG2RAD(rot.y+90.0f));
 	pos.z -= speed * dt * (float)cosf(DEG2RAD(rot.y+90.0f));
    }
    if(GetKey(KEY_ARROW_DOWN) || GetKey('s'))
    {
	pos.x -= speed * dt * (float)sinf(DEG2RAD(rot.y));
	pos.y += speed * dt * (float)sinf(DEG2RAD(rot.x));
	pos.z += speed * dt * (float)cosf(DEG2RAD(rot.y));
    }

    // MOUSE INPUT
    rot.y += (mousex-lastx)*0.1f;
    if(fabs(rot.x+(mousey-lasty)*0.1f) < 90.0f)
	rot.x += (mousey-lasty)*0.1f;

    lastx = mousex;
    lasty = mousey;

    if(lastx > MGL_WINDOW_WIDTH-100 || lastx < 100)
    {
	MoveMouse(MGL_WINDOW_WIDTH/2, lasty);
	lastx = MGL_WINDOW_WIDTH/2;
    }
    if(lasty > MGL_WINDOW_HEIGHT-100 || lasty < 100)
    {
	MoveMouse(lastx, MGL_WINDOW_HEIGHT/2);
	lasty = MGL_WINDOW_HEIGHT/2;
    }
}

int MGLTree::SortFunc(const void *a, const void *b)
{
    MGLTree *one = (MGLTree *)a;
    MGLTree *two = (MGLTree *)b;

    if(one->enabled == false)
	return 1;
    else if(two->enabled == false)
	return -1;
    else if((MGLTree::modelView * one->pos).z < (MGLTree::modelView * two->pos).z)
	return -1;

    return 1;
}

void MGLApp::Render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glDisable(GL_BLEND);
    glEnable(GL_TEXTURE_2D);

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    
    glRotatef(rot.x, 1.0f, 0.0f, 0.0f);
    glRotatef(rot.y, 0.0f, 1.0f, 0.0f);
    glTranslatef(-pos.x, -pos.y, -pos.z);

    DrawSky();
    DrawGround();

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    MGLTree::modelView.GetModelView();
    qsort(Trees, 900, sizeof(MGLTree), MGLTree::SortFunc);
    for(int i=0; i < 900; i++)
    {
	if(Trees[i].enabled == false)
	    continue;
	glPushMatrix();
	glTranslatef(Trees[i].pos.x, Trees[i].pos.y, Trees[i].pos.z);
	Trees[i].tree->Draw();
	glPopMatrix();
    }
}

void MGLApp::DrawSky()
{
    glPushMatrix();
    glLoadIdentity();
    glDepthMask(GL_FALSE);
    glColor4f(0.0f, float(200)/255.0f, float(200)/255.0f, 1.0f);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    glVertex3f(-40.0f, -40.0f, -2.0f);
    glVertex3f( 40.0f, -40.0f, -2.0f);
    glVertex3f( 40.0f, 40.0f, -2.0f);
    glVertex3f(-40.0f, 40.0f, -2.0f);
    glEnd();
    glDepthMask(GL_TRUE);
    glPopMatrix();
}

void MGLApp::DrawGround()
{
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    glBindTexture(GL_TEXTURE_2D, grass);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f);
    glVertex3f(-160.0f, 0.0f, -160.0f);
    glTexCoord2f(20.0f, 0.0f);
    glVertex3f( 160.0f, 0.0f, -160.0f);
    glTexCoord2f(20.0f, 20.0f);
    glVertex3f( 160.0f, 0.0f,  160.0f);
    glTexCoord2f(0.0f, 20.0f);
    glVertex3f(-160.0f, 0.0f,  160.0f);
    glEnd();
}
