/*************************************************************************
MGLMatrix.h

Matrix class.

Author   :  Baldur Karlsson

Adapted for More OpenGL Game Programming
*************************************************************************/

#ifndef MGLMATRIX_H_INCLUDED
#define MGLMATRIX_H_INCLUDED

#ifdef WIN32
  #define WIN32_LEAN_AND_MEAN
  #define WIN32_EXTRA_LEAN
  #include <windows.h>
#endif

#include <GL/GLee.h>

//#include "common.hpp"
#include "MGLMath.h"
#include "MGLVector.h"

namespace MGL
{
    
class Matrix4
{
 public:
    Matrix4()
    {
	m[0] = 1.0f; m[4] = 0.0f; m[8] =  0.0f; m[12] = 0.0f;
	m[1] = 0.0f; m[5] = 1.0f; m[9] =  0.0f; m[13] = 0.0f;
	m[2] = 0.0f; m[6] = 0.0f; m[10] = 1.0f; m[14] = 0.0f;
	m[3] = 0.0f; m[7] = 0.0f; m[11] = 0.0f; m[15] = 1.0f;
    }

    Matrix4(const Matrix4 &other)
    {
	*this = other;
    }

    void identity()
    {
	m[0] = 1; m[4] = 0; m[8] =  0; m[12] = 0;
	m[1] = 0; m[5] = 1; m[9] =  0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10] = 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11] = 0; m[15] = 1;
    }
    float* mat() { return m; }

    void translate(float x, float y, float z) { m[12] += x;m[13] += y;m[14] += z; }
    void rotate(float x, float y, float z)
    {
	float cx, cy, cz, sx, sy, sz;

	cx = (float)cos(x);
	sx = (float)sin(x);
	cy = (float)cos(y);
	sy = (float)sin(y);
	cz = (float)cos(z);
	sz = (float)sin(z);

	Matrix4 mat;
    
	mat.m[0] = cy * cz;
	mat.m[1] = cy * sz;
	mat.m[2] = -sy;
    
	mat.m[4] = (sx * sy * cz) - (cx * sz);
	mat.m[5] = (sx * sy * sz) + (cx * cz);
	mat.m[6] = sx * cy;
	
	mat.m[8] = (cx * sy * cz) + (sx * sz);
	mat.m[9] = (cx * sy * sz) - (sx * cz);
	mat.m[10] = cx * cy;
	
	*this = *this * mat;
    }

    Matrix4 GetModelView()
    {
	glGetFloatv(GL_MODELVIEW_MATRIX, m);
	return *this;
    }

    Matrix4 operator *= (Matrix4 matr)
    {
	m[0] = m[0]*matr.m[0] + m[4]*matr.m[1] + m[8]*matr.m[2];
	m[1] = m[1]*matr.m[0] + m[5]*matr.m[1] + m[9]*matr.m[2];
	m[2] = m[2]*matr.m[0] + m[6]*matr.m[1] + m[10]*matr.m[2];
	m[3] = m[3]*matr.m[0] + m[7]*matr.m[1] + m[11]*matr.m[2];

	m[4] = m[0]*matr.m[4] + m[4]*matr.m[5] + m[8]*matr.m[6];
	m[5] = m[1]*matr.m[4] + m[5]*matr.m[5] + m[9]*matr.m[6];
	m[6] = m[2]*matr.m[4] + m[6]*matr.m[5] + m[10]*matr.m[6];
	m[7] = m[3]*matr.m[4] + m[7]*matr.m[5] + m[11]*matr.m[6];

	m[8] = m[0]*matr.m[8] + m[4]*matr.m[9] + m[8]*matr.m[10];
	m[9] = m[1]*matr.m[8] + m[5]*matr.m[9] + m[9]*matr.m[10];
	m[10] = m[2]*matr.m[8] + m[6]*matr.m[9] + m[10]*matr.m[10];
	m[11] = m[3]*matr.m[8] + m[7]*matr.m[9] + m[11]*matr.m[10];

	m[12] = m[0]*matr.m[12] + m[4]*matr.m[13] + m[8]*matr.m[14];
	m[13] = m[1]*matr.m[12] + m[5]*matr.m[13] + m[9]*matr.m[14];
	m[14] = m[2]*matr.m[12] + m[6]*matr.m[13] + m[10]*matr.m[14];
	m[15] = m[3]*matr.m[12] + m[7]*matr.m[13] + m[11]*matr.m[14];

	return *this;
    }
    
    float &operator [](int i)
    {
	return m[i];
    }

    Matrix4 operator * (Matrix4 matr)
    {
	Matrix4 ret;

	ret.m[0] = m[0]*matr.m[0] + m[4]*matr.m[1] + m[8]*matr.m[2];
	ret.m[1] = m[1]*matr.m[0] + m[5]*matr.m[1] + m[9]*matr.m[2];
	ret.m[2] = m[2]*matr.m[0] + m[6]*matr.m[1] + m[10]*matr.m[2];
	ret.m[3] = m[3]*matr.m[0] + m[7]*matr.m[1] + m[11]*matr.m[2];

	ret.m[4] = m[0]*matr.m[4] + m[4]*matr.m[5] + m[8]*matr.m[6];
	ret.m[5] = m[1]*matr.m[4] + m[5]*matr.m[5] + m[9]*matr.m[6];
	ret.m[6] = m[2]*matr.m[4] + m[6]*matr.m[5] + m[10]*matr.m[6];
	ret.m[7] = m[3]*matr.m[4] + m[7]*matr.m[5] + m[11]*matr.m[6];

	ret.m[8] = m[0]*matr.m[8] + m[4]*matr.m[9] + m[8]*matr.m[10];
	ret.m[9] = m[1]*matr.m[8] + m[5]*matr.m[9] + m[9]*matr.m[10];
	ret.m[10] = m[2]*matr.m[8] + m[6]*matr.m[9] + m[10]*matr.m[10];
	ret.m[11] = m[3]*matr.m[8] + m[7]*matr.m[9] + m[11]*matr.m[10];

	ret.m[12] = m[0]*matr.m[12] + m[4]*matr.m[13] + m[8]*matr.m[14] + m[12];
	ret.m[13] = m[1]*matr.m[12] + m[5]*matr.m[13] + m[9]*matr.m[14] + m[13];
	ret.m[14] = m[2]*matr.m[12] + m[6]*matr.m[13] + m[10]*matr.m[14] + m[14];
	ret.m[15] = m[3]*matr.m[12] + m[7]*matr.m[13] + m[11]*matr.m[14] + m[15];

	return ret;
    }

    Vector3 operator * (Vector3 vec)
    {
	float a, b, c;

	a = vec.x * m[0] + vec.y * m[4] + vec.z * m[8] + m[12];
	b = vec.x * m[1] + vec.y * m[5] + vec.z * m[9] + m[13];
	c = vec.x * m[2] + vec.y * m[6] + vec.z * m[10] + m[14];

	return Vector3(a, b, c);
    }

    float m[16];
};

}
#endif
