/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dimi Christopoulos

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "tvector.h"
#include "mathex.h"
#include "glCamera.h"
#include "SkyPlane.h"
#include "MGLApp.h"
#include "MGLTargaImage.h"
#include "CGLFontClass.h"

#define speed 30
#define rotspeed 50

int window_width, window_height;

GLuint texture[2];
GLuint innerLayerTexture[6];
GLuint outerLayerTexture[6];
GLuint astroLayerTexture[6];
float xmouse=0;
float ymouse=0;
float ttrans[2];

//Camera position,direction and up vector
GLCamera glcam;
TVector32 CamTarget;
GLUquadricObj *quadratic;				// Storage For Our Quadratic Objects

int showObjects=1;
int showplane=1;
int help = 0;
CGLFontClass font;

void LoadGLTextures();
void CreateSkyBox(float x, float y, float z, float width, float height, float length, GLuint texture[6]);
int loadSkyBoxTextures (char *front, char *back, char *left, char *right, char *top, char *bottom, GLuint texture[6]);

// Specify skybox Texture Names. NULL means no texture for this face, dont draw it.
char *innerLayer[6] = {"MountainBox/Front.tga", "MountainBox/Back.tga","MountainBox/Left.tga", "MountainBox/Right.tga", NULL, "MountainBox/Bottom.tga"};
char *outerLayer[6] = {"Data/evening_FR.tga", "Data/evening_BK.tga", "Data/evening_RT.tga", "Data/evening_LF.tga", "Data/evening_UP.tga", NULL};


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	if (!GLeeInit())
		return false;

    LoadGLTextures();				/* Load The Texture(s) */
    loadSkyBoxTextures (innerLayer[0], innerLayer[1], innerLayer[2], 
                        innerLayer[3], innerLayer[4], innerLayer[5], innerLayerTexture);
    loadSkyBoxTextures (outerLayer[0], outerLayer[1], outerLayer[2], 
                        outerLayer[3], outerLayer[4], outerLayer[5], outerLayerTexture);
    loadSkyBoxTextures (NULL,"MountainBox/saturn.tga", NULL, NULL, NULL, NULL, astroLayerTexture);

    GenerateSkyPlane(15, 2000, 800, 3, 3, 0.9f, false);

    glEnable(GL_DEPTH_TEST);			/* Enables Depth Testing*/
    glShadeModel(GL_SMOOTH);			/* Enables Smooth Color Shading*/
    glEnable (GL_CULL_FACE);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);	
    glEnable (GL_COLOR_MATERIAL);
  
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glAlphaFunc(GL_GREATER, 0.1f);
    
    glcam.setPos (0.f, 3.f, 0.f);

    GLfloat sundiffuse[4] = {0.7f, 0.7f, 0.7f, 1.f}; 
    GLfloat sunambient[4] = {0.3f, 0.3f, 0.3f, 1.f};
    GLfloat sunposition[] = {0, 1, 0, 0};

    glLightfv (GL_LIGHT0, GL_POSITION, sunposition);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, sundiffuse);
    glLightfv (GL_LIGHT0, GL_AMBIENT, sunambient); 
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);
  
    quadratic=gluNewQuadric();			// Create A Pointer To The Quadric Object ( NEW )
	gluQuadricNormals(quadratic, GLU_SMOOTH);	// Create Smooth Normals ( NEW )
	gluQuadricTexture(quadratic, GL_TRUE);		// Create Texture Coords ( NEW )
	
    showObjects=1;
    showplane=1;
	
    font.Load("font.tga","font.dat");

	return true;
}

bool MGLApp::Shutdown()
{

    ReleaseSkyPlane();
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

    window_width = width;                                // Store current width/height
    window_height = height;  
	
    font.setDimensions(width,height);

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 60.0f, (GLfloat)width/(GLfloat)height, 1.0f, 2000.0f );
}

void MGLApp::Update( double dt )
{
	// Get interval time in seconds
    float intervalTime = (float) dt;
    
    // Animate Clouds
    ttrans[0] += .05f * intervalTime;
    if (ttrans[0] == 1.0f) ttrans[0] = 0.0f;
    ttrans[1] -= .1f * intervalTime;
    if (ttrans[1] <= 0.0f) ttrans[1] = 1.0f;


  	// If UP/DOWN Cursor keys are pressed move.	
    if (GetAsyncKeyState (VK_UP))
    {
        glcam.move (speed*intervalTime);	 
    }
    if (GetAsyncKeyState (VK_DOWN))
    {
       glcam.move (-speed*intervalTime);
    }
 
    if (xmouse!=0.0f) 
    {
        glcam.rotateY(xmouse*(rotspeed*intervalTime));
    }
    
    if (ymouse!=0.0f) 
    {
        glcam.rotateAxis(ymouse*(rotspeed*intervalTime));
    }
}

void MGLApp::Render()
{
	
    float globalmat[16];
	TVector32 CamPos, Direction, Up;
    TVector32 Horizon = glcam.getCamTarget(1000);
    CamPos    = glcam.getPos();
    Direction = glcam.getDir();
    Up        = glcam.getUp();
 
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// set up the camera
	glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();				/* Reset The View*/
    gluLookAt((GLdouble)CamPos.X(),(GLdouble)CamPos.Y(),(GLdouble)CamPos.Z(),
              (GLdouble)Horizon.X(),(GLdouble)Horizon.Y(),(GLdouble)Horizon.Z(), 
              (GLdouble)Up.X(),(GLdouble)Up.Y(),(GLdouble)Up.Z());


    glDisable (GL_LIGHTING);
    glColor3f(1.f, 1.f, 1.f);
    glEnable(GL_TEXTURE_2D);
    
    glDisable(GL_DEPTH_TEST);
    glDepthMask(GL_FALSE);			
    glPushMatrix();

    // Get the Modelview matrix and set its translation part to 0.
    // In the GL Modelviewmatrix entries 12,13,14 are the translation offsets.
    glGetFloatv(GL_MODELVIEW_MATRIX, globalmat);
    globalmat[12] = globalmat[13] = globalmat[14] = 0.f;
    glLoadMatrixf(globalmat);
    CreateSkyBox(0, 25, 0, 400, 200, 400, outerLayerTexture);  // Render Outer Layer
    
    glEnable(GL_BLEND);

    CreateSkyBox(0, 25, 0, 300, 200, 395, astroLayerTexture);  // Render Middle Layer
  
    // Render Sky Plane
    for (int i=0; i<3; i++)
    {
        glPushMatrix();
          glTranslatef (0.f, -430.f+i, 0.f);
          glMatrixMode(GL_TEXTURE);
          glPushMatrix();
            glTranslatef(ttrans[0], ttrans[1], 0.);
            glBindTexture(GL_TEXTURE_2D, texture[1]);   /* choose the texture to use.*/
            RenderSkyPlane(0);
          glPopMatrix();
          glMatrixMode(GL_MODELVIEW);
        glPopMatrix();
    }
    
    glColor3f(1.f, 1.f, 1.f);
    glEnable(GL_ALPHA_TEST);
    CreateSkyBox(0, 5, 0, 390, 200, 390, innerLayerTexture);   // Render Inner Layer
    glDisable(GL_ALPHA_TEST);

    glDisable(GL_BLEND);
    glPopMatrix();
    glDepthMask(GL_TRUE);
    glEnable(GL_DEPTH_TEST);	

    glEnable(GL_TEXTURE_2D);
    if (showplane)
    {    
        // Render Ground plane
        glBindTexture(GL_TEXTURE_2D, texture[0]);   /* choose the texture to use.*/
        glBegin(GL_QUADS);		                /* begin drawing a cube*/
	    glNormal3f(0, 1, 0);
        glTexCoord2f(0.0f, 0.0f); glVertex3f(-100.0f, 0.0f, 100.0f);	/* Bottom Left Of The Texture and Quad*/
        glTexCoord2f(20.0f, 0.0f); glVertex3f( 100.0f, 0.0f, 100.0f);	/* Bottom Right Of The Texture and Quad*/
        glTexCoord2f(20.0f, 15.0f); glVertex3f( 100.0f, 0.0f,-100.0f);	/* Top Right Of The Texture and Quad*/
        glTexCoord2f(0.0f, 15.0f); glVertex3f(-100.0f, 0.0f,-100.0f);	/* Top Left Of The Texture and Quad*/
        glEnd();
    }

    glDisable(GL_TEXTURE_2D);
    glEnable (GL_LIGHTING);
    
    if (showObjects)
    {
        glPushMatrix();
        glTranslatef (5, 1, -5);
        gluSphere(quadratic,1.0f,20,20);
        glPopMatrix();
    
        glPushMatrix();
        glTranslatef (-5, 0, -5);
        glRotatef(-90, 1, 0, 0);
        gluCylinder(quadratic,2.0f,0.0f,3.0f,32,32);
        glPopMatrix();
    
        glPushMatrix();
        glTranslatef (0, 0, 5);
		glRotatef(-90, 1, 0, 0);
        gluCylinder(quadratic,1.0f,1.0f,3.0f,32,32);
		glPopMatrix();
    }

	if (help)
    {
		
        glEnable(GL_TEXTURE_2D);
        glDisable (GL_LIGHTING);
        font.Display("'1','2': Wireframe/Filled rendering",5,90);
        font.Display("'3': Show objects",5,115);
        font.Display("'h':Help on/off",5,265);
        font.Display("'Space':Grass plane on/off",5,295);
        font.Display("Navigate using Mouse and Cursor Keys",5,325);
        font.Display("Demo by Dimi Christopoulos",5,355);
    }
}


#include "dk_grass.h"
MGL::TargaImage tgaImg;

/* Load Bitmaps And Convert To Textures*/
void LoadGLTextures() 
{
   
    glGenTextures(2, &texture[0]);

    /* Create Texture	*/
    glBindTexture(GL_TEXTURE_2D, texture[0]);   /* 2d texture (x and y size)*/
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); /* scale linearly when image bigger than texture*/
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST); /* scale linearly when image smalled than texture*/
    /* 2d texture, level of detail 0 (normal), 3 components (red, green, blue), x size from image, y size from image, */
    /* border 0 (normal), rgb color data, unsigned byte data, and finally the data itself.*/
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 64, 64, GL_RGB, GL_UNSIGNED_BYTE, dk_grass_array);

	if (!tgaImg.Load("Data/clouds.tga"))
	{
        MessageBox(NULL, "Texture DATA/clouds.tga not found", NULL, MB_OK|MB_ICONERROR);
        return;
	}

	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
    glBindTexture(GL_TEXTURE_2D, texture[1]);   /* 2d texture (x and y size)*/
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); /* scale linearly when image bigger than texture*/
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST); /* scale linearly when image smalled than texture*/
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    /* 2d texture, level of detail 0 (normal), 3 components (red, green, blue), x size from image, y size from image, */
    /* border 0 (normal), rgb color data, unsigned byte data, and finally the data itself.*/
	gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA8, tgaImg.GetWidth(), tgaImg.GetHeight(), tgaImg.GetImageFormat() + GL_RGB, GL_UNSIGNED_BYTE, tgaImg.GetImage());
    tgaImg.Release();
}

//////////////////////////////////////////////////////////
//	Function: "CreateSkyBox"
//
//	Input:	x,y,z   		            Center of SkyBox.
//			width, height, length		Size of SkyBox.
//			textures[6]                 Array with texture objects for mapping the 6 faces of the SkyBox. 
//                                      The order in the array should be Front, Back, Left, Right, Top, Bottom.
//
//	Purpose: Create SkyBox.
//////////////////////////////////////////////////////////
void CreateSkyBox(float x, float y, float z, float width, float height, float length, GLuint texture[6])
{
        // This centers the sky box around (x, y, z)
	x = x - width  / 2;
	y = y - height / 2;
	z = z - length / 2;

        if (texture[1])
        {
	    // Bind the BACK texture of the sky map to the BACK side of the cube
	    glBindTexture(GL_TEXTURE_2D, texture[1]);
	
            // Start drawing the side as a QUAD
	    glBegin(GL_QUADS);		
		// Assign the texture coordinates and vertices for the BACK Side
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x + width, y, z);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x + width, y + height, z); 
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x,	y + height, z);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x,	y, z);
		
	    glEnd();
        }

        if (texture[0])
        {
	    // Bind the FRONT texture of the sky map to the FRONT side of the box
	    glBindTexture(GL_TEXTURE_2D, texture[0]);

	    // Start drawing the side as a QUAD
	    glBegin(GL_QUADS);	
		// Assign the texture coordinates and vertices for the FRONT Side
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,	y, z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,	y + height, z + length);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height, z + length); 
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y, z + length);
	    glEnd();
        }

        if (texture[5])
        {
	    // Bind the BOTTOM texture of the sky map to the BOTTOM side of the box
	    glBindTexture(GL_TEXTURE_2D, texture[5]);

	    // Start drawing the side as a QUAD
	    glBegin(GL_QUADS);		
		// Assign the texture coordinates and vertices for the BOTTOM Side
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,	y, z);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,	y, z + length);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y, z + length); 
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y, z);
	    glEnd();
        }

        if (texture[4])
        {
	    // Bind the TOP texture of the sky map to the TOP side of the box
	    glBindTexture(GL_TEXTURE_2D, texture[4]);
	
	    // Start drawing the side as a QUAD
	    glBegin(GL_QUADS);		
		// Assign the texture coordinates and vertices for the TOP Side
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height, z);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y + height, z + length); 
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,	y + height, z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,	y + height, z);
	    glEnd();
        }

        if (texture[2])
        {
	    // Bind the LEFT texture of the sky map to the LEFT side of the box
	    glBindTexture(GL_TEXTURE_2D, texture[2]);
	
	    // Start drawing the side as a QUAD
	    glBegin(GL_QUADS);		
		// Assign the texture coordinates and vertices for the LEFT Side
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x,	y + height, z);	
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x,	y + height, z + length); 
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x,	y, z + length);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x,	y, z);		
	    glEnd();
        }

        if (texture[3])
        {
	    // Bind the RIGHT texture of the sky map to the RIGHT side of the box
	    glBindTexture(GL_TEXTURE_2D, texture[3]);

	    // Start drawing the side as a QUAD
	    glBegin(GL_QUADS);		
		// Assign the texture coordinates and vertices for the RIGHT Side
		glTexCoord2f(0.0f, 0.0f); glVertex3f(x + width, y, z);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(x + width, y, z + length);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(x + width, y + height, z + length); 
		glTexCoord2f(0.0f, 1.0f); glVertex3f(x + width, y + height, z);
	    glEnd();
        }
}

//////////////////////////////////////////////////////////
//	Function: "loadSkyBoxTextures"
//
//	Input:	front,back,left,right,top,bottom   		  Name of the Textures to load for the six faces of the SkyBox.
//			textures[6]                               Array where the texture objects created are going to be stored in the same order as passed. 
//                                   
//	Purpose: Load and create texture objects.
//////////////////////////////////////////////////////////
int loadSkyBoxTextures (char *front, char *back, char *left, char *right, char *top, char *bottom, GLuint texture[6])
{
    char *filename[6];
    int i;

    filename[0] = front;
    filename[1] = back;
    filename[2] = left;
    filename[3] = right;
    filename[4] = top;
    filename[5] = bottom;

    for (i=0; i<6; i++)
    {
        if (filename[i] != NULL)
        {
            glGenTextures(1, &texture[i]);

            /* Create Texture	*/
            if (!tgaImg.Load(filename[i]))
			{
                MessageBox(NULL, filename[i], NULL, MB_OK|MB_ICONERROR);
                return 0;
			}

            glBindTexture(GL_TEXTURE_2D, texture[i]);   /* 2d texture (x and y size)*/
            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); /* scale linearly when image bigger than texture*/
            glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST); /* scale linearly when image smalled than texture*/
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            /* 2d texture, level of detail 0 (normal), 3 components (red, green, blue), x size from image, y size from image, */
            /* border 0 (normal), rgb color data, unsigned byte data, and finally the data itself.*/
           	gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA8, tgaImg.GetWidth(), tgaImg.GetHeight(), tgaImg.GetImageFormat() + GL_RGB, GL_UNSIGNED_BYTE, tgaImg.GetImage());
			tgaImg.Release();
        }
        else
            texture[i] = 0;
    }

	return 1;
}
