/*************************************************************************
tvector.h

Vector class.

Author   :   Dimi Christopoulos

*************************************************************************/
#ifndef tvector32_h
#define tvector32_h

#include <math.h>
#include "mathex.h"

class TVector32 {
	public:
		enum TStatus { INVALID, DEFAULT, UNIT };

	private:
		float _x, _y, _z;
		TStatus _Status;

		// Constructors
		TVector32(float x, float y, float z, TStatus s) : _x(x), _y(y), _z(z), _Status(s) {}

	public:
		// Constructors
		TVector32() : _x(0.0), _y(0.0), _z(0.0), _Status(INVALID) {}
		TVector32(float x, float y, float z) : _x(x), _y(y), _z(z), _Status(DEFAULT) {}

		// Selectors
		float X() const { return _x; }
		float Y() const { return _y; }
		float Z() const { return _z; }
		int isUnit() const { return _Status==UNIT; }
		int isDefault() const { return _Status==DEFAULT; }
		int isValid() const { return _Status!=INVALID; }

		// Change the status of a vector
		TVector32 &unit();
		static TVector32 &unit(const TVector32 &v, TVector32 &result) { result = v; return result.unit(); }
		static TVector32 unit(const TVector32 &v) { return TVector32(v).unit(); }

		TVector32 &Default();
		static TVector32 Default(const TVector32 &v, TVector32 &result) { result = v; return result.Default(); }
		static TVector32 Default(const TVector32 &v) { return TVector32(v).Default(); }

		// Magnitude
		float mag() const { return (float)(isValid() ? (isUnit() ? 1.0 : sqrt(sqr(X()) + sqr(Y()) + sqr(Z()))) : 0.0f); }
		float mag2() const { return (float)(isValid() ? (isUnit() ? 1.0 : sqrt(sqr(X()) + sqr(Y()) )) : 0.0); }
		float magSqr() const { return (float)(isValid() ? (isUnit() ? 1.0 : sqr(X()) + sqr(Y()) + sqr(Z())) : 0.0f); }

		// Dot or scalar product
		float dot(const TVector32 &v) const { return ((isValid() && v.isValid()) ? (X()*v.X() + Y()*v.Y() + Z()*v.Z()) : 0.0f); }
		static float dot(const TVector32 &v1, const TVector32 &v2) { return v1.dot(v2); }

		// Distance between two vectors
		float dist(const TVector32 &v) const { return (*this-v).mag(); }
		float distSqr(const TVector32 &v) const { return (*this-v).magSqr(); }

		// Optimised arithmetic methods
		static TVector32 &add(const TVector32 &v1, const TVector32 &v2, TVector32 &result);
		static TVector32 &subtract(const TVector32 &v1, const TVector32 &v2, TVector32 &result);
		static TVector32 &cross(const TVector32 &v1, const TVector32 &v2, TVector32 &result);
		static TVector32 &invert(const TVector32 &v1, TVector32 &result);
		static TVector32 &multiply(const TVector32 &v1, const float &scale, TVector32 &result);

		// Vector arithmetic, addition, subtraction and vector product
		TVector32 operator-() const { TVector32 tv; return invert(*this, tv); }
		TVector32 &operator+=(const TVector32 &v) { return add(*this, v, *this); }
		TVector32 &operator-=(const TVector32 &v) { return subtract(*this, v, *this); }
		TVector32 &operator*=(const TVector32 &v) { TVector32 tv(*this); return cross(tv, v, *this); }
		TVector32 &operator*=(const float &scale) { return multiply(*this, scale, *this); }
		TVector32 operator+(const TVector32 &v) const { TVector32 tv; return add(*this, v, tv); }
		TVector32 operator-(const TVector32 &v) const { TVector32 tv; return subtract(*this, v, tv); }
		TVector32 operator*(const TVector32 &v) const { TVector32 tv; return cross(*this, v, tv); }
		TVector32 operator*(const float &scale) const { TVector32 tv; return multiply(*this, scale, tv); }

};
#endif
