/*************************************************************************
MGLSkyDome.cpp

SkyDome Geometry computation and rendering.

Author   :   Dimi Christopoulos, Andrei Stojan

Written for More OpenGL Game Programming
*************************************************************************/

#include "MGLApp.h"

#define GL_CLAMP_TO_EDGE                  0x812F

CSkyDome::CSkyDome()
{
	SkyRadius = 10.0f;
	ColoringMode = COLOR_MODEL;
}

int CSkyDome::Initialize(float Radius, int Slices, int Sides, bool exponential, float dampening)
{
	NumSides = Sides;
	NumSlices = Slices;
	NumIndices = Slices * (Sides + 1) * 2;
	SkyRadius = Radius;
	
	VertexBuffer = new Vector3[(Slices + 1) * (Sides + 1)];
	ColorBuffer = new Color4[(Slices + 1) * (Sides + 1)];
	TexCoordBuffer = new Vector2[(Slices + 1) * (Sides + 1)];
	SkyMapTexCoordBuffer = new Vector2[(Slices + 1) * (Sides + 1)];

	IndexBuffer = new unsigned short [NumIndices];
	
	float polyAng = 2.0f * PI / Sides, ang;

	exponential = false; //use to tesselate
		
	float vx, vy, vz;
	
	int i, j;
	for (j = 0; j <= Slices; j++)
	{
		ang = j * ( (PI / 2) / Slices );
		for (i = 0; i <= Sides; i++)
		{						
			vx = cos(i * polyAng) * cos(ang);
			vy = sin(ang) * dampening;
			vz = sin(i * polyAng) * cos(ang);

			VertexBuffer[j * (Sides + 1) + i].x = vx * SkyRadius;
			VertexBuffer[j * (Sides + 1) + i].z = vz * SkyRadius;
			VertexBuffer[j * (Sides + 1) + i].y = vy * SkyRadius;	

			TexCoordBuffer[j * (Sides + 1) + i].x = (float)(i) / (float)(Sides);
			TexCoordBuffer[j * (Sides + 1) + i].y = (float)(j) / (float)(Slices);

            SkyMapTexCoordBuffer[j * (Sides + 1) + i].y = (float)(j) / (float)(Slices);
		}
	}
	
	int ind = 0;
	for (j = 1; j <= Slices; j++)
	{
		for (i = 0; i <= Sides; i++)
		{			
			IndexBuffer[ind++] =  j      * (Sides + 1) + i;
			IndexBuffer[ind++] = (j - 1) * (Sides + 1) + i;
		}
	}
	
	TargaImage img;
	if (img.Load("flare0.tga") == false)
	{
		return 1;
	}
	
	glGenTextures(1, &FlareTex0);
	glBindTexture(GL_TEXTURE_2D, FlareTex0);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, img.GetWidth(), 
		img.GetHeight(), 0, img.GetImageFormat() + GL_RGB, 
		GL_UNSIGNED_BYTE, img.GetImage());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);	

	img.Release();

	if (img.Load("moon.tga") == false)
	{
		return 1;
	}
	
	glGenTextures(1, &MoonTex);
	glBindTexture(GL_TEXTURE_2D, MoonTex);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, img.GetWidth(), 
		img.GetHeight(), 0, img.GetImageFormat() + GL_RGB, 
		GL_UNSIGNED_BYTE, img.GetImage());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);	
		
	img.Release();

	if (img.Load("sky.tga") == false)
	{
		return 1;
	}
	
	glGenTextures(1, &Texture);
	glBindTexture(GL_TEXTURE_2D, Texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, img.GetWidth(), 
		img.GetHeight(), 0, img.GetImageFormat() + GL_RGB,
		GL_UNSIGNED_BYTE, img.GetImage());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);	
		
	img.Release();

        if (img.Load("skymap.tga") == false)
	{
		return 1;
	}
	
	glGenTextures(1, &SkyMap);
	glBindTexture(GL_TEXTURE_2D, SkyMap);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, img.GetWidth(), 
		img.GetHeight(), 0, img.GetImageFormat() + GL_RGB,
		GL_UNSIGNED_BYTE, img.GetImage());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);	
		
	img.Release();

        //load the color table from the file
	if (ColorTable.Load()) return 1;

        //generate star positions
	NumStars = 1000;
	Stars = new Vector3[NumStars];
	StarColors = new Color4[NumStars];
	
	float st, sp, c; //star theta, star phi

#ifdef WIN32
	srand(GetTickCount());
#endif

	for (i = 0; i < NumStars; i++)
	{
		st = (rand() % 90) / 180.0f * PI;
		sp = (rand() % 360) / 180.0f * PI;

		Stars[i].x = cos(sp) * cos(st) * SkyRadius;
		Stars[i].y = sin(st) * SkyRadius * dampening;
		Stars[i].z = sin(sp) * cos(st) * SkyRadius;
		
		c = (rand() % 256) / 255.0f;

		StarColors[i] = Color4(c, c, c, c);
	}

	return 0;
}

void CSkyDome::SetCoordinates(float Lat, float Long, float Time, float Day)
{
	Latitude = DEG2RAD(Lat);
	Longitude = DEG2RAD(Long);
	TimeOfDay = (Time) * 3600.0f;
	JulianDay = Day;
}

void CSkyDome::Destroy()
{
	delete [] VertexBuffer;
	delete [] ColorBuffer;
	delete [] IndexBuffer;
	delete [] TexCoordBuffer;
	delete [] SkyMapTexCoordBuffer;
}

void CSkyDome::Update(float deltatime)
{
      	TimeOfDay += deltatime;
	if (TimeOfDay > 3600 * 24) //next day?
	{
		JulianDay += 1.0f;
		TimeOfDay -= 3600 * 24;
	}
	
	float hour = TimeOfDay / 3600.0f;
	float solarTime = hour + 0.170 * sin( 4 * PI * (JulianDay - 80) / 373 ) - 0.129 * sin( 2 * PI * (JulianDay - 8) / 355) + 12 * (Latitude - Longitude) / PI;
	float declination = 0.4093 * sin( 2 * PI * (JulianDay - 81) / 368 );
	
	SunTheta = PI / 2 - asin(sin(Latitude) * sin(declination) - cos(Latitude) * cos(declination) * cos(PI * solarTime / 12));
	SunPhi   = atan( -cos(declination) * sin(PI * solarTime / 12) / ((cos(Latitude) * cos(declination) - sin(Latitude) * sin(declination) * sin(PI * solarTime / 12)))); 

	if (SunPhi < 0) SunPhi = 2 * PI + SunPhi;
	
        MoonPhi = PI / 2;
        MoonTheta = PI / 3;

	int ind, num;
	
	float halfPI = PI / 2;
	float SliceAng = (PI / 2) / NumSlices;
	float SideAng = (PI * 2) / NumSides;
	
	float theta, phi;
    float turbidity = 2.0f;
	int index = 0;	
	Shader.SetInfo(SunTheta, turbidity);
	
	for (ind = 0; ind <= NumSlices; ind++)
	{
		for (num = 0; num <= NumSides; num++)
		{
			theta = halfPI - SliceAng * ind;
			phi = num * SideAng;

			if (ColoringMode == COLOR_MODEL)
				Shader.GetVertexColor(theta, phi, SunPhi, &ColorBuffer[index]);
			if (ColoringMode == COLOR_TABLE)
				ColorTable.GetColor(hour, phi, halfPI - theta, &ColorBuffer[index]);
			if (ColoringMode == COLOR_SKYMAP)
				SkyMapTexCoordBuffer[ind * (NumSides + 1) + num].x = (hour/24.0f);
		
			index++;
		}
	}
}

void CSkyDome::Render(Vector3 Camera)
{
	glPushMatrix();
	//glTranslatef(Camera.x, Camera.y, Camera.z); // Uncomment this if you want the dome to follow the user so its extends cant be reached
	glDisable(GL_DEPTH_TEST);
	if (ColoringMode != COLOR_TEXTURE)
            RenderElements(true);   // Render Moon Billboard and Stars if no static SkyTexture is used for the Dome
	glEnable(GL_DEPTH_TEST);
	RenderDome();
	glDisable(GL_DEPTH_TEST);
	if (ColoringMode != COLOR_TEXTURE)
	    RenderElements(false);    // Render Sun Billboard if no static SkyTexture is used for the Dome
	glEnable(GL_DEPTH_TEST);
	glPopMatrix();
}

Color4 CSkyDome::GetHorizonColor(Vector3 CameraView)
{	
	float ang = atan2(CameraView.z, CameraView.x);
	if (ang < 0) ang = 2 * PI + ang;
	int ri = (int)((ang / (2 * PI)) * NumSides);
	
	return Color4(ColorBuffer[ri].R, ColorBuffer[ri].G, ColorBuffer[ri].B, 1.0f);
}

void CSkyDome::RenderElements(bool initial)
{
	float mat[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, mat);    // Get modelview amtrix to render Sun,Moon Billboards
		
	Vector3 vx(mat[0], mat[4], mat[8] );
	Vector3 vy(mat[1], mat[5], mat[9] );
	Vector3 Vertices[4];

	float SunSize = 70, MoonSize = 40;

	glEnable(GL_TEXTURE_2D);

	if (initial)
	{
		glDisable(GL_TEXTURE_2D);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_COLOR_ARRAY);
		
		glVertexPointer(3, GL_FLOAT, 0, Stars);
		glColorPointer(4, GL_FLOAT, 0, StarColors);
		glDrawArrays(GL_POINTS, 0, NumStars);
		
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_COLOR_ARRAY);
                glColor3f(1,1,1);
  	        glEnable(GL_TEXTURE_2D);

		glEnable(GL_ALPHA_TEST);
		glAlphaFunc(GL_GREATER, 0.0f);

		glBindTexture(GL_TEXTURE_2D, MoonTex);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
                
            	Vector3 MoonPos(sin(MoonTheta) * cos(MoonPhi) * SkyRadius,
			    		cos(MoonTheta) * SkyRadius,
			    		sin(MoonTheta) * sin(MoonPhi) * SkyRadius);

		Vertices[0] = (MoonPos + (vx + vy) * -MoonSize*3);		
		Vertices[1] = (MoonPos + (vx - vy) *  MoonSize*3);
		Vertices[2] = (MoonPos + (vx + vy) *  MoonSize*3); 			
		Vertices[3] = (MoonPos + (vy - vx) *  MoonSize*3);

		glBegin(GL_QUADS);
			glTexCoord2f(1.0f , 0.0f); 
			glVertex3fv(Vertices[0].v);
			glTexCoord2f(0.0f , 0.0f); 
			glVertex3fv(Vertices[1].v);
			glTexCoord2f(0.0f , 1.0f); 
			glVertex3fv(Vertices[2].v);
			glTexCoord2f(1.0f , 1.0f); 
			glVertex3fv(Vertices[3].v);
		glEnd();	

		glDisable(GL_BLEND);
	}
	else
	{
   		Vector3 SunPos(	sin(SunTheta) * cos(SunPhi) * SkyRadius,
			    		cos(SunTheta) * SkyRadius,
			    		sin(SunTheta) * sin(SunPhi) * SkyRadius);
					
		Vertices[0] = (SunPos + (vx + vy) * -SunSize * 3);		
		Vertices[1] = (SunPos + (vx - vy) *  SunSize * 3);
		Vertices[2] = (SunPos + (vx + vy) *  SunSize * 3); 			
		Vertices[3] = (SunPos + (vy - vx) *  SunSize * 3);
			
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D, FlareTex0);
		
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);

		glColor4f(1.0f, 1.0f, 1.0f, 0.8f);
			
		glBegin(GL_QUADS);
			glTexCoord2f(1.0f , 0.0f); 
			glVertex3fv(Vertices[0].v);
			glTexCoord2f(0.0f , 0.0f); 
			glVertex3fv(Vertices[1].v);
			glTexCoord2f(0.0f , 1.0f); 
			glVertex3fv(Vertices[2].v);
			glTexCoord2f(1.0f , 1.0f); 
			glVertex3fv(Vertices[3].v);
		glEnd();	

		glDisable(GL_BLEND);
   	}
	
	glDisable(GL_TEXTURE_2D);
}

void CSkyDome::RenderDome()
{
	int i;
	
	bool UseTexture = ((ColoringMode == COLOR_TEXTURE) || (ColoringMode == COLOR_SKYMAP));

        glColor3f(1,1,1);
	
        glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if (UseTexture) 
	{ 
	    glEnable(GL_TEXTURE_2D);
        if (ColoringMode == COLOR_TEXTURE)
	        glBindTexture(GL_TEXTURE_2D, Texture);
        if (ColoringMode == COLOR_SKYMAP)
	        glBindTexture(GL_TEXTURE_2D, SkyMap);
	}
	
	glEnableClientState(GL_VERTEX_ARRAY);

	if (UseTexture)	
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	else
            glEnableClientState(GL_COLOR_ARRAY);

	glVertexPointer(3, GL_FLOAT, 0, &VertexBuffer[0]);	

	if (UseTexture)
    {
            if (ColoringMode == COLOR_TEXTURE) 
                glTexCoordPointer(2, GL_FLOAT, 0, &TexCoordBuffer[0]);
            if (ColoringMode == COLOR_SKYMAP)
                glTexCoordPointer(2, GL_FLOAT, 0, &SkyMapTexCoordBuffer[0]);
	}
    else 
            glColorPointer(4, GL_FLOAT, 0, &ColorBuffer[0]);

	for (i = 0; i < NumSlices; i++)
	{
		glDrawElements(GL_TRIANGLE_STRIP, (NumSides + 1) * 2, GL_UNSIGNED_SHORT, &IndexBuffer[i * (NumSides + 1) * 2]);
	}
	
	glDisableClientState(GL_VERTEX_ARRAY);	
	if (UseTexture) 
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	else 	
            glDisableClientState(GL_COLOR_ARRAY);

	if (UseTexture) 
            glDisable(GL_TEXTURE_2D);

	glDisable(GL_BLEND);
}

