/*************************************************************************
SkyPlane.cpp

SkyPlane implementation.

Author   :   Dimi Christopoulos

*************************************************************************/
#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/gl.h>	/* Header File For The OpenGL32 Library*/
#include <GL/glu.h>	/* Header File For The GLu32 Library*/

#include <math.h>
#include <stdio.h>      /* Header file for standard file i/o.*/

#include "SkyPlane.h"


typedef struct 
{
	float x,y,z;
	float alpha;
	float u, v;
} VERTEX;


VERTEX *PlaneVertices;
int NumPlaneVertices;
int *Indices;
int NumIndices;

//////////////////////////////////////////////////////////
//	Function: "SkyPlane"
//
//	Input:	divisions   		How much to subdivide the plane.
//			PlanetRadius		Size of plane, distance from the center to a corner.
//			Atmosphere radius	Define curvature of plane. 
//			hTile,vTile			Horizontal,verticla tiling of texture.
//          alphaPercent        Define Transparency fromt the center of plane to the edges. 1 means no transp, 0 means the whole plane is transparent.
//          exponentialFadeout  Enable exponential fadeout of transparency.
//
//
//	Purpose: Create SKyplane.
//////////////////////////////////////////////////////////
void GenerateSkyPlane(int divisions, float PlanetRadius, float AtmosphereRadius, float hTile, float vTile, float alphaPercent, bool exponentialFadeout)
{
	// Make sure our vertex array is clear
	if (PlaneVertices) 
	{
		delete PlaneVertices;
		PlaneVertices = NULL;
	}

    // Make sure our index array is clear
	if (Indices)
	{
		delete Indices;
		Indices = NULL;
	}

    if (alphaPercent > 1.f)
        alphaPercent = 1.f;
  
	// Set the number of divisions into a valid range
	int divs = divisions;
	if (divisions < 1) 
		divs = 1;

	// Initialize the Vertex and Indices arrays
	NumPlaneVertices = (divs + 1) * (divs + 1);   // 1 division would give 4 verts
	NumIndices  = divs * divs * 2 * 3;       // 1 division would give 6 indices for 2 tris

	PlaneVertices = new VERTEX[NumPlaneVertices];
	Indices = new int[NumIndices];
        

	// Calculate values for dome generation and texturing 
	float plane_size = (float) sqrt(((2.0f*PlanetRadius)*(2.0f*PlanetRadius))*0.5);
	float delta = plane_size/(float)divs;
	float tex_delta = 1.0f/(float)divs;

	// Variables we'll use during the dome's generation
	float x_dist   = 0.0f;
	float z_dist   = 0.0f;
	float x_height = 0.0f;
	float z_height = 0.0f;
	float height   = 0.0f;

	float alpharadius_squared= (plane_size*alphaPercent*0.5f)*(plane_size*alphaPercent*0.5f);
    float planetradius_squared = PlanetRadius*PlanetRadius;

	VERTEX SV; // temporary vertex
    int i;

	for (i=0;i <= divs;i++)
	{
	    for (int j=0; j <= divs; j++)
		{
			x_dist = (-0.5f * plane_size) + ((float)j*delta);
			z_dist = (-0.5f * plane_size) + ((float)i*delta);

            x_height = (x_dist*x_dist);
			z_height = (z_dist*z_dist);
            height = (x_height + z_height)/planetradius_squared;
			SV.x = x_dist;
            SV.y = (1.f - height) * AtmosphereRadius;
			SV.z = z_dist;

            SV.alpha = (x_height + z_height)/ alpharadius_squared;

			if (exponentialFadeout)
                SV.alpha = SV.alpha * SV.alpha * SV.alpha;
            
			SV.alpha = 1 - SV.alpha;
            
            if (SV.alpha < 0.f )
                SV.alpha = 0.f;

			// Calculate the texture coordinates
			SV.u = hTile*((float)j * tex_delta);
			SV.v = vTile*((float)i * tex_delta);

			PlaneVertices[i*(divs+1)+j] = SV;
		}
	}

	// Calculate the indices
	int index = 0;
	for (i=0; i < divs;i++)
	{
		for (int j=0; j < divs; j++)
		{
			int startvert = (i*(divs+1) + j);

		    // first triangle 
			Indices[index++] = startvert;
			Indices[index++] = startvert+1;
			Indices[index++] = startvert+divs+1;

			// second triangle
			Indices[index++] = startvert+1;
			Indices[index++] = startvert+divs+2;
			Indices[index++] = startvert+divs+1;
		}
	}
}


int RenderSkyPlane(int alpha)
{
	glBegin(GL_TRIANGLES);

	for (int i=0; i < NumIndices; i++)
	{
        if (alpha)
            glColor4f(1.f, 1.f, 1.f, PlaneVertices[Indices[i]].alpha);

		glTexCoord2f(PlaneVertices[Indices[i]].u, PlaneVertices[Indices[i]].v);
		glVertex3f(PlaneVertices[Indices[i]].x, PlaneVertices[Indices[i]].y, PlaneVertices[Indices[i]].z);
	}

	glEnd();

	return 1;
}


void ReleaseSkyPlane()
{
	if (PlaneVertices) 
	{
		delete[] PlaneVertices;
		PlaneVertices = NULL;
	}

	if (Indices)
	{
		delete[] Indices;
		Indices = NULL;
	}
}
