////////////////////////////////////////////////////////////////////////////
///	\File "CTerrainTexture.cpp"
//
///	Author: Brian Story
//
///	Creation Date: 5/27/2005
//
///	Purpose: Implimentation of the CTerrainTexture class.
////////////////////////////////////////////////////////////////////////////





#include "CTerrainTexture.h"
#include <stdio.h>
#include <assert.h>
#include <io.h>
#include <math.h>
#include <gl/glu.h>
#include "MGLTargaImage.h"



//////////////////////////////////////////////////////////
///	Function: "loadTexture"
//
///	FirstModified: 9/27/2005
//
///	\param char *filename		Name of the file to load.
//
///	Output: None
//
///	\return int  Texture object holding the loaded texture.
//
///	Purpose: Loads a targa to be used as a terrain texture and returns
//			 the texture object.
//////////////////////////////////////////////////////////
int CTerrainTexture::loadTexture(char *filename)
{
	// Load the image file.
	MGL::TargaImage image;
	if (!image.Load(filename))
		return 0;


	image.FlipVertical();


	// Create the texture.
	GLuint texture;
	glGenTextures(1, &texture);

	glBindTexture(GL_TEXTURE_2D, texture);

	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, 
									 image.GetWidth(), 
									 image.GetHeight(),  
									 GL_RGB, 
									 GL_UNSIGNED_BYTE, 
									 image.GetImage());

	// Set up filtering.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// Enable largest anisotropic filtering.
	if (GLEE_EXT_texture_filter_anisotropic)
	{
		GLfloat largest_supported_anisotropy;
		glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &largest_supported_anisotropy);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, largest_supported_anisotropy);
	}

	// Set up the wrap.
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);


	// Release the image memory.
	image.Release();

	// Return the texture.
	return texture;
}


//////////////////////////////////////////////////////////
///	Function: "createTextureCoords"
//
///	FirstModified: 9/27/2005
//
///	\param None
//
///	Output: None
//
///	\return None
//
///	Purpose: Creates texture coordinates for each point in the heightmap.
//////////////////////////////////////////////////////////
void CTerrainTexture::createTextureCoords(void)
{
	float xcoord = 0;
	float zcoord = 0;

	// The texture coordinates increase along the x and z value by the
	// x and z steps, which we get by dividing the number of times we
	// want the texture to tile along the x and z axis by the width and
	// height of the heightmap.
	float xstep = TextureTileAmount/(float)Width;
	float zstep = TextureTileAmount/(float)Height;

	// Loop through each point and calculate the texture coordinates.
	for (int z = 0; z < Height; z++)
	{
		xcoord = 0;
		for (int x = 0; x < Width; x++)
		{			
			data[x][z].u = xcoord;
			data[x][z].v = zcoord;

			xcoord += xstep;
		}
		zcoord += zstep;
	}

	// Create texture coords for lighting.
	for (z = 0; z < Height; z++)
	{
		for (int x = 0; x < Width; x++)
		{			
			data[x][z].lightU = (float)x/Width;
			data[x][z].lightV = (float)z/Height;
		}
	}
}


//////////////////////////////////////////////////////////
///	Function: "createHorizonMap"
//
///	FirstModified: 5/12/2005
//
///	\param None
//
///	Output: None
//
///	\return None
//
///	Purpose: Sets the horizon slope at each point in the heightmap.
//////////////////////////////////////////////////////////
void CTerrainTexture::createHorizonMap(void)
{
	// This function will create a horizon map for the heightmap.  A
	// horizon map is a set of numbers defining the position of the 
	// horizon in a number of directions from each point in the heightmap.
	//
	// This horizon map will just include one direction: Negative X.  This
	// will allow for shadows to be cast if the light source is on the -X
	// side of the heightmap.

	// Loop through each value in the heightmap.
	for (int z = 0; z < Height; z++)
	{
		for (int x = 0; x < Width; x++)
		{
			float maxSlope = 0;

			// Calculate the slope of the line between this point and every
			// point to the -x of this point.  The largest slope will be
			// the value we save.
			for (int i = x; i >= 0; i--)
			{
				float slope = 0;
				
				// Calculate the slope at this point.
				if (i == x)
					slope = 0;
				else
					slope = float(data[i][z].y - data[x][z].y) / (i - x);
				
				// Negate the slope calculated since we're moving in the -x direction.
				slope = slope * -1;
				
				// If the calculated slope is larger, save it.
				maxSlope = slope > maxSlope ? slope : maxSlope;
			}
			// Record the largest slope found.
			data[x][z].horizon = maxSlope;
		}
	}
}


//////////////////////////////////////////////////////////
///	Function: "createColors"
//
///	FirstModified: 9/27/2005
//
///	\param None
//
///	Output: None
//
///	\return None
//
///	Purpose: Sets the vertex color for each point in the heightmap.
//////////////////////////////////////////////////////////
void CTerrainTexture::createColors(void)
{
	for (int y = 0; y < Height; y++)
	{
		for (int x = 0; x < Width; x++)
		{
			float slope = data[x][y].normal.y;

			// Calculate our color values.
			GLubyte color = 255 - GLubyte(255 * max(0, min(1,((data[x][y].y-10)/30.0f))));
			GLubyte alpha = 255 - GLubyte(255 * max(0, (min(1, ((1 - slope - .1f) / .3f))))); 

			// Assign the color value to this point.
			data[x][y].color.r = color;
			data[x][y].color.g = color;
			data[x][y].color.b = color;
			data[x][y].color.a = alpha;
		}
	}
}


//////////////////////////////////////////////////////////
///	Function: "createLightmap"
//
///	FirstModified: 5/12/2005
//
///	\param float lightX		X position of the light.
///	\param float lightY		Y position of the light.
///	\param float lightZ		Z position of the light.
//
///	Output: None
//
///	\return None
//
///	Purpose: Creates a lightmap from the given light position.
//////////////////////////////////////////////////////////
void CTerrainTexture::createLightmap(float lightX, float lightY, float lightZ)
{
	// Create a normalized vector from the light position.
	double len = sqrt(lightX*lightX + lightY*lightY + lightZ*lightZ);

	float lightingX = float(lightX / len);
	float lightingY = float(lightY / len);
	float lightingZ = float(lightZ / len);


	// Since the lightmap is most likely not the same size as the heightmap,
	// we divide the width and height by the lightmaps width and height to
	// figure out how often we need to sample normal and horizon information
	// to create the texture.
	float xstep = (float)Width / LightmapTextureWidth;
	float ystep = (float)Height / LightmapTextureHeight;
	   

	// Allocate memory for our texture data.
	GLubyte *lightmap;
	lightmap = new GLubyte[LightmapTextureWidth*LightmapTextureHeight];


	// Loop through each value of our *Lightmap*, calculating a light value
	// using a sample point found through our xstep and ystep.
	int actualX, actualZ;
	for (int y = 0; y < LightmapTextureHeight; y++)
	{
		// Calculate the z value of our sample point.
		actualZ = int(y*ystep);
		
		for (int x = 0; x < LightmapTextureWidth; x++)
		{
			// Calculate the x value of our sample point.
			actualX = int(x*xstep);
			float slope = 999999999;
			float slope2 = 999999999;

			if (lightX - actualX != 0)
			{
				// Calculate the slope to the top of the light source.
				slope  = float((lightY        - data[actualX][actualZ].y)) / (lightX - actualX) * -1;
				
				// Calculate the slope to the bottom of the light source.
				slope2 = float(((lightY + 25) - data[actualX][actualZ].y)) / (lightX - actualX) * -1;
			}

			// Calculate a shadow value between 100 and 255  (255 is full light,
			// and 100 is full shadow.  This keeps the it from getting too dark).
			//
			// If the bottom of the light source is above the horizon, use full
			// light.  If not, calculate a value between 100 and 255 based on the
			// amount of the light source showing.
			int shadow = slope2 > data[actualX][actualZ].horizon ? 255 : (100 + 155 * max(0, min(1, (slope2-slope) / (data[actualX][actualZ].horizon-slope)))); 

			// Calculate the lightmap value by calculating the light value using
			// the formula in the red book and combining it with the shadow value.
			lightmap[y*LightmapTextureWidth+x] = max(0, lightingX*data[actualX][actualZ].normal.x  +
														lightingY*data[actualX][actualZ].normal.y  +
														lightingZ*data[actualX][actualZ].normal.z) * 
														shadow;								 												
		}
	}


	// Create a texture object for the lightmap.
	if (lightmapTexture > 0)
		glDeleteTextures(1, &lightmapTexture);
	glGenTextures(1, &lightmapTexture);


	// Create the lightmap using the data created above.
	glBindTexture(GL_TEXTURE_2D, lightmapTexture);
	glTexImage2D(GL_TEXTURE_2D, 0,  1, 
									LightmapTextureWidth, 
									LightmapTextureHeight,
									0,
									GL_LUMINANCE,
									GL_UNSIGNED_BYTE,
									lightmap);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// Set up the wrap.
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_CLAMP);

	// Delete the lightmap data.
	delete [] lightmap;
}


//////////////////////////////////////////////////////////
///	Function: "enableTextures"
//
///	FirstModified: 9/27/2005
//
///	\param None
//
///	Output: None
//
///	\return None
//
///	Purpose: Set up and enable the texture units to texture the heightmap.
//////////////////////////////////////////////////////////
void CTerrainTexture::enableTextures(void)
{
	// Lighting effects are calculated before the textures are blended,
	// so lighting wouldn't do us any good.
	glDisable(GL_LIGHTING);

	// Enable the first texture unit.
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, groundTextures[unita]);

	// Enable the second texture unit.
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, groundTextures[unitb]);

	// Enable the third texture unit.
	glActiveTextureARB(GL_TEXTURE2_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, groundTextures[unitc]);

	// Enable the fourth texture unit.
	glActiveTextureARB(GL_TEXTURE3_ARB);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, lightmapTexture);

}


//////////////////////////////////////////////////////////
///	Function: "disableTextures"
//
///	FirstModified: 9/27/2005
//
///	\param None
//
///	Output: None
//
///	\return None
//
///	Purpose: Disable the texture units.
//////////////////////////////////////////////////////////
void CTerrainTexture::disableTextures(void)
{
	// Disable the texture units.
	glActiveTextureARB(GL_TEXTURE3_ARB);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE2_ARB);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE1_ARB);
	glDisable(GL_TEXTURE_2D);
	glActiveTextureARB(GL_TEXTURE0_ARB);
	glDisable(GL_TEXTURE_2D);

	glEnable(GL_LIGHTING);
}

