/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"


MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init(HDC g_hDC)
{	
	ghDC = g_hDC;

	if (!GLeeInit())
		return false;

	// Enable Depth Testing.
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	// Enable Culling.
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);

	// Enable Fogging.
	glEnable(GL_FOG);
	glFogf(GL_FOG_START, 600); 
	glFogf(GL_FOG_END, 1000);

	GLfloat fogcolor[4] = {.6196078f, .6588235f, .7647058f, 0.0f};

	glFogfv(GL_FOG_COLOR, fogcolor);
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_DENSITY, 0.5f);
	glHint(GL_FOG_HINT, GL_NICEST);

	// Set the clear color to a nice sky blue.
	glClearColor(.7372549f, .7803921f, .9647058f, 1.0f);


	// Load our font.
	if (!font.load("font.tga", "font.dat"))
		return false;

	// Initialize the heightmap by creating a midpoint displacement heightmap.
	heightmap.createHeightmap(CHeightClass::Midpoint);

	// Initialize the texture data.
	texture.init(heightmap.getHeightData());

	// Create our mesh from the heightmap.
	terrain.init(heightmap.getHeightData());

	// Create the lightmap.
	texture.createLightmap(-1, 25, -1);

	// Initialize the toggle variables.
	wireframe = false;
	boundingBox = false;

	// Initialize the movement variables.
	rotateX = 0;
	rotateY = 0;
	translateX = 0;
	translateZ = 0;

	timeSinceInput = 0.0;

	return true;
}

bool MGLApp::Shutdown()
{
	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;					

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );

	// Give the window dimentions to the font class.
	font.setDimentions(width, height);

	// Save the width and height for later use.
	sWidth = width;
	sHeight = height;
}

void MGLApp::Update( double dt )
{
	timeSinceInput += dt;

	// If the '1' key is pressed, create a heightmap using the hill algorithm.
	if ((GetAsyncKeyState('1') & 0x8000))
	{
		glClear(GL_COLOR_BUFFER_BIT);
		font.display("Creating Heightmap With:", 1, 30);
		font.display("Hill Algorithm", 1, 70);
		SwapBuffers(ghDC);

		heightmap.createHeightmap(CHeightClass::Hill);
		texture.init(heightmap.getHeightData());
		terrain.init(heightmap.getHeightData());
		texture.createLightmap(-1, 25, -1);
	}

	// If the '2' key is pressed, create a heightmap using the fault line algorithm.
	if ((GetAsyncKeyState('2') & 0x8000))
	{
		glClear(GL_COLOR_BUFFER_BIT);
		font.display("Creating Heightmap With:", 1, 30);
		font.display("Fault Line Algorithm", 1, 70);
		SwapBuffers(ghDC);

		heightmap.createHeightmap(CHeightClass::FaultLine);
		texture.init(heightmap.getHeightData());
		terrain.init(heightmap.getHeightData());
		texture.createLightmap(-1, 25, -1);
	}

	// If the '3' key is pressed, create a heightmap using midpoint displacement.
	if ((GetAsyncKeyState('3') & 0x8000))
	{
		glClear(GL_COLOR_BUFFER_BIT);
		font.display("Creating Heightmap With:", 1, 30);
		font.display("Midpoint Algorithm", 1, 70);
		SwapBuffers(ghDC);

		heightmap.createHeightmap(CHeightClass::Midpoint);
		texture.init(heightmap.getHeightData());
		terrain.init(heightmap.getHeightData());
		texture.createLightmap(-1, 25, -1);
	}

	// If the F1 key is pressed, toggle wireframe on and off.
	if ((GetAsyncKeyState(VK_F1) & 0x8000) &&
		timeSinceInput > .25)
	{
		timeSinceInput = 0;

		wireframe = !wireframe;
		if (wireframe)
			glPolygonMode(GL_FRONT, GL_LINE);	
		else
			glPolygonMode(GL_FRONT, GL_FILL);
	}


	// Movement input.
	if ((GetAsyncKeyState(VK_UP) & 0x8000))
		rotateX += dt*18;
	if ((GetAsyncKeyState(VK_DOWN) & 0x8000))
		rotateX -= dt*18;
	if ((GetAsyncKeyState(VK_LEFT) & 0x8000))
		rotateY += dt*18;
	if ((GetAsyncKeyState(VK_RIGHT) & 0x8000))
		rotateY -= dt*18;
	if ((GetAsyncKeyState('W') & 0x8000))
		translateZ += dt*100;
	if ((GetAsyncKeyState('S') & 0x8000))
		translateZ -= dt*100;
	if ((GetAsyncKeyState('A') & 0x8000))
		translateX += dt*100;
	if ((GetAsyncKeyState('D') & 0x8000))
		translateX -= dt*100;


	// Input to change the texture in texture unit 0.
	if (GetAsyncKeyState('Z') & 0x8000 &&
		timeSinceInput > .25)
	{
		timeSinceInput = 0;

		texture.changeTexture(GL_TEXTURE0_ARB);
	}

	// Input to change the texture in texture unit 1.
	if (GetAsyncKeyState('X') & 0x8000 &&
		timeSinceInput > .25)
	{
		timeSinceInput = 0;

		texture.changeTexture(GL_TEXTURE1_ARB);
	}
	
	// Input to change the texture in texture unit 2.
	if (GetAsyncKeyState('C') & 0x8000 &&
		timeSinceInput > .25)
	{
		timeSinceInput = 0;

		texture.changeTexture(GL_TEXTURE2_ARB);
	}		

}

void MGLApp::Render()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );


	// set up the camera
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	// Rotate to face the heightmap.
	glRotatef(180, 0, 1, 0);

	// Add the rotation we got from input.
	glRotatef(rotateX, 1, 0, 0);
	glRotatef(rotateY, 0, 1, 0);

	// Start in a position to see the heightmap.
	glTranslatef(-512, -128, 0);

	// Add the movement we got from input.
	glTranslatef(translateX, 0.0f, translateZ);

	// Display the terrain mesh.
	texture.enableTextures();
	terrain.draw(boundingBox);
	texture.disableTextures();
}
