////////////////////////////////////////////////////////////////////////////
///	\File "CTerrainQuadLOD.h"
//
///	Author: Brian Story
//
///	Creation Date: 5/25/2005
//
///	Purpose: Definition of the CTerrainQuadLOD class.
////////////////////////////////////////////////////////////////////////////





#ifndef _CTERRAINQUADLOD_H_
#define _CTERRAINQUADLOD_H_

#include <math.h>
#include <assert.h>
#include "CHeightClass.h"





///////////////////////////////////
//
///	Class CTerrainQuadLOD
//
///	Created: 5/25/2005
///
///////////////////////////////////

class CTerrainQuadLOD
{
	/////////////////////
	// Private Members //
	/////////////////////

	private:

		// Vector Structure
		struct vector
		{
			int x, z;
			float y;
		};


		// Orientation of the corners, sides, children, and adjacent nodes:
		//
		//	   Corners	   Sides	  Children	  Adjacent
		//
		//	   1--.--0	  .--0--.	  .- . -.
		//	   |\   /|	  |  |  |	  | 1|0 |		 0
		//	   .  .  .	  1--.--3	  .--.--.		1x3
		//	   |/   \|	  |  |  |	  | 2|3 |		 2
		//	   2--.--3	  .--2--.	  .--.--.
		//


		// Structure defining a polygon using three vectors and it's plane
		// formula components.
		struct triStruct
		{
			SVector vect[3];		// The three vectors of a polygon.
			float a, b, c, d;		// The four variables used for the plane formula.
		};


		// Quadtree Node
		struct quadNode
		{			
			vector center;			// Center point.
			vector corner[4];		// Corner points.
			bool activeSide[4];		// Sides that are active.
			vector side[4];			// Side points.
			quadNode* children[4];	// Children nodes.
			quadNode *parent;		// parent node.
			int level;				// How deep this node is.
		};

		// The root node of the quadtree.
		quadNode *root;

		// Width and height of the mesh.
		int Width, Height;
		
		// The x and z (left and top) values of the upper left hand corner of the mesh.
		//    We can use these variables to create a VIPM of a section of the heightmap
		//    rather than the whole thing, which would be handy for Chunked LOD.
		int left, top;

		// The heightmap data this mesh is based on.
		heightData **data;

		// The maximum amount of difference between the heightmap values and the
		// mesh values.
		float deltaHeight;



		//////////////////////////////////////////////////////////
		///	Function: "createColors"
		//
		///	FirstModified: 5/27/2005
		//
		///	\param None
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Creates vertex colors for the heightmap.
		//////////////////////////////////////////////////////////
		void createColors(void);


		//////////////////////////////////////////////////////////
		///	Function: "cleanQuadtree"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param quadNode *node		The node to clean.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Cleans a node and it's children, recursively.
		//////////////////////////////////////////////////////////
		void cleanQuadtree(quadNode *node)
		{
			if (node == NULL)
				return;

			// Clean children first.
			cleanQuadtree(node->children[0]);
			cleanQuadtree(node->children[1]);
			cleanQuadtree(node->children[2]);
			cleanQuadtree(node->children[3]);

			// Clean node.
			delete node;
			node = NULL;
		}


		//////////////////////////////////////////////////////////
		///	Function: "vectorEqual"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param vector a			The first vector to compare
		//		   vector b			The second vector to compare
		//
		///	Output: None
		//
		///	\return True if the vectors are equal.
		//
		///	Purpose:  Checks to see if two vectors are equal.
		//////////////////////////////////////////////////////////
		bool vectorEqual(vector a, vector b)
		{
			// Floating points numbers aren't always rounded perfectly, so
			// instead of checking if they are equal, we check if they are
			// close enough.  Only the y portion is a floating point value.
			if (a.x == b.x && 
				a.y - b.y < 0.00001 &&
				a.z == b.z)
			{
				return true;
			}
			return false;
		}


		//////////////////////////////////////////////////////////
		///	Function: "initNode"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param quadNode *node		The node to initialize.
		//		   vector upperLeft		The upper left corner vector of this node.
		//		   int level			The level of this node.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Initializes the given node.
		//////////////////////////////////////////////////////////
		void initNode(quadNode *node, vector upperLeft, int level)
		{
			// Initialize level and make sure everything checks out so far.
			if (node == NULL || (node->parent == NULL && node != root))
			{
				assert(0 && "We have a null node or parent on a child quadtree node.");
			}
			node->level = level;


			// Find the width of this node.
			int width = 1;
			for (int i = 0; i < level; i++)
			{
				width *= 2;	
			}
			width = (Width-1)/width;

			// Make sure width checks out.
			assert(width && "We have a width of 0; quadtree went down too far.");


			// Initialize children and sides.
			for (i = 0; i < 4; i++)
			{
				node->children[i] = NULL;
				node->activeSide[i] = false;
			}
			
			// Initialize corners.
			node->corner[0].x = upperLeft.x + width; 	
			node->corner[0].z = upperLeft.z; 
			node->corner[0].y =	data[node->corner[0].x][node->corner[0].z].y;

			node->corner[1].x = upperLeft.x; 	
			node->corner[1].z = upperLeft.z;
			node->corner[1].y =	data[node->corner[1].x][node->corner[1].z].y;

			node->corner[2].x =	upperLeft.x;
			node->corner[2].z =	upperLeft.z + width;
			node->corner[2].y =	data[node->corner[2].x][node->corner[2].z].y;

			node->corner[3].x =	upperLeft.x + width;
			node->corner[3].z =	upperLeft.z + width;
			node->corner[3].y =	data[node->corner[3].x][node->corner[3].z].y;

			// Initialize Sides.
			node->side[0].x = upperLeft.x + (width/2);		
			node->side[0].z	= upperLeft.z;
			node->side[0].y	= data[node->side[0].x][node->side[0].z].y;
			
			node->side[1].x	= upperLeft.x;
			node->side[1].z	= upperLeft.z + (width/2);
			node->side[1].y	= data[node->side[1].x][node->side[1].z].y;

			node->side[2].x	= upperLeft.x + (width/2);
			node->side[2].z	= upperLeft.z + width;
			node->side[2].y	= data[node->side[2].x][node->side[2].z].y;

			node->side[3].x	= upperLeft.x + width;
			node->side[3].z	= upperLeft.z + (width/2);
			node->side[3].y	= data[node->side[3].x][node->side[3].z].y;
			
			// Initialize center.
			node->center.x = upperLeft.x + (width/2);
			node->center.z = upperLeft.z + (width/2);
			node->center.y = data[node->center.x][node->center.z].y;
		}


		//////////////////////////////////////////////////////////
		///	Function: "turnOnEdges"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param quadNode *node		The node who's edges are to be activated.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Recursively turns on the edges at the side of the mesh.
		//			 This reduces cracks at the edges of two VIPM meshes placed
		//			 next to each other, as in Chunked LOD.
		//////////////////////////////////////////////////////////
		void turnOnEdges(quadNode *node)
		{
			if (node == NULL)
				return;

			// Make sure sides on the edges are turned on.
			if (node->corner[1].x == left)
				node->activeSide[1] = true;

			if (node->corner[1].z == top)
				node->activeSide[0] = true;
			
			if (node->corner[0].x == left+Width-1)
				node->activeSide[3] = true;
			
			if (node->corner[3].z == top+Height-1)
				node->activeSide[2] = true;

			// Turn on child edges.
			for (int i = 0; i < 4; i++)
				turnOnEdges(node->children[i]);
		}


		//////////////////////////////////////////////////////////
		///	Function: "activateChild"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param quadNode *node		The node that owns the child to activate.
		//		   int child			The child to activate.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Activates the given child of the given node.
		//////////////////////////////////////////////////////////
		void activateChild(quadNode *node, int child)
		{
			// make sure our child input is within range.
			if (child < 0 || child > 3)
				return;

			// Make sure the node isn't null.
			if (node == NULL)
				return;

			// Check to see if child is already active.
			if (node->children[child] != NULL)
				return;


			// First, create and initialize the child.
			node->children[child] = new quadNode;
			node->children[child]->parent = node;
			node->children[child]->level = node->level + 1;
			if (child == 0)
			{
				initNode(node->children[child], node->side[0], node->level + 1);
			}
			if (child == 1)
			{
				initNode(node->children[child], node->corner[1], node->level + 1);
			}
			if (child == 2)
			{
				initNode(node->children[child], node->side[1], node->level + 1);
			}
			if (child == 3)
			{
				initNode(node->children[child], node->center, node->level + 1);
			}


			// Initialize the parent.
			int childMinusOne = (child-1) < 0 ? 3 : child-1;
			int childPlusOne  = (child+1) > 3 ? 0 : child+1;

			node->activeSide[child] = true;
			node->activeSide[childMinusOne] = true;


			// Find the adjacent nodes and activate any children needed to
			// prevent cracks.
			quadNode *tmp = findAdjacent(node, child);
			if (tmp != NULL)
			{
				int childPlusTwo = child+2 <= 3 ? child+2 : child+2-4;
				tmp->activeSide[childPlusTwo] = true;
			}

			tmp = findAdjacent(node, childMinusOne);
			if (tmp != NULL)
			{
				int childPlusTwo = childMinusOne+2 <= 3 ? childMinusOne+2 : childMinusOne+2-4;
				tmp->activeSide[childPlusTwo] = true;
			}
		}


		//////////////////////////////////////////////////////////
		///	Function: "findAdjacent"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param quadNode *node	The node who's adjacent sides we want to find.
		//		   int direction	The direction of the adjacent node.
		//
		///	Output: None
		//
		///	\return quadnode *		The adjacent node.
		//
		///	Purpose: Finds the node adjacent to the given node in the given direction.
		//////////////////////////////////////////////////////////
		quadNode *findAdjacent(quadNode *node, int direction)
		{
			// Check to see if an adjacent node exists.
			if ((direction == 0 && node->corner[1].z == top) ||
				(direction == 1 && node->corner[1].x == left) ||
				(direction == 2 && node->corner[3].z == top+Height-1) ||
				(direction == 3 && node->corner[3].x == left+Width-1))
			{
				return NULL;
			}

			
			// Start at the root node and work our way down.
			quadNode *tmpNode = root;
			int ulx, ulz;


			// Find the length of the node.
			int nodeLength = node->corner[0].x - node->corner[1].x;
			
			// Find the corner of the original node to match up.
			int cornerA = direction;
			
			// Find the corner of the new node to match up.
			int cornerB = direction-1;
			if (cornerB < 0)
				cornerB = 3;

			// Find the upper left hand corner of the new node.
			if (direction == 0)
			{
				ulx = node->corner[1].x;
				ulz = node->corner[1].z - nodeLength;	
			}
			if (direction == 1)
			{
				ulx = node->corner[1].x - nodeLength;
				ulz = node->corner[1].z;
			}
			if (direction == 2)
			{
				ulx = node->corner[2].x;
				ulz = node->corner[2].z;
			}
			if (direction == 3)
			{
				ulx = node->corner[0].x;
				ulz = node->corner[0].z;
			}

			// Start at the top and work down until we find our adjacent node.
			while (nodeLength != (tmpNode->corner[0].x - tmpNode->corner[1].x) ||
				   !vectorEqual(node->corner[cornerA], tmpNode->corner[cornerB]))
			{
				int active = 0;
				
				// Find out which child of tmpNode we need to check next.
				if (ulx < tmpNode->center.x)
				{
					if (ulz < tmpNode->center.z)
						active = 1;
					else
						active = 2;
				}
				else
				{
					if (ulz < tmpNode->center.z)
						active = 0;
					else
						active = 3;
				}

				// If the child doesn't exist, create it.
				if (tmpNode->children[active] == NULL)
					activateChild(tmpNode, active);

				tmpNode = tmpNode->children[active];
			}

			return tmpNode;
		}

		
		//////////////////////////////////////////////////////////
		///	Function: "renderQuad"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param quadNode *node		The node to render.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Renders the given node and it's children recursively.
		//////////////////////////////////////////////////////////
		void renderQuad(quadNode *node)
		{
			if (node == NULL)
				return;

			// Render this node if it has less than four children.
			if (!node->children[0] ||
				!node->children[1] ||
				!node->children[2] ||
				!node->children[3])
			{
				glBegin(GL_TRIANGLE_FAN);

				// Start with the center point.
				glColor4ub(data[node->center.x][node->center.z].color.r, data[node->center.x][node->center.z].color.g, data[node->center.x][node->center.z].color.b, data[node->center.x][node->center.z].color.a);
				glVertex3f(node->center.x, node->center.y, node->center.z);

				for (int i = 0; i < 4; i++)
				{
					// Add the corner if it's child doesn't exist.
					if (!node->children[i])
					{
						glColor4ub(data[node->corner[i].x][node->corner[i].z].color.r, data[node->corner[i].x][node->corner[i].z].color.g, data[node->corner[i].x][node->corner[i].z].color.b, data[node->corner[i].x][node->corner[i].z].color.a);
						glVertex3f(node->corner[i].x, node->corner[i].y, node->corner[i].z);
					}

					// Add the side if it's active.
					if (node->activeSide[i])
					{
						glColor4ub(data[node->side[i].x][node->side[i].z].color.r, data[node->side[i].x][node->side[i].z].color.g, data[node->side[i].x][node->side[i].z].color.b, data[node->side[i].x][node->side[i].z].color.a);
						glVertex3f(node->side[i].x, node->side[i].y, node->side[i].z);
					}
					int next = i+1;
					if (next > 3)
						next = 0;
											 
					// Repeat the center point to skip around the corner if it
					// has a child.
					if (node->children[next])
					{
						glColor4ub(data[node->center.x][node->center.z].color.r, data[node->center.x][node->center.z].color.g, data[node->center.x][node->center.z].color.b, data[node->center.x][node->center.z].color.a);
						glVertex3f(node->center.x, node->center.y, node->center.z);	
					}
				}

				// End the fan at the starting point.
				if (!node->children[0])
				{
					glColor4ub(data[node->corner[0].x][node->corner[0].z].color.r, data[node->corner[0].x][node->corner[0].z].color.g, data[node->corner[0].x][node->corner[0].z].color.b, data[node->corner[0].x][node->corner[0].z].color.a);
					glVertex3f(node->corner[0].x, node->corner[0].y, node->corner[0].z);
				}

				glEnd();
			}

			// Render node's children.
			for (int i = 0; i < 4; i++)
				renderQuad(node->children[i]);
		}


		//////////////////////////////////////////////////////////
		///	Function: "findDifference"
		//
		///	FirstModified: 6/27/2005
		//
		///	\param quadNode *node		The node to check.
		//		   int child			The child area of the node to check.
		//
		///	Output: None
		//
		///	\return the difference between the heightmap and the mesh.
		//
		///	Purpose: Find the greatest difference between the current mesh
		//			 and the heightmap data in the given area of the given node.
		//////////////////////////////////////////////////////////
		float findDifference(quadNode *node, int child)
		{

			SVector upperLeft;
			SVector lowerRight;

			// Find the upper left hand vector and lower right hand vector
			// of the given area.
			if (child == 0)
			{
				upperLeft.x = node->side[0].x;
				upperLeft.z = node->side[0].z;
				lowerRight.x = node->side[3].x;
				lowerRight.z = node->side[3].z;

			}

			if (child == 1)
			{
				upperLeft.x = node->corner[1].x;
				upperLeft.z = node->corner[1].z;
				lowerRight.x = node->center.x;
				lowerRight.z = node->center.z;
			}

			if (child == 2)
			{
				upperLeft.x = node->side[1].x;
				upperLeft.z = node->side[1].z;
				lowerRight.x = node->side[2].x;
				lowerRight.z = node->side[2].z;
			}

			if (child == 3)
			{
				upperLeft.x = node->center.x;
				upperLeft.z = node->center.z;
				lowerRight.x = node->corner[3].x;
				lowerRight.z = node->corner[3].z;
			}


			// Find polygons created by this section of the mesh.
			triStruct triA;
			triStruct triB;

			int childMinusOne = child-1 < 0 ? 3 : child-1;
			int childPlusOne = child+1 > 3 ? 0 : child+1;

			// Triangle A
			triA.vect[0].x = node->center.x;
			triA.vect[0].y = node->center.y;
			triA.vect[0].z = node->center.z;
			if (node->activeSide[childMinusOne])
			{
				triA.vect[1].x = node->side[childMinusOne].x;
				triA.vect[1].y = node->side[childMinusOne].y;
				triA.vect[1].z = node->side[childMinusOne].z;
			}
			else
			{
				triA.vect[1].x = node->corner[childMinusOne].x;
				triA.vect[1].y = node->corner[childMinusOne].y;
				triA.vect[1].z = node->corner[childMinusOne].z;
			}
			triA.vect[2].x = node->corner[child].x;
			triA.vect[2].y = node->corner[child].y;
			triA.vect[2].z = node->corner[child].z;

			// Triangle B
			triB.vect[0].x = node->center.x;
			triB.vect[0].y = node->center.y;
			triB.vect[0].z = node->center.z;
			triB.vect[1].x = node->corner[child].x;
			triB.vect[1].y = node->corner[child].y;
			triB.vect[1].z = node->corner[child].z;
			if (node->activeSide[child])
			{
				triB.vect[2].x = node->side[child].x;
				triB.vect[2].y = node->side[child].y;
				triB.vect[2].z = node->side[child].z;
			}
			else
			{
				triB.vect[2].x = node->corner[childPlusOne].x;
				triB.vect[2].y = node->corner[childPlusOne].y;
				triB.vect[2].z = node->corner[childPlusOne].z;
			}

			// Create planes from triangles.

			// Formula of a plane is (ax + by + cz + d = 0)
			// where (a, b, c) is the normal of the plane
			// and d = -ax -by -cz where (x,y,z) is the point
			// the normal is created from.
			// This means Y = (-d-ax-cz)/b
			
			// Triangle A			
			triA.a = ((triA.vect[1].y - triA.vect[0].y) * (triA.vect[2].z - triA.vect[0].z)) -
					 ((triA.vect[1].z - triA.vect[0].z) * (triA.vect[2].y - triA.vect[0].y));
			triA.b = ((triA.vect[1].z - triA.vect[0].z) * (triA.vect[2].x - triA.vect[0].x)) -
					 ((triA.vect[1].x - triA.vect[0].x) * (triA.vect[2].z - triA.vect[0].z)); 
			triA.c = ((triA.vect[1].x - triA.vect[0].x) * (triA.vect[2].y - triA.vect[0].y)) -
					 ((triA.vect[1].y - triA.vect[0].y) * (triA.vect[2].x - triA.vect[0].x));
 
			triA.d = (-triA.a * triA.vect[0].x) - 
					 ( triA.b * triA.vect[0].y) - 
					 ( triA.c * triA.vect[0].z); 


			// Triangle B
			triB.a = ((triB.vect[1].y - triB.vect[0].y) * (triB.vect[2].z - triB.vect[0].z)) -
					 ((triB.vect[1].z - triB.vect[0].z) * (triB.vect[2].y - triB.vect[0].y));
			triB.b = ((triB.vect[1].z - triB.vect[0].z) * (triB.vect[2].x - triB.vect[0].x)) -
					 ((triB.vect[1].x - triB.vect[0].x) * (triB.vect[2].z - triB.vect[0].z)); 
			triB.c = ((triB.vect[1].x - triB.vect[0].x) * (triB.vect[2].y - triB.vect[0].y)) -
					 ((triB.vect[1].y - triB.vect[0].y) * (triB.vect[2].x - triB.vect[0].x));
 
			triB.d = (-triB.a * triB.vect[0].x) - 
					 ( triB.b * triB.vect[0].y) - 
					 ( triB.c * triB.vect[0].z); 

			

			// Loop through the heightmap points inside the bounds, and find
			// our the max difference inside of this section of mesh.
			float maxDif = 0;
			for (int z = upperLeft.z; z <= lowerRight.z; z++)
			{
				for (int x = upperLeft.x; x <= lowerRight.x; x++)
				{
					// Find which triangle we're in.

					// Tri A
					if ( ((triA.vect[1].z - triA.vect[0].z)*(x - triA.vect[0].x)) -
						 ((triA.vect[1].x - triA.vect[0].x)*(z - triA.vect[0].z)) >= 0 &&
						 ((triA.vect[2].z - triA.vect[1].z)*(x - triA.vect[1].x)) - 
						 ((triA.vect[2].x - triA.vect[1].x)*(z - triA.vect[1].z)) >= 0 &&
						 ((triA.vect[0].z - triA.vect[2].z)*(x - triA.vect[2].x)) - 
						 ((triA.vect[0].x - triA.vect[2].x)*(z - triA.vect[2].z)) >= 0)
					{
						// Use the plane formula to find the y value at this point.
						float y = (-triA.d - (triA.a*x) - (triA.c*z))/triA.b;
						if (maxDif < (abs(y - data[x][z].y)))
						{
							maxDif = abs(y - data[x][z].y);
						}
					}

					// Tri B
					else
					{
						// Use the plane formula to find the y value at this point.
						float y = (-triB.d - (triB.a*x) - (triB.c*z))/triB.b;
						if (maxDif < (abs(y - data[x][z].y)))
						{
							maxDif = abs(y - data[x][z].y);
						}
					}
				}
			}

			return maxDif;
		}


		//////////////////////////////////////////////////////////
		///	Function: "createMesh"
		//
		///	FirstModified: 5/25/2005
		//
		///	\param quadNode *node		The node to create the mesh for.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Recursively creates the whole mesh.
		//////////////////////////////////////////////////////////
		void createMesh(quadNode *node)
		{
			// If we've gone as far as we can, stop.
			int width = node->corner[0].x - node->corner[1].x;
			if (width <= 3)
			{
				return;
			}

			// Loop through all four children.
			for (int i = 0; i < 4; i++)
			{
				// Check to see how close the mesh is to the heightmap data
				// for each child.  If it's not less than deltaHeight, activate
				// the child.
				if (findDifference(node, i) > deltaHeight)
				{
					activateChild(node, i);
					createMesh(node->children[i]);
				}
			}
		}


	////////////////////
	// Public Members //
	////////////////////

	public:


	/////////////////
	// Constructor //
	/////////////////

		//////////////////////////////////////////////////////////
		///	Function: CTerrainQuadLOD
		//
		///	LastModified: 5/25/2005
		//
		///	\param	None
		//
		///	Output:	None
		//
		///	\return None
		//
		///	Purpose: Constructor for the class
		//////////////////////////////////////////////////////////
		CTerrainQuadLOD(void)
		{
			// Initialize the member varaibles.
			root = NULL;
			data = NULL;
			deltaHeight = 0.0f;

			Width = 0;
			Height = 0;
			left = 0;
			top = 0;
		}


	////////////////
	// Destructor //
	////////////////

		//////////////////////////////////////////////////////////
		///	Function: ~CTerrainQuadLOD
		//
		///	LastModified: 5/25/2005
		//
		///	\param	None
		//
		///	Output:	None
		//
		///	\return None
		//
		///	Purpose: Destructor for the class
		//////////////////////////////////////////////////////////
		virtual ~CTerrainQuadLOD(void)
		{
			cleanQuadtree(root);
			root = NULL;
			data = NULL;
		}


		//////////////////////////////////////////////////////////
		///	Function: "draw"
		//
		///	FirstModified: 5/26/2005
		//
		///	\param None
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Renders the mesh.
		//////////////////////////////////////////////////////////
		void draw(void)
		{
			renderQuad(root);
		}


		//////////////////////////////////////////////////////////
		///	Function: "init"
		//
		///	FirstModified: 5/26/2005
		//
		///	\param heightData ***heights	The heightmap data.
		//		   float heightDifference	The max difference between the mesh and heightmap data allowed.
		//		   int p_width				The width of the mesh to create.
		//		   int p_height				The height of the mesh to create.
		//		   int p_left				The left hand (x) value to start at.
		//		   int p_top				The top (z) value to start at.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Initializes the VIPM mesh.
		//////////////////////////////////////////////////////////
		void init(heightData ***heights, float heightDifference,
				  int p_width, int p_height, int p_left, int p_top)
		{
			// Save local copies of the data.
			data = *heights;
			deltaHeight = heightDifference;

			Width = p_width;
			Height = p_height;
			left = p_left;
			top = p_top;

			// Create the color values.
			createColors();

			// Clean any mesh we currently have.
			cleanQuadtree(root);
			root = NULL;

			// Create the root node.
			root = new quadNode;
			root->parent = NULL;
			root->level = 0;
			
			// Initialize the root node.
			vector rootVect;
			rootVect.x = left;
			rootVect.z = top;
			rootVect.y = data[left][top].y;
			initNode(root, rootVect, 0);
			
			// Create the mesh.
			createMesh(root);
			turnOnEdges(root);
		}


		//////////////////////////////////////////////////////////
		///	Function: "resetTeselation"
		//
		///	FirstModified: 5/26/2005
		//
		///	\param float heightDifference		The maximum allowed difference 
		//										between the mesh and the height data.
		//
		///	Output: None
		//
		///	\return None
		//
		///	Purpose: Recreate the mesh with the given height difference.
		//////////////////////////////////////////////////////////
		void resetTeselation(float heightDifference)
		{
			deltaHeight = heightDifference;

			// Clean up the current mesh.
			cleanQuadtree(root);
			root = NULL;

			// Create the root node.
			root = new quadNode;
			root->parent = NULL;
			root->level = 0;
			
			// Initialize the root node.
			vector rootVect;
			rootVect.x = left;
			rootVect.z = top;
			rootVect.y = data[left][top].y;
			initNode(root, rootVect, 0);
			
			// Create the mesh.
			createMesh(root);
			turnOnEdges(root);
		}


		//////////////////////////////////////////////////////////
		///	Function: "getHeight"
		//
		///	FirstModified: 6/27/2005
		//
		///	\param int x		The x position.
		//		   int z		The z position.
		//
		///	Output: None
		//
		///	\return The height at the given point.
		//
		///	Purpose: Returns the mesh height at the given x z point on the mesh.
		//////////////////////////////////////////////////////////
		float getHeight(int x, int z)
		{
			quadNode *node = root;
			int child = 0;

			// Find which node and child area the point is in.
			while (true)
			{
				if (x < node->center.x)
					child = z < node->center.z ? 1 : 2;
				else
					child = z < node->center.z ? 0 : 3;
				
				if (node->children[child] != NULL)
					node = node->children[child];
				else
					break;
			}


			// Find polygons inside of the given node at the given child area.
			triStruct triA;
			triStruct triB;

			int childMinusOne = child-1 < 0 ? 3 : child-1;
			int childPlusOne = child+1 > 3 ? 0 : child+1;

			// Triangle A
			triA.vect[0].x = node->center.x;
			triA.vect[0].y = node->center.y;
			triA.vect[0].z = node->center.z;
			if (node->activeSide[childMinusOne])
			{
				triA.vect[1].x = node->side[childMinusOne].x;
				triA.vect[1].y = node->side[childMinusOne].y;
				triA.vect[1].z = node->side[childMinusOne].z;
			}
			else
			{
				triA.vect[1].x = node->corner[childMinusOne].x;
				triA.vect[1].y = node->corner[childMinusOne].y;
				triA.vect[1].z = node->corner[childMinusOne].z;
			}
			triA.vect[2].x = node->corner[child].x;
			triA.vect[2].y = node->corner[child].y;
			triA.vect[2].z = node->corner[child].z;

			// Triangle B
			triB.vect[0].x = node->center.x;
			triB.vect[0].y = node->center.y;
			triB.vect[0].z = node->center.z;
			triB.vect[1].x = node->corner[child].x;
			triB.vect[1].y = node->corner[child].y;
			triB.vect[1].z = node->corner[child].z;
			if (node->activeSide[child])
			{
				triB.vect[2].x = node->side[child].x;
				triB.vect[2].y = node->side[child].y;
				triB.vect[2].z = node->side[child].z;
			}
			else
			{
				triB.vect[2].x = node->corner[childPlusOne].x;
				triB.vect[2].y = node->corner[childPlusOne].y;
				triB.vect[2].z = node->corner[childPlusOne].z;
			}

			// Find which of the two triangles holds the point.
			if ( ((triA.vect[1].z - triA.vect[0].z)*(x - triA.vect[0].x)) -
				 ((triA.vect[1].x - triA.vect[0].x)*(z - triA.vect[0].z)) >= 0 &&
				 ((triA.vect[2].z - triA.vect[1].z)*(x - triA.vect[1].x)) - 
				 ((triA.vect[2].x - triA.vect[1].x)*(z - triA.vect[1].z)) >= 0 &&
				 ((triA.vect[0].z - triA.vect[2].z)*(x - triA.vect[2].x)) - 
				 ((triA.vect[0].x - triA.vect[2].x)*(z - triA.vect[2].z)) >= 0)
			{
				// Triangle A			
				triA.a = ((triA.vect[1].y - triA.vect[0].y) * (triA.vect[2].z - triA.vect[0].z)) -
						 ((triA.vect[1].z - triA.vect[0].z) * (triA.vect[2].y - triA.vect[0].y));
				triA.b = ((triA.vect[1].z - triA.vect[0].z) * (triA.vect[2].x - triA.vect[0].x)) -
						 ((triA.vect[1].x - triA.vect[0].x) * (triA.vect[2].z - triA.vect[0].z)); 
				triA.c = ((triA.vect[1].x - triA.vect[0].x) * (triA.vect[2].y - triA.vect[0].y)) -
						 ((triA.vect[1].y - triA.vect[0].y) * (triA.vect[2].x - triA.vect[0].x));
 
				triA.d = (-triA.a * triA.vect[0].x) - 
						 ( triA.b * triA.vect[0].y) - 
						 ( triA.c * triA.vect[0].z); 
			 
				// Return the y value using the plane formula.
				return (-triA.d - (triA.a*x) - (triA.c*z))/triA.b;
			 }
			 else
			 {
				// Triangle B
				triB.a = ((triB.vect[1].y - triB.vect[0].y) * (triB.vect[2].z - triB.vect[0].z)) -
						 ((triB.vect[1].z - triB.vect[0].z) * (triB.vect[2].y - triB.vect[0].y));
				triB.b = ((triB.vect[1].z - triB.vect[0].z) * (triB.vect[2].x - triB.vect[0].x)) -
						 ((triB.vect[1].x - triB.vect[0].x) * (triB.vect[2].z - triB.vect[0].z)); 
				triB.c = ((triB.vect[1].x - triB.vect[0].x) * (triB.vect[2].y - triB.vect[0].y)) -
						 ((triB.vect[1].y - triB.vect[0].y) * (triB.vect[2].x - triB.vect[0].x));
 
				triB.d = (-triB.a * triB.vect[0].x) - 
						 ( triB.b * triB.vect[0].y) - 
						 ( triB.c * triB.vect[0].z); 

				// Return the y value using the plane formula.
				return (-triB.d - (triB.a*x) - (triB.c*z))/triB.b;
			}
		}
};

#endif