/****************************************************************
*																*
* C++ Vector and Matrix Algebra routines						*
* Author: Jean-Francois DOUE									*
* Version 3.1 --- October 1993									*
*---------------------------------------------------------------*
* Modified and updated --- May 2005 by Wael El Oraiby			*
****************************************************************/
#include "algebra3.h"

#ifndef M_PI
	#define M_PI 3.14159265358979323846f
#endif

// error handling macro
#define V_ERROR(E) { std::cerr << E; exit(1); }

/****************************************************************
*																*
*		    vec2 Member functions								*
*																*
****************************************************************/

// CONSTRUCTORS

vec2::vec2() {}

vec2::vec2(const float _x, const float _y)
{ x = _x; y = _y; }

vec2::vec2(const float d)
{ x = y = d; }

vec2::vec2(const vec2& v)
{ x = v.x; y = v.y; }

vec2::vec2(const vec3& v) // it is up to caller to avoid divide-by-zero
{ x = v.x/v.z; y = v.y/v.z; };

vec2::vec2(const vec3& v, int dropAxis) {
    switch (dropAxis) {
	case 0: x = v.y; y = v.z; break;
	case 1: x = v.x; y = v.z; break;
	default: x = v.x; y = v.y; break;
    }
}


// ASSIGNMENT OPERATORS

vec2& vec2::operator = (const vec2& v)
{ x = v.x; y = v.y; return *this; }

vec2& vec2::operator += ( const vec2& v )
{ x += v.x; y += v.y; return *this; }

vec2& vec2::operator -= ( const vec2& v )
{ x -= v.x; y -= v.y; return *this; }

vec2& vec2::operator *= ( const float d )
{ x *= d; y *= d; return *this; }

vec2& vec2::operator /= ( const float d )
{ float d_inv = 1.0f/d; x *= d_inv; y *= d_inv; return *this; }

float& vec2::operator [] ( int i) {
    if (i < 0 || i > 1)
	V_ERROR("vec2 [] operator: illegal access; index = " << i << '\n')
    return n[i];
}


// SPECIAL FUNCTIONS

float vec2::length() const
{ return sqrt(length2()); }

float vec2::length2() const
{ return x*x + y*y; }

vec2& vec2::normalize() // it is up to caller to avoid divide-by-zero
{ *this /= length(); return *this; }


// FRIENDS

vec2 operator - (const vec2& a)
{ return vec2(-a.x,-a.y); }

vec2 operator + (const vec2& a, const vec2& b)
{ return vec2(a.x+ b.x, a.y + b.y); }

vec2 operator - (const vec2& a, const vec2& b)
{ return vec2(a.x-b.x, a.y-b.y); }

vec2 operator * (const vec2& a, const float d)
{ return vec2(d*a.x, d*a.y); }

vec2 operator * (const float d, const vec2& a)
{ return a*d; }

vec2 operator * (const mat3& a, const vec2& v) {
    vec3 av;

    av.x = a.v[0].x*v.x + a.v[0].y*v.y + a.v[0].z;
    av.y = a.v[1].x*v.x + a.v[1].y*v.y + a.v[1].z;
    av.z = a.v[2].x*v.x + a.v[2].y*v.y + a.v[2].z;
    return av;
}

vec2 operator * (const vec2& v, mat3& a)
{ return a.transpose() * v; }

float operator * (const vec2& a, const vec2& b)
{ return (a.x*b.x + a.y*b.y); }

vec2 operator / (const vec2& a, const float d)
{ float d_inv = 1.0f/d; return vec2(a.x*d_inv, a.y*d_inv); }

vec3 operator ^ (const vec2& a, const vec2& b)
{ return vec3(0.0, 0.0, a.x * b.y - b.x * a.y); }

int operator == (const vec2& a, const vec2& b)
{ return (a.x == b.x) && (a.y == b.y); }

int operator != (const vec2& a, const vec2& b)
{ return !(a == b); }

void swap(vec2& a, vec2& b)
{ vec2 tmp(a); a = b; b = tmp; }

vec2 min(const vec2& a, const vec2& b)
{ return vec2(MIN(a.x, b.x), MIN(a.y, b.y)); }

vec2 max(const vec2& a, const vec2& b)
{ return vec2(MAX(a.x, b.x), MAX(a.y, b.y)); }

vec2 prod(const vec2& a, const vec2& b)
{ return vec2(a.x * b.x, a.y * b.y); }

/****************************************************************
*																*
*		    vec3 Member functions								*
*																*
****************************************************************/

// CONSTRUCTORS

vec3::vec3() {}

vec3::vec3(const float _x, const float _y, const float _z)
{ x = _x; y = _y; z = _z; }

vec3::vec3(const float d)
{ x = y = z = d; }

vec3::vec3(const vec3& v)
{ x = v.x; y = v.y; z = v.z; }

vec3::vec3(const vec2& v)
{ x = v.x; y = v.y; z = 1.0; }

vec3::vec3(const vec2& v, float d)
{ x = v.x; y = v.y; z = d; }

vec3::vec3(const vec4& v) // it is up to caller to avoid divide-by-zero
{ x = v.x / v.w; y = v.y / v.w;
  z = v.z / v.w; }

vec3::vec3(const vec4& v, int dropAxis) {
    switch (dropAxis) {
	case 0: x = v.y; y = v.z; z = v.w; break;
	case 1: x = v.x; y = v.z; z = v.w; break;
	case 2: x = v.x; y = v.y; z = v.w; break;
	default: x = v.x; y = v.y; z = v.z; break;
    }
}


// ASSIGNMENT OPERATORS

vec3& vec3::operator = (const vec3& v)
{ x = v.x; y = v.y; z = v.z; return *this; }

vec3& vec3::operator += ( const vec3& v )
{ x += v.x; y += v.y; z += v.z; return *this; }

vec3& vec3::operator -= ( const vec3& v )
{ x -= v.x; y -= v.y; z -= v.z; return *this; }

vec3& vec3::operator *= ( const float d )
{ x *= d; y *= d; z *= d; return *this; }

vec3& vec3::operator /= ( const float d )
{ float d_inv = 1.0f/d; x *= d_inv; y *= d_inv; z *= d_inv;
  return *this; }

float& vec3::operator [] ( int i) {
    if (i < 0 || i > 2)
	V_ERROR("vec3 [] operator: illegal access; index = " << i << '\n')
    return n[i];
}


// SPECIAL FUNCTIONS

float vec3::length() const
{  return sqrt(length2()); }

float vec3::length2() const
{  return x*x + y*y + z*z; }

vec3& vec3::normalize() // it is up to caller to avoid divide-by-zero
{ *this /= length(); return *this; }

// FRIENDS

vec3 operator - (const vec3& a)
{  return vec3(-a.x,-a.y,-a.z); }

vec3 operator + (const vec3& a, const vec3& b)
{ return vec3(a.x+ b.x, a.y + b.y, a.z + b.z); }

vec3 operator - (const vec3& a, const vec3& b)
{ return vec3(a.x-b.x, a.y-b.y, a.z-b.z); }

vec3 operator * (const vec3& a, const float d)
{ return vec3(d*a.x, d*a.y, d*a.z); }

vec3 operator * (const float d, const vec3& a)
{ return a*d; }

vec3 operator * (const mat4& a, const vec3& v)
{ return a * vec4(v); }

vec3 operator * (const vec3& v, mat4& a)
{ return a.transpose() * v; }

float operator * (const vec3& a, const vec3& b)
{ return (a.x*b.x + a.y*b.y + a.z*b.z); }

vec3 operator / (const vec3& a, const float d)
{ float d_inv = 1.0f/d; return vec3(a.x*d_inv, a.y*d_inv,
  a.z*d_inv); }

vec3 operator ^ (const vec3& a, const vec3& b) {
    return vec3(a.y*b.z - a.z*b.y,
		a.z*b.x - a.x*b.z,
		a.x*b.y - a.y*b.x);
}

int operator == (const vec3& a, const vec3& b)
{ return (a.x == b.x) && (a.y == b.y) && (a.z == b.z);
}

int operator != (const vec3& a, const vec3& b)
{ return !(a == b); }

void swap(vec3& a, vec3& b)
{ vec3 tmp(a); a = b; b = tmp; }

vec3 min(const vec3& a, const vec3& b)
{ return vec3(MIN(a.x, b.x), MIN(a.y, b.y), MIN(a.z,
  b.z)); }

vec3 max(const vec3& a, const vec3& b)
{ return vec3(MAX(a.x, b.x), MAX(a.y, b.y), MAX(a.z,
  b.z)); }

vec3 prod(const vec3& a, const vec3& b)
{ return vec3(a.x * b.x, a.y * b.y, a.z * b.z); }


/****************************************************************
*																*
*		    vec4 Member functions								*
*																*
****************************************************************/

// CONSTRUCTORS

vec4::vec4() {}

vec4::vec4(const float _x, const float _y, const float _z, const float _w)
{ x = _x; y = _y; z = _z; w = _w; }

vec4::vec4(const float d)
{  x = y = z = w = d; }

vec4::vec4(const vec4& v)
{ x = v.x; y = v.y; z = v.z; w = v.w; }

vec4::vec4(const vec3& v)
{ x = v.x; y = v.y; z = v.z; w = 1.0; }

vec4::vec4(const vec3& v, const float d)
{ x = v.x; y = v.y; z = v.z;  w = d; }


// ASSIGNMENT OPERATORS

vec4& vec4::operator = (const vec4& v)
{ x = v.x; y = v.y; z = v.z; w = v.w;
return *this; }

vec4& vec4::operator += ( const vec4& v )
{ x += v.x; y += v.y; z += v.z; w += v.w;
return *this; }

vec4& vec4::operator -= ( const vec4& v )
{ x -= v.x; y -= v.y; z -= v.z; w -= v.w;
return *this; }

vec4& vec4::operator *= ( const float d )
{ x *= d; y *= d; z *= d; w *= d; return *this; }

vec4& vec4::operator /= ( const float d )
{ float d_inv = 1.0f/d; x *= d_inv; y *= d_inv; z *= d_inv;
  w *= d_inv; return *this; }

float& vec4::operator [] ( int i) {
    if (i < 0 || i > 3)
	V_ERROR("vec4 [] operator: illegal access; index = " << i << '\n')
    return n[i];
}


// SPECIAL FUNCTIONS

float vec4::length() const
{ return sqrt(length2()); }

float vec4::length2() const
{ return x*x + y*y + z*z + w*w; }

vec4& vec4::normalize() // it is up to caller to avoid divide-by-zero
{ *this /= length(); return *this; }

// FRIENDS

vec4 operator - (const vec4& a)
{ return vec4(-a.x,-a.y,-a.z,-a.w); }

vec4 operator + (const vec4& a, const vec4& b)
{ return vec4(a.x + b.x, a.y + b.y, a.z + b.z,
  a.w + b.w); }

vec4 operator - (const vec4& a, const vec4& b)
{  return vec4(a.x - b.x, a.y - b.y, a.z - b.z,
   a.w - b.w); }

vec4 operator * (const vec4& a, const float d)
{ return vec4(d*a.x, d*a.y, d*a.z, d*a.w ); }

vec4 operator * (const float d, const vec4& a)
{ return a*d; }

vec4 operator * (const mat4& a, const vec4& v) {
    #define ROWCOL(i) a.v[i].n[0]*v.x + a.v[i].n[1]*v.y \
    + a.v[i].n[2]*v.z + a.v[i].n[3]*v.w
    return vec4(ROWCOL(0), ROWCOL(1), ROWCOL(2), ROWCOL(3));
    #undef ROWCOL
}

vec4 operator * (const vec4& v, mat4& a)
{ return a.transpose() * v; }

float operator * (const vec4& a, const vec4& b)
{ return (a.x*b.x + a.y*b.y + a.z*b.z +
  a.w*b.w); }

vec4 operator / (const vec4& a, const float d)
{ float d_inv = 1.0f/d; return vec4(a.x*d_inv, a.y*d_inv, a.z*d_inv,
  a.w*d_inv); }

int operator == (const vec4& a, const vec4& b)
{ return (a.x == b.x) && (a.y == b.y) && (a.z == b.z)
  && (a.w == b.w); }

int operator != (const vec4& a, const vec4& b)
{ return !(a == b); }


void swap(vec4& a, vec4& b)
{ vec4 tmp(a); a = b; b = tmp; }

vec4 min(const vec4& a, const vec4& b)
{ return vec4(MIN(a.x, b.x), MIN(a.y, b.y), MIN(a.z,
  b.z), MIN(a.w, b.w)); }

vec4 max(const vec4& a, const vec4& b)
{ return vec4(MAX(a.x, b.x), MAX(a.y, b.y), MAX(a.z,
  b.z), MAX(a.w, b.w)); }

vec4 prod(const vec4& a, const vec4& b)
{ return vec4(a.x * b.x, a.y * b.y, a.z * b.z,
  a.w * b.w); }


/****************************************************************
*																*
*		    mat3 member functions								*
*																*
****************************************************************/

// CONSTRUCTORS

mat3::mat3() { identity(); }

mat3::mat3(const vec3& v0, const vec3& v1, const vec3& v2)
{ v[0] = v0; v[1] = v1; v[2] = v2; }

mat3::mat3(const float d)
{ v[0] = v[1] = v[2] = vec3(d); }

mat3::mat3(const mat3& m)
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; }


// ASSIGNMENT OPERATORS

mat3& mat3::operator = ( const mat3& m )
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; return *this; }

mat3& mat3::operator += ( const mat3& m )
{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; return *this; }

mat3& mat3::operator -= ( const mat3& m )
{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; return *this; }

mat3& mat3::operator *= ( const float d )
{ v[0] *= d; v[1] *= d; v[2] *= d; return *this; }

mat3& mat3::operator /= ( const float d )
{ v[0] /= d; v[1] /= d; v[2] /= d; return *this; }

vec3& mat3::operator [] ( int i) {
    if (i < 0 || i > 2)
	V_ERROR("mat3 [] operator: illegal access; index = " << i << '\n')
    return v[i];
}


// SPECIAL FUNCTIONS
mat3& mat3::identity()
{
	v[0]	= vec3(1.0f, 0.0, 0.0);
	v[1]	= vec3(0.0, 1.0f, 0.0);
	v[2]	= vec3(0.0, 0.0, 1.0f);

	return *this;
}

mat3 mat3::transpose() {
    return mat3(vec3(v[0][0], v[1][0], v[2][0]),
		vec3(v[0][1], v[1][1], v[2][1]),
		vec3(v[0][2], v[1][2], v[2][2]));
}

mat3 mat3::inverse()	    // Gauss-Jordan elimination with partial pivoting
{
	mat3 a(*this),	    // As a evolves from original mat into identity
		b;	   // b evolves from identity into inverse(a)
	int	 i, j, i1;

	// Loop over cols of a from left to right, eliminating above and below diag
	for (j=0; j<3; j++) {   // Find largest pivot in column j among rows j..2
		i1 = j;		    // Row with largest pivot candidate
		for (i=j+1; i<3; i++)
			if (fabs(a.v[i].n[j]) > fabs(a.v[i1].n[j]))
				i1 = i;

		// Swap rows i1 and j in a and b to put pivot on diagonal
		swap(a.v[i1], a.v[j]);
		swap(b.v[i1], b.v[j]);

		// Scale row j to have a unit diagonal
		if (a.v[j].n[j]==0.)
			V_ERROR("mat3::inverse: singular matrix; can't invert\n")
		b.v[j] /= a.v[j].n[j];
		a.v[j] /= a.v[j].n[j];

		// Eliminate off-diagonal elems in col j of a, doing identical ops to b
		for (i=0; i<3; i++)
			if (i!=j) {
				b.v[i] -= a.v[i].n[j]*b.v[j];
				a.v[i] -= a.v[i].n[j]*a.v[j];
			}
	}
	return b;
}

// FRIENDS

mat3 operator - (const mat3& a)
{ return mat3(-a.v[0], -a.v[1], -a.v[2]); }

mat3 operator + (const mat3& a, const mat3& b)
{ return mat3(a.v[0] + b.v[0], a.v[1] + b.v[1], a.v[2] + b.v[2]); }

mat3 operator - (const mat3& a, const mat3& b)
{ return mat3(a.v[0] - b.v[0], a.v[1] - b.v[1], a.v[2] - b.v[2]); }

mat3 operator * (mat3& a, mat3& b) {
    #define ROWCOL(i, j) \
    a.v[i].n[0]*b.v[0][j] + a.v[i].n[1]*b.v[1][j] + a.v[i].n[2]*b.v[2][j]
    return mat3(vec3(ROWCOL(0,0), ROWCOL(0,1), ROWCOL(0,2)),
		vec3(ROWCOL(1,0), ROWCOL(1,1), ROWCOL(1,2)),
		vec3(ROWCOL(2,0), ROWCOL(2,1), ROWCOL(2,2)));
    #undef ROWCOL
}

mat3 operator * (const mat3& a, const float d)
{ return mat3(a.v[0] * d, a.v[1] * d, a.v[2] * d); }

mat3 operator * (const float d, const mat3& a)
{ return a*d; }

mat3 operator / (const mat3& a, const float d)
{ return mat3(a.v[0] / d, a.v[1] / d, a.v[2] / d); }

int operator == (const mat3& a, const mat3& b)
{ return (a.v[0] == b.v[0]) && (a.v[1] == b.v[1]) && (a.v[2] == b.v[2]); }

int operator != (const mat3& a, const mat3& b)
{ return !(a == b); }


void swap(mat3& a, mat3& b)
{ mat3 tmp(a); a = b; b = tmp; }


/****************************************************************
*																*
*		    mat4 member functions								*
*																*
****************************************************************/

// CONSTRUCTORS

mat4::mat4() { identity(); }

mat4::mat4(const vec4& v0, const vec4& v1, const vec4& v2, const vec4& v3)
{ v[0] = v0; v[1] = v1; v[2] = v2; v[3] = v3; }

mat4::mat4(const float d)
{ v[0] = v[1] = v[2] = v[3] = vec4(d); }

mat4::mat4(const mat4& m)
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3]; }


// ASSIGNMENT OPERATORS

mat4& mat4::operator = ( const mat4& m )
{ v[0] = m.v[0]; v[1] = m.v[1]; v[2] = m.v[2]; v[3] = m.v[3];
return *this; }

mat4& mat4::operator += ( const mat4& m )
{ v[0] += m.v[0]; v[1] += m.v[1]; v[2] += m.v[2]; v[3] += m.v[3];
return *this; }

mat4& mat4::operator -= ( const mat4& m )
{ v[0] -= m.v[0]; v[1] -= m.v[1]; v[2] -= m.v[2]; v[3] -= m.v[3];
return *this; }

mat4& mat4::operator *= ( const float d )
{ v[0] *= d; v[1] *= d; v[2] *= d; v[3] *= d; return *this; }

mat4& mat4::operator /= ( const float d )
{ v[0] /= d; v[1] /= d; v[2] /= d; v[3] /= d; return *this; }

vec4& mat4::operator [] ( int i) {
    if (i < 0 || i > 3)
	V_ERROR("mat4 [] operator: illegal access; index = " << i << '\n')
    return v[i];
}

// SPECIAL FUNCTIONS;

mat4& mat4::identity() {
	v[0]	= vec4(1.0f, 0.0, 0.0, 0.0);
	v[1]	= vec4(0.0, 1.0f, 0.0, 0.0);
	v[2]	= vec4(0.0, 0.0, 1.0f, 0.0);
	v[3]	= vec4(0.0, 0.0, 0.0, 1.0f);
	return *this;
}
mat4 mat4::transpose() {
    return mat4(vec4(v[0][0], v[1][0], v[2][0], v[3][0]),
		vec4(v[0][1], v[1][1], v[2][1], v[3][1]),
		vec4(v[0][2], v[1][2], v[2][2], v[3][2]),
		vec4(v[0][3], v[1][3], v[2][3], v[3][3]));
}

mat4 mat4::inverse()	    // Gauss-Jordan elimination with partial pivoting
{
	mat4 a(*this),	    // As a evolves from original mat into identity
		b;   // b evolves from identity into inverse(a)
	int i, j, i1;

	// Loop over cols of a from left to right, eliminating above and below diag
	for (j=0; j<4; j++) {   // Find largest pivot in column j among rows j..3
		i1 = j;		    // Row with largest pivot candidate
		for (i=j+1; i<4; i++)
			if (fabs(a.v[i].n[j]) > fabs(a.v[i1].n[j]))
				i1 = i;

		// Swap rows i1 and j in a and b to put pivot on diagonal
		swap(a.v[i1], a.v[j]);
		swap(b.v[i1], b.v[j]);

		// Scale row j to have a unit diagonal
		if (a.v[j].n[j]==0.)
			V_ERROR("mat4::inverse: singular matrix; can't invert\n");
		b.v[j] /= a.v[j].n[j];
		a.v[j] /= a.v[j].n[j];

		// Eliminate off-diagonal elems in col j of a, doing identical ops to b
		for (i=0; i<4; i++)
			if (i!=j) {
				b.v[i] -= a.v[i].n[j]*b.v[j];
				a.v[i] -= a.v[i].n[j]*a.v[j];
			}
	}
	return b;
}

// FRIENDS

mat4 operator - (const mat4& a)
{ return mat4(-a.v[0], -a.v[1], -a.v[2], -a.v[3]); }

mat4 operator + (const mat4& a, const mat4& b)
{ return mat4(a.v[0] + b.v[0], a.v[1] + b.v[1], a.v[2] + b.v[2],
  a.v[3] + b.v[3]);
}

mat4 operator - (const mat4& a, const mat4& b)
{ return mat4(a.v[0] - b.v[0], a.v[1] - b.v[1], a.v[2] - b.v[2], a.v[3] - b.v[3]); }

mat4 operator * (mat4& a, mat4& b) {
    #define ROWCOL(i, j) a.v[i].n[0]*b.v[0][j] + a.v[i].n[1]*b.v[1][j] + \
    a.v[i].n[2]*b.v[2][j] + a.v[i].n[3]*b.v[3][j]
    return mat4(
    vec4(ROWCOL(0,0), ROWCOL(0,1), ROWCOL(0,2), ROWCOL(0,3)),
    vec4(ROWCOL(1,0), ROWCOL(1,1), ROWCOL(1,2), ROWCOL(1,3)),
    vec4(ROWCOL(2,0), ROWCOL(2,1), ROWCOL(2,2), ROWCOL(2,3)),
    vec4(ROWCOL(3,0), ROWCOL(3,1), ROWCOL(3,2), ROWCOL(3,3))
    );
}

mat4 operator * (const mat4& a, const float d)
{ return mat4(a.v[0] * d, a.v[1] * d, a.v[2] * d, a.v[3] * d); }

mat4 operator * (const float d, const mat4& a)
{ return a*d; }

mat4 operator / (const mat4& a, const float d)
{ return mat4(a.v[0] / d, a.v[1] / d, a.v[2] / d, a.v[3] / d); }

int operator == (const mat4& a, const mat4& b)
{ return ((a.v[0] == b.v[0]) && (a.v[1] == b.v[1]) && (a.v[2] == b.v[2]) &&
  (a.v[3] == b.v[3])); }

int operator != (const mat4& a, const mat4& b)
{ return !(a == b); }

void swap(mat4& a, mat4& b)
{ mat4 tmp(a); a = b; b = tmp; }


/****************************************************************
*																*
*	       2D functions and 3D functions						*
*																*
****************************************************************/

mat3 identity2D()
{   return mat3(vec3(1.0f, 0.0, 0.0),
		vec3(0.0, 1.0f, 0.0),
		vec3(0.0, 0.0, 1.0f)); }

mat3 translation2D(vec2& v)
{   return mat3(vec3(1.0f, 0.0, v.x),
		vec3(0.0, 1.0f, v.y),
		vec3(0.0, 0.0, 1.0f)); }

mat3 rotation2D(vec2& Center, const float angleDeg) {
    float  angleRad = angleDeg * M_PI / 180.0f,
	    c = cos(angleRad),
	    s = sin(angleRad);

    return mat3(vec3(c, -s, Center.x * (1.0f-c) + Center.y * s),
		vec3(s, c, Center.y * (1.0f-c) - Center.x * s),
		vec3(0.0, 0.0, 1.0f));
}

mat3 scaling2D(vec2& scaleVector)
{   return mat3(vec3(scaleVector.x, 0.0, 0.0),
		vec3(0.0, scaleVector.y, 0.0),
		vec3(0.0, 0.0, 1.0f)); }

mat4 identity3D()
{   return mat4(vec4(1.0f, 0.0, 0.0, 0.0),
		vec4(0.0, 1.0f, 0.0, 0.0),
		vec4(0.0, 0.0, 1.0f, 0.0),
		vec4(0.0, 0.0, 0.0, 1.0f)); }

mat4 translation3D(vec3& v)
{   return mat4(vec4(1.0f, 0.0, 0.0, v.x),
		vec4(0.0, 1.0f, 0.0, v.y),
		vec4(0.0, 0.0, 1.0f, v.z),
		vec4(0.0, 0.0, 0.0, 1.0f)); }

mat4 rotation3D(vec3& Axis, const float angleDeg) {
    float  angleRad = angleDeg * M_PI / 180.0f,
	    c = cos(angleRad),
	    s = sin(angleRad),
	    t = 1.0f - c;

    Axis.normalize();
    return mat4(vec4(t * Axis.x * Axis.x + c,
		     t * Axis.x * Axis.y - s * Axis.z,
		     t * Axis.x * Axis.z + s * Axis.y,
		     0.0f),
		vec4(t * Axis.x * Axis.y + s * Axis.z,
		     t * Axis.y * Axis.y + c,
		     t * Axis.y * Axis.z - s * Axis.x,
		     0.0f),
		vec4(t * Axis.x * Axis.z - s * Axis.y,
		     t * Axis.y * Axis.z + s * Axis.x,
		     t * Axis.z * Axis.z + c,
		     0.0f),
		vec4(0.0f, 0.0f, 0.0f, 1.0f));
}

mat4 scaling3D(vec3& scaleVector)
{   return mat4(vec4(scaleVector.x, 0.0, 0.0, 0.0),
		vec4(0.0, scaleVector.y, 0.0, 0.0),
		vec4(0.0, 0.0, scaleVector.z, 0.0),
		vec4(0.0, 0.0, 0.0, 1.0f)); }

mat4 perspective3D(const float d)
{   return mat4(vec4(1.0f, 0.0, 0.0, 0.0),
		vec4(0.0, 1.0f, 0.0, 0.0),
		vec4(0.0, 0.0, 1.0f, 0.0),
		vec4(0.0, 0.0, 1.0f/d, 0.0)); }

