/****************************************************************
*																*
* C++ Vector and Matrix Algebra routines						*
* Author: Jean-Francois DOUE									*
* Version 3.1 --- October 1993									*
*---------------------------------------------------------------*
* Modified and updated --- May 2005 by Wael El Oraiby			*
****************************************************************/
#pragma once

#include <iostream>
#include <cmath>

#ifndef M_PI
	#define M_PI 3.14159265358979323846f
#endif

// min-max macros
#ifndef MIN
	#define MIN(A,B) ((A) < (B) ? (A) : (B))
#endif

#ifndef MAX
	#define MAX(A,B) ((A) > (B) ? (A) : (B))
#endif

#ifdef min
	#undef min
#endif

#ifdef max
	#undef max
#endif

struct vec2;
struct vec3;
struct vec4;
struct mat3;
struct mat4;

/****************************************************************
*																*
*			    2D Vector										*
*																*
****************************************************************/

struct vec2
{
	union {
		struct {
            float n[2];
		};

		struct {
			float u, v;
		};

		struct {
			float x, y;
		};
	};



	// Constructors

	vec2();
	vec2(const float _x, const float _y);
	vec2(const float d);
	vec2(const vec2& v);			// copy constructor
	vec2(const vec3& v);			// cast v3 to v2
	vec2(const vec3& v, int dropAxis);	// cast v3 to v2

	// Assignment operators

	vec2& operator	= ( const vec2& v );	// assignment of a vec2
	vec2& operator += ( const vec2& v );	// incrementation by a vec2
	vec2& operator -= ( const vec2& v );	// decrementation by a vec2
	vec2& operator *= ( const float d );	// multiplication by a constant
	vec2& operator /= ( const float d );	// division by a constant
	float& operator [] ( int i);		// indexing

	// special functions

	float length() const;			// length of a vec2
	float length2() const;			// squared length of a vec2
	vec2& normalize();			// normalize a vec2
};

vec2 operator - (const vec2& v);			    // -v1
vec2 operator + (const vec2& a, const vec2& b);	    // v1 + v2
vec2 operator - (const vec2& a, const vec2& b);	    // v1 - v2
vec2 operator * (const vec2& a, const float d);	    // v1 * 3.0
vec2 operator * (const float d, const vec2& a);	    // 3.0 * v1
vec2 operator * (const mat3& a, const vec2& v);	    // M . v
vec2 operator * (const vec2& v, mat3& a);	    // v . M
float operator * (const vec2& a, const vec2& b);    // dot product
vec2 operator / (const vec2& a, const float d);	    // v1 / 3.0
vec3 operator ^ (const vec2& a, const vec2& b);	    // cross product
int operator == (const vec2& a, const vec2& b);	    // v1 == v2 ?
int operator != (const vec2& a, const vec2& b);	    // v1 != v2 ?
void swap(vec2& a, vec2& b);			    // swap v1 & v2
vec2 min(const vec2& a, const vec2& b);		    // min(v1, v2)
vec2 max(const vec2& a, const vec2& b);		    // max(v1, v2)
vec2 prod(const vec2& a, const vec2& b);		    // term by term *
/****************************************************************
*																*
*			    3D Vector										*
*																*
****************************************************************/

struct vec3
{
	union {
		struct {
            float n[3];
		};

		struct {
			float x, y, z;
		};
	};



	// Constructors

	vec3();
	vec3(const float _x, const float _y, const float _z);
	vec3(const float d);
	vec3(const vec3& v);			    // copy constructor
	vec3(const vec2& v);			    // cast v2 to v3
	vec3(const vec2& v, float d);		    // cast v2 to v3
	vec3(const vec4& v);			    // cast v4 to v3
	vec3(const vec4& v, int dropAxis);	    // cast v4 to v3

	// Assignment operators

	vec3& operator	= ( const vec3& v );	    // assignment of a vec3
	vec3& operator += ( const vec3& v );	    // incrementation by a vec3
	vec3& operator -= ( const vec3& v );	    // decrementation by a vec3
	vec3& operator *= ( const float d );	    // multiplication by a constant
	vec3& operator /= ( const float d );	    // division by a constant
	float& operator [] ( int i);		    // indexing

	// special functions

	float length() const;			    // length of a vec3
	float length2() const;			    // squared length of a vec3
	vec3& normalize();			    // normalize a vec3
};

vec3 operator - (const vec3& v);			    // -v1
vec3 operator + (const vec3& a, const vec3& b);	    // v1 + v2
vec3 operator - (const vec3& a, const vec3& b);	    // v1 - v2
vec3 operator * (const vec3& a, const float d);	    // v1 * 3.0
vec3 operator * (const float d, const vec3& a);	    // 3.0 * v1
vec3 operator * (const mat4& a, const vec3& v);	    // M . v
vec3 operator * (const vec3& v, mat4& a);	    // v . M
float operator * (const vec3& a, const vec3& b);    // dot product
vec3 operator / (const vec3& a, const float d);	    // v1 / 3.0
vec3 operator ^ (const vec3& a, const vec3& b);	    // cross product
void swap(vec3& a, vec3& b);			    // swap v1 & v2
vec3 min(const vec3& a, const vec3& b);		    // min(v1, v2)
vec3 max(const vec3& a, const vec3& b);		    // max(v1, v2)
vec3 prod(const vec3& a, const vec3& b);		    // term by term *
vec2 operator * (const mat3& a, const vec2& v);	    // linear transform
mat3 operator * (mat3& a, mat3& b);		    // matrix 3 product

/****************************************************************
*																*
*			    4D Vector										*
*																*
****************************************************************/

struct vec4
{
	union {
		struct {
            float n[4];
		};

		struct {
			float x, y, z, w;
		};
	};


	// Constructors

	vec4();
	vec4(const float _x, const float _y, const float _z, const float _w);
	vec4(const float d);
	vec4(const vec4& v);			    // copy constructor
	vec4(const vec3& v);			    // cast vec3 to vec4
	vec4(const vec3& v, const float d);	    // cast vec3 to vec4

	// Assignment operators

	vec4& operator	= ( const vec4& v );	    // assignment of a vec4
	vec4& operator += ( const vec4& v );	    // incrementation by a vec4
	vec4& operator -= ( const vec4& v );	    // decrementation by a vec4
	vec4& operator *= ( const float d );	    // multiplication by a constant
	vec4& operator /= ( const float d );	    // division by a constant
	float& operator [] ( int i);		    // indexing

	// special functions

	float length() const;			    // length of a vec4
	float length2() const;			    // squared length of a vec4
	vec4& normalize();			    // normalize a vec4
};

vec4 operator - (const vec4& v);			    // -v1
vec4 operator + (const vec4& a, const vec4& b);	    // v1 + v2
vec4 operator - (const vec4& a, const vec4& b);	    // v1 - v2
vec4 operator * (const vec4& a, const float d);	    // v1 * 3.0
vec4 operator * (const float d, const vec4& a);	    // 3.0 * v1
vec4 operator * (const mat4& a, const vec4& v);	    // M . v
vec4 operator * (const vec4& v, mat4& a);	    // v . M
float operator * (const vec4& a, const vec4& b);    // dot product
vec4 operator / (const vec4& a, const float d);	    // v1 / 3.0
int operator == (const vec4& a, const vec4& b);	    // v1 == v2 ?
int operator != (const vec4& a, const vec4& b);	    // v1 != v2 ?
void swap(vec4& a, vec4& b);			    // swap v1 & v2
vec4 min(const vec4& a, const vec4& b);		    // min(v1, v2)
vec4 max(const vec4& a, const vec4& b);		    // max(v1, v2)
vec4 prod(const vec4& a, const vec4& b);		    // term by term *
vec3 operator * (const mat4& a, const vec3& v);	    // linear transform
mat4 operator * (mat4& a, mat4& b);		    // matrix 4 product
/****************************************************************
*																*
*			   3x3 Matrix										*
*																*
****************************************************************/

struct mat3
{
	union {
		struct {
			vec3 v[3];
		};

		struct {
			float m[9];
		};
	};

	// Constructors

	mat3();
	mat3(const vec3& v0, const vec3& v1, const vec3& v2);
	mat3(const float d);
	mat3(const mat3& m);

	// Assignment operators

	mat3& operator	= ( const mat3& m );	    // assignment of a mat3
	mat3& operator += ( const mat3& m );	    // incrementation by a mat3
	mat3& operator -= ( const mat3& m );	    // decrementation by a mat3
	mat3& operator *= ( const float d );	    // multiplication by a constant
	mat3& operator /= ( const float d );	    // division by a constant
	vec3& operator [] ( int i);		    // indexing

	// special functions
	mat3& identity();				// identity identity
	mat3 transpose();			    // transpose
	mat3 inverse();				    // inverse

	// friends

	mat3 operator - (const mat3& a);			    // -m1

};

mat3 operator + (const mat3& a, const mat3& b);	    // m1 + m2
mat3 operator - (const mat3& a, const mat3& b);	    // m1 - m2
mat3 operator * (mat3& a, mat3& b);		    // m1 * m2
mat3 operator * (const mat3& a, const float d);	    // m1 * 3.0
mat3 operator * (const float d, const mat3& a);	    // 3.0 * m1
mat3 operator / (const mat3& a, const float d);	    // m1 / 3.0
int operator == (const mat3& a, const mat3& b);	    // m1 == m2 ?
int operator != (const mat3& a, const mat3& b);	    // m1 != m2 ?
void swap(mat3& a, mat3& b);			    // swap m1 & m2

// necessary declarations

vec3 operator * (const mat3& a, const vec3& v);	    // linear transform
vec2 operator * (const mat3& a, const vec2& v);	    // linear transform
/****************************************************************
*																*
*			   4x4 Matrix										*
*																*
****************************************************************/

struct mat4
{
	union {
		struct {
            vec4 v[4];
		};

		struct {
			float m[16];
		};
	};


	// Constructors

	mat4();
	mat4(const vec4& v0, const vec4& v1, const vec4& v2, const vec4& v3);
	mat4(const float d);
	mat4(const mat4& m);

	// Assignment operators

	mat4& operator	= ( const mat4& m );	    // assignment of a mat4
	mat4& operator += ( const mat4& m );	    // incrementation by a mat4
	mat4& operator -= ( const mat4& m );	    // decrementation by a mat4
	mat4& operator *= ( const float d );	    // multiplication by a constant
	mat4& operator /= ( const float d );	    // division by a constant
	vec4& operator [] ( int i);		    // indexing

	// special functions

	mat4& identity();
	mat4 transpose();			    // transpose
	mat4 inverse();				    // inverse
};

mat4 operator - (const mat4& a);			    // -m1
mat4 operator + (const mat4& a, const mat4& b);	    // m1 + m2
mat4 operator - (const mat4& a, const mat4& b);	    // m1 - m2
mat4 operator * (mat4& a, mat4& b);		    // m1 * m2
mat4 operator * (const mat4& a, const float d);	    // m1 * 4.0
mat4 operator * (const float d, const mat4& a);	    // 4.0 * m1
mat4 operator / (const mat4& a, const float d);	    // m1 / 3.0
int operator == (const mat4& a, const mat4& b);	    // m1 == m2 ?
int operator != (const mat4& a, const mat4& b);	    // m1 != m2 ?
void swap(mat4& a, mat4& b);			    // swap m1 & m2

// necessary declarations

vec4 operator * (const mat4& a, const vec4& v);	    // linear transform
vec3 operator * (const mat4& a, const vec3& v);	    // linear transform

/****************************************************************
*																*
*	       2D functions and 3D functions						*
*																*
****************************************************************/

mat3 identity2D();					    // identity 2D
mat3 translation2D(vec2& v);				    // translation 2D
mat3 rotation2D(vec2& Center, const float angleDeg);	    // rotation 2D
mat3 scaling2D(vec2& scaleVector);			    // scaling 2D
mat4 identity3D();					    // identity 3D
mat4 translation3D(vec3& v);				    // translation 3D
mat4 rotation3D(vec3& Axis, const float angleDeg);	    // rotation 3D
mat4 scaling3D(vec3& scaleVector);			    // scaling 3D
mat4 perspective3D(const float d);			    // perspective 3D
