/*
** font.cpp
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** basic font class implementation: loading and rendering code
**
** Written for More OpenGL Game Programming
*/

#include "GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>

#include <iostream>
#include <string>
#include <sstream>

#include "font.h"
#include "MGLTypes.h"
#include "MGLTargaImage.h"

using namespace std;
using namespace MGL;

// these are the global variables saved when we are in text rendering
static float modelviewMatrix[16];
static float projectionMatrix[16];

/*
** class constructor: load font texture and setup the font
*/
Font::Font(const char *fontName)
{
	TargaImage	fontImg;

	// load the font
	if ( !fontImg.Load(fontName) )
		throw "Couldn't read the font...";

	// font image must be squared
	if ( fontImg.GetHeight() != fontImg.GetWidth() )
		throw "Fonts should be of the same height and width...";

	// flip the targa image
	fontImg.FlipVertical();

	// setup the opengl texture
	glGenTextures(1,&texId);
	glBindTexture(GL_TEXTURE_2D,texId);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	glTexImage2D(GL_TEXTURE_2D,	0, GL_RGBA,	fontImg.GetWidth(),	fontImg.GetHeight(), 0,	GL_RGBA, GL_UNSIGNED_BYTE, fontImg.GetImage());
	
	// now compute char widths
	float	divisions	= fontImg.GetWidth();
	divisions	/= 16.0f;
	
	computeCharWidth(static_cast<int>(divisions), fontImg.GetImage());

	// setup rendering cursor
	setCursor(0.0f, charHeight);

	// and set the default font drawing size to char height
	drawSize	= charHeight;
}

/*
** class destructor
*/
Font::~Font() {
	// delete the opengl texture
	glDeleteTextures(1,&texId);
}

/*
** set the text cursor
*/
void Font::setCursor(float x, float y)
{
	cursorX	= x;
	cursorY	= y;

	// load the identity matrix
	glLoadIdentity();

	// and translate to the cursor position
	glTranslatef(cursorX, cursorY, 0.0f);
}

/*
** compute each character width:
**	after loading the texture, we loop through it computing the width of each
**	character (saving it). There is a special case: the "i" char case as it's
**	preceeded with "j" char. "j" char is kind of special as it's displaced to
**	the "i" char zone.
*/
void Font::computeCharWidth(int size, unsigned char *data)
{
	charHeight	= static_cast<float>(size);

	// loop through all 256 character
	for ( int c = 0; c < 256; c++ )
	{
		charWidth[c]	= 0.0f;
		
		// now see where each character ends
		int		startX	= ((c & 0x0F) * size );
		for ( int x = startX; x < (startX + size /* need fixing "j" problem */ ); x++ ) {
			int		startY	= ((c >> 4) * size);
			for ( int y = startY; y < (startY + size);  y++ ) {
				if ( data[(x + y * size * 16) * 4 + 3] >= 64 )
					if ( charWidth[c] < (x - startX) )
						charWidth[c]	= static_cast<float>(x - startX);
			}
		}
	}
}

/*
** setup text drawing mode
*/
void textBegin(int width, int height)
{
	// get both projection and model view matrices
	glGetFloatv(GL_PROJECTION_MATRIX, projectionMatrix);
	glGetFloatv(GL_MODELVIEW_MATRIX, modelviewMatrix);

	// setup the new text matrices
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// orthogonal projection should do
	glOrtho(0.0f, width, height, 0.0f, -1.0f, 1.0f);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// disable depth testing
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	
	// must enable blending
	glEnable(GL_BLEND);

	// enable texture use
	glEnable(GL_TEXTURE_2D);
}

/*
** finish text mode
*/
void textEnd()
{
	// disable texture, blending
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	// renable depth tests
	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);

	// restore matrices
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(projectionMatrix);
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(modelviewMatrix);
}

/*
** draw a char
*/
void Font::drawChar(unsigned char ch)
{
	float	divisions	= charHeight;
	float	ratio		= drawSize / charHeight;

	// setup char texture coordinates
	float su	= (static_cast<float>( ch & 0x0F) ) * divisions / ( charHeight * 16.0f );
	float sv	= (static_cast<float>( ch >> 4) * divisions) / ( charHeight * 16.0f);
	float eu	= su + static_cast<float>(charWidth[ch] + 1) / (charHeight * 16.0f);
	float ev	= sv + divisions / (charHeight * 16.0f);

	// draw the char as QUAD
	glBegin(GL_QUADS); {
		glTexCoord2f(su, sv);
		glVertex2f(0.0f, 0.0f);
		glTexCoord2f(su, ev);
		glVertex2f(0.0f, charHeight * ratio);
		glTexCoord2f(eu, ev);
		glVertex2f(charWidth[ch] * ratio, charHeight * ratio);
		glTexCoord2f(eu, sv);
		glVertex2f(charWidth[ch] * ratio, 0.0f);
	} glEnd();
}

/*
** draw a text string
*/
void Font::drawText(const char *str)
{
	// assign the font texture
	glBindTexture(GL_TEXTURE_2D,texId);

	// setup the proper blending function
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);

	// setup the cursor position
	setCursor(cursorX, cursorY);

	char *s = const_cast<char*>(str);
	float	stringWidth	= 0.0f;
	float	lines = 0;
	float	ratio = drawSize / charHeight;

	// draw the string char by char
	for ( ; *s; s++ ) {
		if ( *s == '\n' ) {	// return character
			lines++;
			setCursor(cursorX, cursorY);
			glTranslatef(0.0f, charHeight * lines * ratio, 0.0f);
			stringWidth = 0.0f;
		} else if ( *s == '\r' ) {	// return to line
			lines++;
			glTranslatef(0.0f, charHeight * ratio, 0.0f);
		} else if ( *s == '\t' ) {	// tab char
			glTranslatef(charHeight * ratio * 4.0f, 0.0f, 0.0f);
			stringWidth	+= charHeight * ratio * 4.0f;
		} else if ( *s == ' ' ) {	// space
			stringWidth += charHeight * ratio / 2.0f;
			glTranslatef(charHeight * ratio / 2.0f, 0.0f, 0.0f);
		} else {			// any other char
			drawChar(*s);
			// set the position for the next char
			glTranslatef(charWidth[*s] * ratio, 0.0f, 0.0f);
			stringWidth += charWidth[*s] * ratio;
		}
	}
}
