/*
** frustum.cpp
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** frustum class
**
** Written for More OpenGL Game Programming
*/

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#include "algebra3.h"
#include "geom.h"

#include "frustum.h"

/*
** extract the planes from the opengl projection and modelview matrix, the way
** we do it is found on OpenGL.org (viewcull.c),a nd is explained in the following:
**
** Calculates the six view volume planes in object coordinate (OC) space.
**
**       Algorithm:
**
**       A view volume plane in OC is transformed into CC by multiplying it by
**       the inverse of the combined ModelView and Projection matrix (M).
**       Algebraically, this is written:
**              -1
**         P   M   = P
**          oc        cc
**       
**       The resulting six view volume planes in CC are:
**         [ -1  0  0  1 ]
**         [  1  0  0  1 ]
**         [  0 -1  0  1 ]
**         [  0  1  0  1 ]
**         [  0  0 -1  1 ]
**         [  0  0  1  1 ]
**       
**       To transform the CC view volume planes into OC, we simply multiply
**       the CC plane equations by the combined ModelView and Projection matrices
**       using standard vector-matrix multiplication. Algebraically, this is written:  
**         P   M = P
**          cc      oc
**       
**       Since all of the CC plane equation components are 0, 1, or -1, full vector-
**       matrix multiplication is overkill. For example, the first element of the
**       first OC plane equation is computed as:
**         A = -1 * m0 + 0 * m1 + 0 * m2 + 1 * m3
**       This simplifies to:
**         A = m3 - m0
**       
**       Other terms simpliofy similarly. In fact, all six plane equations can be
**       computed as follows:
**         [ m3-m0  m7-m4  m11-m8  m15-m12 ]
**         [ m3+m0  m7+m4  m11+m8  m15+m12 ]
**         [ m3-m1  m7-m5  m11-m9  m15-m13 ]
**         [ m3+m1  m7+m5  m11+m9  m15+m13 ]
**         [ m3-m2  m7-m6  m11-m10 m15-m14 ]
**         [ m3+m2  m7+m6  m11+m10 m15+m14 ]
*/
void Frustum::setupFromGL()
{
	mat4		projMatrix, viewMatrix, finalMatrix;

	// get the matrices from OpenGL
	glGetFloatv(GL_MODELVIEW_MATRIX, viewMatrix.m);
	glGetFloatv(GL_PROJECTION_MATRIX, projMatrix.m);

	// the final matrix
	finalMatrix	= projMatrix * viewMatrix;
	float	*m	= finalMatrix.m;

	// now get the planes
	planes[0].a	= m[3] - m[0];
	planes[0].b	= m[7] - m[4];
	planes[0].c	= m[11] - m[8];
	planes[0].d	= m[15] - m[12];

	planes[1].a	= m[3] + m[0];
	planes[1].b	= m[7] + m[4];
	planes[1].c	= m[11] + m[8];
	planes[1].d	= m[15] + m[12];

	planes[2].a	= m[3] - m[1];
	planes[2].b	= m[7] - m[5];
	planes[2].c	= m[11] - m[9];
	planes[2].d	= m[15] - m[13];

	planes[3].a	= m[3] + m[1];
	planes[3].b	= m[7] + m[5];
	planes[3].c	= m[11] + m[9];
	planes[3].d	= m[15] + m[13];

	planes[4].a	= m[3] - m[2];
	planes[4].b	= m[7] - m[6];
	planes[4].c	= m[11] - m[10];
	planes[4].d	= m[15] - m[14];

	planes[5].a	= m[3] + m[2];
	planes[5].b	= m[7] + m[6];
	planes[5].c	= m[11] + m[10];
	planes[5].d	= m[15] + m[14];

	// normalize the planes
	for ( int i = 0; i < 6; i++ )
		planes[i].normalize();
}

/*
** check if a sphere is inside the frustum
*/
bool Frustum::isInside(const vec3 &center, float radius)
{
	for( int i = 0; i < 6; i++ ) {
		if ( planes[i].distance(center) < -radius )
			return false;
	}
	return true;
}

/*
** check the bounding box
*/
bool Frustum::isInside(const vec3 &min, const vec3 &max)
{
	for ( int i = 0; i < 6; i++ ) {
		if ( planes[i].signedDistance(vec3(min.x, min.y, min.z)) < 0.0f )
			return false;

		if ( planes[i].signedDistance(vec3(min.x, min.y, max.z)) < 0.0f )
			return false;

		if ( planes[i].signedDistance(vec3(min.x, max.y, min.z)) < 0.0f )
			return false;

		if ( planes[i].signedDistance(vec3(min.x, max.y, max.z)) < 0.0f )
			return false;

		if ( planes[i].signedDistance(vec3(max.x, min.y, min.z)) < 0.0f )
			return false;

		if ( planes[i].signedDistance(vec3(max.x, min.y, max.z)) < 0.0f )
			return false;

		if ( planes[i].signedDistance(vec3(max.x, max.y, min.z)) < 0.0f )
			return false;

		if ( planes[i].signedDistance(vec3(max.x, max.y, max.z)) < 0.0f )
			return false;
	}

	return true;
}
