/*
** geom.cpp
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** geometry structures
**
** Written for More OpenGL Game Programming
*/

#include "algebra3.h"
#include "geom.h"

using namespace std;

/******************************************************************************
** plane class member functions
******************************************************************************/
/*
** Name: plane
** Desc: plane constructor
*/
plane::plane(const vec3 &v0, const vec3 &v1, const vec3 &v2)
{
	vec3	e0	= v1 - v0;
	vec3	e1	= v2 - v0;//v2 - v1;

	normal	= e0 ^ e1;
	normal.normalize();

	dist	= -v0 * normal;
}

/*
** Name: normalize
** Desc: normalize plane
*/
void plane::normalize()
{
	float	len	= normal.length();
	normal.normalize();

	dist	/= len;
}

/*
** Name: signedDistance
** Desc: return the signed distance from a point to the plane
**
** NOTE: assuming plane is normalized
*/
float plane::signedDistance(const vec3 &v) const
{
//	normalize();
	return ((normal * v) + dist);
}

/*
** Name: distance
** Desc: return the distance from a point to the plane
*/
float plane::distance(const vec3 &v) const
{
	return static_cast<float>(fabs(signedDistance(v)));
}

/*
** Name: intersection
** Desc: retrive ray plane intersection
*/
float plane::intersection(vec3 &int_point, const vec3 &start, const vec3 &end) const
{

	float	t, numer, denom;
	vec3	direction	= end - start;

	numer	= (normal * start) + dist;
	denom	= (normal * direction);

	if( denom == 0.0f )
	{
		int_point	= start;
		return -10000.0f;
	}

	t	= - numer / denom;

	int_point	= start + (t * direction);

	return t;
}

float	planeThreshold	= 0.001f;

/*
** Name: less_plane
** Desc: operator( )
*/
bool lessplane::operator()( const plane &p0, const plane &p1 ) const
{
	if( p0.a != p1.a )
		return (p0.a < p1.a);
	if( p0.b != p1.b )
		return (p0.b < p1.b);
	if( p0.c != p1.c )
		return (p0.c < p1.c);
	if( p0.d != p1.d )
		return (p0.d < p1.d);

	return false;
}

/*
** Name: setup
** Desc: setup the bounding volume
*/
void BoundingVolume::setup(int numVerts, vec3 *verts)
{
	maxV	= vec3(-99999.9f,-99999.9f,-99999.9f);
	minV	= vec3(99999.9f,99999.9f,99999.9f);
	center	= vec3(0.0f, 0.0f, 0.0f);
	radius	= 0.0f;
	radius2	= 0.0f;

	for ( int i = 0; i < numVerts; i++ ) {
		if (verts[i].x < minV.x)
			minV.x	= verts[i].x;
		if (verts[i].y < minV.y)
			minV.y	= verts[i].y;
		if (verts[i].z < minV.z)
			minV.z	= verts[i].z;

		if (verts[i].x > maxV.x)
			maxV.x	= verts[i].x;
		if (verts[i].y > maxV.y)
			maxV.y	= verts[i].y;
		if (verts[i].z > maxV.z)
			maxV.z	= verts[i].z;
	}

	// compute center
	center = (minV + maxV) / 2.0f;

	// and recompute bounding volume
	maxV -= center;
	minV -= center;

	for ( int i = 0; i < numVerts; i++ ) {
		float	length2	= (center - verts[i]).length2();

		if (radius2 < length2)
			radius2	= length2;
	}

	radius	= sqrtf(radius2);
}