/*
** map_bsp.cpp
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** bsp builder
**
** Written for More OpenGL Game Programming
*/

#include "GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>

#include <iostream>
#include <vector>
#include <map>
#include <string>

#include "algebra3.h"
#include "geom.h"
#include "scene_map.h"

using namespace std;
using namespace SceneMap;

float Map::epsilon	= 1.0f/32.0f;

/*
** classify a polygon to a plane
*/
Map::FACE_CLASS Map::classifyPolyToPlane(plane &cmpPlane, AreaPolygon &poly)
{
	int num = static_cast<int>(poly.verts.size());
	int front = 0, back = 0, on = 0;
	float res = 0;

	for (int cnt = 0; cnt < num; cnt++)
	{
		// compute point distance to plane
		res = (cmpPlane.normal * poly.verts[cnt]) + cmpPlane.dist;

		if (res > epsilon)	// point in front of the plane
			front++;
		else if (res < -epsilon) // on back of the plane
			back++;
		else				// or on the plane
		{
			front++;
			back++;
			on++;
		}
	}

	if (on == num)
		return FACE_ON;
	else if (front == num)
		return FACE_INFRONT;
	else if (back == num)
		return FACE_ONBACK;

	return FACE_SPANNING;
}

/*
** split a polygon by a plane
*/
void Map::splitPoly(plane &splitter, AreaPolygon &poly, AreaPolygon &front, AreaPolygon &back)
{
	int		count = static_cast<int>(poly.verts.size());
	int		front_c = 0, back_c = 0;
	vec3	ptA, ptB;
	float	sideA, sideB;

	// we start by taking the last point and finding its side
	ptA = poly.verts[count - 1];
	sideA = (splitter.normal * ptA) + splitter.dist;

	front.verts.resize(0);
	back.verts.resize(0);

	for (int i = 0; i < count; i++)
	{
		// then we take the point that follow on the polygon (which is convex)
		ptB = poly.verts[i];
		sideB = (splitter.normal * ptB) + splitter.dist;

		// now start comparing both ends of the edge formed by the 2 points ptA & ptB
		if (sideB > epsilon)	// ptB is in front of the plane
		{
			if (sideA < -epsilon)	// and ptA is on the back of the plane
			{
				vec3 vt = ptB - ptA;
				// split the edge
				float sect = -((splitter.normal * ptA) + splitter.dist) / (splitter.normal * vt);
				// and add the new point to both front and back polygons
				front.verts.push_back(ptA + (sect * vt));
				back.verts.push_back(ptA + (sect * vt));
			}
			// add ptB to the front polygon
			front.verts.push_back(ptB);
		}
		else if (sideB < -epsilon)	// the other case around (ptB is on the back)
		{
			if (sideA > epsilon)	// and ptA is in front
			{
				vec3 vt = ptB - ptA;
				// split and add to both polygons
				float sect = -((splitter.normal * ptA) + splitter.dist) / (splitter.normal * vt);
				front.verts.push_back(ptA + (sect * vt));
				back.verts.push_back(ptA + (sect * vt));
			}
			// and add ptB to the front polygon
			back.verts.push_back(ptB);
		}
		else {	// ptB is on the plane so add to both polygons
			front.verts.push_back(ptB);
			back.verts.push_back(ptB);
		}
		// next point
		ptA = ptB;
		sideA = sideB;
	}

	front.p			= poly.p;
	front.areaId	= poly.areaId;

	back.p			= poly.p;
	back.areaId		= poly.areaId;
}

/*
** find best splitter
*/
int Map::findSplitter(vector<AreaPolygon> &polys, plane &splitter)
{
	AreaPolygon			cmp_face;	// comparaision face
	AreaPolygon			curr;		// current face we are on
	int					best_id	= -1;	// start with an arbitrary plane
	int					best_score	= 999999;
	int					score;//		= 999999;
	int					res, i, j, num_front, num_back, num_spanning, num_on;

	for ( i = 0; i < static_cast<int>(polys.size()); i++ )
	{
		cmp_face	= polys[i];

		num_front	= 0;
		num_back		= 0;
		num_spanning	= 0;
		num_on		= 0;

		// count front, back, spanning and on faces
		for ( j = 0; j < static_cast<int>(polys.size()); j++ )      
		{
			curr	= polys[j];

			if( i != j )
			{
				res = classifyPolyToPlane(cmp_face.p, curr);
				switch( res )
				{
				case FACE_INFRONT:
					num_front++;
					break;
				case FACE_ONBACK:
					num_back++;
					break;
				case FACE_ON:
					num_on++;
					break;
				case FACE_SPANNING:
					num_spanning++;
					break;
				}
			}
		}

		// estimate current plane score
		score = ((2 * num_spanning) + abs(num_front - num_back) + num_on); 

		// update best plane if needed
		if (score != 0 && score < best_score) 
		{
			best_score	= score;
			splitter	= polys[i].p;
			best_id		= i;
		}
	}

	if ( best_id == -1 && polys.size() > 1 )
	{
		splitter	= polys[0].p;
	}

//	cout << "best id: " << best_id << endl;
	return best_id;
}

/*
** build a bsp node
*/
void Map::buildBspNode(vector<AreaPolygon> &polys, int nodeId)
{
	plane			splitter;
	AreaPolygon		front, back;
	int				frontNodeId, backNodeId, i;
	BspNode			n;

	vector<AreaPolygon>	onFaces;
	vector<AreaPolygon>	front_faces;
	vector<AreaPolygon>	back_faces;

	// show progress
	cout << ".";

	// try to find a splitter
	findSplitter(polys, splitter);

	bspNodes[nodeId].splitPlane	= splitter;

	for( i = 0; i < static_cast<int>(polys.size()); i++ )
	{
		switch(classifyPolyToPlane(splitter, polys[i]))
		{
		case FACE_INFRONT:
			front_faces.push_back(polys[i]);
			break;
		case FACE_ONBACK:
			back_faces.push_back(polys[i]);
			break;
		case FACE_ON:
			onFaces.push_back(polys[i]);
			break;
		case FACE_SPANNING:
			splitPoly(splitter, polys[i], front, back);
			front_faces.push_back(front);
			back_faces.push_back(back);
			break;
		}
	}

	if ( front_faces.size() == 0 && back_faces.size() == 0 )
	{
		bspNodes[nodeId].frontNode	= -onFaces[0].areaId;
		bspNodes[nodeId].backNode	= SOLID_AREA;
	}


	if ( front_faces.size() == 0 && back_faces.size() != 0 )
	{
		// we shouldn't get here if everything is normal
		bspNodes[nodeId].frontNode	= -onFaces[0].areaId;
		backNodeId	= static_cast<int>(bspNodes.size());
		bspNodes[nodeId].backNode	= backNodeId;
		bspNodes.push_back(n);

		buildBspNode(back_faces, backNodeId);
	}

	if ( back_faces.size() == 0 && front_faces.size() != 0 )
	{
		bspNodes[nodeId].backNode	= SOLID_AREA;
		frontNodeId	= static_cast<int>(bspNodes.size());
		bspNodes[nodeId].frontNode	= frontNodeId;
		bspNodes.push_back(n);

		buildBspNode(front_faces, frontNodeId);
	}

	if ( back_faces.size() != 0 && front_faces.size() != 0 )
	{
		// copy on faces to front faces
		for ( i = 0; i < static_cast<int>(onFaces.size()); i++ )
			front_faces.push_back(onFaces[i]);

		frontNodeId	= static_cast<int>(bspNodes.size());
		bspNodes[nodeId].frontNode	= frontNodeId;
		bspNodes.push_back(n);
		buildBspNode(front_faces, frontNodeId);
		backNodeId	= static_cast<int>(bspNodes.size());
		bspNodes[nodeId].backNode	= backNodeId;
		bspNodes.push_back(n);
		buildBspNode(back_faces, backNodeId);
	}
}

/*
** build the BSP tree
*/
void Map::buildBsp(vector<AreaPolygon> &polys)
{
	bspNodes.clear();

	cout << "Building the BSP tree...";
	BspNode	n;
	bspNodes.push_back(n);
	buildBspNode(polys, 0);

	cout << "done" << endl << "number of BSP nodes: " 
		<< static_cast<unsigned int>(bspNodes.size()) << endl;
}

/*
** get the point area in a node
*/
int Map::getPointArea(const vec3 &v, int nodeId)
{
	float		pos;

	pos		= bspNodes[nodeId].splitPlane.signedDistance(v);

	// check the case where the point lies in front
	if ( pos >= -epsilon )
	{
		// we have reached an area
		if ( bspNodes[nodeId].frontNode <= 0 )
			return (bspNodes[nodeId].frontNode);

		// else recurse
		return getPointArea(v, bspNodes[nodeId].frontNode);

	}

	// if we lie on or on the back (the else)
	if ( bspNodes[nodeId].backNode <= 0 )
	{
		switch ( bspNodes[nodeId].backNode )
		{
		case SOLID_AREA:
			return 1;	// it'll be inverted eventually
		default:	// shouldn't get here if we have a legal geometry
			return (bspNodes[nodeId].backNode);
		}
	}
		
	return getPointArea(v, bspNodes[nodeId].backNode);
}

/*
** just to ease things for us
*/
int Map::getPointArea(const vec3 &v)
{
	return (-getPointArea(v, 0));
}

/*
** assign the areas to portals (each portal belongs to 2 areas)
*/
void Map::assignAreasToPortals()
{
	// we are going to loop all the portal vertices and find to which areas
	// they belong
	for ( int p = 0; p < numPortals; p++ ) {
		portals[p].areaIds[0]	= -1;
		portals[p].areaIds[1]	= -1;

		for ( int v = 0; v < portals[p].mesh.getNumVerts(); v++ ) {
			int	areaId	= getPointArea(portals[p].mesh.getVerts()[v]);
			//cout << "area id " << areaId << endl;

			if ( portals[p].areaIds[0] != areaId && portals[p].areaIds[1] != areaId ) {
				if ( portals[p].areaIds[0] == -1 )
					portals[p].areaIds[0]	= areaId;
				else if ( portals[p].areaIds[1] == -1 )
					portals[p].areaIds[1]	= areaId;
				//else if ( portals[p].areaIds[0] != -1 ||
				//	portals[p].areaIds[1] == -1 )
				//	throw "Portal belonging to more than 2 areas...";
			}
		}
	}
}