/*
** map_decomposer.cpp
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** mesh decomposion
**
** Written for More OpenGL Game Programming
*/

#include "GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>

#include <iostream>
#include <vector>
#include <map>
#include <string>

#include "algebra3.h"
#include "geom.h"
#include "scene_map.h"

using namespace std;
using namespace SceneMap;

/*
** decompose all areas mesh into areas (basically by marking all the polygons)
*/
struct VertexFaces
{
	static const int MAX_VERTEX_FACES	= 32;
	bool	used;
	int		numFaces;
	int		faces[MAX_VERTEX_FACES];

	VertexFaces(): numFaces(0), used(false) {}
	void	add(int i) { 
		if ((numFaces + 1) > MAX_VERTEX_FACES) throw "exceeded MAX_VERTEX_FACES"; 
		faces[numFaces++]	= i;
	}
};

/*
** walk a vertex in an area: basically mark all faces adjacent to this vertex
*/
void walkVertex(int v, int area, const int *faces, vector<VertexFaces> &areasVerts, vector<int> &areasFaces)
{
	int		i, currFace;

	areasVerts[v].used	= true;

	for ( i = 0; i < areasVerts[v].numFaces; i++ )
	{
		currFace	= areasVerts[v].faces[i];
		if (areasFaces[currFace] == -1)
		{
			areasFaces[currFace]	= area;
			if (!areasVerts[faces[currFace * 3]].used)
				walkVertex(faces[currFace * 3], area, faces, areasVerts, areasFaces);

			if (!areasVerts[faces[currFace * 3 + 1]].used)
				walkVertex(faces[currFace * 3 + 1], area, faces, areasVerts, areasFaces);

			if (!areasVerts[faces[currFace * 3 + 2]].used)
				walkVertex(faces[currFace * 3 + 2], area, faces, areasVerts, areasFaces);

		} else if (areasFaces[currFace] != area)
		{
			cout << "found area " << areasFaces[currFace] << ", we are at a:" << area << "-f:" << currFace << endl;
			int j;
			for ( j = 0; j < static_cast<int>(areasFaces.size()); j++ )
				cout << "face " << i << ": " << areasFaces[i] << endl;

			throw "Invalid ALL_AREAS_MESH topology!!!";
		}
	}
}

/*
** the main method for that
*/
void Map::decomposeAllAreasMesh()
{
	vector<VertexFaces>		areasVerts;	// contains all index of faces refering to a vertex
	vector<int>				areasFaces;	// marked faces according to areas
	vector<AreaPolygon>		bspFaces;		// face for BSP

	int						i, j, k;
	const int*				faces;
	const vec3*				verts;

	faces	= allAreasMesh.getFaces();
	verts	= allAreasMesh.getVerts();

	// reserve memory for our work and initiate
	areasVerts.reserve(allAreasMesh.getNumVerts());
	areasFaces.reserve(allAreasMesh.getNumFaces());

	VertexFaces		f;

	for ( i = 0; i < allAreasMesh.getNumVerts(); i++ )
		areasVerts.push_back(f);

	for ( i = 0; i < allAreasMesh.getNumFaces(); i++ )
		areasFaces.push_back(-1);	// all faces to -1

	// fill in the vertices with the faces
	for ( i = 0; i < allAreasMesh.getNumFaces(); i++ )
	{
		areasVerts[faces[i * 3]].add(i);
		areasVerts[faces[i * 3 + 1]].add(i);
		areasVerts[faces[i * 3 + 2]].add(i);
	}

	// next we will start removing every vertex, and assign all faces incident in this
	// vertex to an area
	int		area	= 0;
	for ( i = 0; i < allAreasMesh.getNumVerts(); i++ )
	{
		if (!areasVerts[i].used)
		{
			walkVertex(i, area, faces, areasVerts, areasFaces);

			// we should have walked all verts and faces now belonging to an area so
			// we can increment the area number
			area++;
		}
	}
	
	//cout << "number of areas: " << (area - 1) << endl;

	// now we build the areas
	numAreas	= area;
	areas		= new Area[numAreas];

	for ( i = 0; i < numAreas; i++ )
	{
		areas[i].areaId	= i;
	}

	// walk all the faces and compute each areas polygons
	for ( i = 0; i < static_cast<int>(areasFaces.size()); i++ )
	{
		AreaPolygon		p;
		p.areaId	= areasFaces[i];
		p.verts.push_back(verts[faces[i * 3]]);
		p.verts.push_back(verts[faces[i * 3 + 1]]);
		p.verts.push_back(verts[faces[i * 3 + 2]]);
		p.p	= plane(p.verts[0], p.verts[1], p.verts[2]);

		areas[p.areaId].polys.push_back(p);
		bspFaces.push_back(p);
	}

	// compute the bounding volumes for the areas
	for ( i = 0; i < numAreas; i++ )
	{
		areas[i].bv.maxV	= vec3(-99999.9f,-99999.9f,-99999.9f);
		areas[i].bv.minV	= vec3(99999.9f,99999.9f,99999.9f);
		areas[i].bv.center	= vec3(0.0f, 0.0f, 0.0f);
		areas[i].bv.radius	= 0.0f;
		areas[i].bv.radius2	= 0.0f;
		for ( j = 0; j < static_cast<int>(areas[i].polys.size()); j++ )
		{
			for ( k = 0; k < 3; k++ )
			{
				vec3	v = areas[i].polys[j].verts[k];

				if (areas[i].bv.maxV.x < v.x)
					areas[i].bv.maxV.x	= v.x;
				if (areas[i].bv.maxV.y < v.y)
					areas[i].bv.maxV.y	= v.y;
				if (areas[i].bv.maxV.z < v.z)
					areas[i].bv.maxV.z	= v.z;

				if (areas[i].bv.minV.x > v.x)
					areas[i].bv.minV.x	= v.x;
				if (areas[i].bv.minV.y > v.y)
					areas[i].bv.minV.y	= v.y;
				if (areas[i].bv.minV.z > v.z)
					areas[i].bv.minV.z	= v.z;
			}
		}

		areas[i].bv.center	= (areas[i].bv.minV + areas[i].bv.maxV) / 2.0f;

		// and recompute bv
		areas[i].bv.maxV -= areas[i].bv.center;
		areas[i].bv.minV -= areas[i].bv.center;

		areas[i].bv.radius2	= ((areas[i].bv.maxV - areas[i].bv.minV)/2.0f).length2();
		areas[i].bv.radius	= ((areas[i].bv.maxV - areas[i].bv.minV)/2.0f).length();

		areas[i].color	= vec3((1.0f + static_cast<float>(rand() % 255)) / 256.0f,
							(1.0f + static_cast<float>(rand() % 255)) / 256.0f,
							(1.0f + static_cast<float>(rand() % 255)) / 256.0f);
	}

	//  build the BSP
	buildBsp(bspFaces);

	// and assign areas to portal
	assignAreasToPortals();
}