/*
** map_loader.cpp
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** map loading
**
** Written for More OpenGL Game Programming
*/

#include "GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <string>

#include "algebra3.h"
#include "geom.h"
#include "scene_map.h"

using namespace std;
using namespace SceneMap;

/*
** overloaded istream operator to load mesh
*/
std::istream& operator >> (std::istream &is, BasicMesh &mm)
{
	int		numVerts, numFaces, i;
	vec3*	verts;
	int*	faces;

	string	token;

	is >> token;
	if (token.compare("BEGIN_MESH"))
		throw "expected BEGIN_MESH";

	is >> token;
	if (token.compare("NUM_VERTS"))
		throw "expected NUM_VERTS";

	is >> numVerts;

	verts	= new vec3[numVerts];

	for (i = 0; i < numVerts; i++)
	{
		is >> verts[i].x;
		is >> token;
		if (token.compare(","))
			throw "expected ,";

		is >> verts[i].y;
		is >> token;
		if (token.compare(","))
			throw "expected ,";

		is >> verts[i].z;
	}

	is >> token;
	if (token.compare("NUM_FACES"))
		throw "expected NUM_FACES";

	is >> numFaces;

	faces	= new int[numFaces * 3];

	for (i = 0; i < numFaces; i++)
	{
		is >> faces[i * 3];
		is >> token;
		if (token.compare(","))
			throw "expected ,";

		is >> faces[i * 3 + 1];
		is >> token;
		if (token.compare(","))
			throw "expected ,";

		is >> faces[i * 3 + 2];
	}

	is >> token;
	if (token.compare("END_MESH"))
		throw "expected END_MESH";

	mm.setup(numVerts, verts, numFaces, faces);

	return is;
}

/*
** load a map
*/
void Map::load(char *fileName)
{
	string		token;
	int			i;

	ifstream	is(fileName);
	if ( !is.is_open() ) {
		throw "unable to find map file";
	}

	// first start with portals
	is >> token;
	if (token != "BEGIN_PORTALS")
		throw "expected BEGIN_PORTALS";

	is >> numPortals;
	cout << "num portals: " << numPortals << endl;
	portals	= new Portal[numPortals];
	
	for (i = 0; i < numPortals; i++)
	{
		is >> token;
		if (token != "PORTAL")
			throw "expected PORTAL";

		is >> portals[i].name;

		cout << "loaded portal \"" << portals[i].name << "\"" << endl;

		is >> portals[i].mesh;
	}

	is >> token;
	if (token != "END_PORTALS")
		throw "expected END_PORTALS";

	// continue to occluders
	is >> token;
	if (token != "BEGIN_OCCLUDERS")
		throw "expected BEGIN_OCCLUDERS";

	is >> numOccluders;
	occluders	= new Occluder[numOccluders];

	cout << "num occluders: " << numOccluders << endl;
	
	for (i = 0; i < numOccluders; i++)
	{
		is >> token;
		if (token != "OCCLUDER")
			throw "expected OCCLUDER";

		is >> occluders[i].name;

		cout << "loaded occluder \"" << occluders[i].name << "\"" << endl;

		is >> occluders[i].mesh;
	}

	is >> token;
	if (token != "END_OCCLUDERS")
		throw "expected END_OCCLUDERS";

	// load all areas mesh (important part for BSP)
	is >> token;
	if (token != "BEGIN_AREAS_MESH")
		throw "expected BEGIN_AREAS_MESH";
	
	is >> allAreasMesh;

	is >> token;
	if (token != "END_AREAS_MESH")
		throw "expected END_AREAS_MESH";

	is.close();

	// compute the global bounding volume
	bv.maxV	= vec3(-99999.9f,-99999.9f,-99999.9f);
	bv.minV	= vec3(99999.9f,99999.9f,99999.9f);
	bv.center	= vec3(0.0f, 0.0f, 0.0f);
	bv.radius	= 0.0f;
	bv.radius2	= 0.0f;
	
	// build for portals portals
	for ( i = 0; i < numPortals; i++ )
	{
		vec3	maxB, minB;
		maxB	= portals[i].mesh.getBoundingVolume().maxV + 
			portals[i].mesh.getBoundingVolume().center;
		minB	= portals[i].mesh.getBoundingVolume().minV + 
			portals[i].mesh.getBoundingVolume().center;

		if (bv.maxV.x < maxB.x)
			bv.maxV.x	= maxB.x;
		if (bv.maxV.y < maxB.y)
			bv.maxV.y	= maxB.y;
		if (bv.maxV.z < maxB.z)
			bv.maxV.z	= maxB.z;

		if (bv.minV.x > minB.x)
			bv.minV.x	= minB.x;
		if (bv.minV.y > minB.y)
			bv.minV.y	= minB.y;
		if (bv.minV.z > minB.z)
			bv.minV.z	= minB.z;
	}

	// and add occluders
	for ( i = 0; i < numOccluders; i++ )
	{
		vec3	maxB, minB;
		maxB	= occluders[i].mesh.getBoundingVolume().maxV + 
			occluders[i].mesh.getBoundingVolume().center;
		minB	= occluders[i].mesh.getBoundingVolume().minV + 
			occluders[i].mesh.getBoundingVolume().center;

		if (bv.maxV.x < maxB.x)
			bv.maxV.x	= maxB.x;
		if (bv.maxV.y < maxB.y)
			bv.maxV.y	= maxB.y;
		if (bv.maxV.z < maxB.z)
			bv.maxV.z	= maxB.z;

		if (bv.minV.x > minB.x)
			bv.minV.x	= minB.x;
		if (bv.minV.y > minB.y)
			bv.minV.y	= minB.y;
		if (bv.minV.z > minB.z)
			bv.minV.z	= minB.z;
	}

	// compute center
	bv.center	= (bv.minV + bv.maxV) / 2.0f;

	// and recompute bv
	bv.maxV -= bv.center;
	bv.minV -= bv.center;

	bv.radius2	= ((bv.maxV - bv.minV)/2.0f).length2();
	bv.radius	= ((bv.maxV - bv.minV)/2.0f).length();

	// decompose the all area meshes into areas
	decomposeAllAreasMesh();
}