/*
** map_vis.cpp
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** occlusion query code
**
** Written for More OpenGL Game Programming
*/

#include "GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

#include "algebra3.h"
#include "geom.h"
#include "scene_map.h"
#include "frustum.h"

using namespace std;
using namespace SceneMap;

void Map::setupVisSystem()
{
	visPortals	= new GLuint[numPortals];

	// generate queries
	glGenQueries(numPortals, visPortals);
}

void Map::issueVisQueries(int w, int h, vec3 &eyePos, vec3 &viewDir, vec3 &up)
{
	int		i;
	float	matView[16];

	// first of all set the proper viewport
	glMatrixMode(GL_MODELVIEW);
	glGetFloatv(GL_MODELVIEW_MATRIX, matView);
	glLoadIdentity();
	glViewport(0, 0, w, h);
	glScissor(0, 0, w, h);
	glEnable(GL_SCISSOR_TEST);

	gluLookAt(0.0f, 0.0f, 0.0f,
		viewDir.x, viewDir.y, viewDir.z,
		0.0f, 0.0f, 1.0f);

	// set the area the eye is in
	eyeArea	= getPointArea(eyePos);


	// center the view
	glTranslatef(-eyePos.x,
		-eyePos.y,
		-eyePos.z);

	// clear the viewport area
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// setup the view frustum
	Frustum	f;
	f.setupFromGL();

	// render all occluder in wireframe (just to be able to distingusih them
	// as we are not using textures)
	glColor3f(0.0f, 0.0f, 0.0f);
	glLineWidth(4.0f);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	for ( i = 0; i < numOccluders; i++ )
		occluders[i].mesh.glRender();
	glLineWidth(1.0f);

	// next render all occluder	in fill mode
	glColor3f(1.0f, 1.0f, 0.0f);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	for ( i = 0; i < numOccluders; i++ )
	{
		const BoundingVolume	bv	= occluders[i].mesh.getBoundingVolume();
		// only render occluders that bypass frustum test
		if ( f.isInside(bv.center, bv.radius) )
            occluders[i].mesh.glRender();
	}

	// disable writing to the color buffer and depth buffer
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask(GL_FALSE);

	// reset portal queries
	portalQueries.resize(0);

	// now issue the queries for portals
	for ( i = 0; i < numPortals; i++ )
	{
		const BoundingVolume	bv	= portals[i].mesh.getBoundingVolume();
		// should cull by frustum here also to limit the number of issued queries
		if ( f.isInside(bv.center, bv.radius) ) {
			// push the query id to the list
			portalQueries.push_back(i);

			glBeginQueryARB(GL_SAMPLES_PASSED_ARB, visPortals[i]);
			portals[i].mesh.glRender();
			glEndQueryARB(GL_SAMPLES_PASSED_ARB);
		}
	}

	// re-enable writing to color and depth buffer
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDepthMask(GL_TRUE);
	glLoadMatrixf(matView);
}

void Map::getVisResults(vector<int> &visPortalsId, vector<int> &visAreasId)
{
	bool	*visibleAreas	= new bool[numAreas];

	for ( int i = 0; i < numAreas; i++ )
		visibleAreas[i]	= false;
	
	// the area where we are is always visible (unless it's void space)
	if ( eyeArea != - 1 )
        visibleAreas[eyeArea]	= true;

	// clear the vectors
	visPortalsId.resize(0);
	visAreasId.resize(0);

	// get visible portals
	for ( int i = 0; i < portalQueries.size(); i++ ) {
		GLuint	sampleCount;	// number of pixels
		glGetQueryObjectuivARB(visPortals[portalQueries[i]], GL_QUERY_RESULT_ARB, &sampleCount);
		if ( sampleCount > 0 ) {	// push only visible portals
			visPortalsId.push_back(i);

			// flag the visible areas
			if ( portals[i].areaIds[0] != - 1 )
				visibleAreas[portals[i].areaIds[0]]	= true;
			if ( portals[i].areaIds[1] != - 1 )
				visibleAreas[portals[i].areaIds[1]]	= true;
		}
	}

	// finally fill in visible areas vector
	for ( int i = 0; i < numAreas; i++ )
		if ( visibleAreas[i] )
			visAreasId.push_back(i);

	delete[] visibleAreas;
}
