/*
** scene_map.h
**
** Copyright (C) 2005 Wael EL ORAIBY
**
** scene map structures and classes definitions
**
** Written for More OpenGL Game Programming
*/

#pragma once

namespace SceneMap {

	/*****************************************************************************
	*																			 *
	* BasicMesh class: basic mesh stuff, no textures and normals inside the mesh *
	*																			 *
	*****************************************************************************/
	class BasicMesh {
	private:
		BoundingVolume			bv;
		int						numVerts;
		int						numFaces;
		vec3*					verts;
		int*					faces;

	public:
		BasicMesh(): verts(NULL), faces(NULL), numVerts(0), numFaces(0) {}
		~BasicMesh() { if (verts) delete[] verts; if (faces) delete[] faces; }

		void					setup(int numVerts, vec3* verts, int numFaces, int* faces);
		const vec3*				getVerts() { return verts; }
		const int*				getFaces() { return faces; }
		int						getNumVerts() { return numVerts; }
		int						getNumFaces() { return numFaces; }
		const BoundingVolume&	getBoundingVolume() const { return bv; }
		virtual void			glRender() const;
	};

	/*****************************************************************************
	*																			 *
	* Portal structure: a portal between two areas/zones						 *
	*																			 *
	*****************************************************************************/
	struct Portal
	{
		std::string				name;		// portal name
		BasicMesh				mesh;		// portal mesh
		int						areaIds[2];	// the areas to which the portal belongs
	};

	/*****************************************************************************
	*																			 *
	* Occluder structure: occluding mesh										 *
	*																			 *
	*****************************************************************************/
	struct Occluder
	{
		std::string				name;
		BasicMesh				mesh;
	};

	const int SOLID_AREA	= -9999999;

	/*****************************************************************************
	*																			 *
	* Area polygon structure: it's only needed to build the BSP tree			 *
	*																			 *
	*****************************************************************************/
	struct AreaPolygon
	{
		//static const int MAX_POLYGON_VERTS	= 32;
		int						areaId;					// area to where the polygon belongs
		//int						numVerts;				// number of polygons verts
		std::vector<vec3>		verts;					// polygon verts
		plane					p;

		AreaPolygon(): areaId(-1) {}
	};

	/*****************************************************************************
	*																			 *
	* Area structure: the area or zones											 *
	*																			 *
	*****************************************************************************/
	struct Area
	{
		int						areaId;					// area id
		BoundingVolume			bv;
		// those are used for visualization only
		std::vector<AreaPolygon>		polys;
		vec3					color;

		// stuff that are present in an area/sector (bullets, players, meshes, triggers)

		void glRender() const
		{
			int		i, j;
			for ( i = 0; i < static_cast<int>(polys.size()); i++ )
			{
				glBegin(GL_TRIANGLE_FAN );
				for ( j = 0; j < static_cast<int>(polys[i].verts.size()); j++ )
					glVertex3f(polys[i].verts[j].x, polys[i].verts[j].y, polys[i].verts[j].z);
				glEnd();
			}

		}

		Area(): areaId(-1) {}
	};

	/*****************************************************************************
	*																			 *
	* Bsp node structure: holds necessary information for BSP tree structure	 *
	*																			 *
	*****************************************************************************/
	struct BspNode
	{
		plane					splitPlane;		// the splitter plane

		int						frontNode;		// front node index (<0 indicates area)
		int						backNode;		// back node index (<0 indicates area)
	};

	/*****************************************************************************
	*																			 *
	* Map class: holds all information related to a map: occluders, portals, bsp *
	*		entities and all related stuff...									 *
	*																			 *
	*****************************************************************************/
	class Map
	{
	private:
		static float			epsilon;
		BoundingVolume			bv;
		int						numPortals;
		int						numOccluders;
		Portal*					portals;
		Occluder*				occluders;
		BasicMesh				allAreasMesh;
		int						numAreas;
		Area*					areas;
		int						eyeArea;		// the area where we are now

		GLuint*					visPortals;		// portals occlusion queries
		std::vector<int>		portalQueries;	// visible portal ids


		// BSP nodes (make it constant size array if you don't like STL)
		std::vector<BspNode>	bspNodes;

		void					decomposeAllAreasMesh();	// decompose all area mesh into areas
		int						findSplitter(std::vector<AreaPolygon> &polys, plane &splitter);	// -1 no splitter found
		void					splitPoly(plane &splitter, AreaPolygon &poly, AreaPolygon &front, AreaPolygon &back);
		void					buildBspNode(std::vector<AreaPolygon> &polys, int nodeId);
		void					buildBsp(std::vector<AreaPolygon> &polys);

		enum FACE_CLASS { FACE_INFRONT, FACE_ONBACK, FACE_ON, FACE_SPANNING };
		FACE_CLASS				classifyPolyToPlane(plane &cmpPlane, AreaPolygon &poly);


		int						getPointArea(const vec3 &v, int nodeId);
		void					assignAreasToPortals();

	public:
		Map(): numPortals(0), numOccluders(0), numAreas(0), visPortals(NULL), 
			portals(NULL), occluders(NULL), areas(NULL) {}

		~Map() { delete[] portals; delete[] occluders; delete[] areas;
		if ( visPortals ) glDeleteQueries(numPortals, visPortals); delete[] visPortals; }

		int						getNumPortals() { return numPortals; }
		int						getNumOccluders() { return numOccluders; }
		const Portal*			getPortals() { return portals; }
		const Occluder*			getOccluders() { return occluders; }
		const BoundingVolume&	getBoundingVolume() { return bv; }
		const BasicMesh&		getAllAreasMesh() { return allAreasMesh; }
		int						getNumAreas() { return numAreas; }
		const Area*				getAreas() { return areas; }

		// spatial position (which area the point is in)
		int						getPointArea(const vec3 &v);

		// load and build map
		void					load(char* fileName);

		// occlusion query stuff
		void					setupVisSystem();

		void					issueVisQueries(int w, int h, vec3 &eyePos, vec3 &atPos, vec3 &up);		// make occlusion queries in a defined window
		void					getVisResults(std::vector<int> &visPortalsId, std::vector<int> &visAreasId);	// get the queries results
	};
};