/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: CCamera.cpp                                               |*/
/*| Author: Joel Villagrana  (joel.villagrana@gmail.com)                |*/
/*|                                                                     |*/
/*| Implemented Classes:                                                |*/
/*| CCamera                                                             |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yy   | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 11/09/04   | JVillagrana       | Initial version                    |*/
/*************************************************************************/

#include "CCamera.h"

#define PITCH_AMOUNT 7.15f
#define YAW_AMOUNT   9.15f
#define ROLL_AMOUNT  7.15f
#define PANNING_DELTA  4.081f
#define ZOOMING_DELTA  6.6f

/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
=====================================================================================
				IMPLEMENTATION OF THE CCamera CLASS METHODS						  
=====================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/

// ==================================================================================
// CONSTRUCTOR
// ==================================================================================
CCamera::CCamera( MGL::Vector3 eye, float pitch, float yaw, float roll )
{
	m_virtualObs= 8.5f;
	m_pitch		= pitch;
	m_pitchRad	= (float)(pitch * PI_OVER_180);
	m_yaw		= yaw;
	m_yawRad	= (float)(yaw * PI_OVER_180);
	m_roll		= roll;
	m_rollRad	= (float)(roll * PI_OVER_180);
	m_Speed		= 0.0f;
}

// ==================================================================================
// INITIALIZE
// ==================================================================================
bool CCamera::InitAll()
{	
	m_virtualObs= 30.0f;
	m_roll		= 0.0f;
	m_rollRad	= 0.0f;
	m_yaw		= 0.0f;
	m_yawRad	= 0.0f;
	m_pitch		= 0.0f;
	m_pitchRad	= 0.0f;
	m_Speed		= 0.0f;

	// Reset the quaternion
	m_quaternion.Reset();
	
	// Reset vectors
	//m_currentWorldPos.Set(10.0f, 3.0f, 15.0f);
	m_currentWorldPos.Set(0.0f, 0.0f, 0.0f);
	m_lastWorldPos.Set(0.0f, 0.0f, 0.0f);
	m_directionVector.Set(0.0f, 0.0f, -1.0f);
	m_rightVector.Set(0.0f, 0.0f, 0.0f);
	m_lookAtVector.Set(0.0f, 0.0f, 0.0f);
	m_upVector.Set(0.0f, 1.0f, 0.0f);
	m_initialDirV.Set(0.0f, 0.0f, -1.0f);

	Set_YawPitchRoll(10.0f, 5.0f, 0.0f);
	// Update vectors with quaternion
	UpdateCurrentWorldPos();	

	return true;
}

// ==================================================================================
// YAW, PITCH, ROLL
// ==================================================================================
void CCamera::Set_YawPitchRoll( float yaw, float pitch, float roll )
{			
	if( (yaw==0.0f)&&(pitch==0.0f)&&(roll==0.0f) )
		return;

	if(yaw!=0.0f)
	{
		m_yaw+=yaw;
		if(m_yaw>=360.0f)m_yaw=0;	
		if(m_yaw<0.0f)m_yaw=360.0f+m_yaw;
		m_yawRad=(GLfloat)(m_yaw*PI_OVER_180);
	}	
	if(pitch!=0.0f)
	{
		m_pitch+=pitch;
		if(m_pitch>=360.0f)m_pitch=0;	
		if(m_pitch<0.0f)m_pitch=360.0f+m_pitch;
		m_pitchRad=(GLfloat)(m_pitch*PI_OVER_180);	
	}	
	if(roll!=0.0f)
	{
		m_roll+=roll;
		if(m_roll>=360.0f)m_roll=0;	
		if(m_roll<0.0f)m_roll=360.0f+m_roll;
		m_rollRad=(GLfloat)(m_roll*PI_OVER_180);
	}
	
	// Update vectors with quaternion
	UpdateRotation();
}

// ==================================================================================
// YAW, PITCH, ROLL
// ==================================================================================
void CCamera::Get_YawPitchRoll( float *yaw, float *pitch, float *roll )
{
	*yaw = m_yaw;
	*pitch = m_pitch;
	*roll = m_roll;
}

// ==================================================================================
// PAN CAMERA
// ==================================================================================
void CCamera::PanCamera(MGL::Vector3 newMove)
{
	// MOVE CAMERA IN LOCAL X AXIS
	m_currentWorldPos	+= ((m_rightVector)*(newMove.x));
	m_lookAtVector		+= ((m_rightVector)*(newMove.x));

	// MOVE CAMERA IN LOCAL Y AXIS
	m_currentWorldPos	+= ((m_upVector)*(newMove.y));
	m_lookAtVector		+= ((m_upVector)*(newMove.y));
}

// ==================================================================================
// UPDATE THE CAMERA EVERY FRAME
// ==================================================================================
void CCamera::Tick()
{
	// Save the current world pos
	m_lastWorldPos = m_currentWorldPos;

	// Apply the translation
	m_currentWorldPos += (m_directionVector * m_Speed);
	m_lookAtVector    += (m_directionVector * m_Speed);
}

// ==================================================================================
// ROTATE THE CAMERA WITH THE QUATERNION
// ==================================================================================
void CCamera::UpdateRotation()
{	
	// IF THERE'S NO ROTATION TO PERFORM, RETURN...
	if((m_pitch==0.0f)&&(m_yaw==0.0f)&&(m_roll==0.0f))return;

	// TEMPORARY QUATERNIONS
	GLQuaternion qy(0.0f, 270.0f, 0.0f);
	GLQuaternion qx(90.0f, 0.0f, 0.0f);
	
	// TEMPORARY VECTOR
	GLVector4 tempDir;

	// QUATERNION TO REPRESENT THE NEW ROTATION (no need to normalize as it's done internally)
	GLQuaternion newRotation(m_pitch, m_yaw, m_roll);

	// QUATERNION TO REPRESENT THE INVERSE OF THE ROTATION QUATERNION
	GLQuaternion inverseQuat;

	// THE RESULTING QUATERNION
	GLQuaternion resultQuat;

	// NEED TO RESET DIRECTION VECTOR SO ROTATION WORKS ON INITIAL DIR
	m_directionVector = m_initialDirV;

	// SET THE QUATERNION
	m_quaternion = newRotation;

	tempDir.setValues(m_directionVector.x, m_directionVector.y, m_directionVector.z);

	// --------------------------
	// CALCULATE DIRECTION VECTOR
	// --------------------------
	GLQuaternion initialDirectionQ(	0.0f, m_directionVector.x, m_directionVector.y,	m_directionVector.z);	
	inverseQuat = newRotation.Get_Inverse();
	resultQuat = (newRotation * initialDirectionQ) * inverseQuat;
	m_directionVector.Set(resultQuat.x, resultQuat.y, resultQuat.z);

	// -----------------------
	// CALCULATE THE UP VECTOR
	// -----------------------
	// RESET DIR AS INITIAL DIR TO CALCULATE THE UP VECTOR
	initialDirectionQ.Set_Values(0.0f, tempDir.x, tempDir.y, tempDir.z);
	resultQuat  = newRotation*qx;
	inverseQuat = resultQuat.Get_Inverse();
	resultQuat  = (resultQuat * initialDirectionQ) * inverseQuat;
	m_upVector.Set(resultQuat.x, resultQuat.y, resultQuat.z);
	
	// --------------------------
	// CALCULATE THE RIGHT VECTOR
	// --------------------------
	// RESET DIR AGAIN TO CALCULATE THE RIGHT VECTOR
	initialDirectionQ.Set_Values(0.0f, tempDir.x, tempDir.y, tempDir.z);
	resultQuat  = newRotation*qy;
	inverseQuat = resultQuat.Get_Inverse();
	resultQuat  = (resultQuat * initialDirectionQ) * inverseQuat;
	m_rightVector.Set(resultQuat.x, resultQuat.y, resultQuat.z);
}

// ==================================================================================
// SET THE CURRENT POSITION BY THE ROTATION OF THE QUATERNION 
// ==================================================================================
void CCamera::UpdateCurrentWorldPos()
{
	MGL::Vector3 tempPos, newWorldPos;

	float dist = m_virtualObs;

	if(((m_directionVector.x==0.0f)&&(m_directionVector.y==0.0f)&&(m_directionVector.z==0.0f))||(dist==0.0f))
		return;

	tempPos = m_directionVector;
	tempPos.Normalize();
	newWorldPos.Set(tempPos.x*dist, tempPos.y*dist, tempPos.z*dist);
	newWorldPos = newWorldPos + m_lookAtVector;
	m_currentWorldPos.Set(newWorldPos.x, newWorldPos.y, newWorldPos.z);	
}

// ==================================================================================
// ZOOM CAMERA
// ==================================================================================
void CCamera::ZoomCamera(float amtZoom)
{
	m_virtualObs += amtZoom;

	if(m_virtualObs < 1.0f)
		m_virtualObs = 1.0f;

	UpdateCurrentWorldPos();
}

// ==================================================================================
// MOUSEMOVE
// ==================================================================================
void CCamera::MouseMove(bool leftButton, bool rightButton, float deltaX, float deltaY)
{
	bool shiftPressed = false;
	short int shiftState;
	shiftState = GetKeyState(VK_SHIFT);
	if(shiftState < 0)
		shiftPressed = true;

	if(leftButton)
	{		
		if(shiftPressed)
		{
			MGL::Vector3 newMove;
			float amtX = (float)(deltaX * PANNING_DELTA);
			float amtY = (float)(deltaY * PANNING_DELTA); 
			newMove.Set(-amtX, -amtY, 0.0f);
			PanCamera(newMove);
		}
		else
		{			
			float amtYaw   = deltaX * YAW_AMOUNT;
			float amtPitch = deltaY * PITCH_AMOUNT;
			Set_YawPitchRoll(amtYaw, -amtPitch, 0.0f);
			UpdateCurrentWorldPos();			
		}
	}
	else if(rightButton)
	{		
		ZoomCamera((deltaY * ZOOMING_DELTA));
	}
}

// -------------------------------- END OF FILE -----------------------------------
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



