/*************************************************************************
MGLApp.cpp

Main application class.

Author   :   Dave Astle

Written for More OpenGL Game Programming
-------------------------------------------------------------------------
| dd/mm/yyyy  |  Programmer        | Explanation of change              |
-------------------------------------------------------------------------
| 09/10/2005  |  J.Villagrana   @1 | -Added libraries links, texture    | 
|             |                    |  manager singleton def and a bunch |
|             |                    |  of changes(MGLApp,Render,Update..)| 
*************************************************************************/

#ifdef WIN32
	#define WIN32_LEAN_AND_MEAN
	#define WIN32_EXTRA_LEAN
	#include <windows.h>
#endif

#include "MGLApp.h"

// ----------------------------@1A BEGINS ------------------------------
#include "./Texture/textureManager.h"
#include "./Scene/Scene.h"
#include "./Camera/CCamera.h"

// SEARCH FOR LIBRARIES AT COMPILE TIME
// OPENGL LIBRARIES
#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")
#pragma comment(lib, "glaux.lib")
// LIBRARY FOR timeGetTime
#pragma comment(lib, "winmm.lib")
// LIBRARY FOR checking OpenGL extensions
#pragma comment(lib, "glee.lib")

// EXPLICITLY DEFINE STATIC SINGLETON OBJECTS
CTextureManager*	CTextureManager::m_pGTextureManager;
CScene*				CScene::m_pScene;
CCamera*			CCamera::m_pCamera;
// ----------------------------@1A ENDS --------------------------------

MGLApp::MGLApp()
{
}

MGLApp::~MGLApp()
{
}

bool MGLApp::Init()
{	
	// DEFAULT LIGHT ATTRIBUTES
	float	defLight_position[4];		// @1A - Light position
	float	defLight_diffuse[4];		// @1A - Diffuse color
	float	defLight_specular[4];		// @1A - Specular color
	float	defLight_ambient[4];		// @1A - Ambient color
	// LIGHT MODEL ATTRIBUTES
	float	ligthingModel_ambient[4];	// @1A - Global ambient light

	if (!GLeeInit())
		return false;

	// ****************** @1A BEGINS *************************

	defLight_position[0]	= -1.5f;
	defLight_position[1]	= 10.0f;
	defLight_position[2]	= 3.0f;
	defLight_position[3]	= 1.0f;
	defLight_diffuse[0]		= 0.17f;
	defLight_diffuse[1]		= 0.60f; 
	defLight_diffuse[2]		= 0.89f; 
	defLight_diffuse[3]		= 1.0f;
	defLight_specular[0]	= 0.887f;
	defLight_specular[1]	= 0.40f;
	defLight_specular[2]	= 0.40f;
	defLight_specular[3]	= 1.0f;
	defLight_ambient[0]		= 0.06f;
	defLight_ambient[1]		= 0.06f;
	defLight_ambient[2]		= 0.08916f;
	defLight_ambient[3]		= 1.0f;
	ligthingModel_ambient[0]= 0.1f;
	ligthingModel_ambient[1]= 0.1f;
	ligthingModel_ambient[2]= 0.1f;
	ligthingModel_ambient[3]= 1.0f;

	// --------------------
	// Init texture manager	
	// --------------------
	CTextureManager::SetTextureManager(new CTextureManager);
	if(!_TextureManager())
	{
		MessageBox(NULL, "Cannot create texture manager class instance", "MGLApp::Init", MB_OK|MB_ICONSTOP);
		return false;
	}	
	else
	{
		_TextureManager()->LoadTexture("Data/Images/background.tga", GL_LINEAR);
	}

	// --------------------
	// Create font class
	// --------------------
	if(!(m_font.CreateGLFont("Data/Images/tx_font.tga")))
		return false;

	// --------------------
	// Init Scene
	// --------------------
	CScene::SetScene(new CScene);
	if(!_Scene())
	{
		MessageBox(NULL, "Cannot create scene class instance", "MGLApp::Init", MB_OK|MB_ICONSTOP);
		return false;
	}
	else
	{
		if(!(_Scene()->InitAll()))
			return false;
		if(!(_Scene()->LoadOBJMesh(SCENE_FILENAME)))
			return false;
		if(!(_Scene()->BuildOctreeForScene()))
			return false;
	}

	// --------------------
	// Init Camera
	// --------------------
	CCamera::SetCamera(new CCamera);
	if(!_Camera())
	{
		MessageBox(NULL, "Cannot create camera class instance", "MGLApp::Init", MB_OK|MB_ICONSTOP);
		return false;
	}

	// --------------------
	// Init Timer
	// --------------------
	if(!(m_timer.Init()))
		return false;	

	m_lastTimeFps = 0.0f;
	m_lastFps = 0.0f;

	m_showInfo = true;
	m_renderNodes = true;
	m_renderGeometry = true;
	m_renderSplitOnly = false;
	m_renderWireframe = false;

	glShadeModel(GL_SMOOTH);
	glClearColor( 0.15f, 0.15f, 0.15f, 1.0f );		

	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);			
	glDepthFunc(GL_LEQUAL);
		
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	
	glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);		
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

	glPolygonMode(GL_BACK, GL_LINE);

	// DEFAULT LIGHTING SETUP
	// SET PARAMETERS FOR LIGHT 0
	glLightfv(GL_LIGHT0, GL_POSITION, defLight_position);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  defLight_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, defLight_specular);
	glLightfv(GL_LIGHT0, GL_AMBIENT,  defLight_ambient);
	// SET THE LIGHTING MODEL
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ligthingModel_ambient);
	// ONLY ENABLE LIGHT0, ENABLE LIGHTING ONLY WHEN NEEDED
	glEnable(GL_LIGHT0);
	
	// ****************** @1A ENDS *************************

	return true;
}

bool MGLApp::Shutdown()
{	
	// ****************** @1A BEGINS *************************
	// Cleanup texture manager
	if(_TextureManager())
	{
		_TextureManager()->DestroyAll();
		delete CTextureManager::GetTextureManager();
	}

	// Cleanup scene
	if(_Scene())
	{
		_Scene()->DestroyAll();
		delete CScene::GetScene();
	}

	// Cleanup camera
	if(_Camera())
	{
		_Camera()->DestroyAll();
		delete CCamera::GetCamera();
	}

	// Cleanup font
	m_font.DestroyAll();
	// ****************** @1A ENDS *************************	

	return true;
}

void MGLApp::SetupProjection(int width, int height)
{
	// avoid div by zero
	if( height == 0 )
		height = 1;	

	// set up GL viewport
	glViewport( 0, 0, width, height );

	// set up a perspective projection
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	gluPerspective( 45.0f, (GLfloat)width/(GLfloat)height, 1.0f, 1000.0f );
}

void MGLApp::Update( double dt )
{
}

void MGLApp::Render()
{
	int		visibleNodes=0, totalNodes=0, trianglesRendered=0, totalTriangles=0;	//@1A
	int		splitTriangles=0;														//@1A
	float	posX=0.0f, posY=1.0f, posZ=10.0f, lookX=0.0f, lookY=0.0f, lookZ=0.0f,	//@1A
			upX=0.0f, upY=1.0f, upZ=0.0f;											//@1A
	MGL::Vector3 cameraPos, lookVector, upVector;									//@1A

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// set up the camera
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();	
	
	// ************* @1A BEGINS *****************	
	if(_Camera())
	{
		cameraPos	= _Camera()->Get_currPos();
		posX		= cameraPos.x;
		posY		= cameraPos.y;
		posZ		= cameraPos.z;

		lookVector	= _Camera()->Get_lookAtVector();
		lookX		= lookVector.x;
		lookY		= lookVector.y;
		lookZ		= lookVector.z;

		upVector	= _Camera()->Get_upVector();
		upX			= upVector.x;
		upY			= upVector.y;
		upZ			= upVector.z;
	}

	gluLookAt(  posX, posY, posZ, lookX, lookY, lookZ, upX, upY, upZ );
	//gluLookAt(5.0f, 2.0f, 10.0f, 7.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
	glColor3f(1.0f, 1.0f, 1.0f);

	glPushMatrix();		
		if(_Scene())
			_Scene()->RenderScene(&visibleNodes, &totalNodes, &trianglesRendered, 
									&totalTriangles, m_renderNodes, &splitTriangles,
									m_renderGeometry, m_renderSplitOnly, m_renderWireframe);
	glPopMatrix();	

	// Display info
	RenderInfo(visibleNodes, totalNodes, trianglesRendered, totalTriangles, splitTriangles);
	// ************** @1A ENDS ******************	
}

// ************************** @1A BEGINS ***************************
void MGLApp::Draw2DBegin()
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, 800, 0, 600, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_DEPTH_BUFFER_BIT);	
}

void MGLApp::Draw2DEnd()
{	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void MGLApp::RenderInfo(int visibleNodes, int totalNodes, int trianglesRendered, 
										int totalTriangles, int splitTriangles)
{	
	char	str[80]	= "";
	float	fps=0, minX=0.0f, minY=0.0f, maxX=800.0f, maxY=600.0f;
	float	now = timeGetTime() / 1000.0f;	
	int		yPos = maxY - 20;

	fps = m_timer.GetFPS();		

	if((now - m_lastTimeFps) >= 0.5f)
	{
		m_lastFps = fps;
		m_lastTimeFps = now;
	}
	else
		fps = m_lastFps;

	sprintf(str,"fps: %.2f",fps);

	Draw2DBegin();	

	glPushMatrix();

		m_font.PrintScreen(0.13f, 0.43f, 0.92f, 1.0f, maxX-100, maxY-25, str);
		m_font.PrintScreen(1.0f, 1.0f, 1.0f, 1.0f, 5, 5, "hide/show info: f1");

		if(m_showInfo)
		{			
			m_font.PrintScreen(0.13f, 0.43f, 0.92f, 1.0f, 5, yPos, "octree");
			sprintf(str, "nodes in octree: %d", totalNodes);
			yPos -= 20;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, str);
			sprintf(str, "nodes rendered: %d", visibleNodes);
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, str);	
			sprintf(str, "tris rendered: %d / %d", trianglesRendered, totalTriangles);
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, str);	
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "hide/show empty nodes: f2");
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "hide/show nodes aabb: f3");
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "hide/show geometry: f4");
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "render split triangles only: f5");
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "render wireframe: f6");
			yPos -= 15;
			sprintf(str,"max triangles per node: %d",MAX_TRIS_IN_OCTREE_NODE);
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, str);
			yPos -= 40;

			m_font.PrintScreen(0.13f, 0.43f, 0.92f, 1.0f, 5, yPos, "camera");
			yPos -= 20;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "rotate camera: left btn");	
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "move camera: shift+left btn");	
			yPos -= 15;
			m_font.PrintScreen(0.30f, 0.62f, 0.96f, 1.0f, 5, yPos, "zoom camera: right btn");	
		}		
	
	glPopMatrix();

	Draw2DEnd();
}

// ************************** @1A ENDS ***************************







