/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: Octree.cpp                                                |*/
/*| Author: Joel Villagrana   (joel.villagrana@gmail.com)               |*/
/*|                                                                     |*/
/*| Classes Implemented:                                                |*/
/*| Octree, OctreeNode                                                  |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yy   | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 20/08/05   | JVillagrana       | -Rewritten                         |*/
/*************************************************************************/
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <glee/glee.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <iostream>

#include "Octree.h"


/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
==================================================================================
				IMPLEMENTATION OF THE OCTREENODE METHODS
==================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/

/* =================================================================================
	INIT ALL
==================================================================================== */
bool OctreeNode::InitAll(OC3NODEID id)
{
	m_ID		=	id;
	m_pParent	=	NULL;
	m_pIndices	=	NULL;
	m_numIndices=	0;
	m_size		=	0.0f;
	m_isLeafNode=	false;	

	m_AABB.Clear();
	m_center.Set(0.0f, 0.0f, 0.0f);
	for(int j=0; j<8; ++j)
		m_pSubNodes[j]=NULL;

	return true;
}

/* =================================================================================
	CLEANUP
==================================================================================== */
void OctreeNode::DestroyAll()
{
	// IF IT'S A LEAF NODE
	if(m_isLeafNode)
	{
		// CLEANUP INDICES (IF ANY) IN THIS NODE
		if((m_numIndices>0) && (m_pIndices!=NULL))
		{
			delete [] m_pIndices;
			m_pIndices = NULL;
			m_numIndices = 0;
		}
	}
	else
	{
		// CLEAN UP NODES RECURSIVELY		
		for(int j=0; j<8; ++j)
		{
			if(m_pSubNodes[j]!=NULL)
			{
				m_pSubNodes[j]->DestroyAll();
				delete m_pSubNodes[j];
			}
			m_pSubNodes[j]=NULL;
		}
	}
}

/* =================================================================================
	CALCULATE THE CENTER OF A NEW NODE
==================================================================================== */
void OctreeNode::Get_CenterOfNextNode(MGL::Vector3 currCenter, MGL::Vector3 *newCenter,
														  float width, OC3NODEID nextID)
{
	MGL::Vector3 tempV;
	float w2 = width/2.0f;	

	newCenter->Set(0.0f, 0.0f, 0.0f);	

	if(nextID != ROOT)
	{		
		switch(nextID)
		{
			case TOP_LEFT_FRONT:
				newCenter->Set(currCenter.x-w2, currCenter.y+w2, currCenter.z+w2);
				break;
			case BOTTOM_LEFT_FRONT:
				newCenter->Set(currCenter.x-w2, currCenter.y-w2, currCenter.z+w2);
				break;
			case BOTTOM_RIGHT_FRONT:
				newCenter->Set(currCenter.x+w2, currCenter.y-w2, currCenter.z+w2);
				break;
			case TOP_RIGHT_FRONT:
				newCenter->Set(currCenter.x+w2, currCenter.y+w2, currCenter.z+w2);
				break;
			case TOP_LEFT_BACK:
				newCenter->Set(currCenter.x-w2, currCenter.y+w2, currCenter.z-w2);
				break;
			case BOTTOM_LEFT_BACK:
				newCenter->Set(currCenter.x-w2, currCenter.y-w2, currCenter.z-w2);
				break;
			case BOTTOM_RIGHT_BACK:
				newCenter->Set(currCenter.x+w2, currCenter.y-w2, currCenter.z-w2);
				break;
			case TOP_RIGHT_BACK:
				newCenter->Set(currCenter.x+w2, currCenter.y+w2, currCenter.z-w2);
				break;
		}
	}
	else
		assert("INVALID OCTREE ID NODE, OctreeNode::Get_CentreOfNextNode()");	
}

/* =================================================================================
	RENDER NODE'S BOUNDING BOX
==================================================================================== */
void OctreeNode::RenderNodeAABB(GLFrustum *viewFrustum, bool renderEmptyAABB)
{
	// IF NODE IS VISIBLE
	if(viewFrustum->AABBinFrustum(m_AABB))
	{
		// IF IT'S A LEAF NODE
		if(m_isLeafNode)
		{
			// IF THIS NODE HAS AN INVALID SIZE OR IT'S EMPTY AND WE'RE
			// ASKED NOT TO RENDER EMPTY NODES, RETURN...
			if((m_size<=0) || ((!renderEmptyAABB) && (m_numIndices==0))) return;

			glPushMatrix();
				glTranslatef(m_center.x, m_center.y, m_center.z);
				glBegin(GL_LINES);					
					(m_numIndices>0)?(glColor3ub(248, 191, 36)):(glColor3ub(0, 230, 0));
					// from p1 to p2
					glVertex3f(-m_size, m_size,  -m_size);glVertex3f(-m_size, -m_size, -m_size);
					// from p2 to p3
					glVertex3f(-m_size, -m_size, -m_size);glVertex3f(m_size,  -m_size, -m_size);
					// from p3 to p4
					glVertex3f(m_size,  -m_size, -m_size);glVertex3f(m_size,  m_size,  -m_size);
					// from p4 to p1
					glVertex3f(m_size,  m_size,  -m_size);glVertex3f(-m_size, m_size,  -m_size);
					// from p5 to p6
					glVertex3f(-m_size, m_size,  m_size); glVertex3f(-m_size, -m_size, m_size);
					// from p6 to p7
					glVertex3f(-m_size, -m_size, m_size); glVertex3f(m_size,  -m_size, m_size);
					// from p7 to p8
					glVertex3f(m_size,  -m_size, m_size); glVertex3f(m_size,  m_size,  m_size);
					// from p8 to p5
					glVertex3f(m_size,  m_size,  m_size); glVertex3f(-m_size, m_size,  m_size);
					// from p1 to p5
					glVertex3f(-m_size, m_size,  -m_size);glVertex3f(-m_size, m_size,  m_size);
					// from p4 to p8
					glVertex3f(m_size,  m_size,  -m_size);glVertex3f(m_size,  m_size,  m_size);
					// from p2 to p6
					glVertex3f(-m_size, -m_size, -m_size);glVertex3f(-m_size, -m_size, m_size);
					// from p3 to pp7
					glVertex3f(m_size,  -m_size, -m_size);glVertex3f(m_size,  -m_size, m_size);
				glEnd();
			glPopMatrix();
		}
		else
		{
			// RECURSIVELY RENDER THE SUBNODES OF THE CURRENT NODE
			if( m_pSubNodes[TOP_LEFT_FRONT]!=NULL)
				m_pSubNodes[TOP_LEFT_FRONT]->RenderNodeAABB(viewFrustum, renderEmptyAABB);

			if( m_pSubNodes[BOTTOM_LEFT_FRONT]!=NULL)
				m_pSubNodes[BOTTOM_LEFT_FRONT]->RenderNodeAABB(viewFrustum, renderEmptyAABB);

			if( m_pSubNodes[BOTTOM_RIGHT_FRONT]!=NULL)
				m_pSubNodes[BOTTOM_RIGHT_FRONT]->RenderNodeAABB(viewFrustum, renderEmptyAABB);

			if( m_pSubNodes[TOP_RIGHT_FRONT]!=NULL)
				m_pSubNodes[TOP_RIGHT_FRONT]->RenderNodeAABB(viewFrustum, renderEmptyAABB);

			if( m_pSubNodes[TOP_LEFT_BACK]!=NULL)
				m_pSubNodes[TOP_LEFT_BACK]->RenderNodeAABB(viewFrustum, renderEmptyAABB);

			if( m_pSubNodes[BOTTOM_LEFT_BACK]!=NULL)
				m_pSubNodes[BOTTOM_LEFT_BACK]->RenderNodeAABB(viewFrustum, renderEmptyAABB);

			if( m_pSubNodes[BOTTOM_RIGHT_BACK]!=NULL)
				m_pSubNodes[BOTTOM_RIGHT_BACK]->RenderNodeAABB(viewFrustum, renderEmptyAABB);

			if( m_pSubNodes[TOP_RIGHT_BACK]!=NULL)
				m_pSubNodes[TOP_RIGHT_BACK]->RenderNodeAABB(viewFrustum, renderEmptyAABB);
		} // Not leaf node
	} // Node visible
}

/* =================================================================================
	CREATE A NEW NODE
==================================================================================== */
bool OctreeNode::CreateSubNode(int *numIndices, int *indices, int *numVertices, 
							MGL::Vector3 *vertices, int *nodesCreated, OC3NODEID id, 
							int maxTriangles, int *numNewIndices, int *newIndices, 
									int *numNewVertices, MGL::Vector3 *newVertices )
{
	bool nodeCreated = false;

	assert((id>=0)&&(id<=7));

	// CREATE THE NEW NODE AND INITIALIZE IT	
	m_pSubNodes[id] = new OctreeNode;
	assert(m_pSubNodes[id] != NULL && "OctreeNode::CreateSubNode");

	if(m_pSubNodes[id] != NULL)
	{
		m_pSubNodes[id]->InitAll(id);
	
		// SET SIZE AND CENTER
		MGL::Vector3 newNodeCenter;
		Get_CenterOfNextNode(m_center, &newNodeCenter, m_size, id);

		m_pSubNodes[id]->Set_size((m_size/2.0f));
		m_pSubNodes[id]->Set_center(newNodeCenter);

		// SUBDIVIDE	
		if(m_pSubNodes[id]->SubdivideNode8(numIndices, indices, numVertices, vertices, 
										this, nodesCreated,	maxTriangles, numNewIndices, 
												newIndices, numNewVertices, newVertices))
		{
			nodeCreated = true;
		}
	}

	return nodeCreated;
}

/* =================================================================================
	RECURSIVE FUNCTION THAT CREATES 8 NODES...
==================================================================================== */
bool OctreeNode::SubdivideNode8(int *numIndices, int *indices, int *numVertices, 
			MGL::Vector3 *vertices, OctreeNode *parent, int *nodesCreated, int maxTriangles, 
		int *numNewIndices, int *newIndices, int *numNewVertices, MGL::Vector3 *newVertices)
{
	MGL::Vector3	triVertices[3];
	int		i=0, j=0, k=0, tmpTriCount[8], tmpTriAssigned[8], offset=0;
	int		assignedInd[3], nonSplit=0;
	int		*tmpInd_1=NULL, *tmpInd_2=NULL, *tmpInd_3=NULL, *tmpInd_4=NULL, 
			*tmpInd_5=NULL, *tmpInd_6=NULL, *tmpInd_7=NULL, *tmpInd_8=NULL;
	int		partitionTriangle=0, trianglesCreated=0, verticesCreated=0,indicesCreated=0;
	bool	skipTriangle = false, split = false;
	char	msgStr[150];

	m_pParent = parent;
	m_AABB.InitAll(m_center.x, m_center.y, m_center.z, (m_size*2.0f));

	// IF THE INDICES LIST IS NOT EMPTY
	if((*numIndices) > 0)
	{	
		// DO WE NEED TO KEEP SUBDIVIDING?
		if( ((*numIndices)/3) > maxTriangles)
		{
			// IT WON'T BE A LEAF NODE... SUBDIVIDE...
			m_isLeafNode = false;

			// ========================================================================
			// CHECK WHICH INDICES WILL BE ASSIGNED TO WHICH SUB NODES, SPLIT TRIANGLES
			// AS NEEDED AND CREATE A LIST OF TRIANGLES FOR EACH NEW NODE AND 
			// ASSIGN THE TRIANGLES...
			// ========================================================================
			
			memset(tmpTriCount, 0, sizeof(tmpTriCount));
			memset(tmpTriAssigned, 0, sizeof(tmpTriAssigned));

			// COUNT HOW MANY INDICES WE WILL NEED FOR THE NEW LISTS
			// (COUNT WILL BE A BIT MORE BECAUSE OF THE PARTITIONED TRIANGLES)
			for(i=0; i < (*numIndices); i+=3)
			{
				for(j=i; j<(i+3); j++)
				{
					if(indices[j] < (*numVertices))
						triVertices[j-i] = vertices[indices[j]];
					else
						triVertices[j-i] = newVertices[indices[j] - (*numVertices)];

				}

				partitionTriangle = 0;
				for(j=0; j<8; j++)
				{
					if(TriangleInNode((OC3NODEID)j, triVertices[0], triVertices[1], triVertices[2]))
					{
						tmpTriCount[j]+=3;
						++partitionTriangle;
					}
				}

				// IF TRIANGLE IS IN MORE THAN ONE NODE, PARTITION TRIANGLES
				if(partitionTriangle > 1)
				{
					// PARTITION TRIANGLE
					split = PartitionTriangle(	numIndices, indices, numVertices, vertices,
												numNewIndices, newIndices, 
												numNewVertices, newVertices, 
												triVertices[0], triVertices[1], triVertices[2], 
												&trianglesCreated, &verticesCreated, &indicesCreated );
					if(split)
					{						
						// INVALIDATE ORIGINAL INDEX SO IT'S NOT ASSIGNED TO ANY LIST
						indices[i]	= -1;
						indices[i+1]= -1;
						indices[i+2]= -1;
					}
					else
						nonSplit++;
				}
			}			

			if(nonSplit>0)
			{
				sprintf(msgStr,"%d TRIANGLES WERE SUPPOSED TO BE SPLIT AND THEY WEREN'T", nonSplit);
				MessageBox(NULL, msgStr, "?", MB_OK);
			}

			// FOR THE NEW TRIANGLES CREATED, CHECK WHICH NODE THEY SHOULD BE ASSIGNED TO
			offset = (*numNewIndices);
			for(i=0; i < indicesCreated; i+=3)
			{
				// WHEN THE VALUE AT 'newIndices[j]' IS LESS THAN *numVertices IT MEANS IT'S AN INDEX TO 
				// AN EXISTING VERTEX, OTHERWISE (newIndices[j] >= *numVertices) IT MEANS IT'S AN INDEX TO
				// A NEW VERTEX

				for(j=i; j<(i+3); j++)
				{
					k = newIndices[j + offset];
					if(k < (*numVertices))
						triVertices[j-i] = vertices[k];
					else
					{
						k -= (*numVertices);
						triVertices[j-i] = newVertices[k];
					}
				}

				for(j=0; j<8; j++)
				{
					if(TriangleInNode((OC3NODEID)j, triVertices[0], triVertices[1], triVertices[2]))
						tmpTriCount[j]+=3;
				}
			}

			// ALLOCATE MEMORY FOR INDICES OF THE CHILDREN NODES
			if(tmpTriCount[0]>0)
			{	
				tmpInd_1 = new int [tmpTriCount[0]]; 
				memset(tmpInd_1, 0, (tmpTriCount[0] * sizeof(int)));
			}
			if(tmpTriCount[1]>0)
			{
				tmpInd_2 = new int [tmpTriCount[1]];
				memset(tmpInd_2, 0, (tmpTriCount[1] * sizeof(int)));
			}
			if(tmpTriCount[2]>0)
			{
				tmpInd_3 = new int [tmpTriCount[2]];
				memset(tmpInd_3, 0, (tmpTriCount[2] * sizeof(int)));
			}
			if(tmpTriCount[3]>0)
			{
				tmpInd_4 = new int [tmpTriCount[3]];
				memset(tmpInd_4, 0, (tmpTriCount[3] * sizeof(int)));
			}
			if(tmpTriCount[4]>0)
			{
				tmpInd_5 = new int [tmpTriCount[4]];
				memset(tmpInd_5, 0, (tmpTriCount[4] * sizeof(int)));
			}
			if(tmpTriCount[5]>0)
			{
				tmpInd_6 = new int [tmpTriCount[5]];
				memset(tmpInd_6, 0, (tmpTriCount[5] * sizeof(int)));
			}
			if(tmpTriCount[6]>0)
			{
				tmpInd_7 = new int [tmpTriCount[6]];
				memset(tmpInd_7, 0, (tmpTriCount[6] * sizeof(int)));
			}
			if(tmpTriCount[7]>0)
			{
				tmpInd_8 = new int [tmpTriCount[7]];
				memset(tmpInd_8, 0, (tmpTriCount[7] * sizeof(int)));
			}

			// ASSIGN INDICES FROM BOTH indices AND newIndices
			for(i=0; i < ((*numIndices) + indicesCreated); i+=3)
			{
				skipTriangle = false;

				if(i < (*numIndices))
				{
					// IF ONE OF THE INDICES OF THE LIST IS '-1' IT MEANS TRIANGLE WAS PARTITIONED
					// SO WE NEED TO SKIP IT, NEW CREATED TRIANGLES WILL BE AT THE END OF THE LIST
					// WITH VALID INDICES
					if(indices[i]==-1 || indices[i+1]==-1 || indices[i+2]==-1)
						skipTriangle = true;
				}

				// IF THIS TRIANGLE WAS NOT PARTITIONED, ASSIGN IT
				if( !skipTriangle )
				{		
					// MAKE SURE WE'VE GOT VALID INDICES
					if(i < (*numIndices))
						assert(indices[i]>=0 && indices[i+1]>=0 && indices[i+2]>=0);
					else
					{
						offset = (*numNewIndices) + (i - (*numIndices));
						assert(newIndices[offset]>=0 && newIndices[offset+1]>=0 && newIndices[offset+2]>=0);
					}

					// CHECK IF THE TRIANGLE SHOULD BE COPIED FROM THE LIST WE'VE BEEN PASSED AS PARAMETER
					// OR FROM THE LIST OF NEW TRIANGLES...					
					// ***********************************************************************************
					// THIS LOOKS A BIT WEIRD BUT WHEN THE RECURSION STEP IS > 1, BOTH THE PASSED LIST AND
					// THE NEW LIST CAN HAVE INDICES TO BOTH OLD AND NEW VERTICES
					// ***********************************************************************************
					for(j=i, k=0; j<(i+3); j++, k++)
					{
						// INDEX FROM ORIGINAL LIST
						if(j < (*numIndices)) 
						{						
							// INDEX FROM ORIGINAL LIST POINTING TO VERTEX IN ORIGINAL LIST
							if(indices[j] < (*numVertices))
							{
								triVertices[k] = vertices[indices[j]];
							}
							else // INDEX FROM ORIGINAL LIST POINTING TO VERTEX IN NEW LIST
							{
								offset = indices[j] - (*numVertices);
								triVertices[k] = newVertices[offset];
							}
							assignedInd[k] = indices[j];
						}
						else // INDEX FROM NEW LIST
						{
							offset = (*numNewIndices) + (j - (*numIndices));
							assignedInd[k] = newIndices[offset];

							// INDEX FROM NEW LIST POINTING TO VERTEX IN ORIGINAL LIST
							if(newIndices[offset] < (*numVertices))
							{
								triVertices[k] = vertices[newIndices[offset]];
							}
							else // INDEX FROM NEWL LIST POINTING TO VERTEX IN NEW LIST
							{
								offset = newIndices[offset] - (*numVertices);
								triVertices[k] = newVertices[offset];
							}								
						}
					}
					
					// ASSIGN INDICES
					if(TriangleInNode(TOP_LEFT_FRONT, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[0]>0))
					{					
						offset = tmpTriAssigned[0];
						memcpy(&tmpInd_1[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[0]+= 3;					
					}
					else if(TriangleInNode(BOTTOM_LEFT_FRONT, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[1]>0))
					{
						offset = tmpTriAssigned[1];
						memcpy(&tmpInd_2[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[1]+= 3;
					}
					else if(TriangleInNode(BOTTOM_RIGHT_FRONT, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[2]>0))
					{
						offset = tmpTriAssigned[2];
						memcpy(&tmpInd_3[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[2]+= 3;
					}
					else if(TriangleInNode(TOP_RIGHT_FRONT, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[3]>0))
					{
						offset = tmpTriAssigned[3];
						memcpy(&tmpInd_4[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[3]+= 3;
					}
					else if(TriangleInNode(TOP_LEFT_BACK, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[4]>0))
					{
						offset = tmpTriAssigned[4];
						memcpy(&tmpInd_5[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[4]+= 3;
					}
					else if(TriangleInNode(BOTTOM_LEFT_BACK, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[5]>0))
					{
						offset = tmpTriAssigned[5];
						memcpy(&tmpInd_6[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[5]+= 3;
					}
					else if(TriangleInNode(BOTTOM_RIGHT_BACK, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[6]>0))
					{
						offset = tmpTriAssigned[6];
						memcpy(&tmpInd_7[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[6]+= 3;
					}
					else if(TriangleInNode(TOP_RIGHT_BACK, triVertices[0], triVertices[1], triVertices[2]) && (tmpTriCount[7]>0))
					{
						offset = tmpTriAssigned[7];
						memcpy(&tmpInd_8[offset], assignedInd, 3*sizeof(int));
						tmpTriAssigned[7]+= 3;
					}

				} // if(!skipTriangle)
				
			} // End for

			(*nodesCreated)   += 8;
			(*numNewVertices) += verticesCreated;	// INCREMENT THE NUMBER OF NEW VERTICES CREATED SO FAR
			(*numNewIndices)  += indicesCreated;	// INCREMENT THE NUMBER OF NEW INDICES CREATED SO FAR
	
			// SET UP CHILDREN NODES
			CreateSubNode(&tmpTriAssigned[0], tmpInd_1, numVertices , vertices, nodesCreated, 
							TOP_LEFT_FRONT, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			CreateSubNode(&tmpTriAssigned[1], tmpInd_2, numVertices, vertices, nodesCreated, 
							BOTTOM_LEFT_FRONT, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			CreateSubNode(&tmpTriAssigned[2], tmpInd_3, numVertices, vertices, nodesCreated, 
							BOTTOM_RIGHT_FRONT, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			CreateSubNode(&tmpTriAssigned[3], tmpInd_4, numVertices, vertices, nodesCreated, 
							TOP_RIGHT_FRONT, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			CreateSubNode(&tmpTriAssigned[4], tmpInd_5, numVertices, vertices, nodesCreated, 
							TOP_LEFT_BACK, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			CreateSubNode(&tmpTriAssigned[5], tmpInd_6, numVertices, vertices, nodesCreated, 
							BOTTOM_LEFT_BACK, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			CreateSubNode(&tmpTriAssigned[6], tmpInd_7, numVertices, vertices, nodesCreated, 
							BOTTOM_RIGHT_BACK, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			CreateSubNode(&tmpTriAssigned[7], tmpInd_8, numVertices, vertices, nodesCreated, 
							TOP_RIGHT_BACK, maxTriangles, numNewIndices, newIndices, 
														numNewVertices, newVertices);

			// CLEANUP
			if(tmpTriCount[0]>0 && tmpInd_1!=NULL){ delete [] tmpInd_1; tmpInd_1=NULL; }
			if(tmpTriCount[1]>0 && tmpInd_2!=NULL){ delete [] tmpInd_2; tmpInd_2=NULL; }
			if(tmpTriCount[2]>0 && tmpInd_3!=NULL){ delete [] tmpInd_3; tmpInd_3=NULL; }
			if(tmpTriCount[3]>0 && tmpInd_4!=NULL){ delete [] tmpInd_4; tmpInd_4=NULL; }
			if(tmpTriCount[4]>0 && tmpInd_5!=NULL){ delete [] tmpInd_5; tmpInd_5=NULL; }
			if(tmpTriCount[5]>0 && tmpInd_6!=NULL){ delete [] tmpInd_6; tmpInd_6=NULL; }
			if(tmpTriCount[6]>0 && tmpInd_7!=NULL){ delete [] tmpInd_7; tmpInd_7=NULL; }
			if(tmpTriCount[7]>0 && tmpInd_8!=NULL){ delete [] tmpInd_8; tmpInd_8=NULL; }
		}
		else // ELSE, NO NEED TO KEEP SUBDIVIDING
		{		
			// RECURSION FINISHED, ASSIGN THE INDICES TO THIS NODE			
			m_isLeafNode = true;
			m_numIndices = (*numIndices);
			m_pIndices = new int [m_numIndices];
			assert(m_pIndices != NULL && "OctreeNode::SubdivideNode8");
			memcpy(m_pIndices, indices, (m_numIndices * sizeof(int)));

			// SINCE IT'S A LEAF NODE, SET CHILDREN NODES AS NULL
			for(i=0; i<8; ++i)
				m_pSubNodes[i]=NULL;
		}
	}
	else // ELSE, NUMINDICES IS ZERO
	{
		m_isLeafNode = true;
		m_pIndices = NULL;
		m_numIndices = 0;
		for(i=0; i<8; ++i)
			m_pSubNodes[i]=NULL;
	}

	return true;
}

/* =================================================================================
	PARTITION TRIANGLE
	A bit wacky stuff... but we want to avoid ID tagging hehehe

	*Note*
	numNewIndices and numNewVertices are 'global' to all the process.	
	verticesCreated, indicesCreated, and trianglesCreated are 'local' to the 
	current recursion level only
==================================================================================== */
bool OctreeNode::PartitionTriangle(int *numIndices, int *indices, int *numVertices,
		MGL::Vector3 *vertices,	int *numNewIndices, int *newIndices, int *numNewVertices, 
		MGL::Vector3 *newVertices, MGL::Vector3 v1, MGL::Vector3 v2, MGL::Vector3 v3, 
					int *trianglesCreated, int *verticesCreated, int *indicesCreated)
{
	GLTriangle	theTriangle, 
				*XPlaneTris[3],			//	TRIANGLES CREATED PARTITIONING WITH X PLANE
				*YPlaneTris[3],			//	TRIANGLES CREATED PARTITIONING WITH Y PLANE
				*ZPlaneTris[3];			//	TRIANGLES CREATED PARTITIONING WITH Z PLANE
	GLPlane		XPlane, YPlane, ZPlane;	//	3 PARTITION PLANES
	HALFSPACE	hfX[3], hfY[3], hfZ[3];	//	ENUM VARS
	float		x[3], y[3], z[3];
	int			newTriX=0, newTriY=0, newTriZ=0, i=0, j=0, k=0;
	bool		splitTri=true, addedTris=true;

	// ZERO MEM
	for(i=0; i<3; i++)
	{
		XPlaneTris[i] = NULL;
		YPlaneTris[i] = NULL;
		ZPlaneTris[i] = NULL;
	}

	// SETUP TRIANGLE
	x[0] = v1.x;	y[0] = v1.y;	z[0] = v1.z;
	x[1] = v2.x;	y[1] = v2.y;	z[1] = v2.z;
	x[2] = v3.x;	y[2] = v3.y;	z[2] = v3.z;
	// SETUP TRIANGLE
	theTriangle.Set_Vertices(x, y, z);


	// =============================
	// SETUP THE PARTITIONING PLANES
	// =============================
	
	// SETUP FIRST PLANE (ALONG X AXIS)
	x[0] = m_center.x;			y[0] = m_center.y + m_size;	z[0] = m_center.z;
	x[1] = m_center.x - m_size;	y[1] = m_center.y;			z[1] = m_center.z;
	x[2] = m_center.x + m_size;	y[2] = m_center.y;			z[2] = m_center.z;
	// SETUP PLANE
	XPlane.SetCoefficientsFromPoints(x, y, z);

	// SETUP SECOND PLANE (ALONG Z AXIS)
	x[0] = m_center.x;	y[0] = m_center.y + m_size;	z[0] = m_center.z;
	x[1] = m_center.x;	y[1] = m_center.y;			z[1] = m_center.z - m_size;
	x[2] = m_center.x;	y[2] = m_center.y;			z[2] = m_center.z + m_size;
	// SETUP PLANE
	ZPlane.SetCoefficientsFromPoints(x, y, z);

	// SETUP THIRD PLANE (ALONG Y AXIS)
	x[0] = m_center.x;			y[0] = m_center.y;	z[0] = m_center.z + m_size;
	x[1] = m_center.x + m_size;	y[1] = m_center.y;	z[1] = m_center.z;
	x[2] = m_center.x - m_size;	y[2] = m_center.y;	z[2] = m_center.z;

	// SETUP PLANE
	YPlane.SetCoefficientsFromPoints(x, y, z);
	// -----------------------------------------------------------------------


	// ===============================
	// PARTITION TRIANGLE ALONG X AXIS
	// ===============================
	XPlane.PartitionTriWithPlane(theTriangle, XPlaneTris, &newTriX, hfX);	

	// ==============================================================================
	// IF TRIANGLES WERE CREATED, TRY TO PARTITION EACH ONE WITH THE OTHER TWO PLANES
	// ==============================================================================
	if(newTriX > 0)
	{
		for(i=0; i < newTriX; i++)
		{
			newTriY = 0;

			// GET VERTICES
			XPlaneTris[i]->Get_Vertices(x, y, z);
			// SETUP TRIANGLE
			theTriangle.Set_Vertices(x, y, z);

			// PARTITION THIS NEW TRIANGLE
			YPlane.PartitionTriWithPlane(theTriangle, YPlaneTris, &newTriY, hfY);
		
			// =======================================================================
			// IF NO NEW TRIANGLES WERE CREATED WITH THE SECOND PLANE, TRY THIRD PLANE
			// =======================================================================
			if(newTriY == 0)
			{
				newTriZ = 0;

				// PARTITION THIS TRIANGLE
				ZPlane.PartitionTriWithPlane(theTriangle, ZPlaneTris, &newTriZ, hfZ);

				if(newTriZ == 0)
				{
					// ==================================================
					// THIS TRIANGLE WAS ONLY PARTITIONED WITH THE X AXIS
					// ==================================================

					addedTris = AddNewTriangle(XPlaneTris[i], numIndices, indices, numVertices, vertices,
												numNewVertices, numNewIndices, newVertices,	newIndices, 
													trianglesCreated, verticesCreated, indicesCreated);
				}
				else
				{
					// ========================================================================
					// TRIANGLE CREATED WITH X-AXIS PARTITION WAS PARTITIONED BY Z AXIS AS WELL
					// ASSIGN THE RESULTING TRIANGLES
					// ========================================================================

					for(k=0; k < newTriZ; k++)
					{
						addedTris = AddNewTriangle(ZPlaneTris[k], numIndices, indices, numVertices, 
												vertices, numNewVertices, numNewIndices, newVertices,
										newIndices, trianglesCreated, verticesCreated, indicesCreated);

						delete ZPlaneTris[k];
						ZPlaneTris[k] = NULL;
					}
				} // END newTriZ > 0
			}
			else	
			{	// =====================================================
				// FOR EACH TRIANGLE CREATED, PARTITION WITH THIRD PLANE
				// =====================================================
				for(j=0; j<newTriY; j++)
				{
					newTriZ = 0;

					YPlaneTris[j]->Get_Vertices(x, y, z);
					theTriangle.Set_Vertices(x, y, z);

					// PARTITION THIS TRIANGLE
					ZPlane.PartitionTriWithPlane(theTriangle, ZPlaneTris, &newTriZ, hfZ);

					if(newTriZ == 0)
					{
						// ========================================================
						// THIS TRIANGLE WAS ONLY PARTITIONED WITH THE X AND Y AXIS
						// ========================================================

						addedTris = AddNewTriangle(YPlaneTris[j], numIndices, indices, numVertices, 
												vertices, numNewVertices, numNewIndices, newVertices,
										newIndices, trianglesCreated, verticesCreated, indicesCreated);
					}
					else
					{
						// ========================================================================
						// TRIANGLE WAS PARTITIONED WITH X, Y, AND Z AXIS (PROBABLY WAS NEAR THE
						// CENTRE OF THE NODE...) ASSIGN THE RESULTING TRIANGLES
						// ========================================================================

						for(k=0; k < newTriZ; k++)
						{
							addedTris = AddNewTriangle(ZPlaneTris[k], numIndices, indices, numVertices, 
													vertices, numNewVertices, numNewIndices, newVertices,
											newIndices, trianglesCreated, verticesCreated, indicesCreated);

							delete ZPlaneTris[k];
							ZPlaneTris[k] = NULL;
						}
					}

					delete YPlaneTris[j];
					YPlaneTris[j] = NULL;

				} // END for(j=0; j<newTriY; j++)

			} // END newTriY > 0

			// CLEANUP	
			delete XPlaneTris[i];
			XPlaneTris[i] = NULL;

		} // END for(i=0; i<newTriX; i++)

	} // END if(newTriX > 0)
	else
	{	// =======================================================
		// FIRST PLANE DIDN'T PARTITION TRIANGLE, TRY SECOND PLANE
		// =======================================================
		newTriY = 0;

		// PARTITION THIS NEW TRIANGLE
		YPlane.PartitionTriWithPlane(theTriangle, YPlaneTris, &newTriY, hfY);

		if(newTriY == 0)
		{	
			// =======================================================
			// SECOND PLANE DIDN'T PARTITION TRIANGLE, TRY THIRD PLANE
			// =======================================================
			newTriZ = 0;

			// PARTITION THIS TRIANGLE
			ZPlane.PartitionTriWithPlane(theTriangle, ZPlaneTris, &newTriZ, hfZ);

			if(newTriZ == 0)
			{
				splitTri = false;
				// ======================================================
				// THIRD PLANE DIDN'T PARTITION IT EITHER, THIS SHOULDN'T
				// ACTUALLY HAPPEN... IF WE CALLED THIS METHOD IS BECAUSE
				// WE DETECTED THE TRIANGLE WAS IN MORE THAN ONE NODE
				// ======================================================
			}
			else
			{
				// ===============================================================
				// TRIANGLE WAS PARTITIONED WITH Z AXIS ONLY, ASSIGN NEW TRIANGLES
				// ===============================================================

				for(k=0; k < newTriZ; k++)
				{
					addedTris = AddNewTriangle(ZPlaneTris[k], numIndices, indices, numVertices, vertices,
												numNewVertices, numNewIndices, newVertices,	newIndices, 
													trianglesCreated, verticesCreated, indicesCreated);

					delete ZPlaneTris[k];
					ZPlaneTris[k] = NULL;
				}
			} // END newTriZ > 0
		} // END newTriY == 0
		else
		{	
			for(j=0; j < newTriY; j++)
			{
				newTriZ = 0;

				YPlaneTris[j]->Get_Vertices(x, y, z);
				theTriangle.Set_Vertices(x, y, z);

				// PARTITION THIS TRIANGLE
				ZPlane.PartitionTriWithPlane(theTriangle, ZPlaneTris, &newTriZ, hfZ);

				if(newTriZ == 0)
				{
					// =======================================
					// TRIANGLE WAS PARTITIONED BY Y AXIS ONLY					
					// =======================================
					
					addedTris = AddNewTriangle(YPlaneTris[j], numIndices, indices, numVertices, vertices,
													numNewVertices, numNewIndices, newVertices,	newIndices, 
														trianglesCreated, verticesCreated, indicesCreated);
				}
				else
				{
					// ==========================================
					// TRIANGLE WAS PARTITIONED WITH Y AND Z AXIS
					// ==========================================

					for(k=0; k < newTriZ; k++)
					{
						addedTris = AddNewTriangle(ZPlaneTris[k], numIndices, indices, numVertices, 
												vertices, numNewVertices, numNewIndices, newVertices,
										newIndices, trianglesCreated, verticesCreated, indicesCreated);

						delete ZPlaneTris[k];
						ZPlaneTris[k] = NULL;
					}
				}

				delete YPlaneTris[j];
				YPlaneTris[j] = NULL;

			} // END for(j=0; j<newTriY; j++)

		} // END newTriY > 0

	} // END newTriX == 0

	return (splitTri && addedTris);
}

/* =================================================================================
	ADD A NEW TRIANGLE TO THE LIST
==================================================================================== */
bool OctreeNode::AddNewTriangle(GLTriangle *tri, int *numIndices, int *indices, 
									int *numVertices, MGL::Vector3 *vertices, 
									int *numNewVertices, int *numNewIndices,
									MGL::Vector3 *newVertices, int *newIndices,	
									int *trianglesCreated, int *verticesCreated,int *indicesCreated)
{	
	MGL::Vector3 searchVertex;
	bool	triAdded = true;
	float	x[3], y[3], z[3];
	int		newInd=0, newV=0, vertexIndex=0, i=0, offsetInd=0, offsetVx=0;
	int		maxNewInd=0, maxNewVx=0;

	// GET MAX INDICES/VERTICES WE CAN CREATE
	maxNewInd = m_pParent->Get_MaxNewInd();
	maxNewVx = m_pParent->Get_MaxNewVx();

	// 3 NEW INDICES WILL BE CREATED, CHECK IF THERE'S ROOM ENOUGH FOR THEM
	if(((*numNewIndices) + (*indicesCreated) + 3) < maxNewInd)
	{
		tri->Get_Vertices(x, y, z);

		// ADD EACH VERTEX
		while((i<3) && (triAdded))
		{
			vertexIndex = -1;
			searchVertex.Set(x[i], y[i], z[i]);

			offsetInd = (*numNewIndices) +	// TOTAL NUMBER OF INDICES CREATED SO FAR
						(*indicesCreated) + // INDICES CREATED IN THIS RECURSION STEP
						newInd;				// INDICES CREATED IN THIS LOOP

			offsetVx =  (*numNewVertices) +	// TOTAL NUMBER OF VERTICES CREATED SO FAR
						(*verticesCreated)+	// VERTICES CREATED IN THIS RECURSION STEP
						newV;				// VERTICES CREATED IN THIS LOOP

			// ==============================================
			// IF VERTEX IS IN ORIGINAL LIST, ADD INDEX TO IT
			// ==============================================
			if(IsVertexInList(searchVertex, &vertexIndex, numVertices, vertices))
			{
				newIndices[offsetInd] = vertexIndex;
				newInd++;
			}
			else if(IsVertexInList(searchVertex, &vertexIndex, numNewVertices, newVertices))
			{	
				// =========================================		
				// IF VERTEX IS IN NEW LIST, ADD INDEX TO IT
				// =========================================
				newIndices[offsetInd] = vertexIndex + (*numVertices);
				newInd++;			
			}
			else
			{
				// ===============================================================
				// VERTEX ISN'T IN ANY LIST, CREATE NEW VERTEX AND ADD INDEX TO IT
				// ===============================================================

				// IF THERE'S ROOM ENOUGH FOR NEW VERTEX...
				if(offsetVx < maxNewVx)
				{
					newVertices[offsetVx] = searchVertex;
					newIndices[offsetInd] = (*numVertices) + offsetVx;
					newInd++;
					newV++;
				}
				else
				{
					triAdded = false;
					MessageBox(NULL, "There's no room for more new vertices, try creating the Octree with a higher 'maximum triangles per node' parameter", "Octree creation failed", MB_OK|MB_ICONSTOP);
				}
			}

			i++;
		} // while((i<3) && (triAdded))

		(*trianglesCreated)++;		// INCREMENT THE NUMBER OF TRIANGLES CREATED IN THIS RECURSION
		(*verticesCreated) += newV;	// INCREMENT THE NUMBER OF VERTICES CREATED IN THIS RECURSION
		(*indicesCreated) += newInd;// INCREMENT THE NUMBER OF INDICES CREATED IN THIS RECURSION

	} //if(((*numNewIndices) + (*indicesCreated) + 3) < maxNewInd)
	else
	{
		triAdded = false;
		MessageBox(NULL, "There's no room for more new indices, try creating the Octree with a higher 'maximum triangles per node' parameter", "Octree creation failed", MB_OK|MB_ICONSTOP);
	}

	return triAdded;
}

/* =================================================================================
	SEARCH FOR A VERTEX IN AN EXISTING LIST...
==================================================================================== */
bool OctreeNode::IsVertexInList(MGL::Vector3 newVertex, int *vertexIndex, 
										int *numVertices, MGL::Vector3 *vertices)
{
	MGL::Vector3 thisVertex;
	bool	found=false, sameX=false, sameY=false, sameZ=false;
	int		i=0;

	while((i < (*numVertices)) && !found)
	{
		thisVertex = vertices[i];

		sameX = sameY = sameZ = false;

		if ((fabs(thisVertex.x - newVertex.x)) <= EPSILON)
			sameX = true;
		if ((fabs(thisVertex.y - newVertex.y)) <= EPSILON)
			sameY = true;
		if ((fabs(thisVertex.z - newVertex.z)) <= EPSILON)
			sameZ = true;

		if(sameX && sameY && sameZ)
		{
			found = true;
			*vertexIndex = i;
		}

		i++;
	}

	return found;
}

/* =================================================================================
	CHECK IF A TRIANGLE IS INSIDE A NODE
==================================================================================== */
bool OctreeNode::TriangleInNode(OC3NODEID id, MGL::Vector3 v1, MGL::Vector3 v2, MGL::Vector3 v3)
{
	bool vertex1IN=false, vertex2IN=false, vertex3IN=false, inside=false;

	if(id==TOP_LEFT_FRONT) 
	{
		if((v1.x<m_center.x)&&(v1.y>=m_center.y)&&(v1.z>=m_center.z))vertex1IN=true;
		if((v2.x<m_center.x)&&(v2.y>=m_center.y)&&(v2.z>=m_center.z))vertex2IN=true;
		if((v3.x<m_center.x)&&(v3.y>=m_center.y)&&(v3.z>=m_center.z))vertex3IN=true;
	}
	else if(id==BOTTOM_LEFT_FRONT) 
	{
		if((v1.x<m_center.x)&&(v1.y<m_center.y)&&(v1.z>=m_center.z))vertex1IN=true;
		if((v2.x<m_center.x)&&(v2.y<m_center.y)&&(v2.z>=m_center.z))vertex2IN=true;
		if((v3.x<m_center.x)&&(v3.y<m_center.y)&&(v3.z>=m_center.z))vertex3IN=true;		
	}
	else if(id==BOTTOM_RIGHT_FRONT) 
	{
		if((v1.x>=m_center.x)&&(v1.y<m_center.y)&&(v1.z>=m_center.z))vertex1IN=true;
		if((v2.x>=m_center.x)&&(v2.y<m_center.y)&&(v2.z>=m_center.z))vertex2IN=true;
		if((v3.x>=m_center.x)&&(v3.y<m_center.y)&&(v3.z>=m_center.z))vertex3IN=true;
	}
	else if(id==TOP_RIGHT_FRONT) 
	{
		if((v1.x>=m_center.x)&&(v1.y>=m_center.y)&&(v1.z>=m_center.z))vertex1IN=true;
		if((v2.x>=m_center.x)&&(v2.y>=m_center.y)&&(v2.z>=m_center.z))vertex2IN=true;
		if((v3.x>=m_center.x)&&(v3.y>=m_center.y)&&(v3.z>=m_center.z))vertex3IN=true;
	}
	else if(id==TOP_LEFT_BACK) 
	{
		if((v1.x<m_center.x)&&(v1.y>=m_center.y)&&(v1.z<m_center.z))vertex1IN=true;
		if((v2.x<m_center.x)&&(v2.y>=m_center.y)&&(v2.z<m_center.z))vertex2IN=true;
		if((v3.x<m_center.x)&&(v3.y>=m_center.y)&&(v3.z<m_center.z))vertex3IN=true;
	}
	else if(id==BOTTOM_LEFT_BACK) 
	{
		if((v1.x<m_center.x)&&(v1.y<m_center.y)&&(v1.z<m_center.z))vertex1IN=true;
		if((v2.x<m_center.x)&&(v2.y<m_center.y)&&(v2.z<m_center.z))vertex2IN=true;
		if((v3.x<m_center.x)&&(v3.y<m_center.y)&&(v3.z<m_center.z))vertex3IN=true;
	}
	else if(id==BOTTOM_RIGHT_BACK) 
	{
		if((v1.x>=m_center.x)&&(v1.y<m_center.y)&&(v1.z<m_center.z))vertex1IN=true;
		if((v2.x>=m_center.x)&&(v2.y<m_center.y)&&(v2.z<m_center.z))vertex2IN=true;
		if((v3.x>=m_center.x)&&(v3.y<m_center.y)&&(v3.z<m_center.z))vertex3IN=true;
	}
	else if(id==TOP_RIGHT_BACK) 
	{
		if((v1.x>=m_center.x)&&(v1.y>=m_center.y)&&(v1.z<m_center.z))vertex1IN=true;
		if((v2.x>=m_center.x)&&(v2.y>=m_center.y)&&(v2.z<m_center.z))vertex2IN=true;
		if((v3.x>=m_center.x)&&(v3.y>=m_center.y)&&(v3.z<m_center.z))vertex3IN=true;
	}

	if( vertex1IN || vertex2IN || vertex3IN) 
		inside = true;

	return inside;
}

/* =================================================================================
	GET LIST OF VISIBLE INDICES
==================================================================================== */
void OctreeNode::GetVisibleIndices(	GLFrustum *viewFrustum,	int *numVisibleNodes,	
										int *numVisibleIndices,	int *visibleIndices,
										int maxVisibleIndices)
{
	assert(visibleIndices != NULL && "OctreeNode::GetVisibleIndices");

	// IF NODE IS VISIBLE
	if((viewFrustum->AABBinFrustum(m_AABB)) && (visibleIndices != NULL))
	{
		(*numVisibleNodes)++;

		if(m_isLeafNode)
		{
			if(m_numIndices>0)
			{
				// WE SHOULD HAVE ALLOCATED THE CORRECT AMOUNT, BUT JUST IN CASE
				if((m_numIndices + (*numVisibleIndices)) < maxVisibleIndices)
				{
					memcpy(&visibleIndices[(*numVisibleIndices)], m_pIndices, (m_numIndices * sizeof(int)));
					(*numVisibleIndices) += m_numIndices;
				}
			}
		}
		else
		{
			// CHECK CHILDREN RECURSIVELY
			if( m_pSubNodes[TOP_LEFT_FRONT]!=NULL)
				m_pSubNodes[TOP_LEFT_FRONT]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

			if( m_pSubNodes[BOTTOM_LEFT_FRONT]!=NULL)
				m_pSubNodes[BOTTOM_LEFT_FRONT]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

			if( m_pSubNodes[BOTTOM_RIGHT_FRONT]!=NULL)
				m_pSubNodes[BOTTOM_RIGHT_FRONT]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

			if( m_pSubNodes[TOP_RIGHT_FRONT]!=NULL)
				m_pSubNodes[TOP_RIGHT_FRONT]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

			if( m_pSubNodes[TOP_LEFT_BACK]!=NULL)
				m_pSubNodes[TOP_LEFT_BACK]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

			if( m_pSubNodes[BOTTOM_LEFT_BACK]!=NULL)
				m_pSubNodes[BOTTOM_LEFT_BACK]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

			if( m_pSubNodes[BOTTOM_RIGHT_BACK]!=NULL)
				m_pSubNodes[BOTTOM_RIGHT_BACK]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

			if( m_pSubNodes[TOP_RIGHT_BACK]!=NULL)
				m_pSubNodes[TOP_RIGHT_BACK]->GetVisibleIndices(viewFrustum, numVisibleNodes, 
											numVisibleIndices, visibleIndices, maxVisibleIndices);

		} // NOT A LEAF NODE
	} // NODE IS VISIBLE
}

/* =================================================================================
	GET MAX NEW INDICES
==================================================================================== */
int OctreeNode::Get_MaxNewInd()
{
	int max=0;

	if(m_ID==ROOT)
		max = m_maxTotalNewIndices;
	else if(m_pParent!=NULL)
		max = m_pParent->Get_MaxNewInd();

	return max;		
}

/* =================================================================================
	GET MAX NEW VERTICES
==================================================================================== */
int OctreeNode::Get_MaxNewVx()
{
	int max=0;

	if(m_ID==ROOT)
		max = m_maxTotalNewVertices;
	else if(m_pParent!=NULL)
		max = m_pParent->Get_MaxNewVx();
	
	return max;		
}

/* =================================================================================
	GET TOTAL TRI ASSIGNED
==================================================================================== */
int OctreeNode::GetTotalTrianglesAssigned()
{
	int total = 0;

	if(m_isLeafNode)
		total = m_numIndices;
	else
	{
		// CHECK CHILDREN RECURSIVELY
		for(int j=0; j<8; j++)
		{
			if(m_pSubNodes[j]!=NULL)
				total += m_pSubNodes[j]->GetTotalTrianglesAssigned();
		}

	} // NOT A LEAF NODE

	return total;
}

/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
=====================================================================================
					IMPLEMENTATION OF THE OCTREE METHODS
=====================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/

/* =================================================================================
	INITIALIZE ALL
==================================================================================== */
bool Octree::InitAll()
{
	bool rootCreated = false;

	m_pOctreeRoot		= NULL;
	m_maxTriangles		= 0;
	m_nodesInTree		= 0;
	m_totalTriangles	= 0;
	m_renderEmptyAABB	= true;

	// ALLOCATE MEMORY FOR THE ROOT OF THE OCTREE
	m_pOctreeRoot = new OctreeNode;
	assert(m_pOctreeRoot!=NULL && "Cannot allocate memory for Octree Root Node, Octree::InitAll()");
	
	if(m_pOctreeRoot != NULL)
		rootCreated = true;

	if(rootCreated)
	{
		// INITIALIZE THE NODE
		m_pOctreeRoot->InitAll(ROOT);
	}

	return rootCreated;
}

/* =================================================================================
	 CLEANUP ALL THE NODES OF THE OCTREE
==================================================================================== */
void Octree::DestroyAll()
{	
	// IF THE ROOT IS NOT NULL
	if(m_pOctreeRoot != NULL)
	{
		// RECURSIVELY CLEANUP
		m_pOctreeRoot->DestroyAll();
		delete m_pOctreeRoot;
		m_pOctreeRoot = NULL;
	}
}

/* =================================================================================
	BUILD THE OCTREE
==================================================================================== */
bool Octree::BuildOctree(int *numIndices, int *indices, int *numVertices, 
		MGL::Vector3 *vertices, int *numNewIndices, int *newIndices, int *numNewVertices, 
				MGL::Vector3 *newVertices, int maxTriangles, int maxNewInd,	int maxNewVx)
{	
	OctreeNode		*pRoot=NULL;
	int				*indicesArray=NULL, nodesCreated=0, i=0;
	MGL::Vector3	*verticesArray=NULL, tempCenter, currVertex;
	float			maxDefSize=0.0f, maxSizeX=0.0f, maxSizeY=0.0f, maxSizeZ=0.0f,
					currentSizeX=0.0f, currentSizeY=0.0f, currentSizeZ=0.0f,
					maxX=-10000.0f, maxY=-10000.0f, maxZ=-10000.0f, 
					minX=10000.0f,  minY=10000.0f,  minZ=10000.0f;
	bool octreeCreated = false;

	// MAKE SURE WE HAVE VALID POINTERS
	assert(m_pOctreeRoot != NULL && "Octree root is null, Octree::BuildOctree");

	// MAKE SURE ALL POINTERS ARE OK
	if(m_pOctreeRoot!=NULL && (*numIndices)>0 && (*numVertices)>0 && 
		indices!=NULL && vertices!=NULL && newVertices!=NULL)
	{
		// SET MAX NEW VERTICES/INDICES THAT CAN BE CREATED
		m_pOctreeRoot->Set_MaxNewInd(maxNewInd);
		m_pOctreeRoot->Set_MaxNewVx(maxNewVx);

		// ALLOCATE MEMORY AND MAKE SURE IT DIDN'T FAIL
		indicesArray = new int [(*numIndices)];
		verticesArray = new MGL::Vector3 [(*numVertices)];

		assert(indicesArray!=NULL && verticesArray!=NULL && "Could not allocate memory for indices/vertices");

		if(indicesArray!=NULL && verticesArray!=NULL)
		{
			// CLEAR ALLOCATED MEMORY
			memset(indicesArray, 0, (sizeof(int) * (*numIndices)));
			memset(verticesArray, 0, (sizeof(MGL::Vector3) * (*numVertices)));
			// COPY INDICES AND VERTICES
			memcpy(indicesArray, indices, (sizeof(int) * (*numIndices)));
			memcpy(verticesArray, vertices, (sizeof(MGL::Vector3) * (*numVertices)));		

			// CALCULATE INITIAL CENTER AND ROOT SIZE
			tempCenter.Set(0.0f, 0.0f, 0.0f);
			for(i=0; i<(*numIndices); i++)
			{
				currVertex = verticesArray[indicesArray[i]];
				tempCenter += currVertex;

				if(currVertex.x > maxX) maxX = currVertex.x;
				if(currVertex.x <= minX) minX = currVertex.x;
				if(currVertex.y > maxY) maxY = currVertex.y;
				if(currVertex.y <= minY) minY = currVertex.y;
				if(currVertex.z > maxZ) maxZ = currVertex.z;
				if(currVertex.z <= minZ) minZ = currVertex.z;
			}
			tempCenter /= (*numIndices);
			maxSizeX = fabs(maxX - minX);
			maxSizeY = fabs(maxY - minY);
			maxSizeZ = fabs(maxZ - minZ);

			tempCenter.Set(((maxX + minX)/2.0f), ((maxY + minY)/2.0f), ((maxZ + minZ)/2.0f));

			// SET THE SIZE OF THE ROOT NODE
			if((maxSizeX>=maxSizeY)&&(maxSizeX>=maxSizeZ))maxDefSize = maxSizeX;
			else if((maxSizeY>=maxSizeX)&&(maxSizeY>=maxSizeZ))maxDefSize = maxSizeY;
			else maxDefSize = maxSizeZ;
			maxDefSize /= 2.0f;

			// SET NODE SIZE
			m_pOctreeRoot->Set_size(maxDefSize);
			
			// SET THIS AS THE CENTER OF THE ROOT
			m_pOctreeRoot->Set_center(tempCenter);

			// BUILD THE OCTREE
			pRoot = m_pOctreeRoot;
			if(m_pOctreeRoot->SubdivideNode8(numIndices, indicesArray, numVertices, 
									verticesArray, pRoot, &nodesCreated, maxTriangles, 
									numNewIndices, newIndices, numNewVertices, newVertices))
			{
				m_nodesInTree = ++nodesCreated; // (+1 bcos of the root node)
				octreeCreated = true;
				// GET THE TOTAL NUMBER OF TRIANGLES ASSIGNED TO NODES
				// WE NEED TO DO THIS BECAUSE OLD AND NEW CREATED TRIANGLES
				// COULD HAVE BEEN SPLIT MORE THAN ONCE
				m_totalTriangles = m_pOctreeRoot->GetTotalTrianglesAssigned();
			}
		}

		// CLEANUP
		if(indicesArray!=NULL)
			delete [] indicesArray;
		if(verticesArray!=NULL)
			delete [] verticesArray;
	}
			
	return octreeCreated;
}

/* =================================================================================
	GET VISIBLE INDICES IN OCTREE
==================================================================================== */
void Octree::GetVisibleIndices( GLFrustum *viewFrustum,	int *numVisibleNodes,	
								int *numVisibleIndices,	int *visibleIndices,
								int maxVisibleIndices)
{
	if(m_pOctreeRoot != NULL)
		m_pOctreeRoot->GetVisibleIndices(viewFrustum, numVisibleNodes, 
							numVisibleIndices,	visibleIndices, maxVisibleIndices);
}

/* =================================================================================
	RENDER THE OCTREE NODES AABB
==================================================================================== */
void Octree::RenderVisibleAABBs( GLFrustum *viewFrustum )
{
	if(m_pOctreeRoot != NULL)
		m_pOctreeRoot->RenderNodeAABB(viewFrustum, m_renderEmptyAABB);
}

// -------------------------------- END OF FILE -----------------------------------
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




