/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: Octree.h                                                  |*/
/*| Author: Joel Villagrana  (joel.villagrana@gmail.com)                |*/
/*|                                                                     |*/
/*| Defined Classes :                                                   |*/
/*| Octree, OctreeNode		                                            |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yy   | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 20/08/05   | JVillagrana       | -Rewritten                         |*/
/*************************************************************************/

#ifndef _OCTREE_H
#define _OCTREE_H

#include <windows.h>
#include <stdio.h>
#include <iostream>

#include "../Common/MGLVector.h"
#include "../Math/helper_math.h"                
#include "../Util/helper_defs.h"
#include "../Texture/textureManager.h"

/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
================================================================================
					Definition of the OctreeNode Class
================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
class OctreeNode
{
private:
	OC3NODEID		m_ID;								// NodeID, ie. TOPLEFT, BOTTOMLEFT..., 
	OctreeNode		*m_pSubNodes[8], *m_pParent;		// Pointers to children and parent	
	int				*m_pIndices;						// A list of the indices inside 
	unsigned short	m_numIndices;						// Number of indices inside this node												
	MGL::Vector3	m_center;							// Center position of the node
	float			m_size;								// Size of the node
	bool			m_isLeafNode;						// Is it a leaf node?
	GLBoundingBox	m_AABB;								// Bounding Box of the node
	int				m_maxTotalNewIndices;				//   Maximum new indices/vertices that 
	int				m_maxTotalNewVertices;				// can be created in the Octree construction
														// process
public:
	OctreeNode(){}
	~OctreeNode(){}

	bool	InitAll(OC3NODEID id);						// Init Node
	void	DestroyAll();								// Cleanup

	bool	SubdivideNode8(	int *numIndices,			// Num indices passed to this node
							int *indices,				// Indices
							int *numVertices,			// Number of vertices passed to this node
							MGL::Vector3 *vertices,		// Vertices
							OctreeNode *parent,			// Pointer to parent
							int *nodesCreated,			// Number of nodes created
							int maxTriangles,			// Max triangles allowed in a node
							int *numNewIndices,			// Num of new indices created
							int *newIndices,			// New indices created
							int *numNewVertices,		// Num of new vertices created
							MGL::Vector3 *newVertices);	// New vertices created

	bool	CreateSubNode(	int *numIndices,			// Num indices to be inside sub node
							int *indices,				// Indices
							int *numVertices,			// Num of vertices to be inside sub node
							MGL::Vector3 *vertices,		// Vertices
							int *nodesCreated,			// Number of nodes created
							OC3NODEID id,				// ID of this sub node to be created
							int maxTriangles,			// Max triangles allowed in a node
							int *numNewIndices,			// Num of new indices created
							int *newIndices,			// New indices created
							int *numNewVertices,		// Num of new vertices created
							MGL::Vector3 *newVertices);	// New vertices created
	
	bool	PartitionTriangle(	int *numIndices,		// Number of indices
								int *indices,			// List of indices
								int *numVertices,		// Number of vertices
								MGL::Vector3 *vertices,	// List of vertices
								int *numNewIndices,		// Num of new indices
								int *newIndices,		// Pointer to new indices
								int *numNewVertices,	// Num of new vertices
								MGL::Vector3 *newVertices, // Pointer to new vertices
														// Triangle to be partitioned
								MGL::Vector3 v1,		// Vertex 1
								MGL::Vector3 v2,		// Vertex 2
								MGL::Vector3 v3,		// Vertex 3
								int *trianglesCreated,	// Num of triangles created
								int *verticesCreated,	// Num of vertices created
								int *indicesCreated);	// Num of indices created

	bool	IsVertexInList(	MGL::Vector3 newVertex,		// Candidate vertex 
							int *vertexIndex,			// Vertex index, if exists
							int *numVertices,			// Number of vertices in list
							MGL::Vector3 *vertices);	// List of vertices where to search for

	bool	AddNewTriangle(	GLTriangle *tri,			// New triangle 
							int *numIndices,			// Number of indices
							int *indices,				// List of indices
							int *numVertices,			// Number of vertices
							MGL::Vector3 *vertices,		// List of vertices
							int *numNewVertices,		// Number of new vertices
							int *numNewIndices,			// Number of new indices
							MGL::Vector3 *newVertices,	// List of new vertices
							int *newIndices,			// List of new indices
							int *trianglesCreated,		// Number of triangles created in the current recursion
							int *verticesCreated,		// Number of vertices created in the current recursion
							int *indicesCreated);		// Number of indices created in the current recursion

	void	Get_CenterOfNextNode(MGL::Vector3 currCenter,// Center of current node
								 MGL::Vector3 *newCenter,// New center
								 float width,			// Width
								 OC3NODEID nID);		// ID of the new node

	bool	TriangleInNode(OC3NODEID id,				// Node ID
							MGL::Vector3 v1,			// Vertex 1
							MGL::Vector3 v2,			// Vertex 2
							MGL::Vector3 v3);			// Vertex 3	

	void	RenderNodeAABB(GLFrustum *viewFrustum,		// Current viewing frustum
							bool renderEmptyAABB);		// Render AABB if node's empty?

	void	GetVisibleIndices(	GLFrustum *viewFrustum,	// Current viewing frustum
								int *numVisibleNodes,	// Number of visible nodes
								int *numVisibleIndices,	// Number of visible indices
								int *visibleIndices,	// Visible indices
								int maxVisibleIndices); // Maximum visible indices
	
	int		GetTotalTrianglesAssigned();

	// =========================================================================
	void			Set_ID(OC3NODEID id)		{			m_ID = id;			}
	OC3NODEID		Get_ID()					{	return	m_ID;				}
	int				Get_triangleCount()			{	return	m_numIndices;		}	
	void			Set_center(MGL::Vector3 c)	{			m_center = c;		}
	void			Get_center(MGL::Vector3 *c)	{			*c = m_center;		}
	void			Set_size(float s)			{			m_size = s;			}
	float			Get_size()					{	return	m_size;				}
	bool			IsLeafNode()				{	return	m_isLeafNode;		}
	void			Get_AABB(GLBoundingBox *aabb){			*aabb = m_AABB;		}
	void			Set_MaxNewInd(int ind)		{ m_maxTotalNewIndices = ind;	}
	void			Set_MaxNewVx(int vx)		{ m_maxTotalNewVertices = vx;	}
	int				Get_MaxNewInd();
	int				Get_MaxNewVx();
};
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ 


/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
================================================================================
					Definition of the Octree Class
================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/
class Octree
{
private:
	OctreeNode		*m_pOctreeRoot;						// A pointer to the root of the Octree		
	int				 m_maxTriangles;					// Max triangles in a node
	int				 m_nodesInTree;						// Total number of nodes in Octree
	bool			 m_renderEmptyAABB;					// Render AABB of empty nodes?
	int				 m_totalTriangles;					// Total number of triangles assigned to octree

public:
	Octree(){}
	~Octree(){}

	bool	InitAll();									// Init Octree
	void	DestroyAll();								// Cleanup

	bool	BuildOctree(int *numIndices,				// Number of indices in mesh
						int *indices,					// Mesh indices
						int *numVertices,				// Number of vertices in mesh
						MGL::Vector3 *vertices,			// Mesh vertices
						int *numNewIndices,				// Number of new indices created on octree construction
						int *newIndices,				// Pre-allocated array for new indices
						int *numNewVertices,			// Number of new vertices created on octree construction
						MGL::Vector3 *newVertices,		// Pre-allocated array of new vertices
						int maxTriangles,				// Max triangles in a node	
						int maxNewInd,					// Max new indices that can be created
						int maxNewVx);					// Max new vertices that can be created

	// --------------------------------------------------------------------------
	// Rendering is done in Scene, we only get the list of visible indices
	// --------------------------------------------------------------------------
	void	GetVisibleIndices( GLFrustum *viewFrustum,	// Current viewing frustum
								int *numVisibleNodes,	// Number of visible nodes
								int *numVisibleIndices,	// Number of visible indices
								int *visibleIndices,	// Visible indices
								int maxVisibleIndices); // Max visible indices

	void	RenderVisibleAABBs( GLFrustum *viewFrustum);// Current viewing frustum
	
	//
	// =========================================================================
	void	Set_maxTriangles(int t)		{			m_maxTriangles = t;			}
	int		Get_maxTriangles()			{	return	m_maxTriangles;				}
	int		Get_nodeCount()				{	return	m_nodesInTree;				}
	bool	IsOctreeCreated()			{	return (m_pOctreeRoot != NULL);		}
	void	Set_renderEmptyAABB(bool v)	{			m_renderEmptyAABB = v;		}
	bool	Get_renderEmptyAABB()		{	return	m_renderEmptyAABB;			}
	int		GetTotalTrianglesAssigned()	{	return	m_totalTriangles;			}
};
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ 

#endif _OCTREE_H

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% END OF FILE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


	
