/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: Scene.cpp                                                 |*/
/*| Author: Joel Villagrana   (joel.villagrana@gmail.com)               |*/
/*|                                                                     |*/
/*| Classes Implemented:                                                |*/
/*| CScene                                                              |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yy   | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 11/03/04   | JVillagrana       | -Initial Version                   |*/
/*************************************************************************/

#include <glee/glee.h>

#include "Scene.h"

/*$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
=====================================================================================
				IMPLEMENTATION OF THE CScene CLASS METHODS						  
=====================================================================================
$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$*/

// ==================================================================================
// INITIALIZE
// ==================================================================================
bool CScene::InitAll()
{
	m_indices			= NULL;
	m_newIndices		= NULL;
	m_allIndices		= NULL;
	m_visibleIndices	= NULL;
	m_numIndices		= 0;
	m_numNewIndices		= 0;
	m_allVertices		= NULL;
	m_vertices			= NULL;
	m_newVertices		= NULL;
	m_numVertices		= 0;
	m_numNewVertices	= 0;
	m_vertexNormals		= NULL;
	m_OctreeBuilt		= false;	

	return true;
}

// ==================================================================================
// CLEANUP ALL MEMORY
// ==================================================================================
void CScene::DestroyAll()
{
	if(m_OctreeBuilt)
	{
		m_Octree.DestroyAll();
		m_OctreeBuilt = false;
	}

	if(m_indices!=NULL)
	{
		delete [] m_indices;
		m_indices = NULL;
	}
	if(m_newIndices!=NULL)
	{
		delete [] m_newIndices;
		m_newIndices = NULL;
	}
	if(m_allIndices!=NULL)
	{
		delete [] m_allIndices;
		m_allIndices = NULL;
	}
	if(m_visibleIndices!=NULL)
	{
		delete [] m_visibleIndices;
		m_visibleIndices = NULL;
	}
	if(m_vertices!=NULL)
	{
		delete [] m_vertices;
		m_vertices = NULL;
	}
	if(m_newVertices!=NULL)
	{
		delete [] m_newVertices;
		m_newVertices = NULL;
	}
	if(m_allVertices!=NULL)
	{
		delete [] m_allVertices;
		m_allVertices = NULL;
	}
	if(m_vertexNormals!=NULL)
	{
		delete [] m_vertexNormals;
		m_vertexNormals = NULL;
	}

	m_numIndices	= 0;
	m_numNewIndices = 0;
	m_numVertices	= 0;
	m_numNewVertices= 0;
}

// ==================================================================================
// LOAD MESH FOR OCTREE
// ==================================================================================
bool CScene::LoadOBJMesh(const char filename[])
{
	FILE	*pFile	= NULL;
	char	line[200] = "", c[2], *tmpPtr=NULL, *lastSpace=NULL, val[15];
	int		i=0, j=0, copiedVertices=0, numRead=0, verticesRead=0, indicesRead=0,
			intValues[3], copiedIndices=0;
	float	floatValues[3];
	bool	slashFound=false;

	if(filename[0]!='\0')
	{
		// CLEAN UP EXISTING OBJECTS BEFORE LOADING
		DestroyAll();

		// FAIL IF WE CAN'T OPEN THE FILE
		pFile = fopen(filename, "rb");
		if(!pFile)
		{
			MessageBox(NULL, "Could not open input file", "CScene::LoadOBJMesh", MB_OK);
			return false;
		}

		// FIRST COUNT HOW MANY VERTICES/INDICES WE NEED TO ALLOCATE
		numRead = fread(c, sizeof(char), 1, pFile);
		while(numRead > 0)
		{
			i = 0;
			memset(line,0,sizeof(line));

			while((c[0]!='\n') && (numRead > 0))
			{
				if(c[0]!='\n')
					line[i++] = c[0];
				numRead = fread(c, sizeof(char), 1, pFile);
			}
			if((line[0]=='v')&&(line[1]!='t')&&(line[1]!='n'))
				m_numVertices++;
			else if(line[0]=='f')
				m_numIndices += 3;

			if(numRead > 0)
				numRead = fread(c, sizeof(char), 1, pFile);
		}	
		
		fclose(pFile);

		if((m_numIndices==0) || (m_numVertices==0))
		{
			MessageBox(NULL, "(m_numIndices==0) || (m_numVertices==0)", "CScene::LoadOBJMesh", MB_OK);
			return false;
		}
	
		// ALLOCATE MEMORY
		m_indices = new int [m_numIndices];
		m_newIndices = new int [m_numIndices*4];

		m_vertices = new MGL::Vector3 [m_numVertices];
		m_newVertices = new MGL::Vector3 [m_numVertices*6];
		
		assert(m_indices!=NULL && m_newIndices!=NULL && m_vertices!=NULL && m_newVertices!=NULL);

		if(m_indices==NULL || m_newIndices==NULL || m_vertices==NULL || m_newVertices==NULL)
		{
			if(m_indices!=NULL) delete [] m_indices;
			if(m_newIndices!=NULL) delete [] m_newIndices;
			if(m_vertices!=NULL) delete [] m_vertices;
			if(m_newVertices!=NULL) delete [] m_newVertices;

			MessageBox(NULL, "Could not allocate memory", "CScene::LoadOBJMesh", MB_OK);
			return false;
		}

		// CLEAR MEMORY
		memset(m_indices, 0, (sizeof(int) * m_numIndices));
		memset(m_newIndices, 0, (sizeof(int) * (m_numIndices*4)));

		memset(m_vertices,0, (sizeof(MGL::Vector3) * m_numVertices));		
		memset(m_newVertices,0, (sizeof(MGL::Vector3) * (m_numVertices*6)));

		// OPEN FILE AGAIN TO READ VERTICES AND INDICES
		pFile = fopen(filename, "rb");
		if(!pFile)
		{
			MessageBox(NULL, "Could not open input file to get indices/vertices", "CScene::LoadOBJMesh", MB_OK);
			return false;
		}

		// READ VERTICES/INDICES
		numRead = fread(c, sizeof(char), 1, pFile);
		while(numRead > 0)
		{
			i = 0;
			verticesRead = 0;
			indicesRead = 0;
			lastSpace = NULL;

			memset(line,0,sizeof(line));			
			memset(intValues,0,sizeof(intValues));
			memset(floatValues,(float)(0),sizeof(floatValues));

			while((c[0]!='\n') && (numRead > 0))
			{
				if((c[0]!='\n') && (c[0]!=13) && (c[0]!=10))
					line[i++] = c[0];
				numRead = fread(c, sizeof(char), 1, pFile);
			}
			line[i]='\0';
			tmpPtr = line;

			while((*tmpPtr != ' ')&&(*tmpPtr != '\0'))
				tmpPtr++;
			
			if(*tmpPtr==' ')
			{
				lastSpace = tmpPtr;
				tmpPtr++;
				j = 1;

				// READING A VERTEX ('V'), NOT VERTEX TEXTURE ('VT') NOR VERTEX NORMAL ('VN')
				if((line[0]=='v')&&(line[1]!='t')&&(line[1]!='n'))
				{
					while((verticesRead<3)&&(*tmpPtr!='\0'))
					{
						memset(val,0,sizeof(val));
						while((*tmpPtr != ' ')&&(*tmpPtr != '\0'))
						{
							j++;
							tmpPtr++;
						}
						if(j>1)
						{
							lastSpace++;
							memcpy(val, lastSpace, (j-1));
							floatValues[verticesRead++] = atof(val);
							lastSpace = tmpPtr++;
							j = 1;
						}
					}

					m_vertices[copiedVertices++].Set(floatValues[0], floatValues[1], floatValues[2]);
				}
				else if(line[0]=='f') // READING FACE INDICES
				{
					while((indicesRead<3)&&(*tmpPtr!='\0'))
					{
						memset(val,0,sizeof(val));
						slashFound = false;
						while((*tmpPtr != ' ')&&(*tmpPtr != '\0'))
						{
							if(*tmpPtr == '/')
								slashFound = true;

							// WE COULD HAVE 'FACE INDEX / TEXTURE COORD INDEX / VERTEX NORMAL INDEX'
							// WE'RE ONLY READING FACE INDICES IN THIS CASE...
							if(!slashFound) 
								j++;

							tmpPtr++;
						}
						if(j>1)
						{
							lastSpace++;
							memcpy(val, lastSpace, (j-1));
							intValues[indicesRead++] = atoi(val);
							lastSpace = tmpPtr++;
							j = 1;
						}
					}

					// OBJ FILES STORE REFERENCES TO VERTICES STARTING AT 1
					// (VERTEX NUMBER INSTEAD OF VERTEX INDEX)
					// SO WE NEED TO MAKE SURE WE GET THE INDICES OK...
					m_indices[copiedIndices++] = (intValues[0] - 1);
					m_indices[copiedIndices++] = (intValues[1] - 1);
					m_indices[copiedIndices++] = (intValues[2] - 1);
				}
			}
			
			if(numRead > 0)
				numRead = fread(c, sizeof(char), 1, pFile);
		}

		assert((copiedVertices==m_numVertices) && (copiedIndices==m_numIndices));

		fclose(pFile);
	}

	return true;
}

// ==================================================================================
// RENDER SCENE
// ==================================================================================
void CScene::RenderScene(int *visibleNodes, int *totalNodes, int *trianglesRendered, 
						int *totalTriangles, bool renderNodes, int *splitTriangles, 
						bool renderGeometry, bool renderSplitOnly, bool renderWireframe)
{	
	GLMatrix44	projMatrix, mvMatrix;
    float		projection[16]; 
    float		modelview[16];  
	int			numVisibleNodes=0, numVisibleIndices=0, maxVisibleIndices=0;

	// GET CURRENT MATRICES
    glGetFloatv( GL_PROJECTION_MATRIX, projection ); 
    glGetFloatv( GL_MODELVIEW_MATRIX, modelview );
	projMatrix.Set(projection);
	mvMatrix.Set(modelview);

	// UPDATE VIEWING FRUSTUM WITH CURRENT MATRICES
	m_viewFrustum.Update(mvMatrix, projMatrix);
	
	// IF OCTREE WAS BUILT
	if((m_OctreeBuilt)&&(m_visibleIndices!=NULL))
	{	
		maxVisibleIndices = m_numIndices + m_numNewIndices;

		// ===================================================================
		// PERFORM FRUSTUM CULLING AND GET THE VISIBLE INDICES FROM THE OCTREE
		// ===================================================================
		
		memset(m_visibleIndices,0,((m_numIndices+m_numNewIndices) * sizeof(int)));
		m_Octree.GetVisibleIndices(&m_viewFrustum, &numVisibleNodes, &numVisibleIndices, 
												m_visibleIndices, maxVisibleIndices);

		if(numVisibleNodes > 0)
			*visibleNodes = numVisibleNodes - 1; //(we do not count the root node)
		else
			*visibleNodes = 0;

		*totalNodes		= m_Octree.Get_nodeCount() - 1; //(we do not count the root node)		
		*splitTriangles	= (m_numNewIndices/3);
		*totalTriangles	= (m_Octree.GetTotalTrianglesAssigned()) / 3;

		if(renderGeometry)
		{
			*trianglesRendered	= (numVisibleIndices/3);

			glColor3f(0.2f, 0.3f, 0.6f);
			glEnable(GL_LIGHTING);


			// =================================
			// RENDER WITH VERTEX ARRAYS 
			// =================================

			// ENABLE VERTEX AND NORMAL ARRAYS
			glEnableClientState(GL_VERTEX_ARRAY);
			glEnableClientState(GL_NORMAL_ARRAY);

			// SET VERTEX AND NORMAL POINTER 
			// *************************** IMPORTANT NOTE *****************************
			// (STRIDE IS ONLY 3*sizeof(float), NOT 6*sizeof(float) BECAUSE DATA INSIDE 
			// MGL::VECTOR3 IS A 'UNION'), IF YOU USE 6*sizeof(float) YOU'LL GET A NICE
			// MEMORY ERROR...
			// ************************************************************************
			glVertexPointer(3, GL_FLOAT, 3*sizeof(float), &m_allVertices[0]);
			glNormalPointer(GL_FLOAT, 3*sizeof(float), &m_vertexNormals[0]);


			// =====================
			// RENDER!
			// =====================

			// *************************** IMPORTANT NOTE 2 ****************************
			// THIRD PARAMETER OF glDrawElements IS GL_UNSIGNED_INT, BUT THE ARRAY OF
			// INDICES IS OF TYPE INT, THIS DOESN'T MATTER ON WINDOWS SINCE IT SEEMS
			// UNSIGNED INT AND INT ARE BOTH THE SAME SIZE, BUT I'M NOT SURE ON LINUX...
			// *************************************************************************
			if(!renderSplitOnly)
			{				
				// THIS RENDERS ONLY VISIBLE INDICES GOTTEN FROM OCTREE
				glDrawElements(GL_TRIANGLES, numVisibleIndices, GL_UNSIGNED_INT, m_visibleIndices);

				if(renderWireframe)
				{
					// RENDER WHITE WIREFRAME ON TOP
					glDisable(GL_LIGHTING);
					glColor3f(1.0f, 1.0f, 1.0f);
					glPolygonMode(GL_FRONT, GL_LINE);
						glBegin(GL_TRIANGLES);
							for(int k=0; k<numVisibleIndices; k++)
								glArrayElement(m_visibleIndices[k]);
						glEnd();
					glPolygonMode(GL_FRONT, GL_FILL);

					// FOR TESTING PURPOSES... RENDER NORMALS...
					MGL::Vector3 v1, v2;
					glColor3f(1.0f, 0.0f, 0.0f);
					glBegin(GL_LINES);
						for(k=0; k<numVisibleIndices; k++)
						{							
							v1 = m_allVertices[m_visibleIndices[k]];
							v2 = v1 + m_vertexNormals[m_visibleIndices[k]] * 0.09f;							
							glVertex3fv(v1.v);
							glVertex3fv(v2.v);
						}							
					glEnd();
				}
			}
			else
			{
				// note this is an estimate, not all new indices were assigned to nodes...
				*totalTriangles	= (m_numNewIndices/3); 

				// THIS RENDERS ONLY NEW TRIANGLES CREATED...
				glDrawElements(GL_TRIANGLES, m_numNewIndices, GL_UNSIGNED_INT, m_newIndices);

				if(renderWireframe)
				{
					glDisable(GL_LIGHTING);
					glColor3f(1.0f, 1.0f, 1.0f);
					glPolygonMode(GL_FRONT, GL_LINE);
						glBegin(GL_TRIANGLES);
							for(int k=0; k<m_numNewIndices; k++)
								glArrayElement(m_newIndices[k]);
						glEnd();
					glPolygonMode(GL_FRONT, GL_FILL);
				}
			}			

			// FINALLY DISABLE VERTEX AND NORMAL ARRAYS
			glDisableClientState(GL_NORMAL_ARRAY);
			glDisableClientState(GL_VERTEX_ARRAY);

			glDisable(GL_LIGHTING);
		}
		else
			*trianglesRendered	= 0;

		// =================================
		// RENDER AABBs
		// =================================

		if(renderNodes)
			m_Octree.RenderVisibleAABBs(&m_viewFrustum);
	}
}

// ==================================================================================
// BUILD OCTREE
// ==================================================================================
bool CScene::BuildOctreeForScene()
{
	// IF ALREADY BUILT, CLEANUP FIRST
	if(m_OctreeBuilt)
	{
		m_Octree.DestroyAll();
		m_OctreeBuilt = false;
	}
	
	if(!m_Octree.InitAll())
	{
		MessageBox(NULL, "Could not init Octree", "CScene::BuildOctreeForScene", MB_OK);
		return false;
	}

	// *NOTE* - newIndices and newVertices were pre-allocated with at least twice the
	// original capacity as m_indices and m_vertices	
	if(!m_Octree.BuildOctree(&m_numIndices, m_indices, &m_numVertices, m_vertices,
							&m_numNewIndices, m_newIndices, &m_numNewVertices, 
							m_newVertices, MAX_TRIS_IN_OCTREE_NODE, (m_numIndices*4),
							(m_numVertices*6)))
	{
		m_Octree.DestroyAll();
		m_OctreeBuilt = false;
		MessageBox(NULL, "Could not create Octree", "CScene::BuildOctreeForScene", MB_OK);
		return false;
	}	
	else
	{
		m_OctreeBuilt = true;
		
		// MAX VISIBLE INDICES
		m_visibleIndices = new int [m_numIndices + m_numNewIndices];
		assert(m_visibleIndices!=NULL && "CScene::BuildOctreeForScene");

		// CREATE A LIST TO MERGE BOTH LISTS OF VERTICES
		m_allVertices = new MGL::Vector3 [m_numVertices + m_numNewVertices];
		assert(m_allVertices!=NULL && "CScene::BuildOctreeForScene");
		memcpy(m_allVertices, m_vertices, (m_numVertices * sizeof(MGL::Vector3)));
		memcpy(&m_allVertices[m_numVertices], m_newVertices, (m_numNewVertices * sizeof(MGL::Vector3)));

		// CREATE A LIST TO MERGE BOTH LISTS OF INDICES
		m_allIndices = new int [m_numIndices + m_numNewIndices];
		assert(m_allIndices!=NULL && "CScene::BuildOctreeForScene");
		memcpy(m_allIndices, m_indices, (m_numIndices * sizeof(int)));
		memcpy(&m_allIndices[m_numIndices], m_newIndices, (m_numNewIndices * sizeof(int)));

		// JUST IN CASE, MAKE SURE ALL OUR NEW INDICES ARE VALID
		for(int k=0; k<(m_numIndices + m_numNewIndices); k++)
		{
			assert(m_allIndices[k] < (m_numVertices + m_numNewVertices));
		}

		// AT THIS POINT, WE WILL MAKE REFERENCE ONLY TO m_allVertices, SO IT'S SAFE TO RELEASE
		// MEMORY OF m_vertices AND m_newVertices
		// THIS IS NOT THE CASE OF THE INDICES... AS WE HAVE A FEW OPTIONS TO DISPLAY DIFFERENT
		// SETS OF TRIANGLES...
		if(m_vertices!=NULL)
		{
			delete [] m_vertices;
			m_vertices = NULL;
		}
		if(m_newVertices!=NULL)
		{
			delete [] m_newVertices;
			m_newVertices = NULL;
		}

		// CALCULATE NORMALS FOR THE TRIANGLES		
		CalculateNormalsForMesh();

		// BY DEFAULT, RENDER AABB OF EMPTY NODES
		m_Octree.Set_renderEmptyAABB(true);
	}

	return true;
}

// ==================================================================================
// CALCULATE NORMALS
// ==================================================================================
void CScene::CalculateNormalsForMesh()
{
	MGL::Vector3 *faceNormals=NULL;
	int i=0, j=0;

	if(m_vertexNormals!=NULL)
	{
		delete [] m_vertexNormals;
		m_vertexNormals = NULL;
	}

	// ALLOCATE TEMP MEMORY FOR FACE NORMALS
	faceNormals = new MGL::Vector3 [((m_numIndices + m_numNewIndices)/3)];
	assert(faceNormals!=NULL && "CalculateNormalsForMesh");

	// ALLOCATE MEMORY FOR VERTEX NORMALS
	m_vertexNormals = new MGL::Vector3 [m_numVertices + m_numNewVertices];
	assert(m_vertexNormals!=NULL && "CalculateNormalsForMesh");

	// CALCULATE NORMALS FOR FACE
	for(i=0, j=0; i<(m_numIndices + m_numNewIndices); i+=3, j++)
		CalculateNormalForFace(i, j, faceNormals);

	// NOW CALCULATE NORMALS FOR VERTICES
	for(i=0; i<(m_numVertices + m_numNewVertices); i++)
		CalculateNormalForVertex(i, faceNormals);

	// FINISHED CALCULATING NORMALS, CLEANUP
	if(faceNormals!=NULL)
	{
		delete [] faceNormals;
		faceNormals = NULL;
	}
}

// ==================================================================================
// CALCULATE THE NORMAL FOR ONE FACE
// ==================================================================================
void CScene::CalculateNormalForFace(int i, int j, MGL::Vector3 *faceNormals)
{	
	MGL::Vector3 v1, v2, v3, v1v2, v1v3, normal;

	v1 = m_allVertices[m_allIndices[i]];
	v2 = m_allVertices[m_allIndices[i+1]];
	v3 = m_allVertices[m_allIndices[i+2]];

	v1v2 = v2 - v1;
	v1v3 = v3 - v1;

	normal = v1v2.CrossProduct(v1v3);
	normal.Normalize();

	faceNormals[j] = normal;
}

// ==================================================================================
// CALCULATE THE NORMAL FOR ONE VERTEX
// ==================================================================================
void CScene::CalculateNormalForVertex(int vertexIndex, MGL::Vector3 *faceNormals)
{
	int *faceIndices=NULL, numFaceIndices=0, indicesFound=0, j=0;
	MGL::Vector3 newNormal, faceNormal;
	float nx=0.0f, ny=0.0f, nz=0.0f;

	// First count the number of faces that share this vertex
	numFaceIndices = GetNumberOfFacesSharingVertex(vertexIndex);	

	// if at least two faces shares this vertex...	
	if(numFaceIndices > 0)
	{
		faceIndices = new int[numFaceIndices];
		assert(faceIndices != NULL);
		memset(faceIndices, 0, ((sizeof(int))*numFaceIndices));

		// now fill the array with the face indices sharing the vertex
		indicesFound = GetFaceIndicesSharingVertex(vertexIndex, faceIndices, numFaceIndices);

		if(numFaceIndices == indicesFound)
		{
			// Average face normals that share the selected vertices
			for(j=0; j<numFaceIndices; ++j)
			{
				// get face normal
				faceNormal = faceNormals[faceIndices[j]/3];								

				nx += faceNormal.x;
				ny += faceNormal.y;
				nz += faceNormal.z;
			}

			// average values
			nx /= (float)(numFaceIndices);
			ny /= (float)(numFaceIndices);
			nz /= (float)(numFaceIndices);

			// set values and normalize
			newNormal.Set(nx, ny, nz);
			newNormal.Normalize();

			// set new normal
			m_vertexNormals[vertexIndex] = newNormal;
		}

		// free array memory
		delete [] faceIndices;
		faceIndices = NULL;
	} // end numFaceIndices >= 2
}

// ==================================================================================
// Get the number of faces sharing a vertex
// ==================================================================================
int CScene::GetNumberOfFacesSharingVertex(int vertexIndex)
{
	int count=0, i=0;

	for(i=0; i<(m_numIndices + m_numNewIndices); i+=3)
	{
		if(	(m_allIndices[i]  == vertexIndex) ||
			(m_allIndices[i+1]== vertexIndex) ||
			(m_allIndices[i+2]== vertexIndex) )
		{		
			++count;
		}
	}

	return count;
}

// ==================================================================================
// Get the face indices of the faces sharing the specified vertex
// ==================================================================================
int CScene::GetFaceIndicesSharingVertex(int vertexIndex, int *indices, int maxF)
{
	int count=0, i=0;	

	for(i=0; i<(m_numIndices + m_numNewIndices); i+=3)
	{
		if(	(m_allIndices[i]  == vertexIndex) ||
			(m_allIndices[i+1]== vertexIndex) ||
			(m_allIndices[i+2]== vertexIndex) )
		{
			if(count<maxF)
			{
				indices[count] = i;				
			}	
			++count;
		}
	}
	return count;
}
 
// -------------------------------- END OF FILE -----------------------------------
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%









