/*************************************************************************/
/*|=====================================================================|*/
/*| Filename: helper_math.cpp                                           |*/
/*| Author: Joel Villagrana   (joel.villagrana@gmail.com)               |*/
/*|                                                                     |*/
/*| Classes Implemented:                                                |*/
/*| -GLLineSeg2D		                                                |*/
/*| -GLVector2                                                          |*/
/*| -GLVector3                                                          |*/
/*| -GLVector4                                                          |*/
/*| -GLPlane                                                            |*/
/*| -GLMatrix44                                                         |*/
/*| -GLBoundingBox                                                      |*/
/*| -GLQuaternion                                                       |*/
/*| -GLBoundingSphere                                                   |*/
/*| -GLRay                                                              |*/
/*| -GLPoint2D                                                          |*/
/*| -GLFrustum                                                          |*/
/*| -GLTriangle                                                         |*/
/*|                                                                     |*/
/*|%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%|*/
/*| dd/mm/yy   | Programmer        | Modifications                      |*/ 
/*|=====================================================================|*/
/*| 12/03/04   | JVillagrana       | Initial Version                    |*/
/*|---------------------------------------------------------------------|*/
/*| 24/03/04   | JVillagrana   @1  | - Added class GLTriangle for BSP   |*/
/*|            |                   |   purposes                         |*/
/*************************************************************************/


#include <windows.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glee/glee.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glaux.h>
#include <float.h>

#include "helper_math.h"

#define ERROR_TOLERANCE 1e-6

// Calculate distance between 2 vertices
GLfloat distanceSquare2Vertices(GLVector4 v1, GLVector4 v2);
GLfloat distance2Points(GLVector3 v1, GLVector3 v2);

// Calculate midPoint between 2 vertices
GLVector4 midPoint2Vertices(GLVector4 v1, GLVector4 v2);

// Calculate the closest point on a line
GLVector4 ClosestPointOnLine(GLVector4 vA, GLVector4 vB, GLVector4 vPoint);

// This returns an intersection point of a polygon and a line (assuming intersects the plane)
GLVector4 IntersectionPoint(GLVector4 vNormal, GLVector4 sPoint, GLVector4 ePoint, double distance);

// This returns true if the intersection point is inside of the polygon
//bool InsidePolygon(GLVector4 vIntersection, CVector3 Poly[], long verticeCount);

// Use this function to test collision between a line and polygon
//bool IntersectedPolygon(CVector3 vPoly[], CVector3 vLine[], int verticeCount);


// Compare 2 vertices
BOOL sameVertex(GLVector4 v1, GLVector4 v2);

// Rotation that will take you from V0 to v1
GLQuaternion RotationArc( GLVector4& v0, GLVector4 v1);

// Smooth Linear Interpolate
GLQuaternion Slerp(GLQuaternion& from, GLQuaternion& to, GLfloat interp);

// Set the LookAt Matrix
//GLMatrix44 SetLookAt(GLVector4& camPos, GLVector4& target, GLVector4& camUp);

// Check if a number is power of 2
bool isPowerOf2(int number);
// Ranged Random
float RangedRandom(float v1, float v2);
// Int Random
int RangedRandomInt(int v1, int v2);

// Functions from O'Rourke Book (Computational Geometry in C++)
// USED FOR POINT IN POLYGON TESTS
int		Area2Int(GLPoint2D a, GLPoint2D b, GLPoint2D c);
float	Area2Float(GLPoint2D a, GLPoint2D b, GLPoint2D c);
bool	Left(GLPoint2D a, GLPoint2D b, GLPoint2D c);
bool	LeftOn(GLPoint2D a, GLPoint2D b, GLPoint2D c);
bool	Collinear(GLPoint2D a, GLPoint2D b, GLPoint2D c);
bool	XOR(bool x, bool y);
bool	Between(GLPoint2D a, GLPoint2D b, GLPoint2D c);
bool	IntersectProp(GLPoint2D a, GLPoint2D b, GLPoint2D c, GLPoint2D d);
bool	Intersect(GLPoint2D a, GLPoint2D b, GLPoint2D c, GLPoint2D d);

// Functions to test Point In Polygon when reduced to 2D domain
bool	Overlaps_XZ(GLfloat clickX, GLfloat clickZ, GLVector4 v0, GLVector4 v1, GLVector4 v2);
bool	Overlaps_XY(GLfloat clickX, GLfloat clickY, GLVector4 v0, GLVector4 v1, GLVector4 v2);
bool	Overlaps_YZ(GLfloat clickY, GLfloat clickZ, GLVector4 v0, GLVector4 v1, GLVector4 v2);

// Determines if an edge from v1 to v2 intersects plane
bool	EdgeIntersectsPlane(GLVector4 v1, GLVector4 v2, GLPlane plane, GLVector4 *intersPt);

// Max Min functions
float vmax(float a, float b);
float vmin(float a, float b);

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLMatrix44 CLASS METHODS							   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLMatrix44::GLMatrix44()
{
	for(int i=0; i<16; ++i)elem[i]=0.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLMatrix44::GLMatrix44( GLfloat a1, GLfloat a2, GLfloat a3, GLfloat a4,
						GLfloat a5, GLfloat a6, GLfloat a7, GLfloat a8, 
						GLfloat a9, GLfloat a10, GLfloat a11, GLfloat a12, 
						GLfloat a13, GLfloat a14, GLfloat a15, GLfloat a16)
{
	elem[0]  = a1;  elem[1]  = a2;  elem[2]  = a3;  elem[3]  = a4;
	elem[4]  = a5;  elem[5]  = a6;  elem[6]  = a7;  elem[7]  = a8;
	elem[8]  = a9;  elem[9]  = a10; elem[10] = a11; elem[11] = a12;
	elem[12] = a13; elem[13] = a14; elem[14] = a15; elem[15] = a16;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLMatrix44::Set( GLfloat *e )
{
	elem[0] = e[0];  elem[1] = e[1];  elem[2] = e[2];  elem[3] = e[3];
	elem[4] = e[4];  elem[5] = e[5];  elem[6] = e[6];  elem[7] = e[7];
	elem[8] = e[8];  elem[9] = e[9];  elem[10]= e[10]; elem[11]= e[11];
	elem[12]= e[12]; elem[13]= e[13]; elem[14]= e[14]; elem[15]= e[15];
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TRANSPOSE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLMatrix44::Transpose(GLMatrix44 *destMatrix)
{
	GLfloat e[16];
	e[0] = elem[0];	e[1] = elem[4]; e[2] = elem[8]; e[3] = elem[12];
	e[4] = elem[1];	e[5] = elem[5]; e[6] = elem[9]; e[7] = elem[13];
	e[8] = elem[2];	e[9] = elem[6]; e[10]= elem[10];e[11]= elem[14];
	e[12]= elem[3];	e[13]= elem[7]; e[14]= elem[11];e[15]= elem[15];

	destMatrix->Set(e);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLMatrix44::~GLMatrix44()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// RESET THE MATRIX TO AN IDENTITY MATRIX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLMatrix44::Op_Identity()
{
	for(int i=0; i<16; ++i)elem[i]=0.0f;
	elem[0]=elem[5]=elem[10]=elem[15]=1.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE = OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLMatrix44::operator =(GLMatrix44 mt)
{
	for(int i=0; i<16; ++i)elem[i]=mt.elem[i];
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE + OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLMatrix44 GLMatrix44::operator +(GLMatrix44& mt)
{
	return GLMatrix44(	elem[0]+mt.elem[0],	  elem[1]+mt.elem[1],   elem[2]+mt.elem[2],   elem[3]+mt.elem[3], 
						elem[4]+mt.elem[4],	  elem[5]+mt.elem[5],   elem[6]+mt.elem[6],   elem[7]+mt.elem[7],
						elem[8]+mt.elem[8],	  elem[9]+mt.elem[9],   elem[10]+mt.elem[10], elem[11]+mt.elem[11],
						elem[12]+mt.elem[12], elem[13]+mt.elem[13], elem[14]+mt.elem[14], elem[15]+mt.elem[15] );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE - OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLMatrix44 GLMatrix44::operator -(GLMatrix44& mt)
{
	return GLMatrix44(	elem[0]-mt.elem[0],	  elem[1]-mt.elem[1],   elem[2]-mt.elem[2],   elem[3]-mt.elem[3], 
						elem[4]-mt.elem[4],	  elem[5]-mt.elem[5],   elem[6]-mt.elem[6],   elem[7]-mt.elem[7],
						elem[8]-mt.elem[8],	  elem[9]-mt.elem[9],   elem[10]-mt.elem[10], elem[11]-mt.elem[11],
						elem[12]-mt.elem[12], elem[13]-mt.elem[13], elem[14]-mt.elem[14], elem[15]-mt.elem[15] );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE * OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLMatrix44::operator *(GLVector4& mt)
{
	GLfloat x, y, z, n;

	x = elem[0]*mt.x + elem[4]*mt.y + elem[8]*mt.z  + elem[12]*mt.n;
	y = elem[1]*mt.x + elem[5]*mt.y + elem[9]*mt.z  + elem[13]*mt.n;
	z = elem[2]*mt.x + elem[6]*mt.y + elem[10]*mt.z + elem[14]*mt.n;
	n = elem[3]*mt.x + elem[7]*mt.y + elem[11]*mt.z + elem[15]*mt.n;

	return GLVector4( x, y, z, n );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE * OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3 GLMatrix44::operator *(GLVector3& mt)
{
	GLfloat x, y, z;

	x = elem[0]*mt.x + elem[3]*mt.y + elem[6]*mt.z;
	y = elem[1]*mt.x + elem[4]*mt.y + elem[7]*mt.z;
	z = elem[2]*mt.x + elem[5]*mt.y + elem[8]*mt.z;

	return GLVector3( x, y, z );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE [] OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLMatrix44::operator [](int i)
{
	assert( (i>=0) && (i<=15) && "Error in GLMatrix44::operator[]" );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE *= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLMatrix44::operator *=(GLMatrix44& m2)
{
	GLMatrix44 newMatrix;

	newMatrix.elem[ 0] = elem[ 0] * m2.elem[ 0] + elem[ 1] * m2.elem[ 4] + elem[ 2] * m2.elem[ 8] + elem[ 3] * m2.elem[12];
	newMatrix.elem[ 1] = elem[ 0] * m2.elem[ 1] + elem[ 1] * m2.elem[ 5] + elem[ 2] * m2.elem[ 9] + elem[ 3] * m2.elem[13];
	newMatrix.elem[ 2] = elem[ 0] * m2.elem[ 2] + elem[ 1] * m2.elem[ 6] + elem[ 2] * m2.elem[10] + elem[ 3] * m2.elem[14];
	newMatrix.elem[ 3] = elem[ 0] * m2.elem[ 3] + elem[ 1] * m2.elem[ 7] + elem[ 2] * m2.elem[11] + elem[ 3] * m2.elem[15];

	newMatrix.elem[ 4] = elem[ 4] * m2.elem[ 0] + elem[ 5] * m2.elem[ 4] + elem[ 6] * m2.elem[ 8] + elem[ 7] * m2.elem[12];
	newMatrix.elem[ 5] = elem[ 4] * m2.elem[ 1] + elem[ 5] * m2.elem[ 5] + elem[ 6] * m2.elem[ 9] + elem[ 7] * m2.elem[13];
	newMatrix.elem[ 6] = elem[ 4] * m2.elem[ 2] + elem[ 5] * m2.elem[ 6] + elem[ 6] * m2.elem[10] + elem[ 7] * m2.elem[14];
	newMatrix.elem[ 7] = elem[ 4] * m2.elem[ 3] + elem[ 5] * m2.elem[ 7] + elem[ 6] * m2.elem[11] + elem[ 7] * m2.elem[15];

	newMatrix.elem[ 8] = elem[ 8] * m2.elem[ 0] + elem[ 9] * m2.elem[ 4] + elem[10] * m2.elem[ 8] + elem[11] * m2.elem[12];
	newMatrix.elem[ 9] = elem[ 8] * m2.elem[ 1] + elem[ 9] * m2.elem[ 5] + elem[10] * m2.elem[ 9] + elem[11] * m2.elem[13];
	newMatrix.elem[10] = elem[ 8] * m2.elem[ 2] + elem[ 9] * m2.elem[ 6] + elem[10] * m2.elem[10] + elem[11] * m2.elem[14];
	newMatrix.elem[11] = elem[ 8] * m2.elem[ 3] + elem[ 9] * m2.elem[ 7] + elem[10] * m2.elem[11] + elem[11] * m2.elem[15];

	newMatrix.elem[12] = elem[12] * m2.elem[ 0] + elem[13] * m2.elem[ 4] + elem[14] * m2.elem[ 8] + elem[15] * m2.elem[12];
	newMatrix.elem[13] = elem[12] * m2.elem[ 1] + elem[13] * m2.elem[ 5] + elem[14] * m2.elem[ 9] + elem[15] * m2.elem[13];
	newMatrix.elem[14] = elem[12] * m2.elem[ 2] + elem[13] * m2.elem[ 6] + elem[14] * m2.elem[10] + elem[15] * m2.elem[14];
	newMatrix.elem[15] = elem[12] * m2.elem[ 3] + elem[13] * m2.elem[ 7] + elem[14] * m2.elem[11] + elem[15] * m2.elem[15];

	*this = newMatrix;
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLVector2 CLASS METHODS							   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2::GLVector2()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2::GLVector2( GLVector2& v )
{
	x = v.x;
	y = v.y;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2::GLVector2( GLfloat _x, GLfloat _y )
{
	x = _x;
	y = _y;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2::~GLVector2()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NORMALIZE VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector2::normalize()
{
	GLfloat thisLength = length();

	assert(thisLength!=0 && "Division by 0 error in GLVector3::normalize()");

	x /= thisLength;
	y /= thisLength;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET VALUES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector2::setValues(GLfloat _x, GLfloat _y)
{
	x = _x;
	y = _y;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE + OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2 GLVector2::operator +(GLVector2& v){
	return GLVector2(x+v.x, y+v.y);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE += OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector2::operator +=(GLVector2& v)
{
	x+=v.x;
	y+=v.y;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE - OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2 GLVector2::operator -(GLVector2& v)
{
	return GLVector2(x-v.x, y-v.y);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE - OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2 GLVector2::operator -(GLfloat c)
{
	return GLVector2(x-c, y-c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE -= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector2::operator -=(GLVector2& v)
{
	x-=v.x;
	y-=v.y;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE = OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector2::operator =(GLVector2& v)
{
	x = v.x;
	y = v.y;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE * OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2 GLVector2::operator *(GLfloat c)
{
	return GLVector2(x*c, y*c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE *= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector2::operator *=(GLfloat c)
{
	x*=c;
	y*=c;
	vecxy[0] = x;
	vecxy[1] = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE == OPERATOR 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLVector2::operator ==(GLVector2& v2)
{
	return ((x==v2.x)&&(y==v2.y));
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE != OPERATOR 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLVector2::operator !=(GLVector2& v2)
{
	if((x==v2.x)&&(y==v2.y))return FALSE;
	else return TRUE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE / OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector2 GLVector2::operator /(GLfloat c)
{
	assert(c!=0);
	return GLVector2(x/c, y/c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE /= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector2::operator /=(GLfloat c)
{
	assert(c!=0);
	x/=c;
	y/=c;
	vecxy[0] = x;
	vecxy[1] = y;
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLVector3 CLASS METHODS							   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3::GLVector3()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3::GLVector3( GLVector3& v )
{
	x = v.x;
	y = v.y;
	z = v.z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3::GLVector3( GLfloat _x, GLfloat _y, GLfloat _z )
{
	x = _x;
	y = _y;
	z = _z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3::~GLVector3()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET VALUES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::setValues(GLfloat _x, GLfloat _y, GLfloat _z)
{
	x = _x;
	y = _y;
	z = _z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET VALUES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::setValues(GLVector3 &v2)
{
	x = v2.x;
	y = v2.y;
	z = v2.z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GET VALUES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::getValues(GLfloat *_x, GLfloat *_y, GLfloat *_z)
{
	*_x = x;
	*_y = y;
	*_z = z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NORMALIZE VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::normalize()
{
	GLfloat thisLength = length();
	//assert(thisLength!=0 && "Division by 0 error in GLVector3::normalize()");

	if((thisLength!=0.0f)&&(thisLength!=1.0f))
	{
		x /= thisLength;
		y /= thisLength;
		z /= thisLength;
	}
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DOT PRODUCT
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat GLVector3::dotProduct(GLVector3& v2)
{
	return (GLfloat)((x*v2.x)+(y*v2.y)+(z*v2.z));
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CROSS PRODUCT
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3 GLVector3::crossProduct(GLVector3& v2)
{
	GLfloat v3x, v3y, v3z;
	
	v3x = y*v2.z - z*v2.y;
	v3y = z*v2.x - x*v2.z;
	v3z = x*v2.y - y*v2.x;

	return GLVector3( v3x, v3y, v3z );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE + OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3 GLVector3::operator +(GLVector3& v){
	return GLVector3(x+v.x, y+v.y, z+v.z);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE += OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::operator +=(GLVector3& v)
{
	x+=v.x;
	y+=v.y;
	z+=v.z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE += OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::operator +=(GLVector4& v)
{
	x+=v.x;
	y+=v.y;
	z+=v.z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE - OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3 GLVector3::operator -(GLVector3& v)
{
	return GLVector3(x-v.x, y-v.y, z-v.z);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE - OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3 GLVector3::operator -(GLfloat c)
{
	return GLVector3(x-c, y-c, z-c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE -= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::operator -=(GLVector3& v)
{
	x-=v.x;
	y-=v.y;
	z-=v.z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE -= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::operator -=(GLVector4& v)
{
	x-=v.x;
	y-=v.y;
	z-=v.z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE = OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::operator =(GLVector3& v)
{
	x = v.x;
	y = v.y;
	z = v.z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE * OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3 GLVector3::operator *(GLfloat c)
{
	return GLVector3(x*c, y*c, z*c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE *= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::operator *=(GLfloat c)
{
	x*=c;
	y*=c;
	z*=c;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE == OPERATOR 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLVector3::operator ==(GLVector3& v2)
{
	return ((x==v2.x)&&(y==v2.y)&&(z==v2.z));
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE != OPERATOR 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLVector3::operator !=(GLVector3& v2)
{
	if((x==v2.x)&&(y==v2.y)&&(z==v2.z))return FALSE;
	else return TRUE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE / OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector3 GLVector3::operator /(GLfloat c)
{
	assert(c!=0);
	return GLVector3(x/c, y/c, z/c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE /= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector3::operator /=(GLfloat c)
{
	assert(c!=0);
	x/=c;
	y/=c;
	z/=c;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DISTANCE BETWEEN TWO VERTICES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat GLVector3::distanceToVertex(GLVector3 &v2)
{
	GLfloat dx, dy, dz, distSq;

	dx	=	v2.x-x;
	dx	*=	dx;
	dy	=	v2.y-y;
	dy	*=	dy;
	dz	=	v2.z-z;
	dz	*=	dz;

	distSq = dx + dy + dz;

	return sqrtf(distSq);
}


// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// PLANE-LINE INTERSECTION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
/*
GLVector3 GLVector3::PlaneLineIntersection(GLVector3 n, float d, GLVector3 p2)
{
	GLVector3 result;
	result.setValues(0.0f, 0.0f, 0.0f);

Vector planelineintersection(Vector n,float d,Vector p1,Vector p2){
	// returns the point where the line p1-p2 intersects the plane n&d
        Vector dif  = p2-p1;
        float dn= n^dif;
        float t = -(d+(n^p1) )/dn;
        return p1 + (dif*t);

	return result;
}
*/

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLVector4 CLASS METHODS							   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4::GLVector4()
{
	x = 0;
	y = 0;
	z = 0;
	n = 1;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4::GLVector4( GLVector4& v )
{
	x = v.x;
	y = v.y;
	z = v.z;
	n = v.n;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4::GLVector4( GLfloat _x, GLfloat _y, GLfloat _z, GLfloat _n )
{
	x = _x;
	y = _y;
	z = _z;
	n = _n;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 4
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4::GLVector4( GLfloat _x, GLfloat _y, GLfloat _z )
{
	x = _x;
	y = _y;
	z = _z;
	n = 1.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4::~GLVector4()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE + OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLVector4::operator +(GLVector4& v){
	return GLVector4(x+v.x, y+v.y, z+v.z, n+v.n);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE + OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLVector4::operator +(GLVector3& v){
	return GLVector4(x+v.x, y+v.y, z+v.z, n);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE += OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::operator +=(GLVector4& v)
{
	x+=v.x;
	y+=v.y;
	z+=v.z;
	n+=v.n;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE += OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::operator +=(GLVector3& v)
{
	x+=v.x;
	y+=v.y;
	z+=v.z;
	n+=0.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE - OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLVector4::operator -(GLVector4& v)
{
	return GLVector4(x-v.x, y-v.y, z-v.z, n-v.n);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE - OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLVector4::operator -(GLfloat c)
{
	return GLVector4(x-c, y-c, z-c, n-c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE -= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::operator -=(GLVector4& v)
{
	x-=v.x;
	y-=v.y;
	z-=v.z;
	n-=v.n;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE = OPERATOR WITH A GLVECTOR3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::operator =(GLVector3& v)
{
	x = v.x;
	y = v.y;
	z = v.z;
	n = 0.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE = OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::operator =(GLVector4& v)
{
	x = v.x;
	y = v.y;
	z = v.z;
	n = v.n;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE * OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLVector4::operator *(GLfloat c){
	return GLVector4(x*c, y*c, z*c, n*c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE *= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::operator *=(GLfloat c)
{
	x*=c;
	y*=c;
	z*=c;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DOT PRODUCT
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat GLVector4::dotProduct(GLVector4& v2)
{
	return (GLfloat)((x*v2.x)+(y*v2.y)+(z*v2.z));
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CROSS PRODUCT
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLVector4::crossProduct(GLVector4& v2)
{
	GLfloat v3x, v3y, v3z;
	
	v3x = y*v2.z - z*v2.y;
	v3y = z*v2.x - x*v2.z;
	v3z = x*v2.y - y*v2.x;

	return GLVector4( v3x, v3y, v3z, 1.0f );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GET THE ANGLE WITH ANOTHER VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat GLVector4::angle(GLVector4& v2)
{
	GLfloat magV1, magV2;
	GLfloat ang;
	GLfloat dotP;

	dotP = dotProduct(v2);

	magV1 = length();
	magV1 = (magV1<0) ? (magV1*(-1)) : magV1;

	magV2 = v2.length();
	magV2 = (magV2<0) ? (magV2*(-1)) : magV2;

	assert(magV1!=0 && magV2!=0 && "Division by 0 error in GLVector4::angle(GLVector4& v2)");

	ang = (GLfloat)(acos((dotP)/(magV1*magV2)));

	// Here we make sure that the angle is not a -1.#IND0000000 number, which means indefinate.
	// If we don't do this check our collision results will sometimes say we are colliding when we aren't.
	// Usually this value is found when the dot product and the maginitude are the same value.
	// Return 0 when this happens.
	if(_isnan(ang))return 0.0f;

	return ang;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NORMALIZE VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::normalize()
{
	GLfloat thisLength = length();
//	assert(thisLength!=0 && "Division by 0 error in GLVector4::normalize()");

	if(thisLength>0.0f)
	{
		x /= thisLength;
		y /= thisLength;
		z /= thisLength;
	}
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE == OPERATOR 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLVector4::operator ==(GLVector4& v2)
{
	return ((x==v2.x)&&(y==v2.y)&&(z==v2.z)&&(n==v2.n));
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE != OPERATOR 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLVector4::operator !=(GLVector4& v2)
{
	if((x==v2.x)&&(y==v2.y)&&(z==v2.z)&&(n==v2.n))return FALSE;
	else return TRUE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE / OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLVector4::operator /(GLfloat c)
{
	assert(c!=0);
	return GLVector4(x/c, y/c, z/c, n/c);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD THE /= OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::operator /=(GLfloat c)
{
	assert(c!=0);
	x/=c;
	y/=c;
	z/=c;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET THE VALUES OF THE VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::setValues(GLfloat _x, GLfloat _y, GLfloat _z )
{
	x = _x;
	y = _y;
	z = _z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// RESET THE VALUES TO ZERO
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::resetValuesToZero()
{
	x = 0.0f;
	y = 0.0f;
	z = 0.0f;
	n = 1.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET VALUES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLVector4::setValues(GLVector4 &v2)
{
	x = v2.x;
	y = v2.y;
	z = v2.z;
	n = v2.n;
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLPlane CLASS METHODS								   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLPlane::GLPlane()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLPlane::GLPlane(GLfloat x[3], GLfloat y[3], GLfloat z[3])
{
	SetCoefficientsFromPoints(x, y, z);	
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CALCULATE THE COEFFICIENTS OF THE PLANE FROM THE 3 VERTICES OF THE TRIANGLE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLPlane::SetCoefficientsFromPoints(GLfloat x[3], GLfloat y[3], GLfloat z[3])
{
	GLVector4 v1, v2, v3, va, vb, normal;
	GLfloat D=0.0f;

	m_pointOnPlane.setValues(x[0], y[0], z[0]);

	v1.setValues(x[0], y[0], z[0]);
	v2.setValues(x[1], y[1], z[1]);
	v3.setValues(x[2], y[2], z[2]);

	// GET A VECTOR FROM v2 to v1
	va = v1 - v2;
	// GET A VECTOR FROM v2 to v3
	vb = v3 - v2;

	// CALCULATE THE NORMAL
	normal = va.crossProduct(vb);
	if((normal.x!=0.0f)&&(normal.y!=0.0f)&&(normal.z!=0.0f))normal.normalize();

	// CALCULATE D
	D = -(normal.dotProduct(v1));

	// SET THE COEFFICIENTS
	Set_Coefficients(normal.x, normal.y, normal.z, D);

	// NORMALIZE THE PLANE
	NormalizeGLPlane();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COPY THE CONTENTS OF A PLANE TO THIS ONE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLPlane::operator =(GLPlane& p2)
{
	m_A = p2.Get_A();
	m_B = p2.Get_B();
	m_C = p2.Get_C();
	m_D = p2.Get_D();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLPlane::GLPlane(GLfloat A, GLfloat B, GLfloat C, GLfloat D)
{
	Set_Coefficients(A, B, C, D);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLPlane::~GLPlane()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET THE COEFFICIENTS
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLPlane::Set_Coefficients(GLfloat A, GLfloat B, GLfloat C, GLfloat D)
{
	m_A = A;
	m_B = B;
	m_C = C;
	m_D = D;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NORMALIZE THE PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLPlane::NormalizeGLPlane()
{
	GLfloat mag;

	mag = sqrtf(m_A * m_A + m_B * m_B + m_C * m_C);
	if(mag==0.0f)mag=1;
	mag = 1/mag;

	m_A *= mag;
	m_B *= mag;
	m_C *= mag;
	m_D *= mag;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CLASSIFY A POINT AS BEING ON THE NEGATIVE OR POSITIVE SIDE OF THE PLANE
// OR ON THE PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
HALFSPACE GLPlane::ClassifyPoint(GLVector4 pt)
{
	GLfloat d;
	
	// IF THE POINT(x,y,z) TIMES THE NORMAL OF THE PLANE(A,B,C) PLUS THE DISTANCE
	// FROM THE PLANE TO THE ORIGIN (D) WILL EQUAL TO 0 IF THE POINT(x,y,z) LIES
	// ON THE PLANE. IF IT'S BEHIND THE PLANE, IT WILL BE A NEGATIVE DISTANCE,
	// IF IT'S IN FRONT OF THE PLANE (THE WAY THE NORMAL IS FACING) IT WILL BE
	// A POSITIVE DISTANCE

	d = m_A*pt.x + m_B*pt.y + m_C*pt.z + m_D;

	if (d < 0) return NEGATIVE; 
	if (d > 0) return POSITIVE; 
	return ON_PLANE; 
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GET THE DISTANCE TO THE POINT
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat GLPlane::SignedDistanceToPoint(GLVector4 pt)
{
	// If the plane is not normalized, returns the signed distance in units
	// of the magnitude of the plane's normal vector

	// Use the plane equation to find the distance (Ax + By + Cz + D = 0)
	// D = -(Ax + By + Cz)

//	return ((GLfloat)(-1)*(m_A*pt.x + m_B*pt.y + m_C*pt.z)); ??
	return ((GLfloat)(m_A*pt.x + m_B*pt.y + m_C*pt.z + m_D));
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST WHETHER AN AXIS ALIGNED BOUNDING BOX IS INSIDE THE PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLPlane::AABBPointsInsidePlane(GLBoundingBox aabb)
{
	GLVector4	point1, point2, point3, point4, point5, point6, point7, point8;

	// Get the 8 points of the Axis Aligned Bounding Box
	point1 = aabb.Get_point1();
	point2 = aabb.Get_point2();
	point3 = aabb.Get_point3();
	point4 = aabb.Get_point4();
	point5 = aabb.Get_point5();
	point6 = aabb.Get_point6();
	point7 = aabb.Get_point7();
	point8 = aabb.Get_point8();

	// Test each point against the plane, if AT LEAST one of the points is 
	// INSIDE the frustum plane (POSITIVE HALFSPACE) then skip the test and consider
	// the aabb inside the frustum

	// Changed != for ==
	if((ClassifyPoint(point1))==POSITIVE)return TRUE;
	if((ClassifyPoint(point2))==POSITIVE)return TRUE;
	if((ClassifyPoint(point3))==POSITIVE)return TRUE;
	if((ClassifyPoint(point4))==POSITIVE)return TRUE;
	if((ClassifyPoint(point5))==POSITIVE)return TRUE;
	if((ClassifyPoint(point6))==POSITIVE)return TRUE;
	if((ClassifyPoint(point7))==POSITIVE)return TRUE;
	if((ClassifyPoint(point8))==POSITIVE)return TRUE;	

	// All the points are outside...
	return FALSE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST WHETHER A POINT IS INSIDE OR OUTSIDE THE PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLPlane::PointInsidePlane(GLVector4 &point)
{
	if((ClassifyPoint(point))==POSITIVE)return TRUE;
	//if((ClassifyPoint(point))==NEGATIVE)return TRUE;
	else return FALSE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST WHETHER A LINE INTERSECTS THE PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLPlane::LineIntersectsPlane(GLVector4 &startPoint, GLVector4 &endPoint)
{
	if( (ClassifyPoint(startPoint)) != (ClassifyPoint(endPoint)) )return TRUE;
	//else if((ClassifyPoint(endPoint))==POSITIVE)return FALSE;
	else return TRUE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GETS THE RESULTING PARTITIONED TRIANGLES OF A TRIANGLE INTERSECTED BY
// THIS PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool GLPlane::PartitionTriWithPlane(GLTriangle tri, GLTriangle* partTris[3], 
														int *numPart, HALFSPACE hf[3])
{
	GLVector4	v1, v2, v3;
	GLVector4	intersectionPt1, intersectionPt2;
	HALFSPACE	side1, side2, side3;
	GLPlane		plane;

	// INITIALIZE TO NULL
	partTris[0] = NULL;
	partTris[1] = NULL;	
	partTris[2] = NULL;

	// GET VERTICES
	tri.Get_Vertices(&v1, &v2, &v3);

	// GET SIDE OF THE VERTICES
	side1 = ClassifyPoint(v1);
	side2 = ClassifyPoint(v2);
	side3 = ClassifyPoint(v3);

	plane.Set_Coefficients(m_A, m_B, m_C, m_D);
	plane.SetPointOnPlane(m_pointOnPlane);

	/* ********************************************************************************* */
	// IF THERE'S AN INTERSECTION POINT FROM FIRST EDGE (v1 -> v2) &&&&&&&&&&&&&&&&&&&&&&&&
	if(EdgeIntersectsPlane(v1, v2, plane, &intersectionPt1))
	{
		// FOUND FIRST INTERSECTION POINT, FIND SECOND

		// FIRST CASE, v2 AND v3 ARE ON THE OTHER SIDE
		if( (side2 != side1) && (side3 == side2) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v1 -> v3
			if(EdgeIntersectsPlane(v1, v3, plane, &intersectionPt2))
			{
				partTris[0] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(intersectionPt2, intersectionPt1, v3);
				hf[1]		= side2;
				partTris[2] = new GLTriangle(v3, intersectionPt1, v2);
				hf[2]		= side2;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...

		} // SECOND CASE, THE OTHER INTERSECTION POINT IS v3
		else if( (side2 != side1) && (side3 == ON_PLANE) )
		{
			intersectionPt2 = v3;

			partTris[0] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
			hf[0]		= side1;
			partTris[1] = new GLTriangle(intersectionPt2, intersectionPt1, v2);
			hf[1]		= side2;
			*numPart	= 2;

		} // LAST CASE,
		else if( (side2 != side1) && (side3 == side1) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v2 -> v3
			if(EdgeIntersectsPlane(v2, v3, plane, &intersectionPt2))
			{
				/*
				partTris[0] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(v1, intersectionPt2, v3);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(intersectionPt1, v2, intersectionPt2);
				hf[2]		= side2;*/
				partTris[0] = new GLTriangle(v1, intersectionPt1, v3);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(v3, intersectionPt1, intersectionPt2);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(intersectionPt2, intersectionPt1, v2);
				hf[2]		= side2;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...
		}
		else return false; // SHOULD NEVER GET TO THIS POINT...

	} // STILL NOT FOUND FIRST INT POINT, TRY THE SECOND EDGE (v2 -> v3) &&&&&&&&&&&&&&&&&&
	/* ********************************************************************************** */
	else if(EdgeIntersectsPlane(v2, v3, plane, &intersectionPt1))
	{
		// FOUND FIRST INTERSECTION POINT, FIND SECOND

		// FIRST CASE, v1 AND v2 ARE ON THE OTHER SIDE
		if( (side2 != side3) && (side1 == side2) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v3 -> v1
			if(EdgeIntersectsPlane(v3, v1, plane, &intersectionPt2))
			{
				/*
				partTris[0] = new GLTriangle(v3, intersectionPt2, intersectionPt1);
				hf[0]		= side3;
				partTris[1] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(intersectionPt1, v1, v2);
				hf[2]		= side2;*/
				partTris[0] = new GLTriangle(v2, intersectionPt1, v1);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(intersectionPt2, intersectionPt1, v3);
				hf[2]		= side3;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...

		} // SECOND CASE, THE OTHER INTERSECTION POINT IS v1
		else if( (side2 != side3) && (side1 == ON_PLANE) )
		{
			intersectionPt2 = v1;

			/*
			partTris[0] = new GLTriangle(v1, intersectionPt1, v3);
			hf[0]		= side3;
			partTris[1] = new GLTriangle(v1, v2, intersectionPt1);
			hf[1]		= side2;*/
			partTris[0] = new GLTriangle(v2, intersectionPt1, v1);
			hf[0]		= side2;
			partTris[1] = new GLTriangle(v1, intersectionPt1, v3);
			hf[1]		= side3;
			*numPart	= 2;

		} // LAST CASE,
		else if( (side2 != side3) && (side1 == side3) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v1 -> v2 ??
			if(EdgeIntersectsPlane(v1, v2, plane, &intersectionPt2))
			{
				/*
				partTris[0] = new GLTriangle(intersectionPt2, v2, intersectionPt1);
				hf[0]		= side2;
				partTris[1] = new GLTriangle(v1, intersectionPt2, intersectionPt1);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(v1, intersectionPt1, v3);
				hf[2]		= side3;*/
				partTris[0] = new GLTriangle(v2, intersectionPt1, intersectionPt2);
				hf[0]		= side2;
				partTris[1] = new GLTriangle(intersectionPt2, intersectionPt1, v1);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(v1, intersectionPt1, v3);
				hf[2]		= side3;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...
		}
		else return false; // SHOULD NEVER GET TO THIS POINT...

	} // STILL NOT FOUND FIRST INT POINT, TRY THE LAST EDGE (v3 -> v1) &&&&&&&&&&&&&&&&&&&&
	/* ********************************************************************************** */
	else if(EdgeIntersectsPlane(v3, v1, plane, &intersectionPt1))
	{
		// FOUND FIRST INTERSECTION POINT, FIND SECOND

		// FIRST CASE, v1 AND v2 ARE ON THE OTHER SIDE
		if( (side3 != side1) && (side1 == side2) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v2 -> v3
			if(EdgeIntersectsPlane(v2, v3, plane, &intersectionPt2))
			{
				/*
				partTris[0] = new GLTriangle(intersectionPt1, intersectionPt2, v3);
				hf[0]		= side3;
				partTris[1] = new GLTriangle(v1, intersectionPt2, intersectionPt1);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(v1, v2, intersectionPt2);
				hf[2]		= side2;*/
				partTris[0] = new GLTriangle(v3, intersectionPt1, intersectionPt2);
				hf[0]		= side3;
				partTris[1] = new GLTriangle(intersectionPt2, intersectionPt1, v2);
				hf[1]		= side2;
				partTris[2] = new GLTriangle(v2, intersectionPt1, v1);
				hf[2]		= side1;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...

		} // SECOND CASE, THE OTHER INTERSECTION POINT IS v2
		else if( (side3 != side1) && (side2 == ON_PLANE) )
		{
			intersectionPt2 = v2;

			/*
			partTris[0] = new GLTriangle(intersectionPt1, intersectionPt2, v3);
			hf[0]		= side3;
			partTris[1] = new GLTriangle(v1, v2, intersectionPt1);
			hf[1]		= side1;*/
			partTris[0] = new GLTriangle(v3, intersectionPt1, v2);
			hf[0]		= side3;
			partTris[1] = new GLTriangle(v2, intersectionPt1, v1);
			hf[1]		= side1;
			*numPart	= 2;

		} // LAST CASE,
		else if( (side3 != side1) && (side2 == side3) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v1 -> v2
			if(EdgeIntersectsPlane(v1, v2, plane, &intersectionPt2))
			{
				/*
				partTris[0] = new GLTriangle(v1, intersectionPt2, intersectionPt1);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(intersectionPt1, intersectionPt2, v2);
				hf[1]		= side2;
				partTris[2] = new GLTriangle(intersectionPt1, v2, v3);
				hf[2]		= side3;*/
				partTris[0] = new GLTriangle(v1, intersectionPt2, intersectionPt1);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(intersectionPt1, intersectionPt2, v3);
				hf[1]		= side3;
				partTris[2] = new GLTriangle(v3, intersectionPt2, v2);
				hf[2]		= side2;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...
		}
		else return false; // SHOULD NEVER GET TO THIS POINT...

	}
	/* ********************************************************************************** */

	return true;
}


/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLFrustum CLASS METHODS							   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLFrustum::GLFrustum()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLFrustum::~GLFrustum()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// UPDATE THE SIX FRUSTUM PLANES WITH THE MATRICES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLFrustum::Update(GLMatrix44 modelViewMatrix, GLMatrix44 projectionMatrix)
{
	GLfloat		a, b, c, d;
	GLPlane		leftP, rightP, bottomP, topP, nearP, farP;
	GLMatrix44	M;

	// Once we have the projection and modelview matrix, the combination will give
	// us the clipping planes

	M  = modelViewMatrix;
	M *= projectionMatrix;


	// EXTRACTION AS DEFINED BY BEN HUMPHREY FROM GAMETUTORIALS.COM
	// ========================================================================
	//
	// Left plane coefficients
	a = (M.elem[3]  + M.elem[0]);
	b = (M.elem[7]  + M.elem[4]);
	c = (M.elem[11] + M.elem[8]);
	d = (M.elem[15] + M.elem[12]);
	leftP.Set_Coefficients(a, b, c, d);
	Set_leftPlane(leftP);

	// Right plane coefficients
	a = (M.elem[3]  - M.elem[0]);
	b = (M.elem[7]  - M.elem[4]);
	c = (M.elem[11] - M.elem[8]);
	d = (M.elem[15] - M.elem[12]);
	rightP.Set_Coefficients(a, b, c, d);
	Set_rightPlane(rightP);

	// Bottom plane coefficients
	a = (M.elem[3]  + M.elem[1]);
	b = (M.elem[7]  + M.elem[5]);
	c = (M.elem[11] + M.elem[9]);
	d = (M.elem[15] + M.elem[13]);
	bottomP.Set_Coefficients(a, b, c, d);
	Set_bottomPlane(bottomP);

	// Top plane coefficients
	a = (M.elem[3]  - M.elem[1]);
	b = (M.elem[7]  - M.elem[5]);
	c = (M.elem[11] - M.elem[9]);
	d = (M.elem[15] - M.elem[13]);
	topP.Set_Coefficients(a, b, c, d);
	Set_topPlane(topP);

	// Near plane coefficients
	a = (M.elem[3]  + M.elem[2] );
	b = (M.elem[7]  + M.elem[6] );
	c = (M.elem[11] + M.elem[10]);
	d = (M.elem[15] + M.elem[14]);
	nearP.Set_Coefficients(a, b, c, d);
	Set_nearPlane(nearP);

	// Far plane coefficients
	a = (M.elem[3]  - M.elem[2] );
	b = (M.elem[7]  - M.elem[6] );
	c = (M.elem[11] - M.elem[10]);
	d = (M.elem[15] - M.elem[14]);
	farP.Set_Coefficients(a, b, c, d);
	Set_farPlane(farP);

	// Let v be a vertex and M a 4x4 projection matrix, transforming the
	// vertex with M results in u
	// After this transformation the point is in homogeneous clipping space
	// In this space, the frustum is an axis aligned bounding box, whose size
	// is API-specific.
	// If the vertex u is inside this box, then the untrasformed vertex v is
	// inside the untransformed viewing frustum
	// The vertex u is inside the box if all these inequalities are true:
	//	-w < x < w
	//	-w < y < w
	//	-w < z < w
	// if -w < x	x is in the inside halfspace of the left clipping plane
	// if x < w		x is in the inside halfspace of the right clipping plane
	// if -w < y	y is in the inside halfspace of the bottom clipping plane
	// if y < w		y is in the inside halfspace of the top clipping plane
	// if -w < z	z is in the inside halfspace of the near clipping plane
	// if z < w		z is in the inside halfspace of the far clipping plane
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST WHETHER AN AXIS ALIGNED BOUNDING BOX IS INSIDE THE FRUSTUM
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLFrustum::AABBinFrustum(GLBoundingBox aabb)
{
	// Test the 8 points of the aabb against the 6 frustum planes, if the 8 points are
	// outside of AT LEAST one plane, then the aabb is outside the frustum
	
	// TEST AGAINST THE LEFT PLANE
	
	if(!m_leftPlane.AABBPointsInsidePlane(aabb))return FALSE;
	// TEST AGAINST THE RIGHT PLANE
	if(!m_rightPlane.AABBPointsInsidePlane(aabb))return FALSE;
	// TEST AGAINST THE TOP PLANE
	if(!m_topPlane.AABBPointsInsidePlane(aabb))return FALSE;
	// TEST AGAINST THE BOTTOM PLANE
	if(!m_bottomPlane.AABBPointsInsidePlane(aabb))return FALSE;
	// TEST AGAINST THE NEAR PLANE
	if(!m_nearPlane.AABBPointsInsidePlane(aabb))return FALSE;
	// TEST AGAINST THE FAR PLANE
	if(!m_farPlane.AABBPointsInsidePlane(aabb))return FALSE; 

	return TRUE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST WHETHER A BOUNDING SPHERE IS INSIDE THE FRUSTUM
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLFrustum::BSinFrustum(GLBoundingSphere bs)
{
	GLfloat		radius, distance;
	GLVector4	center;
	GLPlane		plane;
	
	// IF THE DISTANCE TO THE CENTER IS NEGATIVE, THE SPHERE CENTRE IS INSIDE THE
	// FRUSTUM
	// IF THE DISTANCE TO THE CENTER IS POSITIVE AND GREATER THAN THE RADIUS THEN
	// THE SPHERE IS OUTSIDE THE FRUSTUM
	// IF THE DISTANCE TO THE CENTER IS POSITIVE BUT LESS OR EQUAL THAN THE RADIUS
	// THE SPHERE IS INDIDE THE FRUSTUM

	center = bs.Get_Center();
	radius = bs.Get_Radius();

	// TEST THE SPHERE AGAINS THE LEFT PLANE
	plane	 = m_leftPlane;
	plane.NormalizeGLPlane();
	distance = plane.SignedDistanceToPoint(center);
	if(distance <= -radius)return FALSE;
	//if(distance>radius)return FALSE;


	// TEST THE SPHERE AGAINS THE RIGHT PLANE
	plane	 = m_rightPlane;
	plane.NormalizeGLPlane();
	distance = plane.SignedDistanceToPoint(center);
	if(distance <= -radius)return FALSE;
	//if(distance>radius)return FALSE;

	// TEST THE SPHERE AGAINS THE TOP PLANE
	plane	 = m_topPlane;
	plane.NormalizeGLPlane();
	distance = plane.SignedDistanceToPoint(center);
	if(distance <= -radius)return FALSE;
	//if(distance>radius)return FALSE;

	// TEST THE SPHERE AGAINS THE BOTTOM PLANE
	plane	 = m_bottomPlane;
	plane.NormalizeGLPlane();
	distance = plane.SignedDistanceToPoint(center);
	if(distance <= -radius)return FALSE;
	//if(distance>radius)return FALSE;

	// TEST THE SPHERE AGAINS THE NEAR PLANE
	plane	 = m_nearPlane;
	plane.NormalizeGLPlane();
	distance = plane.SignedDistanceToPoint(center);
	if(distance <= -radius)return FALSE;
	//if(distance>radius)return FALSE;

	// TEST THE SPHERE AGAINS THE FAR PLANE
	plane	 = m_farPlane;
	plane.NormalizeGLPlane();
	distance = plane.SignedDistanceToPoint(center);
	if(distance <= -radius)return FALSE;
	//if(distance>radius)return FALSE;

	return TRUE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST WHETHER A SINGLE POINT IS INSIDE THE VIEWING FRUSTUM
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLFrustum::PointInFrustum(GLfloat x, GLfloat y, GLfloat z)
{
	HALFSPACE result;
	GLVector4 point;

	point.setValues(x, y, z);

	// Calculate the plane equation to check the point position
	result = m_leftPlane.ClassifyPoint(point);
	if(result==NEGATIVE)return FALSE;
	
	result = m_rightPlane.ClassifyPoint(point);
	if(result==NEGATIVE)return FALSE;

	result = m_topPlane.ClassifyPoint(point);
	if(result==NEGATIVE)return FALSE;

	result = m_bottomPlane.ClassifyPoint(point);
	if(result==NEGATIVE)return FALSE;

	result = m_nearPlane.ClassifyPoint(point);
	if(result==NEGATIVE)return FALSE;

	result = m_farPlane.ClassifyPoint(point);
	if(result==NEGATIVE)return FALSE;

	return TRUE;
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLQuaternion CLASS METHODS						   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

/*
	Game Programming Gems 1
	===================================================================================
	If we consider all rotation matrices to represent a rotation of an angle @
	about an axis A(XA, YA, ZA), the Quaternion would be

	Q = [ vXA, vYA, vZa, s ]

	v = sin(@/2)
	s = cos(@/2)  @ = 2*(acos(s))

	It is easy to extract the angle of rotation as being twice the inverse cosine
	of the w term. It is similarly easy to extract the axis of rotation

	Quaternion multiplication takes the form:
	q'' = [w,v][w',v'] = [ww' - v.v', vXv' + wv' + w'v]


	http://http.cs.berkeley.edu/~laura/cs184/quat/quaternion.html
	==================================================================================
	We will compute a rotation about the unit vector, u by an angle theta. 
	The quaternion that computes this rotation is:

	q = [s, v]
	
	v = u*sin(@/2)
	s = cos(@/2)

	@ = 2*acos(s)
	u = v/sin(@/2)

	We will represent a point p in space by the quaternion P=(0,p) We compute 
	the desired rotation of that point by this formula: 

	P rotated = qPq-1

	You may want to confirm that q is a unit quaternion, since that will allow us 
	to use the fact that the inverse of q is q' if q is a unit quaternion. 

	Suppose q1 and q2 are unit quaternions representing two rotations. We want to 
	perform q1 first and then q2. To do this, we apply q2 to the result of q1, 
	regroup the product using associativity, and find that the composite rotation 
	is represented by the quaternion q2*q1. 
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion::GLQuaternion()
{
	 w = 1.0f;
	 x = 0.0f;
	 y = 0.0f;
	 z = 0.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// RESET THE QUATERNION TO THE IDENTITY QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::Reset()
{
	 w = 1.0f;
	 x = 0.0f;
	 y = 0.0f;
	 z = 0.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion::GLQuaternion(GLfloat _w, GLfloat _x, GLfloat _y, GLfloat _z)
{
	w = _w;
	x = _x;
	y = _y;
	z = _z;

	Normalize();		// changed
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR, INITIALIZATING WITH A 4X4 MATRIX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion::GLQuaternion(GLMatrix44& glM)
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR, INITIALIZATING WITH ANOTHER QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion::GLQuaternion(GLQuaternion& q2)
{
	w = q2.w;
	x = q2.x;
	y = q2.y;
	z = q2.z;

	Normalize();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR, INITIALIZATING WITH A UNIT VECTOR AND AN ANGLE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion::GLQuaternion(GLVector4& u, GLfloat angle)
{
	// Make sure the length of the vector is not 0
	GLfloat len = u.length();
	assert( len!=0 && "Quaternion Axis Cannot Be Zero!" );

	// Convert the angle to radians
	GLfloat angleRad = (GLfloat)(angle*PI_OVER_180);

	// One way...
	GLfloat s = sinf(angleRad/2);
	u.normalize();

	// Another way... this way the vector will be normalized when is multiplied by s
	// GLfloat s = sinf(angleRad*0.5f)/len;

	Set_Values( cosf(angleRad/2), u.x * s, u.y * s, u.z * s );

	Normalize();	// changed
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR, INITIALIZATING WITH EULER ANGLES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion::GLQuaternion( GLfloat pitch, GLfloat yaw, GLfloat roll )
{
	Set_Euler( pitch, yaw, roll );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion::~GLQuaternion()
{
	Reset();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COMPUTE THE QUATERNION VALUES WITH EULER ANGLES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::Set_Euler(GLfloat pitch, GLfloat yaw, GLfloat roll)
{
	pitch = (GLfloat)(pitch*(PI_OVER_180));
	yaw	  = (GLfloat)(yaw  *(PI_OVER_180));
	roll  = (GLfloat)(roll *(PI_OVER_180));

	GLfloat cosP = cosf(pitch / 2.0f);
	GLfloat sinP = sinf(pitch / 2.0f);

	GLfloat cosY = cosf(yaw   / 2.0f);
	GLfloat sinY = sinf(yaw   / 2.0f);

	GLfloat cosR = cosf(roll  / 2.0f);
	GLfloat sinR = sinf(roll  / 2.0f);

	// optimize a bit multiplications
	GLfloat cosPcosY = cosP*cosY;
	GLfloat sinPsinY = sinP*sinY;
	GLfloat sinPcosY = sinP*cosY;
	GLfloat cosPsinY = cosP*sinY;

	// THIS WORKS.... DON'T CHANGE!
	Set_Values(
		cosR*cosPcosY + sinR*sinPsinY,	//W
		cosR*sinPcosY + sinR*cosPsinY,	//X
		cosR*cosPsinY - sinR*sinPcosY,	//Y
		sinR*cosPcosY - cosR*sinPsinY	//Z
	);

	/* @REMOVED TEMPORARILY (ROTATION WASN'T WORKING IN PERSPECTIVE)
	... for some reason, x,y,z are interchanged... but it's wrong...
	Set_Values(
		cosR*cosPcosY + sinR*sinPsinY,	//W
		sinR*cosPcosY - cosR*sinPsinY,	//X
		cosR*sinPcosY + sinR*cosPsinY,	//Y
		cosR*cosPsinY - sinR*sinPcosY	//Z
	); */

	/*
	Set_Values(
		cosR*cosP*cosY + sinR*sinP*sinY,
		cosR*sinP*cosY + sinR*cosP*sinY,
		cosR*cosP*sinY - sinR*sinP*cosY,
		sinR*cosP*cosY - cosR*sinP*sinY
	); */

	Normalize(); // changed
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET THE w, x, y, z VALUES IN THE QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::Set_Values(GLfloat _w, GLfloat _x, GLfloat _y, GLfloat _z)
{
	w = _w;
	x = _x;
	y = _y;
	z = _z;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD = OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::operator =(GLQuaternion& q)
{
	w = q.w;
	x = q.x;
	y = q.y;
	z = q.z;

	Normalize();	// changed
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// OVERLOAD / OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion GLQuaternion::operator /(GLfloat c)
{
	assert(c!=0 && "Division by 0 error in GLQuaternion::operator/(GLfloat c)");
	return GLQuaternion( w/c, x/c, y/c, z/c );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// MULTIPLY BY A QUATERNION OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion GLQuaternion::operator *(GLQuaternion& q2)
{
	GLfloat cw, cx, cy, cz;

	// LIKE IN THE REAL-TIME RENDERING BOOK (only reordered operands)
	cx = (y*q2.z) - (z*q2.y) + (q2.w*x) + (w*q2.x);		
	cy = (z*q2.x) - (x*q2.z) + (q2.w*y) + (w*q2.y);		
	cz = (x*q2.y) - (y*q2.x) + (q2.w*z) + (w*q2.z);		
	cw = (w*q2.w) - (x*q2.x) - (y*q2.y) - (z*q2.z);		

//	cx = w*q2.x + x*q2.w + y*q2.z - z*q2.y;
//	cy = w*q2.y + y*q2.w + z*q2.x - x*q2.z;
//	cz = w*q2.z + z*q2.w + x*q2.y - y*q2.x;
//	cw = w*q2.w - x*q2.x - y*q2.y - z*q2.z;

	return GLQuaternion( cw, cx, cy, cz );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// MULTIPLY BY A SCALAR OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion GLQuaternion::operator *(GLfloat s)
{
	return GLQuaternion( w*s, x*s, y*s, z*s );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// MULTIPLY BY A QUATERNION AND ASSIGN OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::operator *=(GLQuaternion& q)
{
	GLQuaternion temp = *this * q;
	*this = temp;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// MULTIPLY BY A SCALAR AND ASSIGN OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::operator *=(GLfloat s)
{
	Set_Values( w*s, x*s, y*s, z*s );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// ADD AND ASSIGN OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::operator +=(GLQuaternion& q)
{
	Set_Values( w+q.w, x+q.x, y+q.y, z+q.z );	
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SUBSTRACT AND ASSIGN OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::operator -=(GLQuaternion& q)
{
	Set_Values( w-q.w, x-q.x, y-q.y, z-q.z );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// ADD OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion GLQuaternion::operator +(GLQuaternion& q)
{
	return GLQuaternion( w+q.w, x+q.x, y+q.y, z+q.z );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CALCULATE THE DOT PRODUCT BETWEEN 2 QUATERNIONS
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat GLQuaternion::Dot_Product(GLQuaternion& q2)
{
	return x*q2.x + y*q2.y + z*q2.z + w*q2.w;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NORMALIZE A QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::Normalize()
{
	GLfloat quatLength = Get_Norm();
	assert(quatLength!=0 && "Division by 0 error in GLQuaternion::Normalize()");

	if(quatLength!=1.0f)
	{
		if(quatLength<0.0001)
			quatLength=0.0001f;			

		quatLength = 1/quatLength;
		*this *= quatLength;
	}
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NEGATE A QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLQuaternion::Negate()
{
	Set_Values( -w, -x, -y, -z );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COMPUTE THE CONJUGATE OF THE QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion GLQuaternion::Get_Conjugate()
{
	return GLQuaternion( w, -x, -y, -z );	
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COMPUTE THE INVERSE OF THE QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion GLQuaternion::Get_Inverse()
{
	GLQuaternion	conjugateQuat, inverseQuat;
	GLfloat			norm;

	conjugateQuat = Get_Conjugate();
	norm		  = Get_Norm();

	assert(norm!=0.0f && "Division by 0 error in GLQuaternion::Get_Inverse()");
	if(norm<0.0001f)
		norm=0.0001f;

	inverseQuat	= conjugateQuat / norm;

	return inverseQuat;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// BUILD A 4X4 MATRIX FROM THE QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLMatrix44 GLQuaternion::Get_MatrixFromQuaternion()
{
	GLMatrix44 newMatrix;

	GLfloat XX=0.0f, YY=0.0f, ZZ=0.0f;
	GLfloat XY=0.0f, XZ=0.0f;
	GLfloat YZ=0.0f;
	GLfloat WX=0.0f, WY=0.0f, WZ=0.0f;

	XX = x*x;
	YY = y*y;
	ZZ = z*z;
	
	XY = x*y;
	XZ = x*z;

	YZ = y*z;

	WX = w*x;
	WY = w*y;
	WZ = w*z;

	// Game Programming Gems 1, page 200-204

	/*
	newMatrix.elem[0]  = 1-(2*(y*y))-(2*(z*z));	
	newMatrix.elem[1]  = 2*((x*y)+(w*z));
	newMatrix.elem[2]  = 2*((x*z)-(w*y));
	newMatrix.elem[3]  = 0;
	newMatrix.elem[4]  = 2*((x*y)-(w*z));
	newMatrix.elem[5]  = 1-(2*(x*x))-(2*(z*z));
	newMatrix.elem[6]  = 2*((y*z)+(w*x));
	newMatrix.elem[7]  = 0;
	newMatrix.elem[8]  = 2*((w*y)+(x*z));
	newMatrix.elem[9]  = 2*((y*z)-(w*x));
	newMatrix.elem[10] = 1-(2*(x*x))-(2*(y*y));
	newMatrix.elem[11] = 0;
	newMatrix.elem[12] = 0;
	newMatrix.elem[13] = 0;
	newMatrix.elem[14] = 0;
	newMatrix.elem[15] = 1; */

	// Real-time rendering, page 48

	newMatrix.elem[0]  = 1 - ( 2 * ( (YY)+(ZZ) ) );	
	newMatrix.elem[1]  = 2 * ( (XY)+(WZ) );
	newMatrix.elem[2]  = 2 * ( (XZ)-(WY) );
	newMatrix.elem[3]  = 0;

	newMatrix.elem[4]  = 2 * ( (XY)-(WZ) );
	newMatrix.elem[5]  = 1 - ( 2 * ( (XX)+(ZZ) ) );
	newMatrix.elem[6]  = 2 * ( (YZ)+(WX) );
	newMatrix.elem[7]  = 0;

	newMatrix.elem[8]  = 2 * ( (XZ)+(WY) );
	newMatrix.elem[9]  = 2 * ( (YZ)-(WX) );
	newMatrix.elem[10] = 1 - ( 2 * ( (XX)+(YY) ) );
	newMatrix.elem[11] = 0;

	newMatrix.elem[12] = 0;
	newMatrix.elem[13] = 0;
	newMatrix.elem[14] = 0;
	newMatrix.elem[15] = 1;

	return newMatrix;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GET THE ANGLE OF ROTATION (IN RADIANS)
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat GLQuaternion::Get_AngleRadOfRotation()
{
	GLfloat angle = (GLfloat)(2*(acos(w)));
	return angle;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GET THE AXIS OF ROTATION OF THE QUATERNION
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLQuaternion::Get_AxisOfRotation()
{
	GLfloat		angleRad = Get_AngleRadOfRotation();
	GLfloat		sinA2;
	
	GLVector4	u;

	// THIS CAN CRASH IF THERE HASN'T BEEN ANY ROTATION BECAUSE THE
	// ANGLE WOULD BE ZERO! AND THEREFORE sinA2 IS ZERO TOO
	// TO AVOID DIVIDING BY ZERO, DIVIDE BY 1
	if(angleRad!=0)
		sinA2=(GLfloat)(sin(angleRad/2));
	else 
		sinA2=1;

	if(angleRad<0.0001)
		angleRad=0.0001f;

	u.x = x/sinA2;
	u.y = y/sinA2;
	u.z = z/sinA2;

	return u;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GET THE DIRECTION VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 GLQuaternion::Get_DirectionVector()
{
	Normalize();

	GLVector4 dir;

	dir.x = 2.0f*(x*z - w*y);
	dir.y = -(2.0f*(y*z - w*x));
	dir.z = -(1.0f - (2.0f*(x*x + y*y)));
	
	return dir;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NOT WORKING...
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion GLQuaternion::Get_Interpolate(GLQuaternion& b, GLfloat b_amt)
{
	return GLQuaternion();
}

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLBoundingSphere CLASS METHODS					   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingSphere::GLBoundingSphere()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingSphere::GLBoundingSphere(GLVector4& center)
{
	m_center	= center;
	m_radius	= 1;
	m_radiusSq	= 1;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingSphere::GLBoundingSphere(GLVector4& center, GLfloat rad)
{
	m_center = center;
	m_radius = rad;
	Set_RadiusSq();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingSphere::~GLBoundingSphere()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COPY THE CONTENTS OF ANOTHER BOUNDING SPHERE INTO THIS ONE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingSphere::operator =(GLBoundingSphere& b)
{
	m_center = b.Get_Center();
	m_radius = b.Get_Radius();
	m_radiusSq = b.Get_RadiusSq();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INITIALISE THE BOUNDING SPHERE WITH THE 3 VERTICES OF A TRIANGLE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingSphere::InitAll(GLVector4& v1, GLVector4& v2, GLVector4& v3)
{
	GLfloat distSqCentreV1=0.0f, distSqCentreV2=0.0f, distSqCentreV3=0.0f;
	GLfloat tempX=0.0f, tempY=0.0f, tempZ=0.0f, longestD=0.0f;

	GLVector4 centreV;
	centreV.setValues(0.0f, 0.0f, 0.0f);

	// CALCULATE THE CENTRE OF THESE 3 VERTICES
	// ADD ALL VALUES OF THE TRIANGLES' VERTICES
	centreV = centreV + v1;
	centreV = centreV + v2;
	centreV = centreV + v3;
	centreV /= 3;	

	// CALCULATE THE RADIUS OF THE BOUNDING SPHERE
	
	// CALCULATE THE DISTANCE OF CENTRE TO V1
	tempX=v1.x - centreV.x;	tempX*=tempX;
	tempY=v1.y - centreV.y;	tempY*=tempY;
	tempZ=v1.z - centreV.z;	tempZ*=tempZ;
	distSqCentreV1=tempX+tempY+tempZ;

	// CALCULATE THE DISTANCE OF CENTRE TO V2
	tempX=v2.x - centreV.x;	tempX*=tempX;
	tempY=v2.y - centreV.y;	tempY*=tempY;
	tempZ=v2.z - centreV.z;	tempZ*=tempZ;
	distSqCentreV2=tempX+tempY+tempZ;

	// CALCULATE THE DISTANCE OF CENTRE TO V3
	tempX=v3.x - centreV.x;	tempX*=tempX;
	tempY=v3.y - centreV.y;	tempY*=tempY;
	tempZ=v3.z - centreV.z;	tempZ*=tempZ;
	distSqCentreV3=tempX+tempY+tempZ;

	if((distSqCentreV1>=distSqCentreV2)&&(distSqCentreV2>=distSqCentreV3))longestD=distSqCentreV1;
	else if((distSqCentreV2>=distSqCentreV3)&&(distSqCentreV3>=distSqCentreV1))longestD=distSqCentreV2;
	else longestD=distSqCentreV3;

	m_center = centreV;
	m_radiusSq = longestD;
	longestD=sqrtf(longestD);
	m_radius = longestD;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CLEAR THE VALUES OF THE SPHERE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingSphere::Clear()
{
	m_center.setValues(0.0f, 0.0f, 0.0f);
	Set_Radius(0.0f);
	Set_RadiusSq();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST IF A SPHERE COLLIDES WITH ANOTHER
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLBoundingSphere::CollidesBS(GLBoundingSphere& bs2)
{
	BOOL collision=FALSE;
	GLfloat distance=0.0f, thisRadius=0.0f, radiusBS2=0.0f;
	GLVector4 centreBS2, thisCentre;

	thisCentre = m_center;
	thisRadius = m_radius;

	centreBS2  = bs2.Get_Center();
	radiusBS2  = bs2.Get_Radius();

	distance   = distanceSquare2Vertices(thisCentre, centreBS2);
	distance   = sqrtf(distance);

	if(distance<(thisRadius+radiusBS2))collision=TRUE;
	else collision=FALSE;

	return collision;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST IF A SPHERE COLLIDES WITH A BOUNDING BOX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLBoundingSphere::CollidesBB(GLBoundingBox &bb2)
{
	BOOL collision=FALSE;

	return collision;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GIVEN THE VERTICES OF A MESH, CALCULATE THE BEST RADIUS FOR ITS BOUNDING SPHERE
// AND ITS CENTRE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
/*
VOID GLBoundingSphere::FindBestRadius(VERTEX3D polygons[], int numPolygons)
{
	GLVector4 minX, maxX, minY, maxY, minZ, maxZ, centreSp, ext1, ext2, moveV, temp;
	GLfloat dist1, dist2, dist3, radius, radiusNotSqrt, distTemp, distMove;
	int i=0;

	minX.setValues(10000.0f, 0.0f, 0.0f);
	maxX.setValues(-10000.0f, 0.0f, 0.0f);
	minY.setValues(0.0f, 10000.0f, 0.0f);
	maxY.setValues(0.0f, -10000.0f, 0.0f);
	minZ.setValues(0.0f, 0.0f, 10000.0f);
	maxZ.setValues(0.0f, 0.0f, -10000.0f);

	// AT LEAST 3 VERTICES ARE NEEDED
	assert(numPolygons>=1 && "At least 3 vertices are needed...");

	// FIND THE MINIMUM AND MAXIMUM VERTICES ALONG EACH AXIS
	for(i=0; i<numPolygons; ++i)
	{
		if(polygons[i].x < minX.x)minX.setValues(polygons[i].x, polygons[i].y, polygons[i].z);
		else if(polygons[i].x > maxX.x)maxX.setValues(polygons[i].x, polygons[i].y, polygons[i].z);

		if(polygons[i].y < minY.y)minY.setValues(polygons[i].x, polygons[i].y, polygons[i].z);
		else if(polygons[i].y > maxY.y)maxY.setValues(polygons[i].x, polygons[i].y, polygons[i].z);

		if(polygons[i].z < minZ.z)minZ.setValues(polygons[i].x, polygons[i].y, polygons[i].z);
		else if(polygons[i].z > maxZ.z)maxZ.setValues(polygons[i].x, polygons[i].y, polygons[i].z);
	}

	// CALCULATE THE DISTANCE BETWEEN MIN AND MAX
	dist1 = distanceSquare2Vertices(maxX, minX);
	dist2 = distanceSquare2Vertices(maxY, minY);
	dist3 = distanceSquare2Vertices(maxZ, minZ);

	// WHICH OF THE 3 DISTANCES IS THE LONGEST?
	// THE CENTRE OF THE SPHERE IS THE MIDPOINT BETWEEN THEM
	// AND THE RADIUS IS THE DISTANCE BETWEEN THEM
	if((dist1>=dist2)&&(dist2>=dist3))
	{
		ext1=minX;
		ext2=maxX;
		centreSp=midPoint2Vertices(minX, maxX);
		radius=sqrtf(dist1);
		radiusNotSqrt=dist1;
	}
	else if(dist2>=dist3)
	{
		ext1=minY;
		ext2=maxY;
		centreSp=midPoint2Vertices(minY, maxY);
		radius=sqrtf(dist2);
		radiusNotSqrt=dist2;
	}
	else 
	{
		ext1=minZ;
		ext2=maxZ;
		centreSp=midPoint2Vertices(minZ, maxZ);
		radius=sqrtf(dist3);
		radiusNotSqrt=dist3;
	}

	// FOR EACH POLYGON WHICH IS NOT ANY OF THE 2 VERTICES THAT HAVE
	// THE LONGEST DISTANCE
	// CALCULATE THE DISTANCE TO THE CENTRE OF THE SPHERE, IF THE CENTRE
	// IS OUTSIDE THE RADIUS MOVE THE CENTER TOWARD THE VERTEX BY (d-rad)/2
	// AND THE RADIUS (d+rad)/2

	// AFTER THIS SECOND TIME, THE SPHERE MUST ENCLOSE ALL VERTICES
	for(i=0; i<numPolygons; ++i)
	{
		temp.setValues(polygons[i].x, polygons[i].y, polygons[i].z);
		if((!sameVertex(temp,ext1))&&(!sameVertex(temp,ext2)))
		{
			distTemp = distanceSquare2Vertices(temp, centreSp);
			if(distTemp > radiusNotSqrt)
			{
				distTemp = sqrtf(distTemp);
				radius = sqrtf(radiusNotSqrt);
				distMove = (distTemp-radius)/2;
				
				moveV=temp;
				moveV*=distMove;
				centreSp+=moveV;

				radius = (distTemp+radius)/2;
				radiusNotSqrt = radius*radius;
			}
		}
	}

	m_center	= centreSp;
	m_radius	= radius;
	m_radiusSq	= radiusNotSqrt;
}
*/

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLBoundingBox CLASS METHODS						   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingBox::GLBoundingBox()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingBox::GLBoundingBox(GLVector4& centerV, GLfloat mag)
{
	InitAll(centerV, mag);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingBox::GLBoundingBox(GLVector4& centerV, GLfloat magX, GLfloat magY, GLfloat magZ)
{
	InitAll(centerV, magX, magY, magZ);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLBoundingBox::~GLBoundingBox()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COPY THE CONTENTS OF ANOTHER BOUNDING BOX INTO THIS ONE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::operator =(GLBoundingBox& b2)
{
	m_center		= b2.m_center;
	m_trcenter		= b2.m_trcenter;
	m_magX			= b2.m_magX;
	m_magY			= b2.m_magY;
	m_magZ			= b2.m_magZ;

	m_c1			= b2.m_c1;
	m_c2			= b2.m_c2;
	m_c3			= b2.m_c3;
	m_c4			= b2.m_c4;
	m_c5			= b2.m_c5;
	m_c6			= b2.m_c6;
	m_c7			= b2.m_c7;
	m_c8			= b2.m_c8;

	m_trc1			= b2.m_trc1;
	m_trc2			= b2.m_trc2;
	m_trc3			= b2.m_trc3;
	m_trc4			= b2.m_trc4;
	m_trc5			= b2.m_trc5;
	m_trc6			= b2.m_trc6;
	m_trc7			= b2.m_trc7;
	m_trc8			= b2.m_trc8;

	ComputePlanes();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INITIALIZE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
HRESULT	GLBoundingBox::InitAll(float x, float y, float z, GLfloat mag)
{
	GLVector4		tempV, centerV;
	GLfloat			magXO2, magYO2, magZO2;

	m_center.setValues(x, y, z);
	centerV			= m_center;
	m_trcenter		= m_center;
	m_magX = m_magY = m_magZ = mag;
	magXO2 = magYO2 = magZO2 = mag/2.0f;

	tempV.setValues(centerV.x - (magXO2), centerV.y + (magYO2), centerV.z - (magZO2));
	m_c1.setValues(tempV.x, tempV.y, tempV.z);
	m_trc1 = m_c1;

	tempV.setValues(centerV.x - (magXO2), centerV.y - (magYO2), centerV.z - (magZO2));
	m_c2.setValues(tempV.x, tempV.y, tempV.z);
	m_trc2 = m_c2;

	tempV.setValues(centerV.x + (magXO2), centerV.y + (magYO2), centerV.z - (magZO2));
	m_c3.setValues(tempV.x, tempV.y, tempV.z);
	m_trc3 = m_c3;

	tempV.setValues(centerV.x + (magXO2), centerV.y - (magYO2), centerV.z - (magZO2));
	m_c4.setValues(tempV.x, tempV.y, tempV.z);
	m_trc4 = m_c4;

	tempV.setValues(centerV.x - (magXO2), centerV.y + (magYO2), centerV.z + (magZO2));
	m_c5.setValues(tempV.x, tempV.y, tempV.z);
	m_trc5 = m_c5;

	tempV.setValues(centerV.x - (magXO2), centerV.y - (magYO2), centerV.z + (magZO2));
	m_c6.setValues(tempV.x, tempV.y, tempV.z);
	m_trc6 = m_c6;

	tempV.setValues(centerV.x + (magXO2), centerV.y + (magYO2), centerV.z + (magZO2));
	m_c7.setValues(tempV.x, tempV.y, tempV.z);
	m_trc7 = m_c7;

	tempV.setValues(centerV.x + (magXO2), centerV.y - (magYO2), centerV.z + (magZO2));
	m_c8.setValues(tempV.x, tempV.y, tempV.z);
	m_trc8 = m_c8;

	ComputePlanes();

	return S_OK;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INITIALIZE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
HRESULT GLBoundingBox::InitAll(GLVector4& centerV, GLfloat mag)
{
	GLVector4		tempV;
	GLfloat			magXO2, magYO2, magZO2;

	m_center		= centerV;
	m_trcenter		= centerV;
	m_magX = m_magY = m_magZ = mag;
	magXO2 = magYO2 = magZO2 = mag/2.0f;

	tempV.setValues(centerV.x - (magXO2), centerV.y + (magYO2), centerV.z - (magZO2));
	m_c1.setValues(tempV.x, tempV.y, tempV.z);
	m_trc1 = m_c1;

	tempV.setValues(centerV.x - (magXO2), centerV.y - (magYO2), centerV.z - (magZO2));
	m_c2.setValues(tempV.x, tempV.y, tempV.z);
	m_trc2 = m_c2;

	tempV.setValues(centerV.x + (magXO2), centerV.y + (magYO2), centerV.z - (magZO2));
	m_c3.setValues(tempV.x, tempV.y, tempV.z);
	m_trc3 = m_c3;

	tempV.setValues(centerV.x + (magXO2), centerV.y - (magYO2), centerV.z - (magZO2));
	m_c4.setValues(tempV.x, tempV.y, tempV.z);
	m_trc4 = m_c4;

	tempV.setValues(centerV.x - (magXO2), centerV.y + (magYO2), centerV.z + (magZO2));
	m_c5.setValues(tempV.x, tempV.y, tempV.z);
	m_trc5 = m_c5;

	tempV.setValues(centerV.x - (magXO2), centerV.y - (magYO2), centerV.z + (magZO2));
	m_c6.setValues(tempV.x, tempV.y, tempV.z);
	m_trc6 = m_c6;

	tempV.setValues(centerV.x + (magXO2), centerV.y + (magYO2), centerV.z + (magZO2));
	m_c7.setValues(tempV.x, tempV.y, tempV.z);
	m_trc7 = m_c7;

	tempV.setValues(centerV.x + (magXO2), centerV.y - (magYO2), centerV.z + (magZO2));
	m_c8.setValues(tempV.x, tempV.y, tempV.z);
	m_trc8 = m_c8;

	ComputePlanes();

	return S_OK;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INITIALIZE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
HRESULT GLBoundingBox::InitAll(GLVector3& centerV, GLfloat mag)
{
	GLVector4	center;
	center.setValues(centerV.x, centerV.y, centerV.z);
	InitAll(center, mag);
	return S_OK;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INITIALIZE 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
HRESULT GLBoundingBox::InitAll(GLVector4& centerV, GLfloat magX, GLfloat magY, GLfloat magZ)
{
	GLVector4		tempV;
	GLfloat			magXO2, magYO2, magZO2;

	m_center		= centerV;
	m_trcenter		= centerV;
	m_magX			= magX;
	m_magY			= magY;
	m_magZ			= magZ;

	magXO2			= magX/2.0f;
	magYO2			= magY/2.0f;
	magZO2			= magZ/2.0f;

	tempV.setValues(centerV.x - (magXO2), centerV.y + (magYO2), centerV.z - (magZO2));
	m_c1.setValues(tempV.x, tempV.y, tempV.z);
	m_trc1 = m_c1;

	tempV.setValues(centerV.x - (magXO2), centerV.y - (magYO2), centerV.z - (magZO2));
	m_c2.setValues(tempV.x, tempV.y, tempV.z);
	m_trc2 = m_c2;

	tempV.setValues(centerV.x + (magXO2), centerV.y + (magYO2), centerV.z - (magZO2));
	m_c3.setValues(tempV.x, tempV.y, tempV.z);
	m_trc3 = m_c3;

	tempV.setValues(centerV.x + (magXO2), centerV.y - (magYO2), centerV.z - (magZO2));
	m_c4.setValues(tempV.x, tempV.y, tempV.z);
	m_trc4 = m_c4;

	tempV.setValues(centerV.x - (magXO2), centerV.y + (magYO2), centerV.z + (magZO2));
	m_c5.setValues(tempV.x, tempV.y, tempV.z);
	m_trc5 = m_c5;

	tempV.setValues(centerV.x - (magXO2), centerV.y - (magYO2), centerV.z + (magZO2));
	m_c6.setValues(tempV.x, tempV.y, tempV.z);
	m_trc6 = m_c6;

	tempV.setValues(centerV.x + (magXO2), centerV.y + (magYO2), centerV.z + (magZO2));
	m_c7.setValues(tempV.x, tempV.y, tempV.z);
	m_trc7 = m_c7;

	tempV.setValues(centerV.x + (magXO2), centerV.y - (magYO2), centerV.z + (magZO2));
	m_c8.setValues(tempV.x, tempV.y, tempV.z);
	m_trc8 = m_c8;

	ComputePlanes();

	return S_OK;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COMPUTE THE 6 PLANES OF THE BOUNDING BOX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::ComputePlanes()
{
	GLfloat x[3], y[3], z[3];

	// SET PLANE 1
	x[0]=m_trc1.x; y[0]=m_trc1.y; z[0]=m_trc1.z;
	x[1]=m_trc2.x; y[1]=m_trc2.y; z[1]=m_trc2.z;
	x[2]=m_trc3.x; y[2]=m_trc3.y; z[2]=m_trc3.z;
	m_planes[0].SetCoefficientsFromPoints(x, y, z);

	// SET PLANE 2
	x[0]=m_trc5.x; y[0]=m_trc5.y; z[0]=m_trc5.z;
	x[1]=m_trc1.x; y[1]=m_trc1.y; z[1]=m_trc1.z;
	x[2]=m_trc7.x; y[2]=m_trc7.y; z[2]=m_trc7.z;
	m_planes[1].SetCoefficientsFromPoints(x, y, z);

	// SET PLANE 3
	x[0]=m_trc7.x; y[0]=m_trc7.y; z[0]=m_trc7.z;
	x[1]=m_trc8.x; y[1]=m_trc8.y; z[1]=m_trc8.z;
	x[2]=m_trc5.x; y[2]=m_trc5.y; z[2]=m_trc5.z;
	m_planes[2].SetCoefficientsFromPoints(x, y, z);

	// SET PLANE 4
	x[0]=m_trc2.x; y[0]=m_trc2.y; z[0]=m_trc2.z;
	x[1]=m_trc6.x; y[1]=m_trc6.y; z[1]=m_trc6.z;
	x[2]=m_trc4.x; y[2]=m_trc4.y; z[2]=m_trc4.z;
	m_planes[3].SetCoefficientsFromPoints(x, y, z);

	// SET PLANE 5
	x[0]=m_trc5.x; y[0]=m_trc5.y; z[0]=m_trc5.z;
	x[1]=m_trc6.x; y[1]=m_trc6.y; z[1]=m_trc6.z;
	x[2]=m_trc1.x; y[2]=m_trc1.y; z[2]=m_trc1.z;
	m_planes[4].SetCoefficientsFromPoints(x, y, z);

	// SET PLANE 6
	x[0]=m_trc3.x; y[0]=m_trc3.y; z[0]=m_trc3.z;
	x[1]=m_trc4.x; y[1]=m_trc4.y; z[1]=m_trc4.z;
	x[2]=m_trc7.x; y[2]=m_trc7.y; z[2]=m_trc7.z;
	m_planes[5].SetCoefficientsFromPoints(x, y, z);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// UPDATE THE SIZE OF THE BOUNDING BOX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::UpdateSizeBB()
{
	GLVector4		tempV;
	GLfloat			magXO2, magYO2, magZO2;

	magXO2			= m_magX/2.0f;
	magYO2			= m_magY/2.0f;
	magZO2			= m_magZ/2.0f;

	tempV.setValues(m_trcenter.x - (magXO2), m_trcenter.y + (magYO2), m_trcenter.z - (magZO2));
	m_c1.setValues(tempV.x, tempV.y, tempV.z);
	m_trc1 = m_c1;

	tempV.setValues(m_trcenter.x - (magXO2), m_trcenter.y - (magYO2), m_trcenter.z - (magZO2));
	m_c2.setValues(tempV.x, tempV.y, tempV.z);
	m_trc2 = m_c2;

	tempV.setValues(m_trcenter.x + (magXO2), m_trcenter.y + (magYO2), m_trcenter.z - (magZO2));
	m_c3.setValues(tempV.x, tempV.y, tempV.z);
	m_trc3 = m_c3;

	tempV.setValues(m_trcenter.x + (magXO2), m_trcenter.y - (magYO2), m_trcenter.z - (magZO2));
	m_c4.setValues(tempV.x, tempV.y, tempV.z);
	m_trc4 = m_c4;

	tempV.setValues(m_trcenter.x - (magXO2), m_trcenter.y + (magYO2), m_trcenter.z + (magZO2));
	m_c5.setValues(tempV.x, tempV.y, tempV.z);
	m_trc5 = m_c5;

	tempV.setValues(m_trcenter.x - (magXO2), m_trcenter.y - (magYO2), m_trcenter.z + (magZO2));
	m_c6.setValues(tempV.x, tempV.y, tempV.z);
	m_trc6 = m_c6;

	tempV.setValues(m_trcenter.x + (magXO2), m_trcenter.y + (magYO2), m_trcenter.z + (magZO2));
	m_c7.setValues(tempV.x, tempV.y, tempV.z);
	m_trc7 = m_c7;

	tempV.setValues(m_trcenter.x + (magXO2), m_trcenter.y - (magYO2), m_trcenter.z + (magZO2));
	m_c8.setValues(tempV.x, tempV.y, tempV.z);
	m_trc8 = m_c8;

	ComputePlanes();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST A POINT AGAINST ALL 6 PLANES OF THE BOUNDING BOX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLBoundingBox::PointInsideBBPlanes(GLVector4 &point)
{
	for(int i=0; i<6; i++)
	{
		if(!(m_planes[i].PointInsidePlane(point)))return FALSE;
	}
	return TRUE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST IF A POINT INTERSECTS THE BB
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLBoundingBox::PointIntersectsBB(GLVector4 pt, int XZ, int YZ, int XY)
{
	BOOL intersects = FALSE;
	int  i=1;
	GLVector4 v0,v1,v2; 

	// CHECK FIRST FACE
	v0 = m_trc1;	v1 = m_trc3;	v2 = m_trc2;
	intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
	if(!intersects)
	{
		v0 = m_trc2;	v1 = m_trc3;	v2 = m_trc4;
		intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
	}
	// CHECK SECOND FACE
	if(!intersects)
	{
		v0 = m_trc5;	v1 = m_trc1;	v2 = m_trc6;
		intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		if(!intersects)
		{
			v0 = m_trc6;	v1 = m_trc1;	v2 = m_trc2;
			intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		}
	}
	// CHECK THIRD FACE
	if(!intersects)
	{
		v0 = m_trc5;	v1 = m_trc7;	v2 = m_trc1;
		intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		if(!intersects)
		{
			v0 = m_trc1;	v1 = m_trc7;	v2 = m_trc3;
			intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		}
	}
	// CHECK FOURTH FACE
	if(!intersects)
	{
		v0 = m_trc2;	v1 = m_trc4;	v2 = m_trc6;
		intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		if(!intersects)
		{
			v0 = m_trc6;	v1 = m_trc4;	v2 = m_trc8;
			intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		}
	}
	// CHECK FIFTH FACE
	if(!intersects)
	{
		v0 = m_trc6;	v1 = m_trc8;	v2 = m_trc5;
		intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		if(!intersects)
		{
			v0 = m_trc5;	v1 = m_trc8;	v2 = m_trc7;
			intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		}
	}
	// CHECK SIXTH FACE
	if(!intersects)
	{
		v0 = m_trc3;	v1 = m_trc7;	v2 = m_trc4;
		intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		if(!intersects)
		{
			v0 = m_trc4;	v1 = m_trc7;	v2 = m_trc8;
			intersects = PointIntersectsFace(pt, v0, v1, v2, XZ, YZ, XY);
		}
	}

	return intersects;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST IF A POINT INTERSECTS ONE FACE OF THE BB
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLBoundingBox::PointIntersectsFace(GLVector4 pt, GLVector4 v0, GLVector4 v1, 
												GLVector4 v2, int XZ, int YZ, int XY)
{
	BOOL intersects = FALSE;

	if(XZ)
	{
		// CHECK THE Y AND Z COORDS OF THE VERTICES AND
		// IGNORE THE X COORDINATE
		if(Overlaps_YZ(pt.y, pt.z, v0, v1, v2))
			intersects = TRUE;
	}
	else if(YZ)
	{
		// CHECK THE X AND Z COORDS OF THE VERTICES AND
		// IGNORE THE Y COORDINATE
		if(Overlaps_XZ(pt.x, pt.z, v0, v1, v2))
			intersects = TRUE;
	}
	else if(XY)
	{
		// CHECK THE X AND Y COORDS OF THE VERTICES AND
		// IGNORE THE Z COORDINATE
		if(Overlaps_XY(pt.x, pt.y, v0, v1, v2))
			intersects = TRUE;
	}

	return intersects;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GIVEN A POINT, CALCULATE TO WHICH SIDE OF THE BOUNDING BOX IS CLOSEST
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
int GLBoundingBox::GetClosestSideToPoint(GLVector4 &point)
{
	int		closestSide		= 7;
	float	leastDistance	= 10000.0f;
	float	dist			= 0.0f;

	for(int i=0; i<6; i++)
	{
		dist = m_planes[i].SignedDistanceToPoint(point);
		if(fabs(dist) <= fabs(leastDistance))
		{
			leastDistance = (float)(fabs(dist));
			closestSide = i+1;
		}
	}

	return (closestSide<7)?(closestSide):(0);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CLEAR THE VALUES OF THE BOX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::Clear()
{
	m_magX = m_magY = m_magZ = 0.0f;

	m_c1.setValues(0.0f, 0.0f, 0.0f);
	m_trc1 = m_c1;
	m_c2.setValues(0.0f, 0.0f, 0.0f);
	m_trc2 = m_c2;
	m_c3.setValues(0.0f, 0.0f, 0.0f);
	m_trc3 = m_c3;
	m_c4.setValues(0.0f, 0.0f, 0.0f);
	m_trc4 = m_c4;
	m_c5.setValues(0.0f, 0.0f, 0.0f);
	m_trc5 = m_c5;
	m_c6.setValues(0.0f, 0.0f, 0.0f);
	m_trc6 = m_c6;
	m_c7.setValues(0.0f, 0.0f, 0.0f);
	m_trc7 = m_c7;
	m_c8.setValues(0.0f, 0.0f, 0.0f);
	m_trc8 = m_c8;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST THE COLLISION OF TWO BOUNDING BOXES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLBoundingBox::CollidesBB(GLBoundingBox &bb2)
{
	// GET THE CORNER POINTS FROM THE SECOND BOUNDING BOX
	GLVector4 b2_p1 = bb2.Get_point1();
	GLVector4 b2_p2 = bb2.Get_point2();
	GLVector4 b2_p3 = bb2.Get_point3();
	GLVector4 b2_p4 = bb2.Get_point4();
	GLVector4 b2_p5 = bb2.Get_point5();
	GLVector4 b2_p6 = bb2.Get_point6();
	GLVector4 b2_p7 = bb2.Get_point7();
	GLVector4 b2_p8 = bb2.Get_point8();

	// TEST FIRST IF ANY OF THE POINTS OF THE SECOND BOUNDING BOX IS INSIDE
	// THE FIRST BOUNDING BOX 
	if(PointInsideBBPlanes(b2_p1))return TRUE;
	if(PointInsideBBPlanes(b2_p2))return TRUE;
	if(PointInsideBBPlanes(b2_p3))return TRUE;
	if(PointInsideBBPlanes(b2_p4))return TRUE;
	if(PointInsideBBPlanes(b2_p5))return TRUE;
	if(PointInsideBBPlanes(b2_p6))return TRUE;
	if(PointInsideBBPlanes(b2_p7))return TRUE;
	if(PointInsideBBPlanes(b2_p8))return TRUE;

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// IF A COLLISION HAS NOT BEEN DETECTED, TEST IF ANY OF THE POINTS OF
	// THE FIRST BOUNDING BOX IS INSIDE THE SECOND BOUNDING BOX, THIS DOUBLE
	// CHECK ENSURES THAT IF A COLLISION EXISTS, IT'S DETECTED
	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	// GET THE CORNER POINTS OF THE FIRST BOUNDING BOX 
	GLVector4 b1_p1 = Get_point1();
	GLVector4 b1_p2 = Get_point2();
	GLVector4 b1_p3 = Get_point3();
	GLVector4 b1_p4 = Get_point4();
	GLVector4 b1_p5 = Get_point5();
	GLVector4 b1_p6 = Get_point6();
	GLVector4 b1_p7 = Get_point7();
	GLVector4 b1_p8 = Get_point8();

	// TEST IF ANY OF THE POINTS OF THE FIRST BOUNDING BOX IS INSIDE THE
	// SECOND BOUNDING BOX
	if(bb2.PointInsideBBPlanes(b1_p1))return TRUE;
	if(bb2.PointInsideBBPlanes(b1_p2))return TRUE;
	if(bb2.PointInsideBBPlanes(b1_p3))return TRUE;
	if(bb2.PointInsideBBPlanes(b1_p4))return TRUE;
	if(bb2.PointInsideBBPlanes(b1_p5))return TRUE;
	if(bb2.PointInsideBBPlanes(b1_p6))return TRUE;
	if(bb2.PointInsideBBPlanes(b1_p7))return TRUE;
	if(bb2.PointInsideBBPlanes(b1_p8))return TRUE;

	return FALSE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// UPDATE THE CORNER POINTS OF THE BOUNDING BOX WITH A TRANSFORMATION MATRIX
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::UpdateCornerPointsWithMatrix(GLMatrix44 &matrix)
{
//	m_trcenter = matrix * m_center;
	m_trc1 = matrix * m_c1;
	m_trc2 = matrix * m_c2;
	m_trc3 = matrix * m_c3;
	m_trc4 = matrix * m_c4;
	m_trc5 = matrix * m_c5;
	m_trc6 = matrix * m_c6;
	m_trc7 = matrix * m_c7;
	m_trc8 = matrix * m_c8;

	ComputePlanes();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
//
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::UpdateWithMatrix_Acc(GLMatrix44 &matrix)
{
	m_trc1 -= m_trcenter;
	m_trc2 -= m_trcenter;
	m_trc3 -= m_trcenter;
	m_trc4 -= m_trcenter;
	m_trc5 -= m_trcenter;
	m_trc6 -= m_trcenter;
	m_trc7 -= m_trcenter;
	m_trc8 -= m_trcenter;

	m_trc1 = matrix * m_trc1;
	m_trc2 = matrix * m_trc2;
	m_trc3 = matrix * m_trc3;
	m_trc4 = matrix * m_trc4;
	m_trc5 = matrix * m_trc5;
	m_trc6 = matrix * m_trc6;
	m_trc7 = matrix * m_trc7;
	m_trc8 = matrix * m_trc8;

	m_trc1 += m_trcenter;
	m_trc2 += m_trcenter;
	m_trc3 += m_trcenter;
	m_trc4 += m_trcenter;
	m_trc5 += m_trcenter;
	m_trc6 += m_trcenter;
	m_trc7 += m_trcenter;
	m_trc8 += m_trcenter;

	ComputePlanes();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// MOVE THE BOUNDING BOX BY A VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::MoveBB(GLVector4& addVector)
{
	m_trcenter	+= addVector;
	m_trc1		+= addVector;
	m_trc2		+= addVector;
	m_trc3		+= addVector;
	m_trc4		+= addVector;
	m_trc5		+= addVector;
	m_trc6		+= addVector;
	m_trc7		+= addVector;
	m_trc8		+= addVector;
	ComputePlanes();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// MOVE THE BOUNDING BOX BY A VECTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::MoveBB_PointsOnly(GLVector4& addVector)
{
	m_trc1		+= addVector;
	m_trc2		+= addVector;
	m_trc3		+= addVector;
	m_trc4		+= addVector;
	m_trc5		+= addVector;
	m_trc6		+= addVector;
	m_trc7		+= addVector;
	m_trc8		+= addVector;
	ComputePlanes();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GET THE 8 CORNER POINTS FROM THIS BB
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLBoundingBox::GetAllPoints(GLVector4 *p1, GLVector4 *p2, GLVector4 *p3,
								 GLVector4 *p4, GLVector4 *p5, GLVector4 *p6,
								 GLVector4 *p7, GLVector4 *p8)
{
	p1->setValues(m_trc1);
	p2->setValues(m_trc2);
	p3->setValues(m_trc3);
	p4->setValues(m_trc4);
	p5->setValues(m_trc5);
	p6->setValues(m_trc6);
	p7->setValues(m_trc7);
	p8->setValues(m_trc8);
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLRay CLASS METHODS								   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLRay::GLRay()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLRay::GLRay(GLVector4& origPt, GLVector4& dirVec)
{
	m_origin	= origPt;
	m_direction	= dirVec;
	m_direction.normalize();
	m_end		= m_origin + m_direction;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLRay::GLRay(GLVector4& origPt, GLVector4& dirVec, GLfloat maxDist)
{
	m_origin	= origPt;
	m_direction	= dirVec;
	m_direction.normalize();
	m_end		= m_origin + (m_direction*maxDist);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLRay::~GLRay()
{
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// TEST WHETHER A RAY INTERSECTS A SPHERE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL GLRay::IntersectsSphere(GLBoundingSphere& sphere)
{
	GLVector4 vectorL, tempVec;
	GLfloat   vectorLLenSq, projLDirec, distSqCentreProj, radiusSquare;

	radiusSquare = sphere.Get_RadiusSq();

	// CALCULATE THE VECTOR L WHICH IS THE VECTOR FROM THE ORIGIN OF THE RAY TO
	// THE CENTRE OF THE SPHERE
	vectorL		 = sphere.Get_Center() - Get_Origin();

	// CALCULATE THE SQUARED LENGTH OF THIS VECTOR (L.L)
	tempVec		 = vectorL;
	vectorLLenSq = vectorL.dotProduct(tempVec);

	// IF vectorLLenSq < radiusSquare IT IMPLIES THAT THE ORIGIN OF THE RAY IS INSIDE
	// THE SPHERE, WHICH MEANS THAT THE RAY IS GUARANTEED TO HIT THE SPHERE...
	if(vectorLLenSq < radiusSquare)return TRUE;

	// CALCULATE THE PROJECTION OF THE VECTOR L ONTO THE DIRECTION OF THE RAY
	projLDirec = vectorL.dotProduct(Get_Direction());

	// IF projLDirec < 0 AND THE ORIGIN OF THE RAY IS OUTSIDE THE SPHERE, THEN
	// THE SPHERE IS BEHIND THE RAY 
	if((projLDirec < 0)&&(vectorLLenSq > radiusSquare))return FALSE;

	// THE SQUARE DISTANCE FROM THE CENTRE OF THE SPHERE TO THE PROJECTION IS
	// COMPUTED USING THE PYTHAGOREAN THEOREM
	distSqCentreProj = vectorLLenSq - (projLDirec*projLDirec);

	// IF distSqCentreProj > radiusSquare THE RAY WILL MISS THE SPHERE
	if(distSqCentreProj > radiusSquare)return FALSE;

	// IF THEY PASS ALL TESTS, THE RAY WILL HIT THE SPHERE, 
	// ..NOT REALLY FINDING THE INTERSECTION POINTS...
	//
	//	RAY/SPHERE INTERSECTION METHOD DESCRIBED IN
	//  Real-Time Rendering, Second Edition, Tomas Akenine-Moller, Eric Haines
	//  ISBN: 156881-182-9
	//	PAGES: 568-571

	return TRUE;
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE OTHER METHODS										   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

GLVector4 ClosestPointOnLine(GLVector4 vA, GLVector4 vB, GLVector4 vPoint)
{
	// CALCULATE CLOSEST POINT, IT WILL BE ONE OF THE ENDPOINTS OR A POINT
	// BETWEEN vA AND vB

	// FIRST CALCULATE A VECTOR FROM vA to vPoint, THEN GET A NORMALIZED VECTOR
	// FROM vA to vB, THE LENGTH IS NOT IMPORTANT, ONLY THE DIRECTION
	// WITH THIS WE FORM 2 VECTORS FROM A PSEUDO CORNER OF A TRIANGLE ON THE PLANE
	// OF THE LINE SEGMENT AND THE POINT

	// NEXT FIND THE MAGNITUDE OF THE LINE SEGMENT, THEN WE USE DOT PRODUCT TO PROJECT
	// THE VECTOR 1 ONTO THE NORMALIZED VECTOR.
	// IF THE RESULT OF THE DOT PRODUCT IS 0, THAT MEANS THE VECTORS WERE PERPENDICULAR
	// AND HAD A 90 DEGREES ANGLE BETWEEN THEM
	
	// THE 0 PART IS THE DISTANCE THE NEW PROJECTED VECTOR IS FROM THE STARTING OF
	// VECTOR 2. IF THE RESULT IS A NEGATIVE NUMBER, IT MEANS THE ANGLE BETWEEN THE 2
	// VECTORS IS GREATER THAN 90 DEGREES, WHICH MEANS THAT THE CLOSEST POINT MUST BE vA
	// BECAUSE IT'S PROJECTED VECTOR IS ON THE OUTSIDE OF THE LINE. SO, IF THE RESULT IS A
	// POSITIVE NUMBER, THE PROJECTED VECTOR IS ON THE RIGHT SIDE OF vA, BUT COULD BE PAST THE
	// RIGHT SIDE OF vB. TO TEST THIS, MAKE SURE THAT THE RESULT OF THE DOT PRODUCT IS NOT
	// GREATER THAN THE DISTANCE D. IF IT IS, THEN THE CLOSEST POINT ON THE PLANE IS vB

	// TO FIND THE POINT BETWEEN THE 2 ENDPOINTS, TAKE THE DISTANCE T FROM POINT vA
	// GIVEN FROM THE DOT PRODUCT OF THE 2 VERCTORS
	// USE THE VECTOR THAT IS GOING THE DIRECTION OF THE LINE SEGMENT, VECTOR 2, AND
	// MULTIPLY IT BY THE DISTANCE SCALAR T, THIS WILL CREATE A VECTOR GOING IN THE DIRECTION OF
	// THE LINE SEGMENT, WITH A DISTANCE OR MAGNITUDE OF THE PROJECTED VECTOR, VECTOR 1, IS FROM vA
	// THEN ADD THIS VECTOR TO vA WHICH GIVES THE POINT ON THE LINE THAT IS CLOSEST TO vPoint

	
	// CREATE THE VECTOR FROM ENDPOINT vA TO THE POINT IN SPACE vPoint
	GLVector4 vVector1 = vPoint - vA;

	// CREATE A NORMALIZED DIRECTION VECTOR FROM ENDPOINT vA to ENDPOINT vB
	GLVector4 vVector2 = vB - vA;
	vVector2.normalize();

	// GET THE DISTANCE TO FIND THE LENGTH OF THE LINE SEGMENT
    float d = distanceSquare2Vertices(vA, vB);
	d = sqrtf(d);

	// USE THE DOT PRODUCT TO PROJECT vVector1 ONTO vVector2
	// THIS GIVES US THE DISTANCE FROM OUR PROJECTED VECTOR FROM vA
    float t = vVector2.dotProduct(vVector1);

	// IF THE PROJECTED DISTANCE FROM vA, "t" IS LESS THAN OR EQUAL TO 0, IT MUST BE
	// CLOSEST TO THE ENDPOINT vA...
    if(t<=0)return vA;

	// IF THE PROJECTED DISTANCE "t" IS GREATER THAN OR EQUAL TO THE MAGNITUDE OF THE LINE
	// SEGMENT, IT MUST BE CLOSEST TO ENDPOINT B
    if(t>=d)return vB;
 
	// IF IT'S NOT CLOSEST TO ANY OF THE ENDPOINTS IT MUST BE BETWEEN THEM
	// CREATE A VECTOR THAT IS OF LENGTH T AND IN THE DIRECTION OF vVector2
	GLVector4 vVector3 = vVector2 * t;

	// TO FIND THE CLOSEST POINT ON THE LINE, WE JUST ADD vVector3 TO THE ORIGINAL ENDPOINT vA
    GLVector4 vClosestPoint = vA + vVector3;

	// RETURN CLOSEST POINT
	return vClosestPoint;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// RETURNS THE INTERSECTION POINT OF THE LINE THAT INTERSECTS THE PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 IntersectionPoint(GLVector4 vNormal, GLVector4 sPoint, GLVector4 ePoint, double distance)
{
	GLVector4 intPoint;

	// ..................................

	return intPoint;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GETS THE SQUARED DISTANCE BETWEEN 2 SPHERE CENTRES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat distanceSquare2Vertices(GLVector4 v1, GLVector4 v2)
{
	GLfloat x, y, z, distSq;

	x	=	v2.x-v1.x;
	x	*=	x;
	y	=	v2.y-v1.y;
	y	*=	y;
	z	=	v2.z-v1.z;
	z	*=	z;

	distSq = x + y + z;

	return distSq;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GETS THE SQUARED DISTANCE BETWEEN 2 SPHERE CENTRES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLfloat distance2Points(GLVector3 v1, GLVector3 v2)
{
	GLfloat x, y, z, distSq;

	x	=	v2.x-v1.x;
	x	*=	x;
	y	=	v2.y-v1.y;
	y	*=	y;
	z	=	v2.z-v1.z;
	z	*=	z;

	distSq = x + y + z;
	distSq = sqrtf(distSq);

	return distSq;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CALCULATE THE MIDDLE POINT BETWEEN 2 VERTICES
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLVector4 midPoint2Vertices(GLVector4 v1, GLVector4 v2)
{
	GLVector4 midP = (v1+v2)/2;
	return midP;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
BOOL sameVertex(GLVector4 v1, GLVector4 v2)
{
	if(v1==v2)return TRUE;
	else return FALSE;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
float vmax(float a, float b)
{
	if(a>=b)return a;
	else return b;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
float vmin(float a, float b)
{
	if(a<=b)return a;
	else return b;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// NOT WORKING...
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion RotationArc( GLVector4& v0, GLVector4 v1)
{
	return GLQuaternion();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CALCULATE THE SPHERICAL LINEAR INTERPOLATION FROM QUATERNION "from" TO QUATERNION "to" 
// IN THE TIME interp WHERE 0 < interp < 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLQuaternion Slerp(GLQuaternion& from, GLQuaternion& to, GLfloat interp)
{
	GLQuaternion temp;

	GLfloat omega, cosO, sinO;
	GLfloat scale0, scale1;

	cosO = from.Dot_Product(to);
	temp = to;

	// adjust signs if necessary
	if(cosO < 0.0)
	{
		cosO = -cosO;
		temp.Negate();
	}

	if((1.0 - cosO) > ERROR_TOLERANCE)
	{
		// standard case SLERP
		omega  = (GLfloat)acos(cosO);
		sinO   = sinf(omega);
		assert(sinO!=0 && "Division by 0 error in GLQuaternion::Slerp()");
		scale0 = sinf((1.0f - interp) * omega) / sinO;
		scale1 = sinf(interp * omega) / sinO;
	}
	else
	{
		// "from" and "to" quaternions are very close, so we can do a linear interpolation
		scale0 = 1.0f - interp;
		scale1 = interp;
	}

	return ((from * scale0) + (temp * scale1));

/*
ANOTHER IMPLEMENTATION...
	Quaternion slerp(Quaternion a,Quaternion b,float interp){
		if((a^b) <0.0) {
			a.r=-a.r;
			a.x=-a.x;
			a.y=-a.y;
			a.z=-a.z;
		}
		float theta = (float)acos(a^b);
		if(theta==0.0f) { return(a);}
		return a*(float)(sin(theta-interp*theta)/sin(theta)) + b*(float)(sin(interp*theta)/sin(theta));
	}
 */
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLLineSeg2D CLASS METHODS							   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DEFAULT CONSTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLLineSeg2D::GLLineSeg2D()
{
	InitAll();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLLineSeg2D::GLLineSeg2D(GLfloat ox, GLfloat oy, GLfloat dx, GLfloat dy)
{
	m_origX = ox;
	m_origY = oy;
	m_destX = dx;
	m_destY = dy;
	if(oy==dy)m_slope=0;
	else if(ox==dx)m_slope=-99999.0f;
	else m_slope = (dy-oy)/(dx-ox);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLLineSeg2D::~GLLineSeg2D()
{
	InitAll();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INIT ALL
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLLineSeg2D::InitAll()
{
	m_origX = 0.0f;
	m_origY = 0.0f;
	m_destX = 1.0f;
	m_destY = 1.0f;
	m_slope = 1.0f;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INTERSECTS ANOTHER LINE SEGMENT
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool GLLineSeg2D::IntersectsLine(GLLineSeg2D &seg2)
{
	GLPoint2D a, b, c, d;

	a.InitAll(m_origX, m_origY);
	b.InitAll(m_destX, m_destY);

	c.InitAll(seg2.m_origX, seg2.m_origY);
	d.InitAll(seg2.m_destX, seg2.m_destY);

	return Intersect(a,b,c,d);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#		IMPLEMENTATION OF THE GLPoint2D CLASS METHODS							   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DEFAULT CONSTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLPoint2D::GLPoint2D()
{
	InitAll(0.0f, 0.0f);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR WITH PARAMETERS
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLPoint2D::GLPoint2D(GLfloat x, GLfloat y)
{
	InitAll(x, y);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLPoint2D::~GLPoint2D()
{
	InitAll(0.0f, 0.0f);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// INIT ALL
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLPoint2D::InitAll(GLfloat x, GLfloat y)
{
	m_X = x;
	m_Y = y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// FUNCTIONS FROM O'ROURKE BOOK (COMPUTATIONAL GEOMETRY IN C++)
// ************************************************************
//
// A DIRECTED LINE IS DETERMINED BY TWO POINTS GIVEN IN A PARTICULAR ORDER (A,B)
// IF POINT C IS TO THE LEFT OF THE LINE DETERMINED BY (A,B), THEN THE TRIPLE (A,B,C)
// FORMS A COUNTERCLOCKWISE CIRCUIT
// C IS TO THE LEFT OF (A,B) IFF THE AREA OF THE COUNTERCLOCKWISE TRIANGLE (A,B,C)
// IS POSITIVE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$


// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
int Area2Int(GLPoint2D a, GLPoint2D b, GLPoint2D c)
{
	GLfloat area2 =	(	((b.Get_X() - a.Get_X())*(c.Get_Y() - a.Get_Y())) - 
						((c.Get_X() - a.Get_X())*(b.Get_Y() - a.Get_Y())) );
	return ((int)area2);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
float Area2Float(GLPoint2D a, GLPoint2D b, GLPoint2D c)
{
	GLfloat area2 =	(	((b.Get_X() - a.Get_X())*(c.Get_Y() - a.Get_Y())) - 
						((c.Get_X() - a.Get_X())*(b.Get_Y() - a.Get_Y())) );
	return (area2);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool Left(GLPoint2D a, GLPoint2D b, GLPoint2D c)
{
	return ((Area2Float(a, b, c)) > 0);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool LeftOn(GLPoint2D a, GLPoint2D b, GLPoint2D c)
{
	return ((Area2Float(a, b, c)) >= 0);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool Collinear(GLPoint2D a, GLPoint2D b, GLPoint2D c)
{
	return ((Area2Float(a, b, c)) == 0);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool XOR(bool x, bool y)
{
	return !x ^ !y;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool Between(GLPoint2D a, GLPoint2D b, GLPoint2D c)
{
	GLPoint2D ba, ca;

	if(!Collinear(a,b,c))return FALSE;

	// IF ab NOT VERTICAL, CHEC BETWEENESS ON X, ELSE ON Y
	if(a.Get_X() != b.Get_X())
	{
		return	( (a.Get_X() <= c.Get_X()) && (c.Get_X() <= b.Get_X())) ||
				( (a.Get_X() >= c.Get_X()) && (c.Get_X() >= b.Get_X()));
	}
	else
	{
		return	( (a.Get_Y() <= c.Get_Y()) && (c.Get_Y() <= b.Get_Y())) ||
				( (a.Get_Y() >= c.Get_Y()) && (c.Get_Y() >= b.Get_Y()));
	}
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool IntersectProp(GLPoint2D a, GLPoint2D b, GLPoint2D c, GLPoint2D d)
{
	// ELIMINATE IMPROPER CASES
	if(	Collinear(a,b,c) ||
		Collinear(a,b,d) ||
		Collinear(c,d,a) ||
		Collinear(c,d,b))
		return FALSE;

	return( XOR(Left(a,b,c), Left(a,b,d)) && XOR(Left(c,d,a), Left(c,d,b)) );
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool Intersect(GLPoint2D a, GLPoint2D b, GLPoint2D c, GLPoint2D d)
{
	if( IntersectProp(a,b,c,d) )return TRUE;
	else if(	Between(a,b,c) ||
				Between(a,b,d) ||
				Between(c,d,a) ||
				Between(c,d,b))return TRUE;
	else return FALSE;
}

/* **************************** @1 CODE BEGINS *********************************** */

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%###################################################################################%
%#																				   #%	
%#				IMPLEMENTATION OF THE GLTriangle CLASS METHODS					   #%
%#																				   #%
####################################################################################%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 1
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLTriangle::GLTriangle()
{
	m_vertex1.setValues(0.0f, 0.0f, 0.0f);
	m_vertex2.setValues(0.0f, 0.0f, 0.0f);
	m_vertex3.setValues(0.0f, 0.0f, 0.0f);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLTriangle::GLTriangle(GLVector4 v1, GLVector4 v2, GLVector4 v3)
{
	m_vertex1 = v1;
	m_vertex2 = v2;
	m_vertex3 = v3;
	ComputePlane();
	ComputeNormal();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CONSTRUCTOR 3
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLTriangle::GLTriangle( float x[3], float y[3], float z[3] )
{
	m_vertex1.setValues(x[0], y[0], z[0]);
	m_vertex2.setValues(x[1], y[1], z[1]);
	m_vertex3.setValues(x[2], y[2], z[2]);
	ComputePlane();
	ComputeNormal();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DESTRUCTOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
GLTriangle::~GLTriangle()
{
	Set_0_Values();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COMPUTE PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLTriangle::ComputePlane()
{
	GLfloat x[3], y[3], z[3];

	x[0]=m_vertex1.x; y[0]=m_vertex1.y; z[0]=m_vertex1.z;
	x[1]=m_vertex2.x; y[1]=m_vertex2.y; z[1]=m_vertex2.z;
	x[2]=m_vertex3.x; y[2]=m_vertex3.y; z[2]=m_vertex3.z;
	m_triPlane.SetCoefficientsFromPoints(x, y, z);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// COMPUTE NORMAL
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLTriangle::ComputeNormal()
{
	GLVector4	va, vb;

	// GET A VECTOR FROM v2 to v1
	va = m_vertex1 - m_vertex2;

	// GET A VECTOR FROM v2 to v3
	vb = m_vertex3 - m_vertex2;

	// CALCULATE THE NORMAL
	m_normal = va.crossProduct(vb);
	if((m_normal.x!=0.0f)&&(m_normal.y!=0.0f)&&(m_normal.z!=0.0f))
		m_normal.normalize();
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DETERMINE WHETHER THE TRIANGLE PASSED AS PARAMETER IS IN FRONT OF THIS TRIANGLE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
POLYGON_SIDE GLTriangle::CalculateSide(GLTriangle tri2)
{
	GLVector4 v1, v2, v3;
	int	numPositive = 0;
	int	numNegative = 0;
	int	numCoincident = 0;
	HALFSPACE side;

	tri2.Get_Vertices(&v1, &v2, &v3);

	// CLASSIFY VERTEX 1
	side = m_triPlane.ClassifyPoint(v1);
	if(side==POSITIVE)numPositive++;
	else if(side==NEGATIVE)numNegative++;
	else numCoincident++;

	// CLASSIFY VERTEX 2
	side = m_triPlane.ClassifyPoint(v2);
	if(side==POSITIVE)numPositive++;
	else if(side==NEGATIVE)numNegative++;
	else numCoincident++;

	// CLASSIFY VERTEX 3
	side = m_triPlane.ClassifyPoint(v3);
	if(side==POSITIVE)numPositive++;
	else if(side==NEGATIVE)numNegative++;
	else numCoincident++;

	if((numPositive>0)&&(numNegative==0))
		return IN_FRONT;
	else if((numPositive==0)&&(numNegative>0))
		return BEHIND;
	else if((numPositive==0)&&(numNegative==0))
		return COINCIDENT;
	else 
		return SPANNING;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CALCULATE SIDE A 3D POINT LIES RELATED TO THIS TRIANGLE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
POLYGON_SIDE GLTriangle::Point_CalculateSide(GLVector4 point)
{
	HALFSPACE side;

	side = m_triPlane.ClassifyPoint(point);

	if(side==POSITIVE)
		return IN_FRONT;
	else if(side==NEGATIVE)
		return BEHIND;
	else 
		return COINCIDENT;

}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// ASSIGN OPERATOR
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLTriangle::operator =(GLTriangle& tri2)
{
	GLVector4	v1, v2, v3;

	tri2.Get_Vertices(&v1, &v2, &v3);

	Set_Vertices(v1, v2, v3);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// SET ALL VERTICES TO 0's
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLTriangle::Set_0_Values()
{
	m_vertex1.setValues(0.0f, 0.0f, 0.0f);
	m_vertex2.setValues(0.0f, 0.0f, 0.0f);
	m_vertex3.setValues(0.0f, 0.0f, 0.0f);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLTriangle::Set_Vertices(float x[3], float y[3], float z[3])
{
	m_vertex1.setValues(x[0], y[0], z[0]);
	m_vertex2.setValues(x[1], y[1], z[1]);
	m_vertex3.setValues(x[2], y[2], z[2]);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
VOID GLTriangle::Get_Vertices(float x[3], float y[3], float z[3])
{
	x[0] = m_vertex1.x;		y[0] = m_vertex1.y;		z[0] = m_vertex1.z;
	x[1] = m_vertex2.x;		y[1] = m_vertex2.y;		z[1] = m_vertex2.z;
	x[2] = m_vertex3.x;		y[2] = m_vertex3.y;		z[2] = m_vertex3.z;
}


/* **************************** @1 CODE ENDS ************************************* */

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CHECKS IF 2D POINT (X,Z)  IS INSIDE A 2D TRIANGLE (Y COORD DISCARDED)
// USEFULL FOR TOP VIEW
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

bool Overlaps_XZ(GLfloat clickX, GLfloat clickZ, GLVector4 v0, GLVector4 v1, GLVector4 v2)
{
	GLfloat x1=0.0f, y1=0.0f, z1=0.0f;
	GLfloat x2=0.0f, y2=0.0f, z2=0.0f;
	GLfloat x3=0.0f, y3=0.0f, z3=0.0f;

	bool LeftSeg1 = false;
	bool LeftSeg2 = false;
	bool LeftSeg3 = false;

	GLPoint2D a, b, c;

	// INITIALIZE VARIABLES WITH PARAMETERS
	x1=v0.x;	y1=v0.y;	z1=v0.z;
	x2=v1.x;	y2=v1.y;	z2=v1.z;
	x3=v2.x;	y3=v2.y;	z3=v2.z;

	// POINT C WILL BE THE SAME IN ALL 3 LEFT SIDE CHECKINGS
	c.InitAll(clickX, -clickZ);

	// CHECK THE FIRST SEGMENT OF THE TRIANGLE
	a.InitAll(x1, -z1);
	b.InitAll(x2, -z2);
	LeftSeg1 = Left(a,b,c);
	if(!LeftSeg1)return false;

	// CHECK THE SECOND SEGMENT OF THE TRIANGLE
	a.InitAll(x2, -z2);
	b.InitAll(x3, -z3);
	LeftSeg2 = Left(a,b,c);
	if(!LeftSeg2)return false;

	// CHECK THE THIRD SEGMENT OF THE TRIANGLE
	a.InitAll(x3, -z3);
	b.InitAll(x1, -z1);
	LeftSeg3 = Left(a,b,c);
	if(!LeftSeg3)return false;

	// IF POINT C IS TO THE LEFT OF ALL 3 TRIANGLE SEGMENTS THEN THE POINT IS INSIDE THE TRIANGLE
	if(LeftSeg1 && LeftSeg2 && LeftSeg3)return true;
	else return false;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CHECKS IF 2D POINT (X,Y)  IS INSIDE A 2D TRIANGLE (Z COORD DISCARDED)
// USEFUL FOR FRONT VIEW
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

bool Overlaps_XY(GLfloat clickX, GLfloat clickY, GLVector4 v0, GLVector4 v1, GLVector4 v2)
{
	GLfloat x1=0.0f, y1=0.0f, z1=0.0f;
	GLfloat x2=0.0f, y2=0.0f, z2=0.0f;
	GLfloat x3=0.0f, y3=0.0f, z3=0.0f;

	bool LeftSeg1 = false;
	bool LeftSeg2 = false;
	bool LeftSeg3 = false;

	GLPoint2D a, b, c;

	// INITIALIZE VARIABLES WITH PARAMETERS
	x1=v0.x;	y1=v0.y;	z1=v0.z;
	x2=v1.x;	y2=v1.y;	z2=v1.z;
	x3=v2.x;	y3=v2.y;	z3=v2.z;

	// POINT C WILL BE THE SAME IN ALL 3 LEFT SIDE CHECKINGS
	c.InitAll(clickX, clickY);

	// CHECK THE FIRST SEGMENT OF THE TRIANGLE
	a.InitAll(x1, y1);
	b.InitAll(x2, y2);
	LeftSeg1 = Left(a,b,c);
	if(!LeftSeg1)return false;

	// CHECK THE SECOND SEGMENT OF THE TRIANGLE
	a.InitAll(x2, y2);
	b.InitAll(x3, y3);
	LeftSeg2 = Left(a,b,c);
	if(!LeftSeg2)return false;

	// CHECK THE THIRD SEGMENT OF THE TRIANGLE
	a.InitAll(x3, y3);
	b.InitAll(x1, y1);
	LeftSeg3 = Left(a,b,c);
	if(!LeftSeg3)return false;

	// IF POINT C IS TO THE LEFT OF ALL 3 TRIANGLE SEGMENTS THEN THE POINT IS INSIDE THE TRIANGLE
	if(LeftSeg1 && LeftSeg2 && LeftSeg3)return true;
	else return false;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CHECKS IF 2D POINT (Y,Z)  IS INSIDE A 2D TRIANGLE (X COORD DISCARDED)
// USEFUL FOR SIDE VIEW, EITHER RIGHT OR LEFT
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

bool Overlaps_YZ(GLfloat clickY, GLfloat clickZ, GLVector4 v0, GLVector4 v1, GLVector4 v2)
{
	GLfloat x1=0.0f, y1=0.0f, z1=0.0f;
	GLfloat x2=0.0f, y2=0.0f, z2=0.0f;
	GLfloat x3=0.0f, y3=0.0f, z3=0.0f;

	bool LeftSeg1 = false;
	bool LeftSeg2 = false;
	bool LeftSeg3 = false;

	GLPoint2D a, b, c;

	// INITIALIZE VARIABLES WITH PARAMETERS
	x1=v0.x;	y1=v0.y;	z1=v0.z;
	x2=v1.x;	y2=v1.y;	z2=v1.z;
	x3=v2.x;	y3=v2.y;	z3=v2.z;

	// POINT C WILL BE THE SAME IN ALL 3 LEFT SIDE CHECKINGS
	c.InitAll(clickZ, clickY);

	// CHECK THE FIRST SEGMENT OF THE TRIANGLE
	a.InitAll(z1, y1);
	b.InitAll(z2, y2);
	LeftSeg1 = Left(a,b,c);
	if(!LeftSeg1)return false;

	// CHECK THE SECOND SEGMENT OF THE TRIANGLE
	a.InitAll(z2, y2);
	b.InitAll(z3, y3);
	LeftSeg2 = Left(a,b,c);
	if(!LeftSeg2)return false;

	// CHECK THE THIRD SEGMENT OF THE TRIANGLE
	a.InitAll(z3, y3);
	b.InitAll(z1, y1);
	LeftSeg3 = Left(a,b,c);
	if(!LeftSeg3)return false;

	// IF POINT C IS TO THE LEFT OF ALL 3 TRIANGLE SEGMENTS THEN THE POINT IS INSIDE THE TRIANGLE
	if(LeftSeg1 && LeftSeg2 && LeftSeg3)return true;
	else return false;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// DETERMINES IF AN EDGE FROM V1 TO V2 INTERSECTS PLANE plane
// IF THERE'S AN INTERSECTION, RETURN INTERSECTION POINT AND true
// TAKEN FROM GRAPHIC GEMS 3 
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$

/*
bool EdgeIntersectsPlane(GLVector4 v1, GLVector4 v2, GLPlane plane, GLVector4 *intersPt)
{
	HALFSPACE	side1, side2;	
	GLVector4	vecFromV1toV2, planeNormal;
	GLfloat		denom, tt;

	// ********************************************************************** 
	// MAKE SURE SIDES ARE DIFFERENT AND NEITHER IS COINCIDENT WITH THE PLANE 
	// ********************************************************************** 

	// GET SIDES OF THE PLANES
	side1 = plane.ClassifyPoint(v1);
	side2 = plane.ClassifyPoint(v2);

	// SINCE WE WON'T CALL THIS FUNCTION IF A VERTEX IS COINCIDENT WITH THE PLANE
	// ONE OF THE SIDES MUST BE POSITIVE AND THE OTHER ONE MUST BE NEGATIVE
	if((side1 == side2) || (side1 == ON_PLANE) || (side2 == ON_PLANE))
		return false;

	// ************************** 
	// COMPUTE INTERSECTION POINT 
	// ************************** 

	// FIRST, GET A VECTOR FROM V1 TO V2, (SUBSTRACT V2 - V1)
	vecFromV1toV2 = v2 - v1;
	// NORMALIZE VECTOR??
	// vecFromV1toV2.normalize();

	// MAKE SURE PLANE IS NORMALIZED
	plane.NormalizeGLPlane();

	// GET PLANE NORMAL
	planeNormal = plane.GetNormal();

	// USE DOT PRODUCT TO PROJECT VECTOR
	denom = planeNormal.dotProduct(vecFromV1toV2);

	assert(denom != 0.0f);

	// CALCULATE TT (DISTANCE)
	tt = (-1)*(plane.SignedDistanceToPoint(v1));
	tt += plane.Get_D();
	tt *= -1;
	tt /= denom;

	// CALCULATE INTERSECTION POINT
	intersPt->setValues((v1.x + (tt * vecFromV1toV2.x)),
						(v1.y + (tt * vecFromV1toV2.y)),
						(v1.z + (tt * vecFromV1toV2.z)));

	return true;
} */

bool EdgeIntersectsPlane(GLVector4 v1, GLVector4 v2, GLPlane plane, GLVector4 *intersPt)
{
	HALFSPACE	side1, side2;	
	GLVector4	vector_U, vector_W, planeNormal, pointOnPlane;
	GLfloat		denom, N, tt;

	// ********************************************************************** 
	// MAKE SURE SIDES ARE DIFFERENT AND NEITHER IS COINCIDENT WITH THE PLANE 
	// ********************************************************************** 

	// GET SIDES OF THE PLANES
	side1 = plane.ClassifyPoint(v1);
	side2 = plane.ClassifyPoint(v2);

	// SINCE WE WON'T CALL THIS FUNCTION IF A VERTEX IS COINCIDENT WITH THE PLANE
	// ONE OF THE SIDES MUST BE POSITIVE AND THE OTHER ONE MUST BE NEGATIVE
	if((side1 == side2) || (side1 == ON_PLANE) || (side2 == ON_PLANE))
		return false;

	// ************************** 
	// COMPUTE INTERSECTION POINT 
	// ************************** 

	// FIRST, GET A VECTOR FROM V1 TO V2, (SUBSTRACT V2 - V1)
	vector_U = v2 - v1;

	// GET VECTOR W
	plane.GetPointOnPlane(&pointOnPlane);
	vector_W = v1 - pointOnPlane;

	// MAKE SURE PLANE IS NORMALIZED
	plane.NormalizeGLPlane();
	// GET PLANE NORMAL
	planeNormal = plane.GetNormal();

	// USE DOT PRODUCT TO PROJECT VECTOR
	denom = planeNormal.dotProduct(vector_U);
	assert(denom != 0.0f);
	if((fabs(denom)) < 0.0000001f)
		denom = 0.0000001f;	

	N = (-1) * (planeNormal.dotProduct(vector_W));

	// CALCULATE TT (DISTANCE)
	tt = N / denom;

	//if (tt < 0.0f || tt > 1.0f)
	//	return false; // no intersection

	// CALCULATE INTERSECTION POINT
	*intersPt = v1 + (vector_U * tt);

	return true;
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// CHECK IF A NUMBER IS POWER OF 2
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
bool isPowerOf2(int number)
{
	if(number==1)
		return true;
	else
	{
		if(number%2==0)
			return isPowerOf2(number/2);
		else
			return false;
	}
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// RANGED RANDOM
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
float RangedRandom(float minV, float maxV)
{
	return minV + (maxV-minV)*((float)rand())/((float)RAND_MAX);
}

// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// RANGED RANDOM
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
int RangedRandomInt(int v1, int v2)
{
	return v1 + (v2-v1)*((int)rand())/((int)RAND_MAX);
}


// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
// ==================================================================================
// GETS THE RESULTING PARTITIONED TRIANGLES OF A TRIANGLE INTERSECTED BY
// THIS PLANE
// ==================================================================================
// $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
/*
bool GLPlane::PartitionTriWithPlane(GLTriangle tri, GLTriangle* partTris[3], 
														int *numPart, int hf[3])
{
	GLVector4	v1, v2, v3;
	GLVector4	intersectionPt1, intersectionPt2;
	HALFSPACE	side1, side2, side3;
	GLPlane		plane;

	// INITIALIZE TO NULL
	partTris[0] = NULL;
	partTris[1] = NULL;	
	partTris[2] = NULL;

	// GET VERTICES
	tri.Get_Vertices(&v1, &v2, &v3);

	// GET SIDE OF THE VERTICES
	side1 = ClassifyPoint(v1);
	side2 = ClassifyPoint(v2);
	side3 = ClassifyPoint(v3);

	plane.Set_Coefficients(m_A, m_B, m_C, m_D);

	// ************************************************************************************ 
	// IF THERE'S AN INTERSECTION POINT FROM FIRST EDGE (v1 -> v2) &&&&&&&&&&&&&&&&&&&&&&&&
	if(EdgeIntersectsPlane(v1, v2, plane, &intersectionPt1))
	{
		// FOUND FIRST INTERSECTION POINT, FIND SECOND

		// FIRST CASE, v2 AND v3 ARE ON THE OTHER SIDE
		if( (side2 != side1) && (side3 == side2) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v1 -> v3
			if(EdgeIntersectsPlane(v1, v3, plane, &intersectionPt2))
			{
				partTris[0] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(intersectionPt2, intersectionPt1, v3);
				hf[1]		= side2;
				partTris[2] = new GLTriangle(v3, intersectionPt1, v2);
				hf[2]		= side2;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...

		} // SECOND CASE, THE OTHER INTERSECTION POINT IS v3
		else if( (side2 != side1) && (side3 == ON_PLANE) )
		{
			intersectionPt2 = v3;

			partTris[0] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
			hf[0]		= side1;
			partTris[1] = new GLTriangle(intersectionPt2, intersectionPt1, v2);
			hf[1]		= side2;
			*numPart	= 2;

		} // LAST CASE,
		else if( (side2 != side1) && (side3 == side1) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v2 -> v3
			if(EdgeIntersectsPlane(v2, v3, plane, &intersectionPt2))
			{
				partTris[0] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(v1, intersectionPt2, v3);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(intersectionPt1, v2, intersectionPt2);
				hf[2]		= side2;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...
		}
		else return false; // SHOULD NEVER GET TO THIS POINT...

	} // STILL NOT FOUND FIRST INT POINT, TRY THE SECOND EDGE (v2 -> v3) &&&&&&&&&&&&&&&&&&
	// ************************************************************************************
	else if(EdgeIntersectsPlane(v2, v3, plane, &intersectionPt1))
	{
		// FOUND FIRST INTERSECTION POINT, FIND SECOND

		// FIRST CASE, v1 AND v2 ARE ON THE OTHER SIDE
		if( (side2 != side3) && (side1 == side2) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v3 -> v1
			if(EdgeIntersectsPlane(v3, v1, plane, &intersectionPt2))
			{
				partTris[0] = new GLTriangle(v3, intersectionPt2, intersectionPt1);
				hf[0]		= side3;
				partTris[1] = new GLTriangle(v1, intersectionPt1, intersectionPt2);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(v1, v2, intersectionPt1);
				hf[2]		= side2;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...

		} // SECOND CASE, THE OTHER INTERSECTION POINT IS v1
		else if( (side2 != side3) && (side1 == ON_PLANE) )
		{
			intersectionPt2 = v1;

			partTris[0] = new GLTriangle(v1, intersectionPt1, v3);
			hf[0]		= side3;
			partTris[1] = new GLTriangle(v1, v2, intersectionPt1);
			hf[1]		= side2;
			*numPart	= 2;

		} // LAST CASE,
		else if( (side2 != side3) && (side1 == side3) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v1 -> v2
			if(EdgeIntersectsPlane(v1, v2, plane, &intersectionPt2))
			{
				partTris[0] = new GLTriangle(intersectionPt2, v2, intersectionPt1);
				hf[0]		= side2;
				partTris[1] = new GLTriangle(v1, intersectionPt2, intersectionPt1);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(v1, intersectionPt1, v3);
				hf[2]		= side3;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...
		}
		else return false; // SHOULD NEVER GET TO THIS POINT...

	} // STILL NOT FOUND FIRST INT POINT, TRY THE LAST EDGE (v3 -> v1) &&&&&&&&&&&&&&&&&&&&
	// ************************************************************************************
	else if(EdgeIntersectsPlane(v3, v1, plane, &intersectionPt1))
	{
		// FOUND FIRST INTERSECTION POINT, FIND SECOND

		// FIRST CASE, v1 AND v2 ARE ON THE OTHER SIDE
		if( (side3 != side1) && (side1 == side2) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v2 -> v3
			if(EdgeIntersectsPlane(v2, v3, plane, &intersectionPt2))
			{
				partTris[0] = new GLTriangle(intersectionPt1, intersectionPt2, v3);
				hf[0]		= side3;
				partTris[1] = new GLTriangle(v1, intersectionPt2, intersectionPt1);
				hf[1]		= side1;
				partTris[2] = new GLTriangle(v1, v2, intersectionPt2);
				hf[2]		= side2;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...

		} // SECOND CASE, THE OTHER INTERSECTION POINT IS v2
		else if( (side3 != side1) && (side2 == ON_PLANE) )
		{
			intersectionPt2 = v2;

			partTris[0] = new GLTriangle(intersectionPt1, intersectionPt2, v3);
			hf[0]		= side3;
			partTris[1] = new GLTriangle(v1, v2, intersectionPt1);
			hf[1]		= side1;
			*numPart	= 2;

		} // LAST CASE,
		else if( (side3 != side1) && (side2 == side3) )
		{
			// THEN, THE OTHER INTERSECTION POINT MUST BE IN THE EDGE v1 -> v2
			if(EdgeIntersectsPlane(v1, v2, plane, &intersectionPt2))
			{
				partTris[0] = new GLTriangle(v1, intersectionPt2, intersectionPt1);
				hf[0]		= side1;
				partTris[1] = new GLTriangle(intersectionPt1, intersectionPt2, v2);
				hf[1]		= side2;
				partTris[2] = new GLTriangle(intersectionPt1, v2, v3);
				hf[2]		= side3;
				*numPart	= 3;
			}
			else return false; // SOMETHING STRANGE...
		}
		else return false; // SHOULD NEVER GET TO THIS POINT...

	}
	// **********************************************************************************

	return true;
}
*/


// -------------------------------- END OF FILE -----------------------------------
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%










